# generated by datamodel-codegen:
#   filename:  processed.json
#   timestamp: 2023-08-02T22:39:49+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field


class Contact(BaseModel):
    name: str
    url: str


class License(Contact):
    pass


class Info(BaseModel):
    version: str
    x_looker_release_version: str = Field(..., alias="x-looker-release-version")
    title: str
    description: str
    contact: Contact
    license: License


class Tag(BaseModel):
    name: str
    description: str


class Schema(BaseModel):
    field_ref: str = Field(..., alias="$ref")


class Parameter(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema] = Field(None, alias="schema")
    type: Optional[str] = None
    format: Optional[str] = None
    x_looker_deprecated: Optional[bool] = Field(None, alias="x-looker-deprecated")


class Field400(BaseModel):
    description: str
    schema_: Schema = Field(..., alias="schema")


class Field404(Field400):
    pass


class Field409(Field400):
    pass


class Field422(Field400):
    pass


class Field429(Field400):
    pass


class Field200(Field400):
    pass


class Responses(BaseModel):
    field_400: Field400 = Field(..., alias="400")
    field_404: Field404 = Field(..., alias="404")
    field_409: Field409 = Field(..., alias="409")
    field_422: Field422 = Field(..., alias="422")
    field_429: Field429 = Field(..., alias="429")
    field_200: Field200 = Field(..., alias="200")


class Post(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter]
    responses: Responses
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldQueryTasks(BaseModel):
    post: Post


class Items(BaseModel):
    type: str


class Parameter1(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str
    items: Items
    collectionFormat: str


class Field4001(Field400):
    pass


class Field4041(Field400):
    pass


class AdditionalProperties(BaseModel):
    type: str
    format: str


class Schema9(BaseModel):
    type: str
    additionalProperties: AdditionalProperties


class Field2001(BaseModel):
    description: str
    schema_: Schema9 = Field(..., alias="schema")


class Responses1(BaseModel):
    field_400: Field4001 = Field(..., alias="400")
    field_404: Field4041 = Field(..., alias="404")
    field_200: Field2001 = Field(..., alias="200")


class Get(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter1]
    responses: Responses1
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldQueryTasksMultiResults(BaseModel):
    get: Get


class Parameter2(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str


class Schema10(Schema):
    pass


class Field4002(BaseModel):
    description: str
    schema_: Schema10 = Field(..., alias="schema")


class Field4042(Field4002):
    pass


class Field2002(Field4002):
    pass


class Responses2(BaseModel):
    field_400: Field4002 = Field(..., alias="400")
    field_404: Field4042 = Field(..., alias="404")
    field_200: Field2002 = Field(..., alias="200")


class Get1(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter2]
    responses: Responses2
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldQueryTasksQueryTaskId(BaseModel):
    get: Get1


class Field4003(Field4002):
    pass


class Field4043(Field4002):
    pass


class Schema15(Items):
    pass


class Field2003(BaseModel):
    description: str
    schema_: Schema15 = Field(..., alias="schema")


class Field204(Field2003):
    pass


class Responses3(BaseModel):
    field_400: Field4003 = Field(..., alias="400")
    field_404: Field4043 = Field(..., alias="404")
    field_200: Field2003 = Field(..., alias="200")
    field_204: Field204 = Field(..., alias="204")


class Get2(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter2]
    responses: Responses3
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldQueryTasksQueryTaskIdResults(BaseModel):
    get: Get2


class Schema17(Schema):
    pass


class Field4004(BaseModel):
    description: str
    schema_: Schema17 = Field(..., alias="schema")


class Field4044(Field4004):
    pass


class Field2004(Field4004):
    pass


class Responses4(BaseModel):
    field_400: Field4004 = Field(..., alias="400")
    field_404: Field4044 = Field(..., alias="404")
    field_200: Field2004 = Field(..., alias="200")


class Get3(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter2]
    responses: Responses4
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldQueriesQueryId(BaseModel):
    get: Get3


class Field4005(Field4004):
    pass


class Field4045(Field4004):
    pass


class Field2005(Field4004):
    pass


class Responses5(BaseModel):
    field_400: Field4005 = Field(..., alias="400")
    field_404: Field4045 = Field(..., alias="404")
    field_200: Field2005 = Field(..., alias="200")


class Get4(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter2]
    responses: Responses5
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldQueriesSlugSlug(BaseModel):
    get: Get4


class Parameter6(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema17] = Field(None, alias="schema")
    type: Optional[str] = None


class Field4006(Field4004):
    pass


class Field4046(Field4004):
    pass


class Field4091(Field4004):
    pass


class Field4221(Field4004):
    pass


class Field4291(Field4004):
    pass


class Field2006(Field4004):
    pass


class Responses6(BaseModel):
    field_400: Field4006 = Field(..., alias="400")
    field_404: Field4046 = Field(..., alias="404")
    field_409: Field4091 = Field(..., alias="409")
    field_422: Field4221 = Field(..., alias="422")
    field_429: Field4291 = Field(..., alias="429")
    field_200: Field2006 = Field(..., alias="200")


class Post1(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter6]
    responses: Responses6
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldQueries(BaseModel):
    post: Post1


class Parameter7(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str
    format: Optional[str] = None


class Field4007(Field4004):
    pass


class Field4047(Field4004):
    pass


class Field4222(Field4004):
    pass


class Field4292(Field4004):
    pass


class Schema34(Items):
    pass


class Field2007(BaseModel):
    description: str
    schema_: Schema34 = Field(..., alias="schema")


class Responses7(BaseModel):
    field_400: Field4007 = Field(..., alias="400")
    field_404: Field4047 = Field(..., alias="404")
    field_422: Field4222 = Field(..., alias="422")
    field_429: Field4292 = Field(..., alias="429")
    field_200: Field2007 = Field(..., alias="200")


class Get5(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter7]
    responses: Responses7
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldQueriesQueryIdRunResultFormat(BaseModel):
    get: Get5


class Schema35(Schema):
    pass


class Parameter8(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema35] = Field(None, alias="schema")
    format: Optional[str] = None


class Field4008(BaseModel):
    description: str
    schema_: Schema35 = Field(..., alias="schema")


class Field4048(Field4008):
    pass


class Field4223(Field4008):
    pass


class Field4293(Field4008):
    pass


class Schema40(Items):
    pass


class Field2008(BaseModel):
    description: str
    schema_: Schema40 = Field(..., alias="schema")


class Responses8(BaseModel):
    field_400: Field4008 = Field(..., alias="400")
    field_404: Field4048 = Field(..., alias="404")
    field_422: Field4223 = Field(..., alias="422")
    field_429: Field4293 = Field(..., alias="429")
    field_200: Field2008 = Field(..., alias="200")


class Post2(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter8]
    responses: Responses8
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldQueriesRunResultFormat(BaseModel):
    post: Post2


class Parameter9(Parameter2):
    pass


class Schema41(Schema):
    pass


class Field4009(BaseModel):
    description: str
    schema_: Schema41 = Field(..., alias="schema")


class Field4049(Field4009):
    pass


class Field4224(Field4009):
    pass


class Field4294(Field4009):
    pass


class Schema45(Items):
    pass


class Field2009(BaseModel):
    description: str
    schema_: Schema45 = Field(..., alias="schema")


class Responses9(BaseModel):
    field_400: Field4009 = Field(..., alias="400")
    field_404: Field4049 = Field(..., alias="404")
    field_422: Field4224 = Field(..., alias="422")
    field_429: Field4294 = Field(..., alias="429")
    field_200: Field2009 = Field(..., alias="200")


class Get6(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter9]
    responses: Responses9
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldQueriesModelsModelNameViewsViewNameRunResultFormat(BaseModel):
    get: Get6


class Schema46(Schema):
    pass


class Field40010(BaseModel):
    description: str
    schema_: Schema46 = Field(..., alias="schema")


class Field40410(Field40010):
    pass


class Field20010(Field40010):
    pass


class Responses10(BaseModel):
    field_400: Field40010 = Field(..., alias="400")
    field_404: Field40410 = Field(..., alias="404")
    field_200: Field20010 = Field(..., alias="200")


class Post3(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    consumes: List[str]
    parameters: List[Parameter9]
    responses: Responses10
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLogin(BaseModel):
    post: Post3


class Field40011(Field40010):
    pass


class Field403(Field40010):
    pass


class Field40411(Field40010):
    pass


class Field20011(Field40010):
    pass


class Responses11(BaseModel):
    field_400: Field40011 = Field(..., alias="400")
    field_403: Field403 = Field(..., alias="403")
    field_404: Field40411 = Field(..., alias="404")
    field_200: Field20011 = Field(..., alias="200")


class Post4(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter9]
    responses: Responses11
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLoginUserId(BaseModel):
    post: Post4


class Field40012(Field40010):
    pass


class Field40412(Field40010):
    pass


class Schema55(Items):
    pass


class Field2041(BaseModel):
    description: str
    schema_: Schema55 = Field(..., alias="schema")


class Responses12(BaseModel):
    field_400: Field40012 = Field(..., alias="400")
    field_404: Field40412 = Field(..., alias="404")
    field_204: Field2041 = Field(..., alias="204")


class Delete(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses12
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLogout(BaseModel):
    delete: Delete


class Schema56(Schema):
    pass


class Field40013(BaseModel):
    description: str
    schema_: Schema56 = Field(..., alias="schema")


class Field40413(Field40013):
    pass


class Field2042(BaseModel):
    description: str


class Responses13(BaseModel):
    field_400: Field40013 = Field(..., alias="400")
    field_404: Field40413 = Field(..., alias="404")
    field_204: Field2042 = Field(..., alias="204")


class Post5(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter9]
    responses: Responses13
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field40014(Field40013):
    pass


class Field40414(Field40013):
    pass


class Field4295(Field40013):
    pass


class Responses14(BaseModel):
    field_400: Field40014 = Field(..., alias="400")
    field_404: Field40414 = Field(..., alias="404")
    field_429: Field4295 = Field(..., alias="429")
    field_204: Field2042 = Field(..., alias="204")


class Delete1(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter9]
    responses: Responses14
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldAlertsAlertIdFollow(BaseModel):
    post: Post5
    delete: Delete1


class Parameter14(Parameter7):
    pass


class Field40015(Field40013):
    pass


class Field40415(Field40013):
    pass


class Field4031(Field40013):
    pass


class Items1(Schema):
    pass


class Schema64(BaseModel):
    type: str
    items: Items1


class Field20012(BaseModel):
    description: str
    schema_: Schema64 = Field(..., alias="schema")


class Responses15(BaseModel):
    field_400: Field40015 = Field(..., alias="400")
    field_404: Field40415 = Field(..., alias="404")
    field_403: Field4031 = Field(..., alias="403")
    field_200: Field20012 = Field(..., alias="200")


class Get7(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter14]
    responses: Responses15
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldAlertsSearch(BaseModel):
    get: Get7


class Parameter15(Parameter2):
    pass


class Schema65(Schema):
    pass


class Field40016(BaseModel):
    description: str
    schema_: Schema65 = Field(..., alias="schema")


class Field40416(Field40016):
    pass


class Field20013(Field40016):
    pass


class Responses16(BaseModel):
    field_400: Field40016 = Field(..., alias="400")
    field_404: Field40416 = Field(..., alias="404")
    field_200: Field20013 = Field(..., alias="200")


class Get8(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter15]
    responses: Responses16
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter16(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema65] = Field(None, alias="schema")


class Field40017(Field40016):
    pass


class Field40417(Field40016):
    pass


class Field4225(Field40016):
    pass


class Field4296(Field40016):
    pass


class Field4032(Field40016):
    pass


class Field20014(Field40016):
    pass


class Responses17(BaseModel):
    field_400: Field40017 = Field(..., alias="400")
    field_404: Field40417 = Field(..., alias="404")
    field_422: Field4225 = Field(..., alias="422")
    field_429: Field4296 = Field(..., alias="429")
    field_403: Field4032 = Field(..., alias="403")
    field_200: Field20014 = Field(..., alias="200")


class Patch(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter16]
    responses: Responses17
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter17(Parameter16):
    pass


class Field40018(Field40016):
    pass


class Field40418(Field40016):
    pass


class Field4226(Field40016):
    pass


class Field4297(Field40016):
    pass


class Field20015(Field40016):
    pass


class Responses18(BaseModel):
    field_400: Field40018 = Field(..., alias="400")
    field_404: Field40418 = Field(..., alias="404")
    field_422: Field4226 = Field(..., alias="422")
    field_429: Field4297 = Field(..., alias="429")
    field_200: Field20015 = Field(..., alias="200")


class Put(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter17]
    responses: Responses18
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter18(Parameter2):
    pass


class Field40019(Field40016):
    pass


class Field40419(Field40016):
    pass


class Field4298(Field40016):
    pass


class Field4033(Field40016):
    pass


class Responses19(BaseModel):
    field_400: Field40019 = Field(..., alias="400")
    field_404: Field40419 = Field(..., alias="404")
    field_429: Field4298 = Field(..., alias="429")
    field_403: Field4033 = Field(..., alias="403")
    field_204: Field2042 = Field(..., alias="204")


class Delete2(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter18]
    responses: Responses19
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldAlertsAlertId(BaseModel):
    get: Get8
    patch: Patch
    put: Put
    delete: Delete2


class Parameter19(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema65 = Field(..., alias="schema")


class Field40020(Field40016):
    pass


class Field40420(Field40016):
    pass


class Field4092(Field40016):
    pass


class Field4227(Field40016):
    pass


class Field4299(Field40016):
    pass


class Field405(Field40016):
    pass


class Field20016(Field40016):
    pass


class Responses20(BaseModel):
    field_400: Field40020 = Field(..., alias="400")
    field_404: Field40420 = Field(..., alias="404")
    field_409: Field4092 = Field(..., alias="409")
    field_422: Field4227 = Field(..., alias="422")
    field_429: Field4299 = Field(..., alias="429")
    field_405: Field405 = Field(..., alias="405")
    field_200: Field20016 = Field(..., alias="200")


class Post6(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter19]
    responses: Responses20
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldAlerts(BaseModel):
    post: Post6


class Parameter20(Parameter2):
    pass


class Field40021(Field40016):
    pass


class Field4034(Field40016):
    pass


class Field40421(Field40016):
    pass


class Field42910(Field40016):
    pass


class Responses21(BaseModel):
    field_400: Field40021 = Field(..., alias="400")
    field_403: Field4034 = Field(..., alias="403")
    field_404: Field40421 = Field(..., alias="404")
    field_429: Field42910 = Field(..., alias="429")
    field_204: Field2042 = Field(..., alias="204")


class Post7(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter20]
    responses: Responses21
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldAlertsAlertIdEnqueue(BaseModel):
    post: Post7


class Parameter21(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str
    format: str


class Field40022(Field40016):
    pass


class Field40422(Field40016):
    pass


class Schema99(Schema64):
    pass


class Field20017(BaseModel):
    description: str
    schema_: Schema99 = Field(..., alias="schema")


class Responses22(BaseModel):
    field_400: Field40022 = Field(..., alias="400")
    field_404: Field40422 = Field(..., alias="404")
    field_200: Field20017 = Field(..., alias="200")


class Get9(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter21]
    responses: Responses22
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldAlertNotifications(BaseModel):
    get: Get9


class Parameter22(Parameter2):
    pass


class Schema100(Schema):
    pass


class Field40023(BaseModel):
    description: str
    schema_: Schema100 = Field(..., alias="schema")


class Field40423(Field40023):
    pass


class Field4228(Field40023):
    pass


class Field42911(Field40023):
    pass


class Field20018(Field40023):
    pass


class Responses23(BaseModel):
    field_400: Field40023 = Field(..., alias="400")
    field_404: Field40423 = Field(..., alias="404")
    field_422: Field4228 = Field(..., alias="422")
    field_429: Field42911 = Field(..., alias="429")
    field_200: Field20018 = Field(..., alias="200")


class Patch1(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter22]
    responses: Responses23
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldAlertNotificationsAlertNotificationId(BaseModel):
    patch: Patch1


class Field40024(Field40023):
    pass


class Field40424(Field40023):
    pass


class Field4229(Field40023):
    pass


class Field42912(Field40023):
    pass


class Field20019(Field40023):
    pass


class Responses24(BaseModel):
    field_400: Field40024 = Field(..., alias="400")
    field_404: Field40424 = Field(..., alias="404")
    field_422: Field4229 = Field(..., alias="422")
    field_429: Field42912 = Field(..., alias="429")
    field_200: Field20019 = Field(..., alias="200")


class Get10(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter22]
    responses: Responses24
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldArtifactUsage(BaseModel):
    get: Get10


class Parameter24(Parameter7):
    pass


class Field40025(Field40023):
    pass


class Field40425(Field40023):
    pass


class Field42913(Field40023):
    pass


class Field42210(Field40023):
    pass


class Schema114(Schema64):
    pass


class Field20020(BaseModel):
    description: str
    schema_: Schema114 = Field(..., alias="schema")


class Responses25(BaseModel):
    field_400: Field40025 = Field(..., alias="400")
    field_404: Field40425 = Field(..., alias="404")
    field_429: Field42913 = Field(..., alias="429")
    field_422: Field42210 = Field(..., alias="422")
    field_200: Field20020 = Field(..., alias="200")


class Get11(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter24]
    responses: Responses25
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldArtifactNamespaces(BaseModel):
    get: Get11


class Parameter25(Parameter2):
    pass


class Schema115(Schema):
    pass


class Field40026(BaseModel):
    description: str
    schema_: Schema115 = Field(..., alias="schema")


class Field40426(Field40026):
    pass


class Field42211(Field40026):
    pass


class Field42914(Field40026):
    pass


class Schema119(Items):
    pass


class Field20021(BaseModel):
    description: str
    schema_: Schema119 = Field(..., alias="schema")


class Responses26(BaseModel):
    field_400: Field40026 = Field(..., alias="400")
    field_404: Field40426 = Field(..., alias="404")
    field_422: Field42211 = Field(..., alias="422")
    field_429: Field42914 = Field(..., alias="429")
    field_200: Field20021 = Field(..., alias="200")


class Get12(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter25]
    responses: Responses26
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldArtifactNamespaceValue(BaseModel):
    get: Get12


class Schema120(Schema):
    pass


class Field40027(BaseModel):
    description: str
    schema_: Schema120 = Field(..., alias="schema")


class Field40427(Field40027):
    pass


class Field42915(Field40027):
    pass


class Responses27(BaseModel):
    field_400: Field40027 = Field(..., alias="400")
    field_404: Field40427 = Field(..., alias="404")
    field_429: Field42915 = Field(..., alias="429")
    field_204: Field2042 = Field(..., alias="204")


class Delete3(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter25]
    responses: Responses27
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldArtifactNamespacePurge(BaseModel):
    delete: Delete3


class Parameter27(Parameter7):
    pass


class Field40028(Field40027):
    pass


class Field40428(Field40027):
    pass


class Field42916(Field40027):
    pass


class Field42212(Field40027):
    pass


class Schema127(Schema64):
    pass


class Field20022(BaseModel):
    description: str
    schema_: Schema127 = Field(..., alias="schema")


class Responses28(BaseModel):
    field_400: Field40028 = Field(..., alias="400")
    field_404: Field40428 = Field(..., alias="404")
    field_429: Field42916 = Field(..., alias="429")
    field_422: Field42212 = Field(..., alias="422")
    field_200: Field20022 = Field(..., alias="200")


class Get13(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter27]
    responses: Responses28
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldArtifactNamespaceSearch(BaseModel):
    get: Get13


class Schema128(Schema):
    pass


class Field40029(BaseModel):
    description: str
    schema_: Schema128 = Field(..., alias="schema")


class Field40429(Field40029):
    pass


class Field42213(Field40029):
    pass


class Field42917(Field40029):
    pass


class Schema132(Schema64):
    pass


class Field20023(BaseModel):
    description: str
    schema_: Schema132 = Field(..., alias="schema")


class Responses29(BaseModel):
    field_400: Field40029 = Field(..., alias="400")
    field_404: Field40429 = Field(..., alias="404")
    field_422: Field42213 = Field(..., alias="422")
    field_429: Field42917 = Field(..., alias="429")
    field_200: Field20023 = Field(..., alias="200")


class Get14(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter27]
    responses: Responses29
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter29(Parameter2):
    pass


class Schema133(Schema):
    pass


class Field40030(BaseModel):
    description: str
    schema_: Schema133 = Field(..., alias="schema")


class Field40430(Field40030):
    pass


class Field42918(Field40030):
    pass


class Responses30(BaseModel):
    field_400: Field40030 = Field(..., alias="400")
    field_404: Field40430 = Field(..., alias="404")
    field_429: Field42918 = Field(..., alias="429")
    field_204: Field2042 = Field(..., alias="204")


class Delete4(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter29]
    responses: Responses30
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldArtifactNamespace(BaseModel):
    get: Get14
    delete: Delete4


class Schema136(Schema64):
    pass


class Parameter30(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema136] = Field(None, alias="schema")


class Schema137(Schema):
    pass


class Field40031(BaseModel):
    description: str
    schema_: Schema137 = Field(..., alias="schema")


class Field40431(Field40031):
    pass


class Field42214(Field40031):
    pass


class Field42919(Field40031):
    pass


class Schema141(Schema64):
    pass


class Field20024(BaseModel):
    description: str
    schema_: Schema141 = Field(..., alias="schema")


class Responses31(BaseModel):
    field_400: Field40031 = Field(..., alias="400")
    field_404: Field40431 = Field(..., alias="404")
    field_422: Field42214 = Field(..., alias="422")
    field_429: Field42919 = Field(..., alias="429")
    field_200: Field20024 = Field(..., alias="200")


class Put1(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter30]
    responses: Responses31
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldArtifactsNamespace(BaseModel):
    put: Put1


class Schema142(Schema):
    pass


class Field40032(BaseModel):
    description: str
    schema_: Schema142 = Field(..., alias="schema")


class Field40432(Field40032):
    pass


class Field20025(Field40032):
    pass


class Responses32(BaseModel):
    field_400: Field40032 = Field(..., alias="400")
    field_404: Field40432 = Field(..., alias="404")
    field_200: Field20025 = Field(..., alias="200")


class Get15(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses32
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter31(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema142 = Field(..., alias="schema")


class Field40033(Field40032):
    pass


class Field40433(Field40032):
    pass


class Field42215(Field40032):
    pass


class Field20026(Field40032):
    pass


class Responses33(BaseModel):
    field_400: Field40033 = Field(..., alias="400")
    field_404: Field40433 = Field(..., alias="404")
    field_422: Field42215 = Field(..., alias="422")
    field_200: Field20026 = Field(..., alias="200")


class Patch2(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter31]
    responses: Responses33
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldCloudStorage(BaseModel):
    get: Get15
    patch: Patch2


class Parameter32(Parameter2):
    pass


class Field40034(Field40032):
    pass


class Field40434(Field40032):
    pass


class Schema152(Schema64):
    pass


class Field20027(BaseModel):
    description: str
    schema_: Schema152 = Field(..., alias="schema")


class Responses34(BaseModel):
    field_400: Field40034 = Field(..., alias="400")
    field_404: Field40434 = Field(..., alias="404")
    field_200: Field20027 = Field(..., alias="200")


class Get16(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter32]
    responses: Responses34
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema153(Schema):
    pass


class Parameter33(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema153 = Field(..., alias="schema")


class Field40035(BaseModel):
    description: str
    schema_: Schema153 = Field(..., alias="schema")


class Field40435(Field40035):
    pass


class Field4093(Field40035):
    pass


class Field42216(Field40035):
    pass


class Field42920(Field40035):
    pass


class Field20028(Field40035):
    pass


class Field4035(Field40035):
    pass


class Responses35(BaseModel):
    field_400: Field40035 = Field(..., alias="400")
    field_404: Field40435 = Field(..., alias="404")
    field_409: Field4093 = Field(..., alias="409")
    field_422: Field42216 = Field(..., alias="422")
    field_429: Field42920 = Field(..., alias="429")
    field_200: Field20028 = Field(..., alias="200")
    field_403: Field4035 = Field(..., alias="403")


class Post8(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter33]
    responses: Responses35
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldColorCollections(BaseModel):
    get: Get16
    post: Post8


class Parameter34(Parameter2):
    pass


class Field40036(Field40035):
    pass


class Field40436(Field40035):
    pass


class Schema163(Schema64):
    pass


class Field20029(BaseModel):
    description: str
    schema_: Schema163 = Field(..., alias="schema")


class Responses36(BaseModel):
    field_400: Field40036 = Field(..., alias="400")
    field_404: Field40436 = Field(..., alias="404")
    field_200: Field20029 = Field(..., alias="200")


class Get17(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter34]
    responses: Responses36
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldColorCollectionsCustom(BaseModel):
    get: Get17


class Schema164(Schema):
    pass


class Field40037(BaseModel):
    description: str
    schema_: Schema164 = Field(..., alias="schema")


class Field40437(Field40037):
    pass


class Schema166(Schema64):
    pass


class Field20030(BaseModel):
    description: str
    schema_: Schema166 = Field(..., alias="schema")


class Responses37(BaseModel):
    field_400: Field40037 = Field(..., alias="400")
    field_404: Field40437 = Field(..., alias="404")
    field_200: Field20030 = Field(..., alias="200")


class Get18(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter34]
    responses: Responses37
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldColorCollectionsStandard(BaseModel):
    get: Get18


class Schema167(Schema):
    pass


class Field40038(BaseModel):
    description: str
    schema_: Schema167 = Field(..., alias="schema")


class Field40438(Field40038):
    pass


class Field42217(Field40038):
    pass


class Field42921(Field40038):
    pass


class Field20031(Field40038):
    pass


class Responses38(BaseModel):
    field_400: Field40038 = Field(..., alias="400")
    field_404: Field40438 = Field(..., alias="404")
    field_422: Field42217 = Field(..., alias="422")
    field_429: Field42921 = Field(..., alias="429")
    field_200: Field20031 = Field(..., alias="200")


class Put2(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter34]
    responses: Responses38
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field40039(Field40038):
    pass


class Field40439(Field40038):
    pass


class Field20032(Field40038):
    pass


class Responses39(BaseModel):
    field_400: Field40039 = Field(..., alias="400")
    field_404: Field40439 = Field(..., alias="404")
    field_200: Field20032 = Field(..., alias="200")


class Get19(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses39
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldColorCollectionsDefault(BaseModel):
    put: Put2
    get: Get19


class Field40040(Field40038):
    pass


class Field40440(Field40038):
    pass


class Field20033(Field40038):
    pass


class Responses40(BaseModel):
    field_400: Field40040 = Field(..., alias="400")
    field_404: Field40440 = Field(..., alias="404")
    field_200: Field20033 = Field(..., alias="200")


class Get20(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter34]
    responses: Responses40
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter38(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema167] = Field(None, alias="schema")


class Field40041(Field40038):
    pass


class Field40441(Field40038):
    pass


class Field42218(Field40038):
    pass


class Field42922(Field40038):
    pass


class Field20034(Field40038):
    pass


class Field4036(Field40038):
    pass


class Responses41(BaseModel):
    field_400: Field40041 = Field(..., alias="400")
    field_404: Field40441 = Field(..., alias="404")
    field_422: Field42218 = Field(..., alias="422")
    field_429: Field42922 = Field(..., alias="429")
    field_200: Field20034 = Field(..., alias="200")
    field_403: Field4036 = Field(..., alias="403")


class Patch3(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter38]
    responses: Responses41
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter39(Parameter2):
    pass


class Field40042(Field40038):
    pass


class Field40442(Field40038):
    pass


class Field42923(Field40038):
    pass


class Field20035(Field2042):
    pass


class Field4037(Field40038):
    pass


class Schema189(Items):
    pass


class Field2048(BaseModel):
    description: str
    schema_: Schema189 = Field(..., alias="schema")


class Responses42(BaseModel):
    field_400: Field40042 = Field(..., alias="400")
    field_404: Field40442 = Field(..., alias="404")
    field_429: Field42923 = Field(..., alias="429")
    field_200: Field20035 = Field(..., alias="200")
    field_403: Field4037 = Field(..., alias="403")
    field_204: Field2048 = Field(..., alias="204")


class Delete5(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter39]
    responses: Responses42
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldColorCollectionsCollectionId(BaseModel):
    get: Get20
    patch: Patch3
    delete: Delete5


class Parameter40(Parameter7):
    pass


class Schema190(Schema):
    pass


class Field40043(BaseModel):
    description: str
    schema_: Schema190 = Field(..., alias="schema")


class Field40443(Field40043):
    pass


class Schema192(Schema64):
    pass


class Field20036(BaseModel):
    description: str
    schema_: Schema192 = Field(..., alias="schema")


class Responses43(BaseModel):
    field_400: Field40043 = Field(..., alias="400")
    field_404: Field40443 = Field(..., alias="404")
    field_200: Field20036 = Field(..., alias="200")


class Get21(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter40]
    responses: Responses43
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentFavoriteSearch(BaseModel):
    get: Get21


class Parameter41(Parameter2):
    pass


class Schema193(Schema):
    pass


class Field40044(BaseModel):
    description: str
    schema_: Schema193 = Field(..., alias="schema")


class Field40444(Field40044):
    pass


class Field20037(Field40044):
    pass


class Responses44(BaseModel):
    field_400: Field40044 = Field(..., alias="400")
    field_404: Field40444 = Field(..., alias="404")
    field_200: Field20037 = Field(..., alias="200")


class Get22(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter41]
    responses: Responses44
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field40045(Field40044):
    pass


class Field40445(Field40044):
    pass


class Field42924(Field40044):
    pass


class Schema199(Items):
    pass


class Field2049(BaseModel):
    description: str
    schema_: Schema199 = Field(..., alias="schema")


class Responses45(BaseModel):
    field_400: Field40045 = Field(..., alias="400")
    field_404: Field40445 = Field(..., alias="404")
    field_429: Field42924 = Field(..., alias="429")
    field_204: Field2049 = Field(..., alias="204")


class Delete6(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter41]
    responses: Responses45
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentFavoriteContentFavoriteId(BaseModel):
    get: Get22
    delete: Delete6


class Schema200(Schema):
    pass


class Parameter43(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema200 = Field(..., alias="schema")


class Field40046(BaseModel):
    description: str
    schema_: Schema200 = Field(..., alias="schema")


class Field40446(Field40046):
    pass


class Field4094(Field40046):
    pass


class Field42219(Field40046):
    pass


class Field42925(Field40046):
    pass


class Field20038(Field40046):
    pass


class Responses46(BaseModel):
    field_400: Field40046 = Field(..., alias="400")
    field_404: Field40446 = Field(..., alias="404")
    field_409: Field4094 = Field(..., alias="409")
    field_422: Field42219 = Field(..., alias="422")
    field_429: Field42925 = Field(..., alias="429")
    field_200: Field20038 = Field(..., alias="200")


class Post9(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter43]
    responses: Responses46
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentFavorite(BaseModel):
    post: Post9


class Parameter44(Parameter2):
    pass


class Field40047(Field40046):
    pass


class Field40447(Field40046):
    pass


class Schema209(Schema64):
    pass


class Field20039(BaseModel):
    description: str
    schema_: Schema209 = Field(..., alias="schema")


class Responses47(BaseModel):
    field_400: Field40047 = Field(..., alias="400")
    field_404: Field40447 = Field(..., alias="404")
    field_200: Field20039 = Field(..., alias="200")


class Get23(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter44]
    responses: Responses47
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentMetadata(BaseModel):
    get: Get23


class Schema210(Schema):
    pass


class Parameter45(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema210] = Field(None, alias="schema")


class Field40048(BaseModel):
    description: str
    schema_: Schema210 = Field(..., alias="schema")


class Field40448(Field40048):
    pass


class Field42220(Field40048):
    pass


class Field42926(Field40048):
    pass


class Field20040(Field40048):
    pass


class Responses48(BaseModel):
    field_400: Field40048 = Field(..., alias="400")
    field_404: Field40448 = Field(..., alias="404")
    field_422: Field42220 = Field(..., alias="422")
    field_429: Field42926 = Field(..., alias="429")
    field_200: Field20040 = Field(..., alias="200")


class Patch4(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter45]
    responses: Responses48
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter46(Parameter2):
    pass


class Field40049(Field40048):
    pass


class Field40449(Field40048):
    pass


class Field20041(Field40048):
    pass


class Responses49(BaseModel):
    field_400: Field40049 = Field(..., alias="400")
    field_404: Field40449 = Field(..., alias="404")
    field_200: Field20041 = Field(..., alias="200")


class Get24(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter46]
    responses: Responses49
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentMetadataContentMetadataId(BaseModel):
    patch: Patch4
    get: Get24


class Parameter47(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema210] = Field(None, alias="schema")
    type: Optional[str] = None


class Field40050(Field40048):
    pass


class Field40450(Field40048):
    pass


class Field4095(Field40048):
    pass


class Field42221(Field40048):
    pass


class Field42927(Field40048):
    pass


class Field20042(Field40048):
    pass


class Responses50(BaseModel):
    field_400: Field40050 = Field(..., alias="400")
    field_404: Field40450 = Field(..., alias="404")
    field_409: Field4095 = Field(..., alias="409")
    field_422: Field42221 = Field(..., alias="422")
    field_429: Field42927 = Field(..., alias="429")
    field_200: Field20042 = Field(..., alias="200")


class Post10(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter47]
    responses: Responses50
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class Parameter48(Parameter2):
    pass


class Field40051(Field40048):
    pass


class Field40451(Field40048):
    pass


class Schema228(Schema64):
    pass


class Field20043(BaseModel):
    description: str
    schema_: Schema228 = Field(..., alias="schema")


class Responses51(BaseModel):
    field_400: Field40051 = Field(..., alias="400")
    field_404: Field40451 = Field(..., alias="404")
    field_200: Field20043 = Field(..., alias="200")


class Get25(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter48]
    responses: Responses51
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentMetadataAccess(BaseModel):
    post: Post10
    get: Get25


class Schema229(Schema):
    pass


class Parameter49(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema229] = Field(None, alias="schema")


class Field40052(BaseModel):
    description: str
    schema_: Schema229 = Field(..., alias="schema")


class Field40452(Field40052):
    pass


class Field42222(Field40052):
    pass


class Field42928(Field40052):
    pass


class Field20044(Field40052):
    pass


class Responses52(BaseModel):
    field_400: Field40052 = Field(..., alias="400")
    field_404: Field40452 = Field(..., alias="404")
    field_422: Field42222 = Field(..., alias="422")
    field_429: Field42928 = Field(..., alias="429")
    field_200: Field20044 = Field(..., alias="200")


class Put3(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter49]
    responses: Responses52
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter50(Parameter2):
    pass


class Field40053(Field40052):
    pass


class Field40453(Field40052):
    pass


class Field42929(Field40052):
    pass


class Schema238(Items):
    pass


class Field20410(BaseModel):
    description: str
    schema_: Schema238 = Field(..., alias="schema")


class Responses53(BaseModel):
    field_400: Field40053 = Field(..., alias="400")
    field_404: Field40453 = Field(..., alias="404")
    field_429: Field42929 = Field(..., alias="429")
    field_204: Field20410 = Field(..., alias="204")


class Delete7(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter50]
    responses: Responses53
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentMetadataAccessContentMetadataAccessId(BaseModel):
    put: Put3
    delete: Delete7


class Parameter51(Parameter7):
    pass


class Schema239(Schema):
    pass


class Field40054(BaseModel):
    description: str
    schema_: Schema239 = Field(..., alias="schema")


class Field40454(Field40054):
    pass


class Schema241(Schema64):
    pass


class Field20045(BaseModel):
    description: str
    schema_: Schema241 = Field(..., alias="schema")


class Responses54(BaseModel):
    field_400: Field40054 = Field(..., alias="400")
    field_404: Field40454 = Field(..., alias="404")
    field_200: Field20045 = Field(..., alias="200")


class Get26(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter51]
    responses: Responses54
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentTerms(BaseModel):
    get: Get26


class Schema242(Schema):
    pass


class Field40055(BaseModel):
    description: str
    schema_: Schema242 = Field(..., alias="schema")


class Field40455(Field40055):
    pass


class Schema244(Items):
    pass


class Field20046(BaseModel):
    description: str
    schema_: Schema244 = Field(..., alias="schema")


class Schema245(Schema):
    pass


class Field42223(BaseModel):
    description: str
    schema_: Schema245 = Field(..., alias="schema")


class Responses55(BaseModel):
    field_400: Field40055 = Field(..., alias="400")
    field_404: Field40455 = Field(..., alias="404")
    field_200: Field20046 = Field(..., alias="200")
    field_422: Field42223 = Field(..., alias="422")


class Get27(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter51]
    responses: Responses55
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentThumbnailTypeResourceId(BaseModel):
    get: Get27


class Parameter53(Parameter2):
    pass


class Field40056(Field42223):
    pass


class Field40456(Field42223):
    pass


class Field42224(Field42223):
    pass


class Field42930(Field42223):
    pass


class Field20047(Field42223):
    pass


class Responses56(BaseModel):
    field_400: Field40056 = Field(..., alias="400")
    field_404: Field40456 = Field(..., alias="404")
    field_422: Field42224 = Field(..., alias="422")
    field_429: Field42930 = Field(..., alias="429")
    field_200: Field20047 = Field(..., alias="200")


class Get28(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter53]
    responses: Responses56
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentValidation(BaseModel):
    get: Get28


class Parameter54(Parameter7):
    pass


class Field40057(Field42223):
    pass


class Field40457(Field42223):
    pass


class Schema253(Schema64):
    pass


class Field20048(BaseModel):
    description: str
    schema_: Schema253 = Field(..., alias="schema")


class Responses57(BaseModel):
    field_400: Field40057 = Field(..., alias="400")
    field_404: Field40457 = Field(..., alias="404")
    field_200: Field20048 = Field(..., alias="200")


class Get29(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter54]
    responses: Responses57
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldContentViewSearch(BaseModel):
    get: Get29


class Schema254(Schema):
    pass


class Field40058(BaseModel):
    description: str
    schema_: Schema254 = Field(..., alias="schema")


class Field40458(Field40058):
    pass


class Field4038(Field40058):
    pass


class Schema257(Schema64):
    pass


class Field20049(BaseModel):
    description: str
    schema_: Schema257 = Field(..., alias="schema")


class Responses58(BaseModel):
    field_400: Field40058 = Field(..., alias="400")
    field_404: Field40458 = Field(..., alias="404")
    field_403: Field4038 = Field(..., alias="403")
    field_200: Field20049 = Field(..., alias="200")


class Get30(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter54]
    responses: Responses58
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldCredentialsEmailSearch(BaseModel):
    get: Get30


class Schema258(Schema):
    pass


class Field40059(BaseModel):
    description: str
    schema_: Schema258 = Field(..., alias="schema")


class Field40459(Field40059):
    pass


class Field20050(Field40059):
    pass


class Responses59(BaseModel):
    field_400: Field40059 = Field(..., alias="400")
    field_404: Field40459 = Field(..., alias="404")
    field_200: Field20050 = Field(..., alias="200")


class Get31(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses59
    deprecated: bool
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter56(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema258] = Field(None, alias="schema")
    type: Optional[str] = None


class Field40060(Field40059):
    pass


class Field40460(Field40059):
    pass


class Field42225(Field40059):
    pass


class Field42931(Field40059):
    pass


class Field20051(Field40059):
    pass


class Responses60(BaseModel):
    field_400: Field40060 = Field(..., alias="400")
    field_404: Field40460 = Field(..., alias="404")
    field_422: Field42225 = Field(..., alias="422")
    field_429: Field42931 = Field(..., alias="429")
    field_200: Field20051 = Field(..., alias="200")


class Patch5(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter56]
    responses: Responses60
    deprecated: bool
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldCustomWelcomeEmail(BaseModel):
    get: Get31
    patch: Patch5


class Parameter57(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema258 = Field(..., alias="schema")


class Field40061(Field40059):
    pass


class Field40461(Field40059):
    pass


class Field42226(Field40059):
    pass


class Field42932(Field40059):
    pass


class Field20052(Field40059):
    pass


class Responses61(BaseModel):
    field_400: Field40061 = Field(..., alias="400")
    field_404: Field40461 = Field(..., alias="404")
    field_422: Field42226 = Field(..., alias="422")
    field_429: Field42932 = Field(..., alias="429")
    field_200: Field20052 = Field(..., alias="200")


class Put4(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter57]
    responses: Responses61
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldCustomWelcomeEmailTest(BaseModel):
    put: Put4


class Parameter58(Parameter2):
    pass


class Field40062(Field40059):
    pass


class Field40462(Field40059):
    pass


class Schema275(Schema64):
    pass


class Field20053(BaseModel):
    description: str
    schema_: Schema275 = Field(..., alias="schema")


class Responses62(BaseModel):
    field_400: Field40062 = Field(..., alias="400")
    field_404: Field40462 = Field(..., alias="404")
    field_200: Field20053 = Field(..., alias="200")


class Get32(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter58]
    responses: Responses62
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema276(Schema):
    pass


class Parameter59(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema276 = Field(..., alias="schema")


class Field40063(BaseModel):
    description: str
    schema_: Schema276 = Field(..., alias="schema")


class Field40463(Field40063):
    pass


class Field4096(Field40063):
    pass


class Field42227(Field40063):
    pass


class Field42933(Field40063):
    pass


class Field20054(Field40063):
    pass


class Responses63(BaseModel):
    field_400: Field40063 = Field(..., alias="400")
    field_404: Field40463 = Field(..., alias="404")
    field_409: Field4096 = Field(..., alias="409")
    field_422: Field42227 = Field(..., alias="422")
    field_429: Field42933 = Field(..., alias="429")
    field_200: Field20054 = Field(..., alias="200")


class Post11(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter59]
    responses: Responses63
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboards(BaseModel):
    get: Get32
    post: Post11


class Parameter60(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str
    format: Optional[str] = None
    x_looker_deprecated: Optional[bool] = Field(None, alias="x-looker-deprecated")


class Field40064(Field40063):
    pass


class Field40464(Field40063):
    pass


class Schema285(Schema64):
    pass


class Field20055(BaseModel):
    description: str
    schema_: Schema285 = Field(..., alias="schema")


class Responses64(BaseModel):
    field_400: Field40064 = Field(..., alias="400")
    field_404: Field40464 = Field(..., alias="404")
    field_200: Field20055 = Field(..., alias="200")


class Get33(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter60]
    responses: Responses64
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsSearch(BaseModel):
    get: Get33


class Schema286(Schema):
    pass


class Parameter61(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema286] = Field(None, alias="schema")


class Field40065(BaseModel):
    description: str
    schema_: Schema286 = Field(..., alias="schema")


class Field40465(Field40065):
    pass


class Field4097(Field40065):
    pass


class Field42228(Field40065):
    pass


class Field42934(Field40065):
    pass


class Field201(Field40065):
    pass


class Field20056(Field40065):
    pass


class Responses65(BaseModel):
    field_400: Field40065 = Field(..., alias="400")
    field_404: Field40465 = Field(..., alias="404")
    field_409: Field4097 = Field(..., alias="409")
    field_422: Field42228 = Field(..., alias="422")
    field_429: Field42934 = Field(..., alias="429")
    field_201: Field201 = Field(..., alias="201")
    field_200: Field20056 = Field(..., alias="200")


class Post12(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter61]
    responses: Responses65
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsLookmlDashboardIdImportSpaceId(BaseModel):
    post: Post12


class Parameter62(Parameter61):
    pass


class Field40066(Field40065):
    pass


class Field40466(Field40065):
    pass


class Field42229(Field40065):
    pass


class Field42935(Field40065):
    pass


class Items19(AdditionalProperties):
    pass


class Schema299(BaseModel):
    type: str
    items: Items19


class Field20057(BaseModel):
    description: str
    schema_: Schema299 = Field(..., alias="schema")


class Responses66(BaseModel):
    field_400: Field40066 = Field(..., alias="400")
    field_404: Field40466 = Field(..., alias="404")
    field_422: Field42229 = Field(..., alias="422")
    field_429: Field42935 = Field(..., alias="429")
    field_200: Field20057 = Field(..., alias="200")


class Patch6(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter62]
    responses: Responses66
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsLookmlDashboardIdSync(BaseModel):
    patch: Patch6


class Parameter63(Parameter2):
    pass


class Schema300(Schema):
    pass


class Field40067(BaseModel):
    description: str
    schema_: Schema300 = Field(..., alias="schema")


class Field40467(Field40067):
    pass


class Field42936(Field40067):
    pass


class Field4051(Field40067):
    pass


class Schema304(Items):
    pass


class Field20411(BaseModel):
    description: str
    schema_: Schema304 = Field(..., alias="schema")


class Responses67(BaseModel):
    field_400: Field40067 = Field(..., alias="400")
    field_404: Field40467 = Field(..., alias="404")
    field_429: Field42936 = Field(..., alias="429")
    field_405: Field4051 = Field(..., alias="405")
    field_204: Field20411 = Field(..., alias="204")


class Delete8(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter63]
    responses: Responses67
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema305(Schema):
    pass


class Parameter64(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema305] = Field(None, alias="schema")


class Field40068(BaseModel):
    description: str
    schema_: Schema305 = Field(..., alias="schema")


class Field40468(Field40068):
    pass


class Field42230(Field40068):
    pass


class Field42937(Field40068):
    pass


class Field4052(Field40068):
    pass


class Field20058(Field40068):
    pass


class Responses68(BaseModel):
    field_400: Field40068 = Field(..., alias="400")
    field_404: Field40468 = Field(..., alias="404")
    field_422: Field42230 = Field(..., alias="422")
    field_429: Field42937 = Field(..., alias="429")
    field_405: Field4052 = Field(..., alias="405")
    field_200: Field20058 = Field(..., alias="200")


class Patch7(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter64]
    responses: Responses68
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter65(Parameter2):
    pass


class Field40069(Field40068):
    pass


class Field40469(Field40068):
    pass


class Field20059(Field40068):
    pass


class Responses69(BaseModel):
    field_400: Field40069 = Field(..., alias="400")
    field_404: Field40469 = Field(..., alias="404")
    field_200: Field20059 = Field(..., alias="200")


class Get34(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter65]
    responses: Responses69
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsDashboardId(BaseModel):
    delete: Delete8
    patch: Patch7
    get: Get34


class Field40070(Field40068):
    pass


class Field40470(Field40068):
    pass


class Field20060(Field40068):
    pass


class Responses70(BaseModel):
    field_400: Field40070 = Field(..., alias="400")
    field_404: Field40470 = Field(..., alias="404")
    field_200: Field20060 = Field(..., alias="200")


class Get35(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter65]
    responses: Responses70
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsAggregateTableLookmlDashboardId(BaseModel):
    get: Get35


class Field40071(Field40068):
    pass


class Field40471(Field40068):
    pass


class Field20061(Field40068):
    pass


class Responses71(BaseModel):
    field_400: Field40071 = Field(..., alias="400")
    field_404: Field40471 = Field(..., alias="404")
    field_200: Field20061 = Field(..., alias="200")


class Get36(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter65]
    responses: Responses71
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsLookmlDashboardId(BaseModel):
    get: Get36


class Field40072(Field40068):
    pass


class Field40472(Field40068):
    pass


class Field42231(Field40068):
    pass


class Field42938(Field40068):
    pass


class Field20062(Field40068):
    pass


class Field2011(Field40068):
    pass


class Responses72(BaseModel):
    field_400: Field40072 = Field(..., alias="400")
    field_404: Field40472 = Field(..., alias="404")
    field_422: Field42231 = Field(..., alias="422")
    field_429: Field42938 = Field(..., alias="429")
    field_200: Field20062 = Field(..., alias="200")
    field_201: Field2011 = Field(..., alias="201")


class Patch8(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter65]
    responses: Responses72
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsDashboardIdMove(BaseModel):
    patch: Patch8


class Parameter69(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema305 = Field(..., alias="schema")


class Field40073(Field40068):
    pass


class Field40473(Field40068):
    pass


class Field4098(Field40068):
    pass


class Field42232(Field40068):
    pass


class Field42939(Field40068):
    pass


class Field20063(Field40068):
    pass


class Responses73(BaseModel):
    field_400: Field40073 = Field(..., alias="400")
    field_404: Field40473 = Field(..., alias="404")
    field_409: Field4098 = Field(..., alias="409")
    field_422: Field42232 = Field(..., alias="422")
    field_429: Field42939 = Field(..., alias="429")
    field_200: Field20063 = Field(..., alias="200")


class Post13(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter69]
    responses: Responses73
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsLookml(BaseModel):
    post: Post13


class Parameter70(Parameter69):
    pass


class Field40074(Field40068):
    pass


class Field40474(Field40068):
    pass


class Field4099(Field40068):
    pass


class Field42233(Field40068):
    pass


class Field42940(Field40068):
    pass


class Field20064(Field40068):
    pass


class Responses74(BaseModel):
    field_400: Field40074 = Field(..., alias="400")
    field_404: Field40474 = Field(..., alias="404")
    field_409: Field4099 = Field(..., alias="409")
    field_422: Field42233 = Field(..., alias="422")
    field_429: Field42940 = Field(..., alias="429")
    field_200: Field20064 = Field(..., alias="200")


class Post14(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter70]
    responses: Responses74
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsFromLookml(BaseModel):
    post: Post14


class Parameter71(Parameter2):
    pass


class Field40075(Field40068):
    pass


class Field40475(Field40068):
    pass


class Field40910(Field40068):
    pass


class Field42234(Field40068):
    pass


class Field42941(Field40068):
    pass


class Field20065(Field40068):
    pass


class Field2012(Field40068):
    pass


class Responses75(BaseModel):
    field_400: Field40075 = Field(..., alias="400")
    field_404: Field40475 = Field(..., alias="404")
    field_409: Field40910 = Field(..., alias="409")
    field_422: Field42234 = Field(..., alias="422")
    field_429: Field42941 = Field(..., alias="429")
    field_200: Field20065 = Field(..., alias="200")
    field_201: Field2012 = Field(..., alias="201")


class Post15(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter71]
    responses: Responses75
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsDashboardIdCopy(BaseModel):
    post: Post15


class Field40076(Field40068):
    pass


class Field40476(Field40068):
    pass


class Items20(Schema):
    pass


class Schema350(BaseModel):
    type: str
    items: Items20


class Field20066(BaseModel):
    description: str
    schema_: Schema350 = Field(..., alias="schema")


class Responses76(BaseModel):
    field_400: Field40076 = Field(..., alias="400")
    field_404: Field40476 = Field(..., alias="404")
    field_200: Field20066 = Field(..., alias="200")


class Get37(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter71]
    responses: Responses76
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardElementsSearch(BaseModel):
    get: Get37


class Schema351(Schema):
    pass


class Field40077(BaseModel):
    description: str
    schema_: Schema351 = Field(..., alias="schema")


class Field40477(Field40077):
    pass


class Field20067(Field40077):
    pass


class Responses77(BaseModel):
    field_400: Field40077 = Field(..., alias="400")
    field_404: Field40477 = Field(..., alias="404")
    field_200: Field20067 = Field(..., alias="200")


class Get38(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter71]
    responses: Responses77
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field40078(Field40077):
    pass


class Field40478(Field40077):
    pass


class Field42942(Field40077):
    pass


class Schema357(Items):
    pass


class Field20412(BaseModel):
    description: str
    schema_: Schema357 = Field(..., alias="schema")


class Responses78(BaseModel):
    field_400: Field40078 = Field(..., alias="400")
    field_404: Field40478 = Field(..., alias="404")
    field_429: Field42942 = Field(..., alias="429")
    field_204: Field20412 = Field(..., alias="204")


class Delete9(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter71]
    responses: Responses78
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema358(Schema):
    pass


class Parameter75(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema358] = Field(None, alias="schema")


class Field40079(BaseModel):
    description: str
    schema_: Schema358 = Field(..., alias="schema")


class Field40479(Field40079):
    pass


class Field42235(Field40079):
    pass


class Field42943(Field40079):
    pass


class Field20068(Field40079):
    pass


class Responses79(BaseModel):
    field_400: Field40079 = Field(..., alias="400")
    field_404: Field40479 = Field(..., alias="404")
    field_422: Field42235 = Field(..., alias="422")
    field_429: Field42943 = Field(..., alias="429")
    field_200: Field20068 = Field(..., alias="200")


class Patch9(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter75]
    responses: Responses79
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardElementsDashboardElementId(BaseModel):
    get: Get38
    delete: Delete9
    patch: Patch9


class Parameter76(Parameter2):
    pass


class Field40080(Field40079):
    pass


class Field40480(Field40079):
    pass


class Schema366(Schema350):
    pass


class Field20069(BaseModel):
    description: str
    schema_: Schema366 = Field(..., alias="schema")


class Responses80(BaseModel):
    field_400: Field40080 = Field(..., alias="400")
    field_404: Field40480 = Field(..., alias="404")
    field_200: Field20069 = Field(..., alias="200")


class Get39(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter76]
    responses: Responses80
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsDashboardIdDashboardElements(BaseModel):
    get: Get39


class Schema367(Schema):
    pass


class Parameter77(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema367] = Field(None, alias="schema")
    type: Optional[str] = None


class Field40081(BaseModel):
    description: str
    schema_: Schema367 = Field(..., alias="schema")


class Field40481(Field40081):
    pass


class Field40911(Field40081):
    pass


class Field42236(Field40081):
    pass


class Field42944(Field40081):
    pass


class Field20070(Field40081):
    pass


class Responses81(BaseModel):
    field_400: Field40081 = Field(..., alias="400")
    field_404: Field40481 = Field(..., alias="404")
    field_409: Field40911 = Field(..., alias="409")
    field_422: Field42236 = Field(..., alias="422")
    field_429: Field42944 = Field(..., alias="429")
    field_200: Field20070 = Field(..., alias="200")


class Post16(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter77]
    responses: Responses81
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardElements(BaseModel):
    post: Post16


class Parameter78(Parameter2):
    pass


class Field40082(Field40081):
    pass


class Field40482(Field40081):
    pass


class Field20071(Field40081):
    pass


class Responses82(BaseModel):
    field_400: Field40082 = Field(..., alias="400")
    field_404: Field40482 = Field(..., alias="404")
    field_200: Field20071 = Field(..., alias="200")


class Get40(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter78]
    responses: Responses82
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field40083(Field40081):
    pass


class Field40483(Field40081):
    pass


class Field42945(Field40081):
    pass


class Schema380(Items):
    pass


class Field20413(BaseModel):
    description: str
    schema_: Schema380 = Field(..., alias="schema")


class Responses83(BaseModel):
    field_400: Field40083 = Field(..., alias="400")
    field_404: Field40483 = Field(..., alias="404")
    field_429: Field42945 = Field(..., alias="429")
    field_204: Field20413 = Field(..., alias="204")


class Delete10(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter78]
    responses: Responses83
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema381(Schema):
    pass


class Parameter80(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema381] = Field(None, alias="schema")


class Field40084(BaseModel):
    description: str
    schema_: Schema381 = Field(..., alias="schema")


class Field40484(Field40084):
    pass


class Field42237(Field40084):
    pass


class Field42946(Field40084):
    pass


class Field20072(Field40084):
    pass


class Responses84(BaseModel):
    field_400: Field40084 = Field(..., alias="400")
    field_404: Field40484 = Field(..., alias="404")
    field_422: Field42237 = Field(..., alias="422")
    field_429: Field42946 = Field(..., alias="429")
    field_200: Field20072 = Field(..., alias="200")


class Patch10(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter80]
    responses: Responses84
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardFiltersDashboardFilterId(BaseModel):
    get: Get40
    delete: Delete10
    patch: Patch10


class Parameter81(Parameter2):
    pass


class Field40085(Field40084):
    pass


class Field40485(Field40084):
    pass


class Schema389(Schema350):
    pass


class Field20073(BaseModel):
    description: str
    schema_: Schema389 = Field(..., alias="schema")


class Responses85(BaseModel):
    field_400: Field40085 = Field(..., alias="400")
    field_404: Field40485 = Field(..., alias="404")
    field_200: Field20073 = Field(..., alias="200")


class Get41(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter81]
    responses: Responses85
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsDashboardIdDashboardFilters(BaseModel):
    get: Get41


class Schema390(Schema):
    pass


class Parameter82(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema390] = Field(None, alias="schema")
    type: Optional[str] = None


class Field40086(BaseModel):
    description: str
    schema_: Schema390 = Field(..., alias="schema")


class Field40486(Field40086):
    pass


class Field40912(Field40086):
    pass


class Field42238(Field40086):
    pass


class Field42947(Field40086):
    pass


class Field20074(Field40086):
    pass


class Responses86(BaseModel):
    field_400: Field40086 = Field(..., alias="400")
    field_404: Field40486 = Field(..., alias="404")
    field_409: Field40912 = Field(..., alias="409")
    field_422: Field42238 = Field(..., alias="422")
    field_429: Field42947 = Field(..., alias="429")
    field_200: Field20074 = Field(..., alias="200")


class Post17(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter82]
    responses: Responses86
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardFilters(BaseModel):
    post: Post17


class Parameter83(Parameter2):
    pass


class Field40087(Field40086):
    pass


class Field40487(Field40086):
    pass


class Field20075(Field40086):
    pass


class Responses87(BaseModel):
    field_400: Field40087 = Field(..., alias="400")
    field_404: Field40487 = Field(..., alias="404")
    field_200: Field20075 = Field(..., alias="200")


class Get42(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter83]
    responses: Responses87
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter84(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema390] = Field(None, alias="schema")


class Field40088(Field40086):
    pass


class Field40488(Field40086):
    pass


class Field42239(Field40086):
    pass


class Field42948(Field40086):
    pass


class Field20076(Field40086):
    pass


class Responses88(BaseModel):
    field_400: Field40088 = Field(..., alias="400")
    field_404: Field40488 = Field(..., alias="404")
    field_422: Field42239 = Field(..., alias="422")
    field_429: Field42948 = Field(..., alias="429")
    field_200: Field20076 = Field(..., alias="200")


class Patch11(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter84]
    responses: Responses88
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardLayoutComponentsDashboardLayoutComponentId(BaseModel):
    get: Get42
    patch: Patch11


class Parameter85(Parameter2):
    pass


class Field40089(Field40086):
    pass


class Field40489(Field40086):
    pass


class Schema408(Schema350):
    pass


class Field20077(BaseModel):
    description: str
    schema_: Schema408 = Field(..., alias="schema")


class Responses89(BaseModel):
    field_400: Field40089 = Field(..., alias="400")
    field_404: Field40489 = Field(..., alias="404")
    field_200: Field20077 = Field(..., alias="200")


class Get43(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter85]
    responses: Responses89
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardLayoutsDashboardLayoutIdDashboardLayoutComponents(BaseModel):
    get: Get43


class Schema409(Schema):
    pass


class Field40090(BaseModel):
    description: str
    schema_: Schema409 = Field(..., alias="schema")


class Field40490(Field40090):
    pass


class Field20078(Field40090):
    pass


class Responses90(BaseModel):
    field_400: Field40090 = Field(..., alias="400")
    field_404: Field40490 = Field(..., alias="404")
    field_200: Field20078 = Field(..., alias="200")


class Get44(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter85]
    responses: Responses90
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field40091(Field40090):
    pass


class Field40491(Field40090):
    pass


class Field42949(Field40090):
    pass


class Schema415(Items):
    pass


class Field20414(BaseModel):
    description: str
    schema_: Schema415 = Field(..., alias="schema")


class Schema416(Schema):
    pass


class Field42240(BaseModel):
    description: str
    schema_: Schema416 = Field(..., alias="schema")


class Responses91(BaseModel):
    field_400: Field40091 = Field(..., alias="400")
    field_404: Field40491 = Field(..., alias="404")
    field_429: Field42949 = Field(..., alias="429")
    field_204: Field20414 = Field(..., alias="204")
    field_422: Field42240 = Field(..., alias="422")


class Delete11(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter85]
    responses: Responses91
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter88(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema416] = Field(None, alias="schema")


class Field40092(Field42240):
    pass


class Field40492(Field42240):
    pass


class Field42241(Field42240):
    pass


class Field42950(Field42240):
    pass


class Field20079(Field42240):
    pass


class Responses92(BaseModel):
    field_400: Field40092 = Field(..., alias="400")
    field_404: Field40492 = Field(..., alias="404")
    field_422: Field42241 = Field(..., alias="422")
    field_429: Field42950 = Field(..., alias="429")
    field_200: Field20079 = Field(..., alias="200")


class Patch12(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter88]
    responses: Responses92
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardLayoutsDashboardLayoutId(BaseModel):
    get: Get44
    delete: Delete11
    patch: Patch12


class Parameter89(Parameter2):
    pass


class Field40093(Field42240):
    pass


class Field40493(Field42240):
    pass


class Schema425(Schema350):
    pass


class Field20080(BaseModel):
    description: str
    schema_: Schema425 = Field(..., alias="schema")


class Responses93(BaseModel):
    field_400: Field40093 = Field(..., alias="400")
    field_404: Field40493 = Field(..., alias="404")
    field_200: Field20080 = Field(..., alias="200")


class Get45(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter89]
    responses: Responses93
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardsDashboardIdDashboardLayouts(BaseModel):
    get: Get45


class Schema426(Schema):
    pass


class Parameter90(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema426] = Field(None, alias="schema")
    type: Optional[str] = None


class Field40094(BaseModel):
    description: str
    schema_: Schema426 = Field(..., alias="schema")


class Field40494(Field40094):
    pass


class Field40913(Field40094):
    pass


class Field42242(Field40094):
    pass


class Field42951(Field40094):
    pass


class Field20081(Field40094):
    pass


class Responses94(BaseModel):
    field_400: Field40094 = Field(..., alias="400")
    field_404: Field40494 = Field(..., alias="404")
    field_409: Field40913 = Field(..., alias="409")
    field_422: Field42242 = Field(..., alias="422")
    field_429: Field42951 = Field(..., alias="429")
    field_200: Field20081 = Field(..., alias="200")


class Post18(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter90]
    responses: Responses94
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDashboardLayouts(BaseModel):
    post: Post18


class Parameter91(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema426 = Field(..., alias="schema")


class Field40095(Field40094):
    pass


class Field40495(Field40094):
    pass


class Field20082(Field40094):
    pass


class Responses95(BaseModel):
    field_400: Field40095 = Field(..., alias="400")
    field_404: Field40495 = Field(..., alias="404")
    field_200: Field20082 = Field(..., alias="200")


class Post19(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter91]
    responses: Responses95
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDataActions(BaseModel):
    post: Post19


class AdditionalProperties1(Items):
    pass


class Schema437(BaseModel):
    type: str
    additionalProperties: AdditionalProperties1


class Parameter92(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema437 = Field(..., alias="schema")


class Schema438(Schema):
    pass


class Field40096(BaseModel):
    description: str
    schema_: Schema438 = Field(..., alias="schema")


class Field40496(Field40096):
    pass


class Field42243(Field40096):
    pass


class Field20083(Field40096):
    pass


class Responses96(BaseModel):
    field_400: Field40096 = Field(..., alias="400")
    field_404: Field40496 = Field(..., alias="404")
    field_422: Field42243 = Field(..., alias="422")
    field_200: Field20083 = Field(..., alias="200")


class Post20(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter92]
    responses: Responses96
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDataActionsForm(BaseModel):
    post: Post20


class Field40097(Field40096):
    pass


class Field40497(Field40096):
    pass


class Schema444(Schema350):
    pass


class Field20084(BaseModel):
    description: str
    schema_: Schema444 = Field(..., alias="schema")


class Responses97(BaseModel):
    field_400: Field40097 = Field(..., alias="400")
    field_404: Field40497 = Field(..., alias="404")
    field_200: Field20084 = Field(..., alias="200")


class Get46(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses97
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDatagroups(BaseModel):
    get: Get46


class Parameter93(Parameter2):
    pass


class Schema445(Schema):
    pass


class Field40098(BaseModel):
    description: str
    schema_: Schema445 = Field(..., alias="schema")


class Field40498(Field40098):
    pass


class Field20085(Field40098):
    pass


class Responses98(BaseModel):
    field_400: Field40098 = Field(..., alias="400")
    field_404: Field40498 = Field(..., alias="404")
    field_200: Field20085 = Field(..., alias="200")


class Get47(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter93]
    responses: Responses98
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter94(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema445] = Field(None, alias="schema")


class Field40099(Field40098):
    pass


class Field40499(Field40098):
    pass


class Field42244(Field40098):
    pass


class Field42952(Field40098):
    pass


class Field40914(Field40098):
    pass


class Field20086(Field40098):
    pass


class Responses99(BaseModel):
    field_400: Field40099 = Field(..., alias="400")
    field_404: Field40499 = Field(..., alias="404")
    field_422: Field42244 = Field(..., alias="422")
    field_429: Field42952 = Field(..., alias="429")
    field_409: Field40914 = Field(..., alias="409")
    field_200: Field20086 = Field(..., alias="200")


class Patch13(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter94]
    responses: Responses99
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDatagroupsDatagroupId(BaseModel):
    get: Get47
    patch: Patch13


class Parameter95(Parameter2):
    pass


class Field400100(Field40098):
    pass


class Field404100(Field40098):
    pass


class Schema457(Schema350):
    pass


class Field20087(BaseModel):
    description: str
    schema_: Schema457 = Field(..., alias="schema")


class Responses100(BaseModel):
    field_400: Field400100 = Field(..., alias="400")
    field_404: Field404100 = Field(..., alias="404")
    field_200: Field20087 = Field(..., alias="200")


class Get48(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter95]
    responses: Responses100
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema458(Schema):
    pass


class Parameter96(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema458 = Field(..., alias="schema")


class Field400101(BaseModel):
    description: str
    schema_: Schema458 = Field(..., alias="schema")


class Field404101(Field400101):
    pass


class Field40915(Field400101):
    pass


class Field42245(Field400101):
    pass


class Field42953(Field400101):
    pass


class Field20088(Field400101):
    pass


class Responses101(BaseModel):
    field_400: Field400101 = Field(..., alias="400")
    field_404: Field404101 = Field(..., alias="404")
    field_409: Field40915 = Field(..., alias="409")
    field_422: Field42245 = Field(..., alias="422")
    field_429: Field42953 = Field(..., alias="429")
    field_200: Field20088 = Field(..., alias="200")


class Post21(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter96]
    responses: Responses101
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldConnections(BaseModel):
    get: Get48
    post: Post21


class Parameter97(Parameter2):
    pass


class Field400102(Field400101):
    pass


class Field404102(Field400101):
    pass


class Field20089(Field400101):
    pass


class Responses102(BaseModel):
    field_400: Field400102 = Field(..., alias="400")
    field_404: Field404102 = Field(..., alias="404")
    field_200: Field20089 = Field(..., alias="200")


class Get49(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter97]
    responses: Responses102
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter98(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema458] = Field(None, alias="schema")


class Field400103(Field400101):
    pass


class Field404103(Field400101):
    pass


class Field42246(Field400101):
    pass


class Field42954(Field400101):
    pass


class Field20090(Field400101):
    pass


class Responses103(BaseModel):
    field_400: Field400103 = Field(..., alias="400")
    field_404: Field404103 = Field(..., alias="404")
    field_422: Field42246 = Field(..., alias="422")
    field_429: Field42954 = Field(..., alias="429")
    field_200: Field20090 = Field(..., alias="200")


class Patch14(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter98]
    responses: Responses103
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter99(Parameter2):
    pass


class Field400104(Field400101):
    pass


class Field404104(Field400101):
    pass


class Field42955(Field400101):
    pass


class Schema477(Items):
    pass


class Field20415(BaseModel):
    description: str
    schema_: Schema477 = Field(..., alias="schema")


class Responses104(BaseModel):
    field_400: Field400104 = Field(..., alias="400")
    field_404: Field404104 = Field(..., alias="404")
    field_429: Field42955 = Field(..., alias="429")
    field_204: Field20415 = Field(..., alias="204")


class Delete12(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter99]
    responses: Responses104
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldConnectionsConnectionName(BaseModel):
    get: Get49
    patch: Patch14
    delete: Delete12


class Schema478(Schema):
    pass


class Field400105(BaseModel):
    description: str
    schema_: Schema478 = Field(..., alias="schema")


class Field404105(Field400105):
    pass


class Field42956(Field400105):
    pass


class Field42247(Field400105):
    pass


class Schema482(Items):
    pass


class Field20416(BaseModel):
    description: str
    schema_: Schema482 = Field(..., alias="schema")


class Responses105(BaseModel):
    field_400: Field400105 = Field(..., alias="400")
    field_404: Field404105 = Field(..., alias="404")
    field_429: Field42956 = Field(..., alias="429")
    field_422: Field42247 = Field(..., alias="422")
    field_204: Field20416 = Field(..., alias="204")


class Delete13(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter99]
    responses: Responses105
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldConnectionsConnectionNameConnectionOverrideOverrideContext(BaseModel):
    delete: Delete13


class Items27(Items):
    pass


class Parameter101(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str
    items: Optional[Items27] = None
    collectionFormat: Optional[str] = None


class Schema483(Schema):
    pass


class Field42957(BaseModel):
    description: str
    schema_: Schema483 = Field(..., alias="schema")


class Field400106(Field42957):
    pass


class Field404106(Field42957):
    pass


class Field42248(Field42957):
    pass


class Items28(Schema):
    pass


class Schema487(BaseModel):
    type: str
    items: Items28


class Field20091(BaseModel):
    description: str
    schema_: Schema487 = Field(..., alias="schema")


class Responses106(BaseModel):
    field_429: Field42957 = Field(..., alias="429")
    field_400: Field400106 = Field(..., alias="400")
    field_404: Field404106 = Field(..., alias="404")
    field_422: Field42248 = Field(..., alias="422")
    field_200: Field20091 = Field(..., alias="200")


class Put5(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter101]
    responses: Responses106
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldConnectionsConnectionNameTest(BaseModel):
    put: Put5


class Schema488(Schema):
    pass


class Items29(Items):
    pass


class Parameter102(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema488] = Field(None, alias="schema")
    type: Optional[str] = None
    items: Optional[Items29] = None
    collectionFormat: Optional[str] = None


class Field400107(BaseModel):
    description: str
    schema_: Schema488 = Field(..., alias="schema")


class Field404107(Field400107):
    pass


class Field42958(Field400107):
    pass


class Items30(Schema):
    pass


class Schema492(BaseModel):
    type: str
    items: Items30


class Field20092(BaseModel):
    description: str
    schema_: Schema492 = Field(..., alias="schema")


class Responses107(BaseModel):
    field_400: Field400107 = Field(..., alias="400")
    field_404: Field404107 = Field(..., alias="404")
    field_429: Field42958 = Field(..., alias="429")
    field_200: Field20092 = Field(..., alias="200")


class Put6(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter102]
    responses: Responses107
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldConnectionsTest(BaseModel):
    put: Put6


class Parameter103(Parameter2):
    pass


class Schema493(Schema):
    pass


class Field400108(BaseModel):
    description: str
    schema_: Schema493 = Field(..., alias="schema")


class Field404108(Field400108):
    pass


class Field42249(Field400108):
    pass


class Field42959(Field400108):
    pass


class Schema497(Items):
    pass


class Field20093(BaseModel):
    description: str
    schema_: Schema497 = Field(..., alias="schema")


class Field20417(Field2042):
    pass


class Responses108(BaseModel):
    field_400: Field400108 = Field(..., alias="400")
    field_404: Field404108 = Field(..., alias="404")
    field_422: Field42249 = Field(..., alias="422")
    field_429: Field42959 = Field(..., alias="429")
    field_200: Field20093 = Field(..., alias="200")
    field_204: Field20417 = Field(..., alias="204")


class Post22(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter103]
    responses: Responses108
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdManifestLockAll(BaseModel):
    post: Post22


class Schema498(Schema):
    pass


class Field400109(BaseModel):
    description: str
    schema_: Schema498 = Field(..., alias="schema")


class Field404109(Field400109):
    pass


class Field20094(Field400109):
    pass


class Responses109(BaseModel):
    field_400: Field400109 = Field(..., alias="400")
    field_404: Field404109 = Field(..., alias="404")
    field_200: Field20094 = Field(..., alias="200")


class Get50(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter103]
    responses: Responses109
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDerivedTableGraphModelModel(BaseModel):
    get: Get50


class Field400110(Field400109):
    pass


class Field404110(Field400109):
    pass


class Field20095(Field400109):
    pass


class Responses110(BaseModel):
    field_400: Field400110 = Field(..., alias="400")
    field_404: Field404110 = Field(..., alias="404")
    field_200: Field20095 = Field(..., alias="200")


class Get51(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter103]
    responses: Responses110
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDerivedTableGraphViewView(BaseModel):
    get: Get51


class Field400111(Field400109):
    pass


class Field404111(Field400109):
    pass


class Field20096(Field400109):
    pass


class Responses111(BaseModel):
    field_400: Field400111 = Field(..., alias="400")
    field_404: Field404111 = Field(..., alias="404")
    field_200: Field20096 = Field(..., alias="200")


class Get52(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter103]
    responses: Responses111
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDerivedTableModelNameViewNameStart(BaseModel):
    get: Get52


class Field400112(Field400109):
    pass


class Field404112(Field400109):
    pass


class Field20097(Field400109):
    pass


class Responses112(BaseModel):
    field_400: Field400112 = Field(..., alias="400")
    field_404: Field404112 = Field(..., alias="404")
    field_200: Field20097 = Field(..., alias="200")


class Get53(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter103]
    responses: Responses112
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDerivedTableMaterializationIdStatus(BaseModel):
    get: Get53


class Field400113(Field400109):
    pass


class Field404113(Field400109):
    pass


class Field20098(Field400109):
    pass


class Responses113(BaseModel):
    field_400: Field400113 = Field(..., alias="400")
    field_404: Field404113 = Field(..., alias="404")
    field_200: Field20098 = Field(..., alias="200")


class Get54(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter103]
    responses: Responses113
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDerivedTableMaterializationIdStop(BaseModel):
    get: Get54


class Field400114(Field400109):
    pass


class Field404114(Field400109):
    pass


class Schema515(Schema492):
    pass


class Field20099(BaseModel):
    description: str
    schema_: Schema515 = Field(..., alias="schema")


class Responses114(BaseModel):
    field_400: Field400114 = Field(..., alias="400")
    field_404: Field404114 = Field(..., alias="404")
    field_200: Field20099 = Field(..., alias="200")


class Get55(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter103]
    responses: Responses114
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDialectInfo(BaseModel):
    get: Get55


class Schema516(Schema):
    pass


class Field400115(BaseModel):
    description: str
    schema_: Schema516 = Field(..., alias="schema")


class Field404115(Field400115):
    pass


class Field200100(Field400115):
    pass


class Responses115(BaseModel):
    field_400: Field400115 = Field(..., alias="400")
    field_404: Field404115 = Field(..., alias="404")
    field_200: Field200100 = Field(..., alias="200")


class Get56(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses115
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter110(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema516 = Field(..., alias="schema")


class Field400116(Field400115):
    pass


class Field404116(Field400115):
    pass


class Field42250(Field400115):
    pass


class Field42960(Field400115):
    pass


class Field200101(Field400115):
    pass


class Responses116(BaseModel):
    field_400: Field400116 = Field(..., alias="400")
    field_404: Field404116 = Field(..., alias="404")
    field_422: Field42250 = Field(..., alias="422")
    field_429: Field42960 = Field(..., alias="429")
    field_200: Field200101 = Field(..., alias="200")


class Patch15(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter110]
    responses: Responses116
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldDigestEmailsEnabled(BaseModel):
    get: Get56
    patch: Patch15


class Field400117(Field400115):
    pass


class Field404117(Field400115):
    pass


class Field42961(Field400115):
    pass


class Field200102(Field400115):
    pass


class Responses117(BaseModel):
    field_400: Field400117 = Field(..., alias="400")
    field_404: Field404117 = Field(..., alias="404")
    field_429: Field42961 = Field(..., alias="429")
    field_200: Field200102 = Field(..., alias="200")


class Post23(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses117
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldDigestEmailSend(BaseModel):
    post: Post23


class Field400118(Field400115):
    pass


class Field4039(Field400115):
    pass


class Field404118(Field400115):
    pass


class Field200103(Field400115):
    pass


class Responses118(BaseModel):
    field_400: Field400118 = Field(..., alias="400")
    field_403: Field4039 = Field(..., alias="403")
    field_404: Field404118 = Field(..., alias="404")
    field_200: Field200103 = Field(..., alias="200")


class Get57(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses118
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldPublicEgressIpAddresses(BaseModel):
    get: Get57


class Parameter111(Parameter110):
    pass


class Field400119(Field400115):
    pass


class Field404119(Field400115):
    pass


class Field40916(Field400115):
    pass


class Field42251(Field400115):
    pass


class Field42962(Field400115):
    pass


class Field200104(Field400115):
    pass


class Responses119(BaseModel):
    field_400: Field400119 = Field(..., alias="400")
    field_404: Field404119 = Field(..., alias="404")
    field_409: Field40916 = Field(..., alias="409")
    field_422: Field42251 = Field(..., alias="422")
    field_429: Field42962 = Field(..., alias="429")
    field_200: Field200104 = Field(..., alias="200")


class Post24(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter111]
    responses: Responses119
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldEmbedConfigSecrets(BaseModel):
    post: Post24


class Parameter112(Parameter2):
    pass


class Field400120(Field400115):
    pass


class Field404120(Field400115):
    pass


class Field42963(Field400115):
    pass


class Schema543(Items):
    pass


class Field20418(BaseModel):
    description: str
    schema_: Schema543 = Field(..., alias="schema")


class Responses120(BaseModel):
    field_400: Field400120 = Field(..., alias="400")
    field_404: Field404120 = Field(..., alias="404")
    field_429: Field42963 = Field(..., alias="429")
    field_204: Field20418 = Field(..., alias="204")


class Delete14(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter112]
    responses: Responses120
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldEmbedConfigSecretsEmbedSecretId(BaseModel):
    delete: Delete14


class Schema544(Schema):
    pass


class Parameter113(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema544 = Field(..., alias="schema")


class Field400121(BaseModel):
    description: str
    schema_: Schema544 = Field(..., alias="schema")


class Field404121(Field400121):
    pass


class Field40917(Field400121):
    pass


class Field42252(Field400121):
    pass


class Field42964(Field400121):
    pass


class Field200105(Field400121):
    pass


class Responses121(BaseModel):
    field_400: Field400121 = Field(..., alias="400")
    field_404: Field404121 = Field(..., alias="404")
    field_409: Field40917 = Field(..., alias="409")
    field_422: Field42252 = Field(..., alias="422")
    field_429: Field42964 = Field(..., alias="429")
    field_200: Field200105 = Field(..., alias="200")


class Post25(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter113]
    responses: Responses121
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldEmbedSsoUrl(BaseModel):
    post: Post25


class Parameter114(Parameter113):
    pass


class Field400122(Field400121):
    pass


class Field404122(Field400121):
    pass


class Field40918(Field400121):
    pass


class Field42253(Field400121):
    pass


class Field42965(Field400121):
    pass


class Field200106(Field400121):
    pass


class Responses122(BaseModel):
    field_400: Field400122 = Field(..., alias="400")
    field_404: Field404122 = Field(..., alias="404")
    field_409: Field40918 = Field(..., alias="409")
    field_422: Field42253 = Field(..., alias="422")
    field_429: Field42965 = Field(..., alias="429")
    field_200: Field200106 = Field(..., alias="200")


class Post26(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter114]
    responses: Responses122
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldEmbedTokenUrlMe(BaseModel):
    post: Post26


class Parameter115(Parameter113):
    pass


class Field400123(Field400121):
    pass


class Field404123(Field400121):
    pass


class Field40919(Field400121):
    pass


class Field42254(Field400121):
    pass


class Field42966(Field400121):
    pass


class Field200107(Field400121):
    pass


class Responses123(BaseModel):
    field_400: Field400123 = Field(..., alias="400")
    field_404: Field404123 = Field(..., alias="404")
    field_409: Field40919 = Field(..., alias="409")
    field_422: Field42254 = Field(..., alias="422")
    field_429: Field42966 = Field(..., alias="429")
    field_200: Field200107 = Field(..., alias="200")


class Post27(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter115]
    responses: Responses123
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldEmbedCookielessSessionAcquire(BaseModel):
    post: Post27


class Parameter116(Parameter2):
    pass


class Field400124(Field400121):
    pass


class Field404124(Field400121):
    pass


class Field42967(Field400121):
    pass


class Schema568(Items):
    pass


class Field20419(BaseModel):
    description: str
    schema_: Schema568 = Field(..., alias="schema")


class Schema569(Schema):
    pass


class Field42255(BaseModel):
    description: str
    schema_: Schema569 = Field(..., alias="schema")


class Responses124(BaseModel):
    field_400: Field400124 = Field(..., alias="400")
    field_404: Field404124 = Field(..., alias="404")
    field_429: Field42967 = Field(..., alias="429")
    field_204: Field20419 = Field(..., alias="204")
    field_422: Field42255 = Field(..., alias="422")


class Delete15(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter116]
    responses: Responses124
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldEmbedCookielessSessionSessionReferenceToken(BaseModel):
    delete: Delete15


class Parameter117(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema569 = Field(..., alias="schema")


class Field400125(Field42255):
    pass


class Field404125(Field42255):
    pass


class Field42256(Field42255):
    pass


class Field42968(Field42255):
    pass


class Field200108(Field42255):
    pass


class Responses125(BaseModel):
    field_400: Field400125 = Field(..., alias="400")
    field_404: Field404125 = Field(..., alias="404")
    field_422: Field42256 = Field(..., alias="422")
    field_429: Field42968 = Field(..., alias="429")
    field_200: Field200108 = Field(..., alias="200")


class Put7(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter117]
    responses: Responses125
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldEmbedCookielessSessionGenerateTokens(BaseModel):
    put: Put7


class Parameter118(Parameter2):
    pass


class Field400126(Field42255):
    pass


class Field404126(Field42255):
    pass


class Schema578(Schema492):
    pass


class Field200109(BaseModel):
    description: str
    schema_: Schema578 = Field(..., alias="schema")


class Responses126(BaseModel):
    field_400: Field400126 = Field(..., alias="400")
    field_404: Field404126 = Field(..., alias="404")
    field_200: Field200109 = Field(..., alias="200")


class Get58(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter118]
    responses: Responses126
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema579(Schema):
    pass


class Parameter119(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema579 = Field(..., alias="schema")


class Field400127(BaseModel):
    description: str
    schema_: Schema579 = Field(..., alias="schema")


class Field404127(Field400127):
    pass


class Field40920(Field400127):
    pass


class Field42257(Field400127):
    pass


class Field42969(Field400127):
    pass


class Field200110(Field400127):
    pass


class Responses127(BaseModel):
    field_400: Field400127 = Field(..., alias="400")
    field_404: Field404127 = Field(..., alias="404")
    field_409: Field40920 = Field(..., alias="409")
    field_422: Field42257 = Field(..., alias="422")
    field_429: Field42969 = Field(..., alias="429")
    field_200: Field200110 = Field(..., alias="200")


class Post28(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter119]
    responses: Responses127
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldExternalOauthApplications(BaseModel):
    get: Get58
    post: Post28


class Parameter120(Parameter119):
    pass


class Field400128(Field400127):
    pass


class Field404128(Field400127):
    pass


class Field40921(Field400127):
    pass


class Field42258(Field400127):
    pass


class Field42970(Field400127):
    pass


class Field200111(Field400127):
    pass


class Responses128(BaseModel):
    field_400: Field400128 = Field(..., alias="400")
    field_404: Field404128 = Field(..., alias="404")
    field_409: Field40921 = Field(..., alias="409")
    field_422: Field42258 = Field(..., alias="422")
    field_429: Field42970 = Field(..., alias="429")
    field_200: Field200111 = Field(..., alias="200")


class Post29(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter120]
    responses: Responses128
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldExternalOauthApplicationsUserState(BaseModel):
    post: Post29


class Parameter121(Parameter2):
    pass


class Field400129(Field400127):
    pass


class Field404129(Field400127):
    pass


class Schema595(Schema492):
    pass


class Field200112(BaseModel):
    description: str
    schema_: Schema595 = Field(..., alias="schema")


class Responses129(BaseModel):
    field_400: Field400129 = Field(..., alias="400")
    field_404: Field404129 = Field(..., alias="404")
    field_200: Field200112 = Field(..., alias="200")


class Get59(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter121]
    responses: Responses129
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdGitBranches(BaseModel):
    get: Get59


class Schema596(Schema):
    pass


class Field400130(BaseModel):
    description: str
    schema_: Schema596 = Field(..., alias="schema")


class Field404130(Field400130):
    pass


class Field200113(Field400130):
    pass


class Responses130(BaseModel):
    field_400: Field400130 = Field(..., alias="400")
    field_404: Field404130 = Field(..., alias="404")
    field_200: Field200113 = Field(..., alias="200")


class Get60(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter121]
    responses: Responses130
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter123(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema596] = Field(None, alias="schema")


class Field400131(Field400130):
    pass


class Field404131(Field400130):
    pass


class Field40922(Field400130):
    pass


class Field42259(Field400130):
    pass


class Field42971(Field400130):
    pass


class Field200114(Field400130):
    pass


class Responses131(BaseModel):
    field_400: Field400131 = Field(..., alias="400")
    field_404: Field404131 = Field(..., alias="404")
    field_409: Field40922 = Field(..., alias="409")
    field_422: Field42259 = Field(..., alias="422")
    field_429: Field42971 = Field(..., alias="429")
    field_200: Field200114 = Field(..., alias="200")


class Post30(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter123]
    responses: Responses131
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter124(Parameter123):
    pass


class Field400132(Field400130):
    pass


class Field404132(Field400130):
    pass


class Field42260(Field400130):
    pass


class Field42972(Field400130):
    pass


class Field200115(Field400130):
    pass


class Responses132(BaseModel):
    field_400: Field400132 = Field(..., alias="400")
    field_404: Field404132 = Field(..., alias="404")
    field_422: Field42260 = Field(..., alias="422")
    field_429: Field42972 = Field(..., alias="429")
    field_200: Field200115 = Field(..., alias="200")


class Put8(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter124]
    responses: Responses132
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdGitBranch(BaseModel):
    get: Get60
    post: Post30
    put: Put8


class Parameter125(Parameter2):
    pass


class Field400133(Field400130):
    pass


class Field404133(Field400130):
    pass


class Field200116(Field400130):
    pass


class Responses133(BaseModel):
    field_400: Field400133 = Field(..., alias="400")
    field_404: Field404133 = Field(..., alias="404")
    field_200: Field200116 = Field(..., alias="200")


class Get61(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter125]
    responses: Responses133
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400134(Field400130):
    pass


class Field404134(Field400130):
    pass


class Field42973(Field400130):
    pass


class Schema618(Items):
    pass


class Field20420(BaseModel):
    description: str
    schema_: Schema618 = Field(..., alias="schema")


class Responses134(BaseModel):
    field_400: Field400134 = Field(..., alias="400")
    field_404: Field404134 = Field(..., alias="404")
    field_429: Field42973 = Field(..., alias="429")
    field_204: Field20420 = Field(..., alias="204")


class Delete16(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter125]
    responses: Responses134
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdGitBranchBranchName(BaseModel):
    get: Get61
    delete: Delete16


class Items34(Items):
    pass


class Parameter127(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str
    format: Optional[str] = None
    x_looker_deprecated: Optional[bool] = Field(None, alias="x-looker-deprecated")
    items: Optional[Items34] = None
    collectionFormat: Optional[str] = None


class Schema619(Schema):
    pass


class Field400135(BaseModel):
    description: str
    schema_: Schema619 = Field(..., alias="schema")


class Field404135(Field400135):
    pass


class Items35(Schema):
    pass


class Schema621(BaseModel):
    type: str
    items: Items35


class Field200117(BaseModel):
    description: str
    schema_: Schema621 = Field(..., alias="schema")


class Responses135(BaseModel):
    field_400: Field400135 = Field(..., alias="400")
    field_404: Field404135 = Field(..., alias="404")
    field_200: Field200117 = Field(..., alias="200")


class Get62(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter127]
    responses: Responses135
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema622(Schema):
    pass


class Parameter128(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema622] = Field(None, alias="schema")
    type: Optional[str] = None


class Field400136(BaseModel):
    description: str
    schema_: Schema622 = Field(..., alias="schema")


class Field404136(Field400136):
    pass


class Field40923(Field400136):
    pass


class Field42261(Field400136):
    pass


class Field42974(Field400136):
    pass


class Field200118(Field400136):
    pass


class Responses136(BaseModel):
    field_400: Field400136 = Field(..., alias="400")
    field_404: Field404136 = Field(..., alias="404")
    field_409: Field40923 = Field(..., alias="409")
    field_422: Field42261 = Field(..., alias="422")
    field_429: Field42974 = Field(..., alias="429")
    field_200: Field200118 = Field(..., alias="200")


class Post31(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter128]
    responses: Responses136
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldGroups(BaseModel):
    get: Get62
    post: Post31


class Parameter129(Parameter7):
    pass


class Field400137(Field400136):
    pass


class Field404137(Field400136):
    pass


class Schema631(Schema621):
    pass


class Field200119(BaseModel):
    description: str
    schema_: Schema631 = Field(..., alias="schema")


class Responses137(BaseModel):
    field_400: Field400137 = Field(..., alias="400")
    field_404: Field404137 = Field(..., alias="404")
    field_200: Field200119 = Field(..., alias="200")


class Get63(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter129]
    responses: Responses137
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldGroupsSearch(BaseModel):
    get: Get63


class Schema632(Schema):
    pass


class Field400138(BaseModel):
    description: str
    schema_: Schema632 = Field(..., alias="schema")


class Field404138(Field400138):
    pass


class Schema634(Schema621):
    pass


class Field200120(BaseModel):
    description: str
    schema_: Schema634 = Field(..., alias="schema")


class Responses138(BaseModel):
    field_400: Field400138 = Field(..., alias="400")
    field_404: Field404138 = Field(..., alias="404")
    field_200: Field200120 = Field(..., alias="200")


class Get64(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter129]
    responses: Responses138
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldGroupsSearchWithRoles(BaseModel):
    get: Get64


class Schema635(Schema):
    pass


class Field400139(BaseModel):
    description: str
    schema_: Schema635 = Field(..., alias="schema")


class Field404139(Field400139):
    pass


class Schema637(Schema621):
    pass


class Field200121(BaseModel):
    description: str
    schema_: Schema637 = Field(..., alias="schema")


class Responses139(BaseModel):
    field_400: Field400139 = Field(..., alias="400")
    field_404: Field404139 = Field(..., alias="404")
    field_200: Field200121 = Field(..., alias="200")


class Get65(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter129]
    responses: Responses139
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldGroupsSearchWithHierarchy(BaseModel):
    get: Get65


class Parameter132(Parameter2):
    pass


class Schema638(Schema):
    pass


class Field400140(BaseModel):
    description: str
    schema_: Schema638 = Field(..., alias="schema")


class Field404140(Field400140):
    pass


class Field200122(Field400140):
    pass


class Responses140(BaseModel):
    field_400: Field400140 = Field(..., alias="400")
    field_404: Field404140 = Field(..., alias="404")
    field_200: Field200122 = Field(..., alias="200")


class Get66(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter132]
    responses: Responses140
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter133(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema638] = Field(None, alias="schema")


class Field400141(Field400140):
    pass


class Field404141(Field400140):
    pass


class Field42262(Field400140):
    pass


class Field42975(Field400140):
    pass


class Field200123(Field400140):
    pass


class Responses141(BaseModel):
    field_400: Field400141 = Field(..., alias="400")
    field_404: Field404141 = Field(..., alias="404")
    field_422: Field42262 = Field(..., alias="422")
    field_429: Field42975 = Field(..., alias="429")
    field_200: Field200123 = Field(..., alias="200")


class Patch16(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter133]
    responses: Responses141
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter134(Parameter2):
    pass


class Field400142(Field400140):
    pass


class Field404142(Field400140):
    pass


class Field42976(Field400140):
    pass


class Field40310(Field400140):
    pass


class Schema651(Items):
    pass


class Field20421(BaseModel):
    description: str
    schema_: Schema651 = Field(..., alias="schema")


class Responses142(BaseModel):
    field_400: Field400142 = Field(..., alias="400")
    field_404: Field404142 = Field(..., alias="404")
    field_429: Field42976 = Field(..., alias="429")
    field_403: Field40310 = Field(..., alias="403")
    field_204: Field20421 = Field(..., alias="204")


class Delete17(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter134]
    responses: Responses142
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldGroupsGroupId(BaseModel):
    get: Get66
    patch: Patch16
    delete: Delete17


class Schema652(Schema):
    pass


class Field400143(BaseModel):
    description: str
    schema_: Schema652 = Field(..., alias="schema")


class Field404143(Field400143):
    pass


class Schema654(Schema621):
    pass


class Field200124(BaseModel):
    description: str
    schema_: Schema654 = Field(..., alias="schema")


class Responses143(BaseModel):
    field_400: Field400143 = Field(..., alias="400")
    field_404: Field404143 = Field(..., alias="404")
    field_200: Field200124 = Field(..., alias="200")


class Get67(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter134]
    responses: Responses143
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema655(Schema):
    pass


class Parameter136(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema655] = Field(None, alias="schema")


class Field400144(BaseModel):
    description: str
    schema_: Schema655 = Field(..., alias="schema")


class Field40311(Field400144):
    pass


class Field404144(Field400144):
    pass


class Field200125(Field400144):
    pass


class Responses144(BaseModel):
    field_400: Field400144 = Field(..., alias="400")
    field_403: Field40311 = Field(..., alias="403")
    field_404: Field404144 = Field(..., alias="404")
    field_200: Field200125 = Field(..., alias="200")


class Post32(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter136]
    responses: Responses144
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldGroupsGroupIdGroups(BaseModel):
    get: Get67
    post: Post32


class Parameter137(Parameter60):
    pass


class Field400145(Field400144):
    pass


class Field404145(Field400144):
    pass


class Schema662(Schema621):
    pass


class Field200126(BaseModel):
    description: str
    schema_: Schema662 = Field(..., alias="schema")


class Responses145(BaseModel):
    field_400: Field400145 = Field(..., alias="400")
    field_404: Field404145 = Field(..., alias="404")
    field_200: Field200126 = Field(..., alias="200")


class Get68(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter137]
    responses: Responses145
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema663(Schema):
    pass


class Parameter138(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema663] = Field(None, alias="schema")


class Field400146(BaseModel):
    description: str
    schema_: Schema663 = Field(..., alias="schema")


class Field40312(Field400146):
    pass


class Field404146(Field400146):
    pass


class Field200127(Field400146):
    pass


class Responses146(BaseModel):
    field_400: Field400146 = Field(..., alias="400")
    field_403: Field40312 = Field(..., alias="403")
    field_404: Field404146 = Field(..., alias="404")
    field_200: Field200127 = Field(..., alias="200")


class Post33(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter138]
    responses: Responses146
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldGroupsGroupIdUsers(BaseModel):
    get: Get68
    post: Post33


class Parameter139(Parameter2):
    pass


class Field400147(Field400146):
    pass


class Field40313(Field400146):
    pass


class Field404147(Field400146):
    pass


class Field20422(Field2042):
    pass


class Responses147(BaseModel):
    field_400: Field400147 = Field(..., alias="400")
    field_403: Field40313 = Field(..., alias="403")
    field_404: Field404147 = Field(..., alias="404")
    field_204: Field20422 = Field(..., alias="204")


class Delete18(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter139]
    responses: Responses147
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldGroupsGroupIdUsersUserId(BaseModel):
    delete: Delete18


class Field400148(Field400146):
    pass


class Field40314(Field400146):
    pass


class Field404148(Field400146):
    pass


class Responses148(BaseModel):
    field_400: Field400148 = Field(..., alias="400")
    field_403: Field40314 = Field(..., alias="403")
    field_404: Field404148 = Field(..., alias="404")
    field_204: Field20422 = Field(..., alias="204")


class Delete19(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter139]
    responses: Responses148
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldGroupsGroupIdGroupsDeletingGroupId(BaseModel):
    delete: Delete19


class Parameter141(Parameter138):
    pass


class Field400149(Field400146):
    pass


class Field404149(Field400146):
    pass


class Field42263(Field400146):
    pass


class Field200128(Field400146):
    pass


class Responses149(BaseModel):
    field_400: Field400149 = Field(..., alias="400")
    field_404: Field404149 = Field(..., alias="404")
    field_422: Field42263 = Field(..., alias="422")
    field_200: Field200128 = Field(..., alias="200")


class Patch17(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter141]
    responses: Responses149
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter142(Parameter2):
    pass


class Field400150(Field400146):
    pass


class Field404150(Field400146):
    pass


class Responses150(BaseModel):
    field_400: Field400150 = Field(..., alias="400")
    field_404: Field404150 = Field(..., alias="404")
    field_204: Field20422 = Field(..., alias="204")


class Delete20(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter142]
    responses: Responses150
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldGroupsGroupIdAttributeValuesUserAttributeId(BaseModel):
    patch: Patch17
    delete: Delete20


class Field400151(Field400146):
    pass


class Field404151(Field400146):
    pass


class Schema683(Schema621):
    pass


class Field200129(BaseModel):
    description: str
    schema_: Schema683 = Field(..., alias="schema")


class Responses151(BaseModel):
    field_400: Field400151 = Field(..., alias="400")
    field_404: Field404151 = Field(..., alias="404")
    field_200: Field200129 = Field(..., alias="200")


class Get69(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter142]
    responses: Responses151
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema684(Schema):
    pass


class Parameter144(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema684] = Field(None, alias="schema")
    type: Optional[str] = None


class Field400152(BaseModel):
    description: str
    schema_: Schema684 = Field(..., alias="schema")


class Field404152(Field400152):
    pass


class Field40924(Field400152):
    pass


class Field42264(Field400152):
    pass


class Field42977(Field400152):
    pass


class Field200130(Field400152):
    pass


class Responses152(BaseModel):
    field_400: Field400152 = Field(..., alias="400")
    field_404: Field404152 = Field(..., alias="404")
    field_409: Field40924 = Field(..., alias="409")
    field_422: Field42264 = Field(..., alias="422")
    field_429: Field42977 = Field(..., alias="429")
    field_200: Field200130 = Field(..., alias="200")


class Post34(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter144]
    responses: Responses152
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldBoards(BaseModel):
    get: Get69
    post: Post34


class Parameter145(Parameter7):
    pass


class Field400153(Field400152):
    pass


class Field404153(Field400152):
    pass


class Schema693(Schema621):
    pass


class Field200131(BaseModel):
    description: str
    schema_: Schema693 = Field(..., alias="schema")


class Responses153(BaseModel):
    field_400: Field400153 = Field(..., alias="400")
    field_404: Field404153 = Field(..., alias="404")
    field_200: Field200131 = Field(..., alias="200")


class Get70(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter145]
    responses: Responses153
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldBoardsSearch(BaseModel):
    get: Get70


class Parameter146(Parameter2):
    pass


class Schema694(Schema):
    pass


class Field400154(BaseModel):
    description: str
    schema_: Schema694 = Field(..., alias="schema")


class Field404154(Field400154):
    pass


class Field200132(Field400154):
    pass


class Responses154(BaseModel):
    field_400: Field400154 = Field(..., alias="400")
    field_404: Field404154 = Field(..., alias="404")
    field_200: Field200132 = Field(..., alias="200")


class Get71(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter146]
    responses: Responses154
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter147(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema694] = Field(None, alias="schema")


class Field400155(Field400154):
    pass


class Field404155(Field400154):
    pass


class Field42265(Field400154):
    pass


class Field42978(Field400154):
    pass


class Field200133(Field400154):
    pass


class Responses155(BaseModel):
    field_400: Field400155 = Field(..., alias="400")
    field_404: Field404155 = Field(..., alias="404")
    field_422: Field42265 = Field(..., alias="422")
    field_429: Field42978 = Field(..., alias="429")
    field_200: Field200133 = Field(..., alias="200")


class Patch18(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter147]
    responses: Responses155
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter148(Parameter2):
    pass


class Field400156(Field400154):
    pass


class Field404156(Field400154):
    pass


class Field42979(Field400154):
    pass


class Schema706(Items):
    pass


class Field20425(BaseModel):
    description: str
    schema_: Schema706 = Field(..., alias="schema")


class Responses156(BaseModel):
    field_400: Field400156 = Field(..., alias="400")
    field_404: Field404156 = Field(..., alias="404")
    field_429: Field42979 = Field(..., alias="429")
    field_204: Field20425 = Field(..., alias="204")


class Delete21(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter148]
    responses: Responses156
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldBoardsBoardId(BaseModel):
    get: Get71
    patch: Patch18
    delete: Delete21


class Schema707(Schema):
    pass


class Field400157(BaseModel):
    description: str
    schema_: Schema707 = Field(..., alias="schema")


class Field404157(Field400157):
    pass


class Schema709(Schema621):
    pass


class Field200134(BaseModel):
    description: str
    schema_: Schema709 = Field(..., alias="schema")


class Responses157(BaseModel):
    field_400: Field400157 = Field(..., alias="400")
    field_404: Field404157 = Field(..., alias="404")
    field_200: Field200134 = Field(..., alias="200")


class Get72(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter148]
    responses: Responses157
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema710(Schema):
    pass


class Parameter150(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema710] = Field(None, alias="schema")
    type: Optional[str] = None


class Field400158(BaseModel):
    description: str
    schema_: Schema710 = Field(..., alias="schema")


class Field404158(Field400158):
    pass


class Field40925(Field400158):
    pass


class Field42266(Field400158):
    pass


class Field42980(Field400158):
    pass


class Field200135(Field400158):
    pass


class Responses158(BaseModel):
    field_400: Field400158 = Field(..., alias="400")
    field_404: Field404158 = Field(..., alias="404")
    field_409: Field40925 = Field(..., alias="409")
    field_422: Field42266 = Field(..., alias="422")
    field_429: Field42980 = Field(..., alias="429")
    field_200: Field200135 = Field(..., alias="200")


class Post35(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter150]
    responses: Responses158
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldBoardItems(BaseModel):
    get: Get72
    post: Post35


class Parameter151(Parameter2):
    pass


class Field400159(Field400158):
    pass


class Field404159(Field400158):
    pass


class Field200136(Field400158):
    pass


class Responses159(BaseModel):
    field_400: Field400159 = Field(..., alias="400")
    field_404: Field404159 = Field(..., alias="404")
    field_200: Field200136 = Field(..., alias="200")


class Get73(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter151]
    responses: Responses159
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter152(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema710] = Field(None, alias="schema")


class Field400160(Field400158):
    pass


class Field404160(Field400158):
    pass


class Field42267(Field400158):
    pass


class Field42981(Field400158):
    pass


class Field200137(Field400158):
    pass


class Responses160(BaseModel):
    field_400: Field400160 = Field(..., alias="400")
    field_404: Field404160 = Field(..., alias="404")
    field_422: Field42267 = Field(..., alias="422")
    field_429: Field42981 = Field(..., alias="429")
    field_200: Field200137 = Field(..., alias="200")


class Patch19(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter152]
    responses: Responses160
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter153(Parameter2):
    pass


class Field400161(Field400158):
    pass


class Field404161(Field400158):
    pass


class Field42982(Field400158):
    pass


class Schema729(Items):
    pass


class Field20426(BaseModel):
    description: str
    schema_: Schema729 = Field(..., alias="schema")


class Responses161(BaseModel):
    field_400: Field400161 = Field(..., alias="400")
    field_404: Field404161 = Field(..., alias="404")
    field_429: Field42982 = Field(..., alias="429")
    field_204: Field20426 = Field(..., alias="204")


class Delete22(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter153]
    responses: Responses161
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldBoardItemsBoardItemId(BaseModel):
    get: Get73
    patch: Patch19
    delete: Delete22


class Schema730(Schema):
    pass


class Field400162(BaseModel):
    description: str
    schema_: Schema730 = Field(..., alias="schema")


class Field404162(Field400162):
    pass


class Schema732(Schema621):
    pass


class Field200138(BaseModel):
    description: str
    schema_: Schema732 = Field(..., alias="schema")


class Responses162(BaseModel):
    field_400: Field400162 = Field(..., alias="400")
    field_404: Field404162 = Field(..., alias="404")
    field_200: Field200138 = Field(..., alias="200")


class Get74(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter153]
    responses: Responses162
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldPrimaryHomepageSections(BaseModel):
    get: Get74


class Schema733(Schema):
    pass


class Field400163(BaseModel):
    description: str
    schema_: Schema733 = Field(..., alias="schema")


class Field404163(Field400163):
    pass


class Schema735(Schema621):
    pass


class Field200139(BaseModel):
    description: str
    schema_: Schema735 = Field(..., alias="schema")


class Responses163(BaseModel):
    field_400: Field400163 = Field(..., alias="400")
    field_404: Field404163 = Field(..., alias="404")
    field_200: Field200139 = Field(..., alias="200")


class Get75(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter153]
    responses: Responses163
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema736(Schema):
    pass


class Parameter156(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema736] = Field(None, alias="schema")
    type: Optional[str] = None


class Field400164(BaseModel):
    description: str
    schema_: Schema736 = Field(..., alias="schema")


class Field404164(Field400164):
    pass


class Field40926(Field400164):
    pass


class Field42268(Field400164):
    pass


class Field42983(Field400164):
    pass


class Field200140(Field400164):
    pass


class Responses164(BaseModel):
    field_400: Field400164 = Field(..., alias="400")
    field_404: Field404164 = Field(..., alias="404")
    field_409: Field40926 = Field(..., alias="409")
    field_422: Field42268 = Field(..., alias="422")
    field_429: Field42983 = Field(..., alias="429")
    field_200: Field200140 = Field(..., alias="200")


class Post36(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter156]
    responses: Responses164
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldBoardSections(BaseModel):
    get: Get75
    post: Post36


class Parameter157(Parameter2):
    pass


class Field400165(Field400164):
    pass


class Field404165(Field400164):
    pass


class Field200141(Field400164):
    pass


class Responses165(BaseModel):
    field_400: Field400165 = Field(..., alias="400")
    field_404: Field404165 = Field(..., alias="404")
    field_200: Field200141 = Field(..., alias="200")


class Get76(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter157]
    responses: Responses165
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter158(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema736] = Field(None, alias="schema")


class Field400166(Field400164):
    pass


class Field404166(Field400164):
    pass


class Field42269(Field400164):
    pass


class Field42984(Field400164):
    pass


class Field200142(Field400164):
    pass


class Responses166(BaseModel):
    field_400: Field400166 = Field(..., alias="400")
    field_404: Field404166 = Field(..., alias="404")
    field_422: Field42269 = Field(..., alias="422")
    field_429: Field42984 = Field(..., alias="429")
    field_200: Field200142 = Field(..., alias="200")


class Patch20(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter158]
    responses: Responses166
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter159(Parameter2):
    pass


class Field400167(Field400164):
    pass


class Field404167(Field400164):
    pass


class Field42985(Field400164):
    pass


class Schema755(Items):
    pass


class Field20427(BaseModel):
    description: str
    schema_: Schema755 = Field(..., alias="schema")


class Responses167(BaseModel):
    field_400: Field400167 = Field(..., alias="400")
    field_404: Field404167 = Field(..., alias="404")
    field_429: Field42985 = Field(..., alias="429")
    field_204: Field20427 = Field(..., alias="204")


class Delete23(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter159]
    responses: Responses167
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldBoardSectionsBoardSectionId(BaseModel):
    get: Get76
    patch: Patch20
    delete: Delete23


class Schema756(Schema):
    pass


class Field400168(BaseModel):
    description: str
    schema_: Schema756 = Field(..., alias="schema")


class Field404168(Field400168):
    pass


class Schema758(Schema621):
    pass


class Field200143(BaseModel):
    description: str
    schema_: Schema758 = Field(..., alias="schema")


class Responses168(BaseModel):
    field_400: Field400168 = Field(..., alias="400")
    field_404: Field404168 = Field(..., alias="404")
    field_200: Field200143 = Field(..., alias="200")


class Get77(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter159]
    responses: Responses168
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema759(Schema):
    pass


class Parameter161(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema759] = Field(None, alias="schema")
    type: Optional[str] = None


class Field400169(BaseModel):
    description: str
    schema_: Schema759 = Field(..., alias="schema")


class Field404169(Field400169):
    pass


class Field40927(Field400169):
    pass


class Field42270(Field400169):
    pass


class Field42986(Field400169):
    pass


class Field200144(Field400169):
    pass


class Responses169(BaseModel):
    field_400: Field400169 = Field(..., alias="400")
    field_404: Field404169 = Field(..., alias="404")
    field_409: Field40927 = Field(..., alias="409")
    field_422: Field42270 = Field(..., alias="422")
    field_429: Field42986 = Field(..., alias="429")
    field_200: Field200144 = Field(..., alias="200")


class Post37(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter161]
    responses: Responses169
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldIntegrationHubs(BaseModel):
    get: Get77
    post: Post37


class Parameter162(Parameter2):
    pass


class Field400170(Field400169):
    pass


class Field404170(Field400169):
    pass


class Field200145(Field400169):
    pass


class Responses170(BaseModel):
    field_400: Field400170 = Field(..., alias="400")
    field_404: Field404170 = Field(..., alias="404")
    field_200: Field200145 = Field(..., alias="200")


class Get78(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter162]
    responses: Responses170
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter163(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema759] = Field(None, alias="schema")


class Field400171(Field400169):
    pass


class Field404171(Field400169):
    pass


class Field42271(Field400169):
    pass


class Field42987(Field400169):
    pass


class Field200146(Field400169):
    pass


class Responses171(BaseModel):
    field_400: Field400171 = Field(..., alias="400")
    field_404: Field404171 = Field(..., alias="404")
    field_422: Field42271 = Field(..., alias="422")
    field_429: Field42987 = Field(..., alias="429")
    field_200: Field200146 = Field(..., alias="200")


class Patch21(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter163]
    responses: Responses171
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class Parameter164(Parameter2):
    pass


class Field400172(Field400169):
    pass


class Field404172(Field400169):
    pass


class Field42988(Field400169):
    pass


class Schema778(Items):
    pass


class Field20428(BaseModel):
    description: str
    schema_: Schema778 = Field(..., alias="schema")


class Responses172(BaseModel):
    field_400: Field400172 = Field(..., alias="400")
    field_404: Field404172 = Field(..., alias="404")
    field_429: Field42988 = Field(..., alias="429")
    field_204: Field20428 = Field(..., alias="204")


class Delete24(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter164]
    responses: Responses172
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldIntegrationHubsIntegrationHubId(BaseModel):
    get: Get78
    patch: Patch21
    delete: Delete24


class Schema779(Schema):
    pass


class Field400173(BaseModel):
    description: str
    schema_: Schema779 = Field(..., alias="schema")


class Field404173(Field400173):
    pass


class Field42272(Field400173):
    pass


class Field200147(Field400173):
    pass


class Responses173(BaseModel):
    field_400: Field400173 = Field(..., alias="400")
    field_404: Field404173 = Field(..., alias="404")
    field_422: Field42272 = Field(..., alias="422")
    field_200: Field200147 = Field(..., alias="200")


class Post38(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter164]
    responses: Responses173
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldIntegrationHubsIntegrationHubIdAcceptLegalAgreement(BaseModel):
    post: Post38


class Field400174(Field400173):
    pass


class Field404174(Field400173):
    pass


class Schema785(Schema621):
    pass


class Field200148(BaseModel):
    description: str
    schema_: Schema785 = Field(..., alias="schema")


class Responses174(BaseModel):
    field_400: Field400174 = Field(..., alias="400")
    field_404: Field404174 = Field(..., alias="404")
    field_200: Field200148 = Field(..., alias="200")


class Get79(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter164]
    responses: Responses174
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldIntegrations(BaseModel):
    get: Get79


class Schema786(Schema):
    pass


class Field400175(BaseModel):
    description: str
    schema_: Schema786 = Field(..., alias="schema")


class Field404175(Field400175):
    pass


class Field200149(Field400175):
    pass


class Responses175(BaseModel):
    field_400: Field400175 = Field(..., alias="400")
    field_404: Field404175 = Field(..., alias="404")
    field_200: Field200149 = Field(..., alias="200")


class Get80(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter164]
    responses: Responses175
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter168(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema786] = Field(None, alias="schema")


class Field400176(Field400175):
    pass


class Field404176(Field400175):
    pass


class Field42273(Field400175):
    pass


class Field42989(Field400175):
    pass


class Field200150(Field400175):
    pass


class Responses176(BaseModel):
    field_400: Field400176 = Field(..., alias="400")
    field_404: Field404176 = Field(..., alias="404")
    field_422: Field42273 = Field(..., alias="422")
    field_429: Field42989 = Field(..., alias="429")
    field_200: Field200150 = Field(..., alias="200")


class Patch22(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter168]
    responses: Responses176
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldIntegrationsIntegrationId(BaseModel):
    get: Get80
    patch: Patch22


class Schema795(Schema437):
    pass


class Parameter169(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema795] = Field(None, alias="schema")


class Schema796(Schema):
    pass


class Field400177(BaseModel):
    description: str
    schema_: Schema796 = Field(..., alias="schema")


class Field404177(Field400177):
    pass


class Field42274(Field400177):
    pass


class Field200151(Field400177):
    pass


class Responses177(BaseModel):
    field_400: Field400177 = Field(..., alias="400")
    field_404: Field404177 = Field(..., alias="404")
    field_422: Field42274 = Field(..., alias="422")
    field_200: Field200151 = Field(..., alias="200")


class Post39(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter169]
    responses: Responses177
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldIntegrationsIntegrationIdForm(BaseModel):
    post: Post39


class Parameter170(Parameter2):
    pass


class Field400178(Field400177):
    pass


class Field404178(Field400177):
    pass


class Field42275(Field400177):
    pass


class Field200152(Field400177):
    pass


class Responses178(BaseModel):
    field_400: Field400178 = Field(..., alias="400")
    field_404: Field404178 = Field(..., alias="404")
    field_422: Field42275 = Field(..., alias="422")
    field_200: Field200152 = Field(..., alias="200")


class Post40(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter170]
    responses: Responses178
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldIntegrationsIntegrationIdTest(BaseModel):
    post: Post40


class Field400179(Field400177):
    pass


class Field404179(Field400177):
    pass


class Field200153(Field400177):
    pass


class Responses179(BaseModel):
    field_400: Field400179 = Field(..., alias="400")
    field_404: Field404179 = Field(..., alias="404")
    field_200: Field200153 = Field(..., alias="200")


class Get81(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses179
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter171(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema796 = Field(..., alias="schema")


class Field400180(Field400177):
    pass


class Field404180(Field400177):
    pass


class Field42276(Field400177):
    pass


class Field42990(Field400177):
    pass


class Field200154(Field400177):
    pass


class Responses180(BaseModel):
    field_400: Field400180 = Field(..., alias="400")
    field_404: Field404180 = Field(..., alias="404")
    field_422: Field42276 = Field(..., alias="422")
    field_429: Field42990 = Field(..., alias="429")
    field_200: Field200154 = Field(..., alias="200")


class Patch23(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter171]
    responses: Responses180
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldInternalHelpResourcesContent(BaseModel):
    get: Get81
    patch: Patch23


class Field400181(Field400177):
    pass


class Field404181(Field400177):
    pass


class Field200155(Field400177):
    pass


class Responses181(BaseModel):
    field_400: Field400181 = Field(..., alias="400")
    field_404: Field404181 = Field(..., alias="404")
    field_200: Field200155 = Field(..., alias="200")


class Get82(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses181
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldInternalHelpResourcesEnabled(BaseModel):
    get: Get82


class Parameter172(Parameter171):
    pass


class Field400182(Field400177):
    pass


class Field404182(Field400177):
    pass


class Field42277(Field400177):
    pass


class Field42991(Field400177):
    pass


class Field200156(Field400177):
    pass


class Responses182(BaseModel):
    field_400: Field400182 = Field(..., alias="400")
    field_404: Field404182 = Field(..., alias="404")
    field_422: Field42277 = Field(..., alias="422")
    field_429: Field42991 = Field(..., alias="429")
    field_200: Field200156 = Field(..., alias="200")


class Patch24(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter172]
    responses: Responses182
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldInternalHelpResources(BaseModel):
    patch: Patch24


class Field40315(Field400177):
    pass


class Field404183(Field400177):
    pass


class Field200157(Field400177):
    pass


class Responses183(BaseModel):
    field_403: Field40315 = Field(..., alias="403")
    field_404: Field404183 = Field(..., alias="404")
    field_200: Field200157 = Field(..., alias="200")


class Get83(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses183
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter173(Parameter171):
    pass


class Field400183(Field400177):
    pass


class Field40316(Field400177):
    pass


class Field404184(Field400177):
    pass


class Field42278(Field400177):
    pass


class Field200158(Field400177):
    pass


class Responses184(BaseModel):
    field_400: Field400183 = Field(..., alias="400")
    field_403: Field40316 = Field(..., alias="403")
    field_404: Field404184 = Field(..., alias="404")
    field_422: Field42278 = Field(..., alias="422")
    field_200: Field200158 = Field(..., alias="200")


class Patch25(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter173]
    responses: Responses184
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLdapConfig(BaseModel):
    get: Get83
    patch: Patch25


class Parameter174(Parameter171):
    pass


class Field400184(Field400177):
    pass


class Field40317(Field400177):
    pass


class Field404185(Field400177):
    pass


class Field42279(Field400177):
    pass


class Field200159(Field400177):
    pass


class Responses185(BaseModel):
    field_400: Field400184 = Field(..., alias="400")
    field_403: Field40317 = Field(..., alias="403")
    field_404: Field404185 = Field(..., alias="404")
    field_422: Field42279 = Field(..., alias="422")
    field_200: Field200159 = Field(..., alias="200")


class Put9(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter174]
    responses: Responses185
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLdapConfigTestConnection(BaseModel):
    put: Put9


class Parameter175(Parameter171):
    pass


class Field400185(Field400177):
    pass


class Field40318(Field400177):
    pass


class Field404186(Field400177):
    pass


class Field42280(Field400177):
    pass


class Field200160(Field400177):
    pass


class Responses186(BaseModel):
    field_400: Field400185 = Field(..., alias="400")
    field_403: Field40318 = Field(..., alias="403")
    field_404: Field404186 = Field(..., alias="404")
    field_422: Field42280 = Field(..., alias="422")
    field_200: Field200160 = Field(..., alias="200")


class Put10(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter175]
    responses: Responses186
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLdapConfigTestAuth(BaseModel):
    put: Put10


class Parameter176(Parameter171):
    pass


class Field400186(Field400177):
    pass


class Field40319(Field400177):
    pass


class Field404187(Field400177):
    pass


class Field42281(Field400177):
    pass


class Field200161(Field400177):
    pass


class Responses187(BaseModel):
    field_400: Field400186 = Field(..., alias="400")
    field_403: Field40319 = Field(..., alias="403")
    field_404: Field404187 = Field(..., alias="404")
    field_422: Field42281 = Field(..., alias="422")
    field_200: Field200161 = Field(..., alias="200")


class Put11(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter176]
    responses: Responses187
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLdapConfigTestUserInfo(BaseModel):
    put: Put11


class Parameter177(Parameter171):
    pass


class Field400187(Field400177):
    pass


class Field40320(Field400177):
    pass


class Field404188(Field400177):
    pass


class Field42282(Field400177):
    pass


class Field200162(Field400177):
    pass


class Responses188(BaseModel):
    field_400: Field400187 = Field(..., alias="400")
    field_403: Field40320 = Field(..., alias="403")
    field_404: Field404188 = Field(..., alias="404")
    field_422: Field42282 = Field(..., alias="422")
    field_200: Field200162 = Field(..., alias="200")


class Put12(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter177]
    responses: Responses188
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLdapConfigTestUserAuth(BaseModel):
    put: Put12


class Field400188(Field400177):
    pass


class Field40321(Field400177):
    pass


class Field404189(Field400177):
    pass


class Schema858(Schema621):
    pass


class Field200163(BaseModel):
    description: str
    schema_: Schema858 = Field(..., alias="schema")


class Responses189(BaseModel):
    field_400: Field400188 = Field(..., alias="400")
    field_403: Field40321 = Field(..., alias="403")
    field_404: Field404189 = Field(..., alias="404")
    field_200: Field200163 = Field(..., alias="200")


class Get84(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses189
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLegacyFeatures(BaseModel):
    get: Get84


class Parameter178(Parameter2):
    pass


class Schema859(Schema):
    pass


class Field400189(BaseModel):
    description: str
    schema_: Schema859 = Field(..., alias="schema")


class Field40322(Field400189):
    pass


class Field404190(Field400189):
    pass


class Field200164(Field400189):
    pass


class Responses190(BaseModel):
    field_400: Field400189 = Field(..., alias="400")
    field_403: Field40322 = Field(..., alias="403")
    field_404: Field404190 = Field(..., alias="404")
    field_200: Field200164 = Field(..., alias="200")


class Get85(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter178]
    responses: Responses190
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter179(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema859] = Field(None, alias="schema")


class Field400190(Field400189):
    pass


class Field40323(Field400189):
    pass


class Field404191(Field400189):
    pass


class Field42283(Field400189):
    pass


class Field42992(Field400189):
    pass


class Field200165(Field400189):
    pass


class Responses191(BaseModel):
    field_400: Field400190 = Field(..., alias="400")
    field_403: Field40323 = Field(..., alias="403")
    field_404: Field404191 = Field(..., alias="404")
    field_422: Field42283 = Field(..., alias="422")
    field_429: Field42992 = Field(..., alias="429")
    field_200: Field200165 = Field(..., alias="200")


class Patch26(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter179]
    responses: Responses191
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLegacyFeaturesLegacyFeatureId(BaseModel):
    get: Get85
    patch: Patch26


class Field400191(Field400189):
    pass


class Field404192(Field400189):
    pass


class Schema872(Schema621):
    pass


class Field200166(BaseModel):
    description: str
    schema_: Schema872 = Field(..., alias="schema")


class Responses192(BaseModel):
    field_400: Field400191 = Field(..., alias="400")
    field_404: Field404192 = Field(..., alias="404")
    field_200: Field200166 = Field(..., alias="200")


class Get86(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses192
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLocales(BaseModel):
    get: Get86


class Parameter180(Parameter2):
    pass


class Schema873(Schema):
    pass


class Field400192(BaseModel):
    description: str
    schema_: Schema873 = Field(..., alias="schema")


class Field404193(Field400192):
    pass


class Schema875(Schema621):
    pass


class Field200167(BaseModel):
    description: str
    schema_: Schema875 = Field(..., alias="schema")


class Responses193(BaseModel):
    field_400: Field400192 = Field(..., alias="400")
    field_404: Field404193 = Field(..., alias="404")
    field_200: Field200167 = Field(..., alias="200")


class Get87(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter180]
    responses: Responses193
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema876(Schema):
    pass


class Parameter181(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema876] = Field(None, alias="schema")
    type: Optional[str] = None


class Field400193(BaseModel):
    description: str
    schema_: Schema876 = Field(..., alias="schema")


class Field404194(Field400193):
    pass


class Field40928(Field400193):
    pass


class Field42284(Field400193):
    pass


class Field42993(Field400193):
    pass


class Field200168(Field400193):
    pass


class Responses194(BaseModel):
    field_400: Field400193 = Field(..., alias="400")
    field_404: Field404194 = Field(..., alias="404")
    field_409: Field40928 = Field(..., alias="409")
    field_422: Field42284 = Field(..., alias="422")
    field_429: Field42993 = Field(..., alias="429")
    field_200: Field200168 = Field(..., alias="200")


class Post41(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter181]
    responses: Responses194
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLooks(BaseModel):
    get: Get87
    post: Post41


class Parameter182(Parameter60):
    pass


class Field400194(Field400193):
    pass


class Field404195(Field400193):
    pass


class Schema885(Schema621):
    pass


class Field200169(BaseModel):
    description: str
    schema_: Schema885 = Field(..., alias="schema")


class Responses195(BaseModel):
    field_400: Field400194 = Field(..., alias="400")
    field_404: Field404195 = Field(..., alias="404")
    field_200: Field200169 = Field(..., alias="200")


class Get88(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter182]
    responses: Responses195
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLooksSearch(BaseModel):
    get: Get88


class Parameter183(Parameter2):
    pass


class Schema886(Schema):
    pass


class Field400195(BaseModel):
    description: str
    schema_: Schema886 = Field(..., alias="schema")


class Field404196(Field400195):
    pass


class Field200170(Field400195):
    pass


class Responses196(BaseModel):
    field_400: Field400195 = Field(..., alias="400")
    field_404: Field404196 = Field(..., alias="404")
    field_200: Field200170 = Field(..., alias="200")


class Get89(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter183]
    responses: Responses196
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter184(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema886] = Field(None, alias="schema")


class Field400196(Field400195):
    pass


class Field404197(Field400195):
    pass


class Field42285(Field400195):
    pass


class Field42994(Field400195):
    pass


class Field200171(Field400195):
    pass


class Responses197(BaseModel):
    field_400: Field400196 = Field(..., alias="400")
    field_404: Field404197 = Field(..., alias="404")
    field_422: Field42285 = Field(..., alias="422")
    field_429: Field42994 = Field(..., alias="429")
    field_200: Field200171 = Field(..., alias="200")


class Patch27(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter184]
    responses: Responses197
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter185(Parameter2):
    pass


class Field400197(Field400195):
    pass


class Field404198(Field400195):
    pass


class Field42995(Field400195):
    pass


class Schema898(Items):
    pass


class Field20429(BaseModel):
    description: str
    schema_: Schema898 = Field(..., alias="schema")


class Responses198(BaseModel):
    field_400: Field400197 = Field(..., alias="400")
    field_404: Field404198 = Field(..., alias="404")
    field_429: Field42995 = Field(..., alias="429")
    field_204: Field20429 = Field(..., alias="204")


class Delete25(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter185]
    responses: Responses198
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLooksLookId(BaseModel):
    get: Get89
    patch: Patch27
    delete: Delete25


class Parameter186(Parameter7):
    pass


class Schema899(Schema):
    pass


class Field400198(BaseModel):
    description: str
    schema_: Schema899 = Field(..., alias="schema")


class Field404199(Field400198):
    pass


class Field42286(Field400198):
    pass


class Field42996(Field400198):
    pass


class Schema903(Items):
    pass


class Field200172(BaseModel):
    description: str
    schema_: Schema903 = Field(..., alias="schema")


class Responses199(BaseModel):
    field_400: Field400198 = Field(..., alias="400")
    field_404: Field404199 = Field(..., alias="404")
    field_422: Field42286 = Field(..., alias="422")
    field_429: Field42996 = Field(..., alias="429")
    field_200: Field200172 = Field(..., alias="200")


class Get90(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter186]
    responses: Responses199
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLooksLookIdRunResultFormat(BaseModel):
    get: Get90


class Parameter187(Parameter2):
    pass


class Schema904(Schema):
    pass


class Field400199(BaseModel):
    description: str
    schema_: Schema904 = Field(..., alias="schema")


class Field404200(Field400199):
    pass


class Field40929(Field400199):
    pass


class Field42287(Field400199):
    pass


class Field42997(Field400199):
    pass


class Field200173(Field400199):
    pass


class Field2013(Field400199):
    pass


class Responses200(BaseModel):
    field_400: Field400199 = Field(..., alias="400")
    field_404: Field404200 = Field(..., alias="404")
    field_409: Field40929 = Field(..., alias="409")
    field_422: Field42287 = Field(..., alias="422")
    field_429: Field42997 = Field(..., alias="429")
    field_200: Field200173 = Field(..., alias="200")
    field_201: Field2013 = Field(..., alias="201")


class Post42(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter187]
    responses: Responses200
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLooksLookIdCopy(BaseModel):
    post: Post42


class Field400200(Field400199):
    pass


class Field404201(Field400199):
    pass


class Field42288(Field400199):
    pass


class Field42998(Field400199):
    pass


class Field200174(Field400199):
    pass


class Field2014(Field400199):
    pass


class Responses201(BaseModel):
    field_400: Field400200 = Field(..., alias="400")
    field_404: Field404201 = Field(..., alias="404")
    field_422: Field42288 = Field(..., alias="422")
    field_429: Field42998 = Field(..., alias="429")
    field_200: Field200174 = Field(..., alias="200")
    field_201: Field2014 = Field(..., alias="201")


class Patch28(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter187]
    responses: Responses201
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLooksLookIdMove(BaseModel):
    patch: Patch28


class Parameter189(Parameter7):
    pass


class Field400201(Field400199):
    pass


class Field404202(Field400199):
    pass


class Schema919(Schema621):
    pass


class Field200175(BaseModel):
    description: str
    schema_: Schema919 = Field(..., alias="schema")


class Responses202(BaseModel):
    field_400: Field400201 = Field(..., alias="400")
    field_404: Field404202 = Field(..., alias="404")
    field_200: Field200175 = Field(..., alias="200")


class Get91(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter189]
    responses: Responses202
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema920(Schema):
    pass


class Parameter190(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema920 = Field(..., alias="schema")


class Field400202(BaseModel):
    description: str
    schema_: Schema920 = Field(..., alias="schema")


class Field404203(Field400202):
    pass


class Field40930(Field400202):
    pass


class Field42289(Field400202):
    pass


class Field42999(Field400202):
    pass


class Field200176(Field400202):
    pass


class Responses203(BaseModel):
    field_400: Field400202 = Field(..., alias="400")
    field_404: Field404203 = Field(..., alias="404")
    field_409: Field40930 = Field(..., alias="409")
    field_422: Field42289 = Field(..., alias="422")
    field_429: Field42999 = Field(..., alias="429")
    field_200: Field200176 = Field(..., alias="200")


class Post43(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter190]
    responses: Responses203
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLookmlModels(BaseModel):
    get: Get91
    post: Post43


class Parameter191(Parameter2):
    pass


class Field400203(Field400202):
    pass


class Field404204(Field400202):
    pass


class Field200177(Field400202):
    pass


class Responses204(BaseModel):
    field_400: Field400203 = Field(..., alias="400")
    field_404: Field404204 = Field(..., alias="404")
    field_200: Field200177 = Field(..., alias="200")


class Get92(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter191]
    responses: Responses204
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter192(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema920] = Field(None, alias="schema")


class Field400204(Field400202):
    pass


class Field404205(Field400202):
    pass


class Field42290(Field400202):
    pass


class Field429100(Field400202):
    pass


class Field200178(Field400202):
    pass


class Responses205(BaseModel):
    field_400: Field400204 = Field(..., alias="400")
    field_404: Field404205 = Field(..., alias="404")
    field_422: Field42290 = Field(..., alias="422")
    field_429: Field429100 = Field(..., alias="429")
    field_200: Field200178 = Field(..., alias="200")


class Patch29(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter192]
    responses: Responses205
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter193(Parameter2):
    pass


class Field400205(Field400202):
    pass


class Field404206(Field400202):
    pass


class Field429101(Field400202):
    pass


class Schema939(Items):
    pass


class Field20430(BaseModel):
    description: str
    schema_: Schema939 = Field(..., alias="schema")


class Responses206(BaseModel):
    field_400: Field400205 = Field(..., alias="400")
    field_404: Field404206 = Field(..., alias="404")
    field_429: Field429101 = Field(..., alias="429")
    field_204: Field20430 = Field(..., alias="204")


class Delete26(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter193]
    responses: Responses206
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLookmlModelsLookmlModelName(BaseModel):
    get: Get92
    patch: Patch29
    delete: Delete26


class Schema940(Schema):
    pass


class Field400206(BaseModel):
    description: str
    schema_: Schema940 = Field(..., alias="schema")


class Field404207(Field400206):
    pass


class Field200179(Field400206):
    pass


class Responses207(BaseModel):
    field_400: Field400206 = Field(..., alias="400")
    field_404: Field404207 = Field(..., alias="404")
    field_200: Field200179 = Field(..., alias="200")


class Get93(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter193]
    responses: Responses207
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldLookmlModelsLookmlModelNameExploresExploreName(BaseModel):
    get: Get93


class Field400207(Field400206):
    pass


class Field404208(Field400206):
    pass


class Field200180(Field400206):
    pass


class Responses208(BaseModel):
    field_400: Field400207 = Field(..., alias="400")
    field_404: Field404208 = Field(..., alias="404")
    field_200: Field200180 = Field(..., alias="200")


class Get94(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter193]
    responses: Responses208
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldMergeQueriesMergeQueryId(BaseModel):
    get: Get94


class Parameter196(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema940] = Field(None, alias="schema")
    type: Optional[str] = None


class Field400208(Field400206):
    pass


class Field404209(Field400206):
    pass


class Field40931(Field400206):
    pass


class Field42291(Field400206):
    pass


class Field429102(Field400206):
    pass


class Field200181(Field400206):
    pass


class Responses209(BaseModel):
    field_400: Field400208 = Field(..., alias="400")
    field_404: Field404209 = Field(..., alias="404")
    field_409: Field40931 = Field(..., alias="409")
    field_422: Field42291 = Field(..., alias="422")
    field_429: Field429102 = Field(..., alias="429")
    field_200: Field200181 = Field(..., alias="200")


class Post44(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter196]
    responses: Responses209
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldMergeQueries(BaseModel):
    post: Post44


class Parameter197(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str
    additionalProperties: Optional[AdditionalProperties1] = None


class Field400209(Field400206):
    pass


class Field404210(Field400206):
    pass


class Field200182(Field400206):
    pass


class Responses210(BaseModel):
    field_400: Field400209 = Field(..., alias="400")
    field_404: Field404210 = Field(..., alias="404")
    field_200: Field200182 = Field(..., alias="200")


class Get95(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter197]
    responses: Responses210
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldModelsModelNameViewsViewNameFieldsFieldNameSuggestions(BaseModel):
    get: Get95


class Parameter198(Parameter2):
    pass


class Field400210(Field400206):
    pass


class Field404211(Field400206):
    pass


class Field200183(Field400206):
    pass


class Responses211(BaseModel):
    field_400: Field400210 = Field(..., alias="400")
    field_404: Field404211 = Field(..., alias="404")
    field_200: Field200183 = Field(..., alias="200")


class Get96(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter198]
    responses: Responses211
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldModelsModelName(BaseModel):
    get: Get96


class Field400211(Field400206):
    pass


class Field404212(Field400206):
    pass


class Field429103(Field400206):
    pass


class Items53(Items):
    pass


class Schema962(BaseModel):
    type: str
    items: Items53


class Field200184(BaseModel):
    description: str
    schema_: Schema962 = Field(..., alias="schema")


class Responses212(BaseModel):
    field_400: Field400211 = Field(..., alias="400")
    field_404: Field404212 = Field(..., alias="404")
    field_429: Field429103 = Field(..., alias="429")
    field_200: Field200184 = Field(..., alias="200")


class Get97(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter198]
    responses: Responses212
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldConnectionsConnectionNameDatabases(BaseModel):
    get: Get97


class Schema963(Schema):
    pass


class Field400212(BaseModel):
    description: str
    schema_: Schema963 = Field(..., alias="schema")


class Field404213(Field400212):
    pass


class Field42292(Field400212):
    pass


class Field429104(Field400212):
    pass


class Field200185(Field400212):
    pass


class Responses213(BaseModel):
    field_400: Field400212 = Field(..., alias="400")
    field_404: Field404213 = Field(..., alias="404")
    field_422: Field42292 = Field(..., alias="422")
    field_429: Field429104 = Field(..., alias="429")
    field_200: Field200185 = Field(..., alias="200")


class Get98(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter198]
    responses: Responses213
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldConnectionsConnectionNameFeatures(BaseModel):
    get: Get98


class Field400213(Field400212):
    pass


class Field404214(Field400212):
    pass


class Field42293(Field400212):
    pass


class Field429105(Field400212):
    pass


class Items54(Schema):
    pass


class Schema972(BaseModel):
    type: str
    items: Items54


class Field200186(BaseModel):
    description: str
    schema_: Schema972 = Field(..., alias="schema")


class Responses214(BaseModel):
    field_400: Field400213 = Field(..., alias="400")
    field_404: Field404214 = Field(..., alias="404")
    field_422: Field42293 = Field(..., alias="422")
    field_429: Field429105 = Field(..., alias="429")
    field_200: Field200186 = Field(..., alias="200")


class Get99(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter198]
    responses: Responses214
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldConnectionsConnectionNameSchemas(BaseModel):
    get: Get99


class Parameter202(Parameter7):
    pass


class Schema973(Schema):
    pass


class Field400214(BaseModel):
    description: str
    schema_: Schema973 = Field(..., alias="schema")


class Field404215(Field400214):
    pass


class Field42294(Field400214):
    pass


class Field429106(Field400214):
    pass


class Schema977(Schema972):
    pass


class Field200187(BaseModel):
    description: str
    schema_: Schema977 = Field(..., alias="schema")


class Responses215(BaseModel):
    field_400: Field400214 = Field(..., alias="400")
    field_404: Field404215 = Field(..., alias="404")
    field_422: Field42294 = Field(..., alias="422")
    field_429: Field429106 = Field(..., alias="429")
    field_200: Field200187 = Field(..., alias="200")


class Get100(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter202]
    responses: Responses215
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldConnectionsConnectionNameTables(BaseModel):
    get: Get100


class Schema978(Schema):
    pass


class Field400215(BaseModel):
    description: str
    schema_: Schema978 = Field(..., alias="schema")


class Field404216(Field400215):
    pass


class Field42295(Field400215):
    pass


class Field429107(Field400215):
    pass


class Schema982(Schema972):
    pass


class Field200188(BaseModel):
    description: str
    schema_: Schema982 = Field(..., alias="schema")


class Responses216(BaseModel):
    field_400: Field400215 = Field(..., alias="400")
    field_404: Field404216 = Field(..., alias="404")
    field_422: Field42295 = Field(..., alias="422")
    field_429: Field429107 = Field(..., alias="429")
    field_200: Field200188 = Field(..., alias="200")


class Get101(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter202]
    responses: Responses216
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldConnectionsConnectionNameColumns(BaseModel):
    get: Get101


class Parameter204(Parameter2):
    pass


class Schema983(Schema):
    pass


class Field400216(BaseModel):
    description: str
    schema_: Schema983 = Field(..., alias="schema")


class Field404217(Field400216):
    pass


class Field42296(Field400216):
    pass


class Field429108(Field400216):
    pass


class Schema987(Schema972):
    pass


class Field200189(BaseModel):
    description: str
    schema_: Schema987 = Field(..., alias="schema")


class Responses217(BaseModel):
    field_400: Field400216 = Field(..., alias="400")
    field_404: Field404217 = Field(..., alias="404")
    field_422: Field42296 = Field(..., alias="422")
    field_429: Field429108 = Field(..., alias="429")
    field_200: Field200189 = Field(..., alias="200")


class Get102(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter204]
    responses: Responses217
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldConnectionsConnectionNameSearchColumns(BaseModel):
    get: Get102


class Schema988(Schema):
    pass


class Parameter205(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema988] = Field(None, alias="schema")


class Field400217(BaseModel):
    description: str
    schema_: Schema988 = Field(..., alias="schema")


class Field404218(Field400217):
    pass


class Field42297(Field400217):
    pass


class Field429109(Field400217):
    pass


class Field200190(Field400217):
    pass


class Responses218(BaseModel):
    field_400: Field400217 = Field(..., alias="400")
    field_404: Field404218 = Field(..., alias="404")
    field_422: Field42297 = Field(..., alias="422")
    field_429: Field429109 = Field(..., alias="429")
    field_200: Field200190 = Field(..., alias="200")


class Post45(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter205]
    responses: Responses218
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldConnectionsConnectionNameCostEstimate(BaseModel):
    post: Post45


class Field400218(Field400217):
    pass


class Field404219(Field400217):
    pass


class Field200191(Field400217):
    pass


class Responses219(BaseModel):
    field_400: Field400218 = Field(..., alias="400")
    field_404: Field404219 = Field(..., alias="404")
    field_200: Field200191 = Field(..., alias="200")


class Get103(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses219
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldMobileSettings(BaseModel):
    get: Get103


class Parameter206(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema988 = Field(..., alias="schema")


class Field400219(Field400217):
    pass


class Field404220(Field400217):
    pass


class Field40932(Field400217):
    pass


class Field42298(Field400217):
    pass


class Field429110(Field400217):
    pass


class Field200192(Field400217):
    pass


class Responses220(BaseModel):
    field_400: Field400219 = Field(..., alias="400")
    field_404: Field404220 = Field(..., alias="404")
    field_409: Field40932 = Field(..., alias="409")
    field_422: Field42298 = Field(..., alias="422")
    field_429: Field429110 = Field(..., alias="429")
    field_200: Field200192 = Field(..., alias="200")


class Post46(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter206]
    responses: Responses220
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldMobileDevice(BaseModel):
    post: Post46


class Parameter207(Parameter2):
    pass


class Field400220(Field400217):
    pass


class Field404221(Field400217):
    pass


class Field42299(Field400217):
    pass


class Field429111(Field400217):
    pass


class Field200193(Field400217):
    pass


class Responses221(BaseModel):
    field_400: Field400220 = Field(..., alias="400")
    field_404: Field404221 = Field(..., alias="404")
    field_422: Field42299 = Field(..., alias="422")
    field_429: Field429111 = Field(..., alias="429")
    field_200: Field200193 = Field(..., alias="200")


class Patch30(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter207]
    responses: Responses221
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400221(Field400217):
    pass


class Field404222(Field400217):
    pass


class Field429112(Field400217):
    pass


class Field20431(Field2042):
    pass


class Responses222(BaseModel):
    field_400: Field400221 = Field(..., alias="400")
    field_404: Field404222 = Field(..., alias="404")
    field_429: Field429112 = Field(..., alias="429")
    field_204: Field20431 = Field(..., alias="204")


class Delete27(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter207]
    responses: Responses222
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldMobileDeviceDeviceId(BaseModel):
    patch: Patch30
    delete: Delete27


class Parameter209(Parameter7):
    pass


class Field400222(Field400217):
    pass


class Field404223(Field400217):
    pass


class Schema1014(Schema972):
    pass


class Field200194(BaseModel):
    description: str
    schema_: Schema1014 = Field(..., alias="schema")


class Responses223(BaseModel):
    field_400: Field400222 = Field(..., alias="400")
    field_404: Field404223 = Field(..., alias="404")
    field_200: Field200194 = Field(..., alias="200")


class Get104(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter209]
    responses: Responses223
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldModelSetsSearch(BaseModel):
    get: Get104


class Parameter210(Parameter2):
    pass


class Schema1015(Schema):
    pass


class Field400223(BaseModel):
    description: str
    schema_: Schema1015 = Field(..., alias="schema")


class Field404224(Field400223):
    pass


class Field200195(Field400223):
    pass


class Responses224(BaseModel):
    field_400: Field400223 = Field(..., alias="400")
    field_404: Field404224 = Field(..., alias="404")
    field_200: Field200195 = Field(..., alias="200")


class Get105(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter210]
    responses: Responses224
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400224(Field400223):
    pass


class Field404225(Field400223):
    pass


class Field4053(Field400223):
    pass


class Schema1021(Items):
    pass


class Field20432(BaseModel):
    description: str
    schema_: Schema1021 = Field(..., alias="schema")


class Responses225(BaseModel):
    field_400: Field400224 = Field(..., alias="400")
    field_404: Field404225 = Field(..., alias="404")
    field_405: Field4053 = Field(..., alias="405")
    field_204: Field20432 = Field(..., alias="204")


class Delete28(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter210]
    responses: Responses225
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1022(Schema):
    pass


class Parameter212(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1022] = Field(None, alias="schema")


class Field400225(BaseModel):
    description: str
    schema_: Schema1022 = Field(..., alias="schema")


class Field404226(Field400225):
    pass


class Field4054(Field400225):
    pass


class Field422100(Field400225):
    pass


class Field200196(Field400225):
    pass


class Responses226(BaseModel):
    field_400: Field400225 = Field(..., alias="400")
    field_404: Field404226 = Field(..., alias="404")
    field_405: Field4054 = Field(..., alias="405")
    field_422: Field422100 = Field(..., alias="422")
    field_200: Field200196 = Field(..., alias="200")


class Patch31(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter212]
    responses: Responses226
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldModelSetsModelSetId(BaseModel):
    get: Get105
    delete: Delete28
    patch: Patch31


class Parameter213(Parameter2):
    pass


class Field404227(Field400225):
    pass


class Schema1029(Schema972):
    pass


class Field200197(BaseModel):
    description: str
    schema_: Schema1029 = Field(..., alias="schema")


class Responses227(BaseModel):
    field_404: Field404227 = Field(..., alias="404")
    field_200: Field200197 = Field(..., alias="200")


class Get106(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter213]
    responses: Responses227
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1030(Schema):
    pass


class Parameter214(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1030 = Field(..., alias="schema")


class Field400226(BaseModel):
    description: str
    schema_: Schema1030 = Field(..., alias="schema")


class Field404228(Field400226):
    pass


class Field40933(Field400226):
    pass


class Field422101(Field400226):
    pass


class Field200198(Field400226):
    pass


class Responses228(BaseModel):
    field_400: Field400226 = Field(..., alias="400")
    field_404: Field404228 = Field(..., alias="404")
    field_409: Field40933 = Field(..., alias="409")
    field_422: Field422101 = Field(..., alias="422")
    field_200: Field200198 = Field(..., alias="200")


class Post47(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter214]
    responses: Responses228
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldModelSets(BaseModel):
    get: Get106
    post: Post47


class Parameter215(Parameter2):
    pass


class Field400227(Field400226):
    pass


class Field404229(Field400226):
    pass


class Schema1038(Schema972):
    pass


class Field200199(BaseModel):
    description: str
    schema_: Schema1038 = Field(..., alias="schema")


class Responses229(BaseModel):
    field_400: Field400227 = Field(..., alias="400")
    field_404: Field404229 = Field(..., alias="404")
    field_200: Field200199 = Field(..., alias="200")


class Get107(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter215]
    responses: Responses229
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldOauthClientApps(BaseModel):
    get: Get107


class Schema1039(Schema):
    pass


class Field400228(BaseModel):
    description: str
    schema_: Schema1039 = Field(..., alias="schema")


class Field404230(Field400228):
    pass


class Field200200(Field400228):
    pass


class Responses230(BaseModel):
    field_400: Field400228 = Field(..., alias="400")
    field_404: Field404230 = Field(..., alias="404")
    field_200: Field200200 = Field(..., alias="200")


class Get108(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter215]
    responses: Responses230
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400229(Field400228):
    pass


class Field404231(Field400228):
    pass


class Field429113(Field400228):
    pass


class Schema1045(Items):
    pass


class Field20433(BaseModel):
    description: str
    schema_: Schema1045 = Field(..., alias="schema")


class Responses231(BaseModel):
    field_400: Field400229 = Field(..., alias="400")
    field_404: Field404231 = Field(..., alias="404")
    field_429: Field429113 = Field(..., alias="429")
    field_204: Field20433 = Field(..., alias="204")


class Delete29(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter215]
    responses: Responses231
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1046(Schema):
    pass


class Parameter218(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1046] = Field(None, alias="schema")


class Field400230(BaseModel):
    description: str
    schema_: Schema1046 = Field(..., alias="schema")


class Field404232(Field400230):
    pass


class Field40934(Field400230):
    pass


class Field422102(Field400230):
    pass


class Field429114(Field400230):
    pass


class Field200201(Field400230):
    pass


class Responses232(BaseModel):
    field_400: Field400230 = Field(..., alias="400")
    field_404: Field404232 = Field(..., alias="404")
    field_409: Field40934 = Field(..., alias="409")
    field_422: Field422102 = Field(..., alias="422")
    field_429: Field429114 = Field(..., alias="429")
    field_200: Field200201 = Field(..., alias="200")


class Post48(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter218]
    responses: Responses232
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter219(Parameter218):
    pass


class Field400231(Field400230):
    pass


class Field404233(Field400230):
    pass


class Field422103(Field400230):
    pass


class Field429115(Field400230):
    pass


class Field200202(Field400230):
    pass


class Responses233(BaseModel):
    field_400: Field400231 = Field(..., alias="400")
    field_404: Field404233 = Field(..., alias="404")
    field_422: Field422103 = Field(..., alias="422")
    field_429: Field429115 = Field(..., alias="429")
    field_200: Field200202 = Field(..., alias="200")


class Patch32(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter219]
    responses: Responses233
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldOauthClientAppsClientGuid(BaseModel):
    get: Get108
    delete: Delete29
    post: Post48
    patch: Patch32


class Parameter220(Parameter2):
    pass


class Field400232(Field400230):
    pass


class Field404234(Field400230):
    pass


class Field429116(Field400230):
    pass


class Schema1062(Items):
    pass


class Field20434(BaseModel):
    description: str
    schema_: Schema1062 = Field(..., alias="schema")


class Responses234(BaseModel):
    field_400: Field400232 = Field(..., alias="400")
    field_404: Field404234 = Field(..., alias="404")
    field_429: Field429116 = Field(..., alias="429")
    field_204: Field20434 = Field(..., alias="204")


class Delete30(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter220]
    responses: Responses234
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldOauthClientAppsClientGuidTokens(BaseModel):
    delete: Delete30


class Schema1063(Schema):
    pass


class Field400233(BaseModel):
    description: str
    schema_: Schema1063 = Field(..., alias="schema")


class Field404235(Field400233):
    pass


class Field422104(Field400233):
    pass


class Field429117(Field400233):
    pass


class Field20435(Field2042):
    pass


class Schema1067(Items):
    pass


class Field200203(BaseModel):
    description: str
    schema_: Schema1067 = Field(..., alias="schema")


class Responses235(BaseModel):
    field_400: Field400233 = Field(..., alias="400")
    field_404: Field404235 = Field(..., alias="404")
    field_422: Field422104 = Field(..., alias="422")
    field_429: Field429117 = Field(..., alias="429")
    field_204: Field20435 = Field(..., alias="204")
    field_200: Field200203 = Field(..., alias="200")


class Post49(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter220]
    responses: Responses235
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1068(Schema):
    pass


class Field400234(BaseModel):
    description: str
    schema_: Schema1068 = Field(..., alias="schema")


class Field404236(Field400234):
    pass


class Field429118(Field400234):
    pass


class Schema1071(Items):
    pass


class Field20436(BaseModel):
    description: str
    schema_: Schema1071 = Field(..., alias="schema")


class Responses236(BaseModel):
    field_400: Field400234 = Field(..., alias="400")
    field_404: Field404236 = Field(..., alias="404")
    field_429: Field429118 = Field(..., alias="429")
    field_204: Field20436 = Field(..., alias="204")


class Delete31(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter220]
    responses: Responses236
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldOauthClientAppsClientGuidUsersUserId(BaseModel):
    post: Post49
    delete: Delete31


class Schema1072(Schema):
    pass


class Field40324(BaseModel):
    description: str
    schema_: Schema1072 = Field(..., alias="schema")


class Field404237(Field40324):
    pass


class Field200204(Field40324):
    pass


class Responses237(BaseModel):
    field_403: Field40324 = Field(..., alias="403")
    field_404: Field404237 = Field(..., alias="404")
    field_200: Field200204 = Field(..., alias="200")


class Get109(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses237
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter223(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1072 = Field(..., alias="schema")


class Field400235(Field40324):
    pass


class Field40325(Field40324):
    pass


class Field404238(Field40324):
    pass


class Field422105(Field40324):
    pass


class Field200205(Field40324):
    pass


class Responses238(BaseModel):
    field_400: Field400235 = Field(..., alias="400")
    field_403: Field40325 = Field(..., alias="403")
    field_404: Field404238 = Field(..., alias="404")
    field_422: Field422105 = Field(..., alias="422")
    field_200: Field200205 = Field(..., alias="200")


class Patch33(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter223]
    responses: Responses238
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldOidcConfig(BaseModel):
    get: Get109
    patch: Patch33


class Parameter224(Parameter2):
    pass


class Field40326(Field40324):
    pass


class Field404239(Field40324):
    pass


class Field200206(Field40324):
    pass


class Responses239(BaseModel):
    field_403: Field40326 = Field(..., alias="403")
    field_404: Field404239 = Field(..., alias="404")
    field_200: Field200206 = Field(..., alias="200")


class Get110(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter224]
    responses: Responses239
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400236(Field40324):
    pass


class Field40327(Field40324):
    pass


class Field404240(Field40324):
    pass


class Schema1087(Items):
    pass


class Field20437(BaseModel):
    description: str
    schema_: Schema1087 = Field(..., alias="schema")


class Responses240(BaseModel):
    field_400: Field400236 = Field(..., alias="400")
    field_403: Field40327 = Field(..., alias="403")
    field_404: Field404240 = Field(..., alias="404")
    field_204: Field20437 = Field(..., alias="204")


class Delete32(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter224]
    responses: Responses240
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldOidcTestConfigsTestSlug(BaseModel):
    get: Get110
    delete: Delete32


class Schema1088(Schema):
    pass


class Parameter226(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1088 = Field(..., alias="schema")


class Field400237(BaseModel):
    description: str
    schema_: Schema1088 = Field(..., alias="schema")


class Field40328(Field400237):
    pass


class Field404241(Field400237):
    pass


class Field422106(Field400237):
    pass


class Field200207(Field400237):
    pass


class Responses241(BaseModel):
    field_400: Field400237 = Field(..., alias="400")
    field_403: Field40328 = Field(..., alias="403")
    field_404: Field404241 = Field(..., alias="404")
    field_422: Field422106 = Field(..., alias="422")
    field_200: Field200207 = Field(..., alias="200")


class Post50(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter226]
    responses: Responses241
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldOidcTestConfigs(BaseModel):
    post: Post50


class Field40329(Field400237):
    pass


class Field400238(Field400237):
    pass


class Field404242(Field400237):
    pass


class Field200208(Field400237):
    pass


class Responses242(BaseModel):
    field_403: Field40329 = Field(..., alias="403")
    field_400: Field400238 = Field(..., alias="400")
    field_404: Field404242 = Field(..., alias="404")
    field_200: Field200208 = Field(..., alias="200")


class Get111(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses242
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter227(Parameter226):
    pass


class Field40330(Field400237):
    pass


class Field400239(Field400237):
    pass


class Field404243(Field400237):
    pass


class Field422107(Field400237):
    pass


class Field429119(Field400237):
    pass


class Field200209(Field400237):
    pass


class Responses243(BaseModel):
    field_403: Field40330 = Field(..., alias="403")
    field_400: Field400239 = Field(..., alias="400")
    field_404: Field404243 = Field(..., alias="404")
    field_422: Field422107 = Field(..., alias="422")
    field_429: Field429119 = Field(..., alias="429")
    field_200: Field200209 = Field(..., alias="200")


class Patch34(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter227]
    responses: Responses243
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldPasswordConfig(BaseModel):
    get: Get111
    patch: Patch34


class Field40331(Field400237):
    pass


class Field400240(Field400237):
    pass


class Field404244(Field400237):
    pass


class Field422108(Field400237):
    pass


class Field429120(Field400237):
    pass


class Schema1110(Items):
    pass


class Field200210(BaseModel):
    description: str
    schema_: Schema1110 = Field(..., alias="schema")


class Responses244(BaseModel):
    field_403: Field40331 = Field(..., alias="403")
    field_400: Field400240 = Field(..., alias="400")
    field_404: Field404244 = Field(..., alias="404")
    field_422: Field422108 = Field(..., alias="422")
    field_429: Field429120 = Field(..., alias="429")
    field_200: Field200210 = Field(..., alias="200")


class Put13(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses244
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldPasswordConfigForcePasswordResetAtNextLoginForAllUsers(BaseModel):
    put: Put13


class Schema1111(Schema):
    pass


class Field400241(BaseModel):
    description: str
    schema_: Schema1111 = Field(..., alias="schema")


class Field404245(Field400241):
    pass


class Schema1113(Schema972):
    pass


class Field200211(BaseModel):
    description: str
    schema_: Schema1113 = Field(..., alias="schema")


class Responses245(BaseModel):
    field_400: Field400241 = Field(..., alias="400")
    field_404: Field404245 = Field(..., alias="404")
    field_200: Field200211 = Field(..., alias="200")


class Get112(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses245
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldPermissions(BaseModel):
    get: Get112


class Parameter228(Parameter7):
    pass


class Schema1114(Schema):
    pass


class Field400242(BaseModel):
    description: str
    schema_: Schema1114 = Field(..., alias="schema")


class Field404246(Field400242):
    pass


class Schema1116(Schema972):
    pass


class Field200212(BaseModel):
    description: str
    schema_: Schema1116 = Field(..., alias="schema")


class Responses246(BaseModel):
    field_400: Field400242 = Field(..., alias="400")
    field_404: Field404246 = Field(..., alias="404")
    field_200: Field200212 = Field(..., alias="200")


class Get113(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter228]
    responses: Responses246
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldPermissionSetsSearch(BaseModel):
    get: Get113


class Parameter229(Parameter2):
    pass


class Schema1117(Schema):
    pass


class Field400243(BaseModel):
    description: str
    schema_: Schema1117 = Field(..., alias="schema")


class Field404247(Field400243):
    pass


class Field200213(Field400243):
    pass


class Responses247(BaseModel):
    field_400: Field400243 = Field(..., alias="400")
    field_404: Field404247 = Field(..., alias="404")
    field_200: Field200213 = Field(..., alias="200")


class Get114(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter229]
    responses: Responses247
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400244(Field400243):
    pass


class Field404248(Field400243):
    pass


class Field429121(Field400243):
    pass


class Field4055(Field400243):
    pass


class Schema1124(Items):
    pass


class Field20438(BaseModel):
    description: str
    schema_: Schema1124 = Field(..., alias="schema")


class Responses248(BaseModel):
    field_400: Field400244 = Field(..., alias="400")
    field_404: Field404248 = Field(..., alias="404")
    field_429: Field429121 = Field(..., alias="429")
    field_405: Field4055 = Field(..., alias="405")
    field_204: Field20438 = Field(..., alias="204")


class Delete33(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter229]
    responses: Responses248
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1125(Schema):
    pass


class Parameter231(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1125] = Field(None, alias="schema")


class Field400245(BaseModel):
    description: str
    schema_: Schema1125 = Field(..., alias="schema")


class Field404249(Field400245):
    pass


class Field422109(Field400245):
    pass


class Field429122(Field400245):
    pass


class Field4056(Field400245):
    pass


class Field200214(Field400245):
    pass


class Responses249(BaseModel):
    field_400: Field400245 = Field(..., alias="400")
    field_404: Field404249 = Field(..., alias="404")
    field_422: Field422109 = Field(..., alias="422")
    field_429: Field429122 = Field(..., alias="429")
    field_405: Field4056 = Field(..., alias="405")
    field_200: Field200214 = Field(..., alias="200")


class Patch35(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter231]
    responses: Responses249
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldPermissionSetsPermissionSetId(BaseModel):
    get: Get114
    delete: Delete33
    patch: Patch35


class Parameter232(Parameter2):
    pass


class Field400246(Field400245):
    pass


class Field404250(Field400245):
    pass


class Schema1134(Schema972):
    pass


class Field200215(BaseModel):
    description: str
    schema_: Schema1134 = Field(..., alias="schema")


class Responses250(BaseModel):
    field_400: Field400246 = Field(..., alias="400")
    field_404: Field404250 = Field(..., alias="404")
    field_200: Field200215 = Field(..., alias="200")


class Get115(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter232]
    responses: Responses250
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1135(Schema):
    pass


class Parameter233(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1135 = Field(..., alias="schema")


class Field400247(BaseModel):
    description: str
    schema_: Schema1135 = Field(..., alias="schema")


class Field404251(Field400247):
    pass


class Field40935(Field400247):
    pass


class Field422110(Field400247):
    pass


class Field429123(Field400247):
    pass


class Field200216(Field400247):
    pass


class Responses251(BaseModel):
    field_400: Field400247 = Field(..., alias="400")
    field_404: Field404251 = Field(..., alias="404")
    field_409: Field40935 = Field(..., alias="409")
    field_422: Field422110 = Field(..., alias="422")
    field_429: Field429123 = Field(..., alias="429")
    field_200: Field200216 = Field(..., alias="200")


class Post51(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter233]
    responses: Responses251
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldPermissionSets(BaseModel):
    get: Get115
    post: Post51


class Parameter234(Parameter2):
    pass


class Field400248(Field400247):
    pass


class Field404252(Field400247):
    pass


class Field422111(Field400247):
    pass


class Field429124(Field400247):
    pass


class Schema1146(Items):
    pass


class Field200217(BaseModel):
    description: str
    schema_: Schema1146 = Field(..., alias="schema")


class Field20439(Field2042):
    pass


class Responses252(BaseModel):
    field_400: Field400248 = Field(..., alias="400")
    field_404: Field404252 = Field(..., alias="404")
    field_422: Field422111 = Field(..., alias="422")
    field_429: Field429124 = Field(..., alias="429")
    field_200: Field200217 = Field(..., alias="200")
    field_204: Field20439 = Field(..., alias="204")


class Post52(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter234]
    responses: Responses252
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdDeployRefToProduction(BaseModel):
    post: Post52


class Schema1147(Schema):
    pass


class Field400249(BaseModel):
    description: str
    schema_: Schema1147 = Field(..., alias="schema")


class Field404253(Field400249):
    pass


class Field422112(Field400249):
    pass


class Field429125(Field400249):
    pass


class Schema1151(Items):
    pass


class Field200218(BaseModel):
    description: str
    schema_: Schema1151 = Field(..., alias="schema")


class Responses253(BaseModel):
    field_400: Field400249 = Field(..., alias="400")
    field_404: Field404253 = Field(..., alias="404")
    field_422: Field422112 = Field(..., alias="422")
    field_429: Field429125 = Field(..., alias="429")
    field_200: Field200218 = Field(..., alias="200")
    field_204: Field20439 = Field(..., alias="204")


class Post53(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter234]
    responses: Responses253
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdDeployToProduction(BaseModel):
    post: Post53


class Schema1152(Schema):
    pass


class Field400250(BaseModel):
    description: str
    schema_: Schema1152 = Field(..., alias="schema")


class Field404254(Field400250):
    pass


class Field422113(Field400250):
    pass


class Field429126(Field400250):
    pass


class Schema1156(Items):
    pass


class Field200219(BaseModel):
    description: str
    schema_: Schema1156 = Field(..., alias="schema")


class Responses254(BaseModel):
    field_400: Field400250 = Field(..., alias="400")
    field_404: Field404254 = Field(..., alias="404")
    field_422: Field422113 = Field(..., alias="422")
    field_429: Field429126 = Field(..., alias="429")
    field_200: Field200219 = Field(..., alias="200")
    field_204: Field20439 = Field(..., alias="204")


class Post54(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter234]
    responses: Responses254
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdResetToProduction(BaseModel):
    post: Post54


class Schema1157(Schema):
    pass


class Field400251(BaseModel):
    description: str
    schema_: Schema1157 = Field(..., alias="schema")


class Field404255(Field400251):
    pass


class Field422114(Field400251):
    pass


class Field429127(Field400251):
    pass


class Schema1161(Items):
    pass


class Field200220(BaseModel):
    description: str
    schema_: Schema1161 = Field(..., alias="schema")


class Responses255(BaseModel):
    field_400: Field400251 = Field(..., alias="400")
    field_404: Field404255 = Field(..., alias="404")
    field_422: Field422114 = Field(..., alias="422")
    field_429: Field429127 = Field(..., alias="429")
    field_200: Field200220 = Field(..., alias="200")
    field_204: Field20439 = Field(..., alias="204")


class Post55(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter234]
    responses: Responses255
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdResetToRemote(BaseModel):
    post: Post55


class Schema1162(Schema):
    pass


class Field400252(BaseModel):
    description: str
    schema_: Schema1162 = Field(..., alias="schema")


class Field404256(Field400252):
    pass


class Schema1164(Schema972):
    pass


class Field200221(BaseModel):
    description: str
    schema_: Schema1164 = Field(..., alias="schema")


class Responses256(BaseModel):
    field_400: Field400252 = Field(..., alias="400")
    field_404: Field404256 = Field(..., alias="404")
    field_200: Field200221 = Field(..., alias="200")


class Get116(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter234]
    responses: Responses256
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1165(Schema):
    pass


class Parameter239(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1165 = Field(..., alias="schema")


class Field400253(BaseModel):
    description: str
    schema_: Schema1165 = Field(..., alias="schema")


class Field404257(Field400253):
    pass


class Field40936(Field400253):
    pass


class Field422115(Field400253):
    pass


class Field429128(Field400253):
    pass


class Field200222(Field400253):
    pass


class Responses257(BaseModel):
    field_400: Field400253 = Field(..., alias="400")
    field_404: Field404257 = Field(..., alias="404")
    field_409: Field40936 = Field(..., alias="409")
    field_422: Field422115 = Field(..., alias="422")
    field_429: Field429128 = Field(..., alias="429")
    field_200: Field200222 = Field(..., alias="200")


class Post56(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter239]
    responses: Responses257
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjects(BaseModel):
    get: Get116
    post: Post56


class Parameter240(Parameter2):
    pass


class Field400254(Field400253):
    pass


class Field404258(Field400253):
    pass


class Field200223(Field400253):
    pass


class Responses258(BaseModel):
    field_400: Field400254 = Field(..., alias="400")
    field_404: Field404258 = Field(..., alias="404")
    field_200: Field200223 = Field(..., alias="200")


class Get117(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter240]
    responses: Responses258
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter241(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1165] = Field(None, alias="schema")


class Field400255(Field400253):
    pass


class Field404259(Field400253):
    pass


class Field422116(Field400253):
    pass


class Field429129(Field400253):
    pass


class Field40937(Field400253):
    pass


class Field500(Field400253):
    pass


class Field200224(Field400253):
    pass


class Responses259(BaseModel):
    field_400: Field400255 = Field(..., alias="400")
    field_404: Field404259 = Field(..., alias="404")
    field_422: Field422116 = Field(..., alias="422")
    field_429: Field429129 = Field(..., alias="429")
    field_409: Field40937 = Field(..., alias="409")
    field_500: Field500 = Field(..., alias="500")
    field_200: Field200224 = Field(..., alias="200")


class Patch36(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter241]
    responses: Responses259
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectId(BaseModel):
    get: Get117
    patch: Patch36


class Parameter242(Parameter2):
    pass


class Field400256(Field400253):
    pass


class Field404260(Field400253):
    pass


class Field200225(Field400253):
    pass


class Responses260(BaseModel):
    field_400: Field400256 = Field(..., alias="400")
    field_404: Field404260 = Field(..., alias="404")
    field_200: Field200225 = Field(..., alias="200")


class Get118(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter242]
    responses: Responses260
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdManifest(BaseModel):
    get: Get118


class Field400257(Field400253):
    pass


class Field404261(Field400253):
    pass


class Field422117(Field400253):
    pass


class Field429130(Field400253):
    pass


class Field40938(Field400253):
    pass


class Schema1191(Items):
    pass


class Field200226(BaseModel):
    description: str
    schema_: Schema1191 = Field(..., alias="schema")


class Responses261(BaseModel):
    field_400: Field400257 = Field(..., alias="400")
    field_404: Field404261 = Field(..., alias="404")
    field_422: Field422117 = Field(..., alias="422")
    field_429: Field429130 = Field(..., alias="429")
    field_409: Field40938 = Field(..., alias="409")
    field_200: Field200226 = Field(..., alias="200")


class Post57(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter242]
    responses: Responses261
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1192(Schema):
    pass


class Field400258(BaseModel):
    description: str
    schema_: Schema1192 = Field(..., alias="schema")


class Field404262(Field400258):
    pass


class Schema1194(Items):
    pass


class Field200227(BaseModel):
    description: str
    schema_: Schema1194 = Field(..., alias="schema")


class Responses262(BaseModel):
    field_400: Field400258 = Field(..., alias="400")
    field_404: Field404262 = Field(..., alias="404")
    field_200: Field200227 = Field(..., alias="200")


class Get119(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter242]
    responses: Responses262
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdGitDeployKey(BaseModel):
    post: Post57
    get: Get119


class Schema1195(Schema):
    pass


class Field400259(BaseModel):
    description: str
    schema_: Schema1195 = Field(..., alias="schema")


class Field404263(Field400259):
    pass


class Field422118(Field400259):
    pass


class Field429131(Field400259):
    pass


class Field200228(Field400259):
    pass


class Responses263(BaseModel):
    field_400: Field400259 = Field(..., alias="400")
    field_404: Field404263 = Field(..., alias="404")
    field_422: Field422118 = Field(..., alias="422")
    field_429: Field429131 = Field(..., alias="429")
    field_200: Field200228 = Field(..., alias="200")


class Post58(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter242]
    responses: Responses263
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400260(Field400259):
    pass


class Field404264(Field400259):
    pass


class Field200229(Field400259):
    pass


class Responses264(BaseModel):
    field_400: Field400260 = Field(..., alias="400")
    field_404: Field404264 = Field(..., alias="404")
    field_204: Field20439 = Field(..., alias="204")
    field_200: Field200229 = Field(..., alias="200")


class Get120(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter242]
    responses: Responses264
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdValidate(BaseModel):
    post: Post58
    get: Get120


class Field400261(Field400259):
    pass


class Field404265(Field400259):
    pass


class Field200230(Field400259):
    pass


class Responses265(BaseModel):
    field_400: Field400261 = Field(..., alias="400")
    field_404: Field404265 = Field(..., alias="404")
    field_200: Field200230 = Field(..., alias="200")


class Get121(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter242]
    responses: Responses265
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdCurrentWorkspace(BaseModel):
    get: Get121


class Field400262(Field400259):
    pass


class Field404266(Field400259):
    pass


class Schema1208(Schema972):
    pass


class Field200231(BaseModel):
    description: str
    schema_: Schema1208 = Field(..., alias="schema")


class Responses266(BaseModel):
    field_400: Field400262 = Field(..., alias="400")
    field_404: Field404266 = Field(..., alias="404")
    field_200: Field200231 = Field(..., alias="200")


class Get122(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter242]
    responses: Responses266
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdFiles(BaseModel):
    get: Get122


class Schema1209(Schema):
    pass


class Field400263(BaseModel):
    description: str
    schema_: Schema1209 = Field(..., alias="schema")


class Field404267(Field400263):
    pass


class Field200232(Field400263):
    pass


class Responses267(BaseModel):
    field_400: Field400263 = Field(..., alias="400")
    field_404: Field404267 = Field(..., alias="404")
    field_200: Field200232 = Field(..., alias="200")


class Get123(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter242]
    responses: Responses267
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdFilesFile(BaseModel):
    get: Get123


class Field400264(Field400263):
    pass


class Field404268(Field400263):
    pass


class Schema1214(Schema972):
    pass


class Field200233(BaseModel):
    description: str
    schema_: Schema1214 = Field(..., alias="schema")


class Responses268(BaseModel):
    field_400: Field400264 = Field(..., alias="400")
    field_404: Field404268 = Field(..., alias="404")
    field_200: Field200233 = Field(..., alias="200")


class Get124(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter242]
    responses: Responses268
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdGitConnectionTests(BaseModel):
    get: Get124


class Schema1215(Schema):
    pass


class Field400265(BaseModel):
    description: str
    schema_: Schema1215 = Field(..., alias="schema")


class Field404269(Field400265):
    pass


class Field422119(Field400265):
    pass


class Field429132(Field400265):
    pass


class Field200234(Field400265):
    pass


class Responses269(BaseModel):
    field_400: Field400265 = Field(..., alias="400")
    field_404: Field404269 = Field(..., alias="404")
    field_422: Field422119 = Field(..., alias="422")
    field_429: Field429132 = Field(..., alias="429")
    field_200: Field200234 = Field(..., alias="200")


class Get125(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter242]
    responses: Responses269
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdGitConnectionTestsTestId(BaseModel):
    get: Get125


class Field400266(Field400265):
    pass


class Field404270(Field400265):
    pass


class Schema1222(Schema972):
    pass


class Field200235(BaseModel):
    description: str
    schema_: Schema1222 = Field(..., alias="schema")


class Responses270(BaseModel):
    field_400: Field400266 = Field(..., alias="400")
    field_404: Field404270 = Field(..., alias="404")
    field_200: Field200235 = Field(..., alias="200")


class Get126(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter242]
    responses: Responses270
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdLookmlTests(BaseModel):
    get: Get126


class Schema1223(Schema):
    pass


class Field400267(BaseModel):
    description: str
    schema_: Schema1223 = Field(..., alias="schema")


class Field404271(Field400267):
    pass


class Field422120(Field400267):
    pass


class Field429133(Field400267):
    pass


class Schema1227(Schema972):
    pass


class Field200236(BaseModel):
    description: str
    schema_: Schema1227 = Field(..., alias="schema")


class Responses271(BaseModel):
    field_400: Field400267 = Field(..., alias="400")
    field_404: Field404271 = Field(..., alias="404")
    field_422: Field422120 = Field(..., alias="422")
    field_429: Field429133 = Field(..., alias="429")
    field_200: Field200236 = Field(..., alias="200")


class Get127(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter242]
    responses: Responses271
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdLookmlTestsRun(BaseModel):
    get: Get127


class Schema1228(Schema):
    pass


class Parameter254(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1228] = Field(None, alias="schema")


class Field400268(BaseModel):
    description: str
    schema_: Schema1228 = Field(..., alias="schema")


class Field404272(Field400268):
    pass


class Field40939(Field400268):
    pass


class Field422121(Field400268):
    pass


class Field429134(Field400268):
    pass


class Field200237(Field400268):
    pass


class Responses272(BaseModel):
    field_400: Field400268 = Field(..., alias="400")
    field_404: Field404272 = Field(..., alias="404")
    field_409: Field40939 = Field(..., alias="409")
    field_422: Field422121 = Field(..., alias="422")
    field_429: Field429134 = Field(..., alias="429")
    field_200: Field200237 = Field(..., alias="200")
    field_204: Field20439 = Field(..., alias="204")


class Post59(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter254]
    responses: Responses272
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsProjectIdTag(BaseModel):
    post: Post59


class Parameter255(Parameter7):
    pass


class Field400269(Field400268):
    pass


class Field404273(Field400268):
    pass


class Field40940(Field400268):
    pass


class Field422122(Field400268):
    pass


class Field429135(Field400268):
    pass


class Field200238(Field400268):
    pass


class Responses273(BaseModel):
    field_400: Field400269 = Field(..., alias="400")
    field_404: Field404273 = Field(..., alias="404")
    field_409: Field40940 = Field(..., alias="409")
    field_422: Field422122 = Field(..., alias="422")
    field_429: Field429135 = Field(..., alias="429")
    field_200: Field200238 = Field(..., alias="200")


class Post60(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter255]
    responses: Responses273
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRenderTasksLooksLookIdResultFormat(BaseModel):
    post: Post60


class Field400270(Field400268):
    pass


class Field404274(Field400268):
    pass


class Field40941(Field400268):
    pass


class Field422123(Field400268):
    pass


class Field429136(Field400268):
    pass


class Field200239(Field400268):
    pass


class Responses274(BaseModel):
    field_400: Field400270 = Field(..., alias="400")
    field_404: Field404274 = Field(..., alias="404")
    field_409: Field40941 = Field(..., alias="409")
    field_422: Field422123 = Field(..., alias="422")
    field_429: Field429136 = Field(..., alias="429")
    field_200: Field200239 = Field(..., alias="200")


class Post61(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter255]
    responses: Responses274
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRenderTasksQueriesQueryIdResultFormat(BaseModel):
    post: Post61


class Parameter257(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1228] = Field(None, alias="schema")
    format: Optional[str] = None


class Field400271(Field400268):
    pass


class Field404275(Field400268):
    pass


class Field40942(Field400268):
    pass


class Field422124(Field400268):
    pass


class Field429137(Field400268):
    pass


class Field200240(Field400268):
    pass


class Responses275(BaseModel):
    field_400: Field400271 = Field(..., alias="400")
    field_404: Field404275 = Field(..., alias="404")
    field_409: Field40942 = Field(..., alias="409")
    field_422: Field422124 = Field(..., alias="422")
    field_429: Field429137 = Field(..., alias="429")
    field_200: Field200240 = Field(..., alias="200")


class Post62(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter257]
    responses: Responses275
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRenderTasksDashboardsDashboardIdResultFormat(BaseModel):
    post: Post62


class Parameter258(Parameter2):
    pass


class Field400272(Field400268):
    pass


class Field404276(Field400268):
    pass


class Field200241(Field400268):
    pass


class Responses276(BaseModel):
    field_400: Field400272 = Field(..., alias="400")
    field_404: Field404276 = Field(..., alias="404")
    field_200: Field200241 = Field(..., alias="200")


class Get128(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter258]
    responses: Responses276
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRenderTasksRenderTaskId(BaseModel):
    get: Get128


class Field400273(Field400268):
    pass


class Field404277(Field400268):
    pass


class Field202(Field2042):
    pass


class Schema1259(Items):
    pass


class Field200242(BaseModel):
    description: str
    schema_: Schema1259 = Field(..., alias="schema")


class Responses277(BaseModel):
    field_400: Field400273 = Field(..., alias="400")
    field_404: Field404277 = Field(..., alias="404")
    field_202: Field202 = Field(..., alias="202")
    field_200: Field200242 = Field(..., alias="200")


class Get129(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter258]
    responses: Responses277
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRenderTasksRenderTaskIdResults(BaseModel):
    get: Get129


class Parameter260(Parameter7):
    pass


class Schema1260(Schema):
    pass


class Field400274(BaseModel):
    description: str
    schema_: Schema1260 = Field(..., alias="schema")


class Field404278(Field400274):
    pass


class Field40943(Field400274):
    pass


class Field422125(Field400274):
    pass


class Field429138(Field400274):
    pass


class Field200243(Field400274):
    pass


class Responses278(BaseModel):
    field_400: Field400274 = Field(..., alias="400")
    field_404: Field404278 = Field(..., alias="404")
    field_409: Field40943 = Field(..., alias="409")
    field_422: Field422125 = Field(..., alias="422")
    field_429: Field429138 = Field(..., alias="429")
    field_200: Field200243 = Field(..., alias="200")


class Post63(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter260]
    responses: Responses278
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRenderTasksDashboardElementsDashboardElementIdResultFormat(BaseModel):
    post: Post63


class Parameter261(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1260] = Field(None, alias="schema")


class Field400275(Field400274):
    pass


class Field404279(Field400274):
    pass


class Field40944(Field400274):
    pass


class Field422126(Field400274):
    pass


class Field429139(Field400274):
    pass


class Field200244(Field400274):
    pass


class Responses279(BaseModel):
    field_400: Field400275 = Field(..., alias="400")
    field_404: Field404279 = Field(..., alias="404")
    field_409: Field40944 = Field(..., alias="409")
    field_422: Field422126 = Field(..., alias="422")
    field_429: Field429139 = Field(..., alias="429")
    field_200: Field200244 = Field(..., alias="200")


class Put14(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter261]
    responses: Responses279
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter262(Parameter2):
    pass


class Field400276(Field400274):
    pass


class Field404280(Field400274):
    pass


class Field429140(Field400274):
    pass


class Schema1276(Items):
    pass


class Field20445(BaseModel):
    description: str
    schema_: Schema1276 = Field(..., alias="schema")


class Responses280(BaseModel):
    field_400: Field400276 = Field(..., alias="400")
    field_404: Field404280 = Field(..., alias="404")
    field_429: Field429140 = Field(..., alias="429")
    field_204: Field20445 = Field(..., alias="204")


class Delete34(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter262]
    responses: Responses280
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsRootProjectIdCredentialCredentialId(BaseModel):
    put: Put14
    delete: Delete34


class Schema1277(Schema):
    pass


class Field400277(BaseModel):
    description: str
    schema_: Schema1277 = Field(..., alias="schema")


class Field404281(Field400277):
    pass


class Schema1279(Schema972):
    pass


class Field200245(BaseModel):
    description: str
    schema_: Schema1279 = Field(..., alias="schema")


class Responses281(BaseModel):
    field_400: Field400277 = Field(..., alias="400")
    field_404: Field404281 = Field(..., alias="404")
    field_200: Field200245 = Field(..., alias="200")


class Get130(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter262]
    responses: Responses281
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldProjectsRootProjectIdCredentials(BaseModel):
    get: Get130


class Items70(Items):
    pass


class Parameter264(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str
    items: Optional[Items70] = None
    collectionFormat: Optional[str] = None


class Schema1280(Schema):
    pass


class Field400278(BaseModel):
    description: str
    schema_: Schema1280 = Field(..., alias="schema")


class Field404282(Field400278):
    pass


class Items71(Schema):
    pass


class Schema1282(BaseModel):
    type: str
    items: Items71


class Field200246(BaseModel):
    description: str
    schema_: Schema1282 = Field(..., alias="schema")


class Responses282(BaseModel):
    field_400: Field400278 = Field(..., alias="400")
    field_404: Field404282 = Field(..., alias="404")
    field_200: Field200246 = Field(..., alias="200")


class Get131(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter264]
    responses: Responses282
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1283(Schema):
    pass


class Parameter265(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1283 = Field(..., alias="schema")


class Field400279(BaseModel):
    description: str
    schema_: Schema1283 = Field(..., alias="schema")


class Field404283(Field400279):
    pass


class Field40945(Field400279):
    pass


class Field422127(Field400279):
    pass


class Field429141(Field400279):
    pass


class Field200247(Field400279):
    pass


class Responses283(BaseModel):
    field_400: Field400279 = Field(..., alias="400")
    field_404: Field404283 = Field(..., alias="404")
    field_409: Field40945 = Field(..., alias="409")
    field_422: Field422127 = Field(..., alias="422")
    field_429: Field429141 = Field(..., alias="429")
    field_200: Field200247 = Field(..., alias="200")


class Post64(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter265]
    responses: Responses283
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRoles(BaseModel):
    get: Get131
    post: Post64


class Parameter266(Parameter7):
    pass


class Field400280(Field400279):
    pass


class Field404284(Field400279):
    pass


class Schema1292(Schema1282):
    pass


class Field200248(BaseModel):
    description: str
    schema_: Schema1292 = Field(..., alias="schema")


class Responses284(BaseModel):
    field_400: Field400280 = Field(..., alias="400")
    field_404: Field404284 = Field(..., alias="404")
    field_200: Field200248 = Field(..., alias="200")


class Get132(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter266]
    responses: Responses284
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRolesSearch(BaseModel):
    get: Get132


class Schema1293(Schema):
    pass


class Field400281(BaseModel):
    description: str
    schema_: Schema1293 = Field(..., alias="schema")


class Field404285(Field400281):
    pass


class Schema1295(Schema1282):
    pass


class Field200249(BaseModel):
    description: str
    schema_: Schema1295 = Field(..., alias="schema")


class Responses285(BaseModel):
    field_400: Field400281 = Field(..., alias="400")
    field_404: Field404285 = Field(..., alias="404")
    field_200: Field200249 = Field(..., alias="200")


class Get133(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter266]
    responses: Responses285
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRolesSearchWithUserCount(BaseModel):
    get: Get133


class Parameter268(Parameter2):
    pass


class Schema1296(Schema):
    pass


class Field400282(BaseModel):
    description: str
    schema_: Schema1296 = Field(..., alias="schema")


class Field404286(Field400282):
    pass


class Field200250(Field400282):
    pass


class Responses286(BaseModel):
    field_400: Field400282 = Field(..., alias="400")
    field_404: Field404286 = Field(..., alias="404")
    field_200: Field200250 = Field(..., alias="200")


class Get134(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter268]
    responses: Responses286
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400283(Field400282):
    pass


class Field404287(Field400282):
    pass


class Field429142(Field400282):
    pass


class Field4057(Field400282):
    pass


class Schema1303(Items):
    pass


class Field20446(BaseModel):
    description: str
    schema_: Schema1303 = Field(..., alias="schema")


class Responses287(BaseModel):
    field_400: Field400283 = Field(..., alias="400")
    field_404: Field404287 = Field(..., alias="404")
    field_429: Field429142 = Field(..., alias="429")
    field_405: Field4057 = Field(..., alias="405")
    field_204: Field20446 = Field(..., alias="204")


class Delete35(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter268]
    responses: Responses287
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1304(Schema):
    pass


class Parameter270(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1304] = Field(None, alias="schema")


class Field400284(BaseModel):
    description: str
    schema_: Schema1304 = Field(..., alias="schema")


class Field404288(Field400284):
    pass


class Field422128(Field400284):
    pass


class Field429143(Field400284):
    pass


class Field4058(Field400284):
    pass


class Field200251(Field400284):
    pass


class Responses288(BaseModel):
    field_400: Field400284 = Field(..., alias="400")
    field_404: Field404288 = Field(..., alias="404")
    field_422: Field422128 = Field(..., alias="422")
    field_429: Field429143 = Field(..., alias="429")
    field_405: Field4058 = Field(..., alias="405")
    field_200: Field200251 = Field(..., alias="200")


class Patch37(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter270]
    responses: Responses288
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRolesRoleId(BaseModel):
    get: Get134
    delete: Delete35
    patch: Patch37


class Parameter271(Parameter2):
    pass


class Field400285(Field400284):
    pass


class Field404289(Field400284):
    pass


class Schema1313(Schema1282):
    pass


class Field200252(BaseModel):
    description: str
    schema_: Schema1313 = Field(..., alias="schema")


class Responses289(BaseModel):
    field_400: Field400285 = Field(..., alias="400")
    field_404: Field404289 = Field(..., alias="404")
    field_200: Field200252 = Field(..., alias="200")


class Get135(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter271]
    responses: Responses289
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Items75(Items):
    pass


class Schema1314(BaseModel):
    type: str
    items: Items75


class Parameter272(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1314] = Field(None, alias="schema")


class Schema1315(Schema):
    pass


class Field400286(BaseModel):
    description: str
    schema_: Schema1315 = Field(..., alias="schema")


class Field404290(Field400286):
    pass


class Field422129(Field400286):
    pass


class Field429144(Field400286):
    pass


class Field40332(Field400286):
    pass


class Items76(Schema):
    pass


class Schema1320(BaseModel):
    type: str
    items: Items76


class Field200253(BaseModel):
    description: str
    schema_: Schema1320 = Field(..., alias="schema")


class Responses290(BaseModel):
    field_400: Field400286 = Field(..., alias="400")
    field_404: Field404290 = Field(..., alias="404")
    field_422: Field422129 = Field(..., alias="422")
    field_429: Field429144 = Field(..., alias="429")
    field_403: Field40332 = Field(..., alias="403")
    field_200: Field200253 = Field(..., alias="200")


class Put15(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter272]
    responses: Responses290
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRolesRoleIdGroups(BaseModel):
    get: Get135
    put: Put15


class Parameter273(Parameter2):
    pass


class Schema1321(Schema):
    pass


class Field400287(BaseModel):
    description: str
    schema_: Schema1321 = Field(..., alias="schema")


class Field404291(Field400287):
    pass


class Schema1323(Schema1320):
    pass


class Field200254(BaseModel):
    description: str
    schema_: Schema1323 = Field(..., alias="schema")


class Responses291(BaseModel):
    field_400: Field400287 = Field(..., alias="400")
    field_404: Field404291 = Field(..., alias="404")
    field_200: Field200254 = Field(..., alias="200")


class Get136(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter273]
    responses: Responses291
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Items78(Items):
    pass


class Schema1324(BaseModel):
    type: str
    items: Items78


class Parameter274(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1324] = Field(None, alias="schema")


class Schema1325(Schema):
    pass


class Field400288(BaseModel):
    description: str
    schema_: Schema1325 = Field(..., alias="schema")


class Field404292(Field400288):
    pass


class Field422130(Field400288):
    pass


class Field429145(Field400288):
    pass


class Field40333(Field400288):
    pass


class Field4059(Field400288):
    pass


class Items79(Schema):
    pass


class Schema1331(BaseModel):
    type: str
    items: Items79


class Field200255(BaseModel):
    description: str
    schema_: Schema1331 = Field(..., alias="schema")


class Responses292(BaseModel):
    field_400: Field400288 = Field(..., alias="400")
    field_404: Field404292 = Field(..., alias="404")
    field_422: Field422130 = Field(..., alias="422")
    field_429: Field429145 = Field(..., alias="429")
    field_403: Field40333 = Field(..., alias="403")
    field_405: Field4059 = Field(..., alias="405")
    field_200: Field200255 = Field(..., alias="200")


class Put16(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter274]
    responses: Responses292
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRolesRoleIdUsers(BaseModel):
    get: Get136
    put: Put16


class Schema1332(Schema):
    pass


class Field404293(BaseModel):
    description: str
    schema_: Schema1332 = Field(..., alias="schema")


class Schema1333(Schema1331):
    pass


class Field200256(BaseModel):
    description: str
    schema_: Schema1333 = Field(..., alias="schema")


class Responses293(BaseModel):
    field_404: Field404293 = Field(..., alias="404")
    field_200: Field200256 = Field(..., alias="200")


class Get137(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses293
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRunningQueries(BaseModel):
    get: Get137


class Parameter275(Parameter2):
    pass


class Schema1334(Schema):
    pass


class Field400289(BaseModel):
    description: str
    schema_: Schema1334 = Field(..., alias="schema")


class Field404294(Field400289):
    pass


class Field422131(Field400289):
    pass


class Schema1337(Items):
    pass


class Field20447(BaseModel):
    description: str
    schema_: Schema1337 = Field(..., alias="schema")


class Responses294(BaseModel):
    field_400: Field400289 = Field(..., alias="400")
    field_404: Field404294 = Field(..., alias="404")
    field_422: Field422131 = Field(..., alias="422")
    field_204: Field20447 = Field(..., alias="204")


class Delete36(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter275]
    responses: Responses294
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldRunningQueriesQueryTaskId(BaseModel):
    delete: Delete36


class Schema1338(Schema):
    pass


class Field40334(BaseModel):
    description: str
    schema_: Schema1338 = Field(..., alias="schema")


class Field404295(Field40334):
    pass


class Field200257(Field40334):
    pass


class Responses295(BaseModel):
    field_403: Field40334 = Field(..., alias="403")
    field_404: Field404295 = Field(..., alias="404")
    field_200: Field200257 = Field(..., alias="200")


class Get138(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses295
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter276(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1338 = Field(..., alias="schema")


class Field400290(Field40334):
    pass


class Field40335(Field40334):
    pass


class Field404296(Field40334):
    pass


class Field422132(Field40334):
    pass


class Field200258(Field40334):
    pass


class Responses296(BaseModel):
    field_400: Field400290 = Field(..., alias="400")
    field_403: Field40335 = Field(..., alias="403")
    field_404: Field404296 = Field(..., alias="404")
    field_422: Field422132 = Field(..., alias="422")
    field_200: Field200258 = Field(..., alias="200")


class Patch38(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter276]
    responses: Responses296
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSamlConfig(BaseModel):
    get: Get138
    patch: Patch38


class Parameter277(Parameter2):
    pass


class Field40336(Field40334):
    pass


class Field404297(Field40334):
    pass


class Field200259(Field40334):
    pass


class Responses297(BaseModel):
    field_403: Field40336 = Field(..., alias="403")
    field_404: Field404297 = Field(..., alias="404")
    field_200: Field200259 = Field(..., alias="200")


class Get139(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter277]
    responses: Responses297
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400291(Field40334):
    pass


class Field40337(Field40334):
    pass


class Field404298(Field40334):
    pass


class Schema1353(Items):
    pass


class Field20448(BaseModel):
    description: str
    schema_: Schema1353 = Field(..., alias="schema")


class Responses298(BaseModel):
    field_400: Field400291 = Field(..., alias="400")
    field_403: Field40337 = Field(..., alias="403")
    field_404: Field404298 = Field(..., alias="404")
    field_204: Field20448 = Field(..., alias="204")


class Delete37(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter277]
    responses: Responses298
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSamlTestConfigsTestSlug(BaseModel):
    get: Get139
    delete: Delete37


class Schema1354(Schema):
    pass


class Parameter279(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1354 = Field(..., alias="schema")


class Field400292(BaseModel):
    description: str
    schema_: Schema1354 = Field(..., alias="schema")


class Field40338(Field400292):
    pass


class Field404299(Field400292):
    pass


class Field422133(Field400292):
    pass


class Field200260(Field400292):
    pass


class Responses299(BaseModel):
    field_400: Field400292 = Field(..., alias="400")
    field_403: Field40338 = Field(..., alias="403")
    field_404: Field404299 = Field(..., alias="404")
    field_422: Field422133 = Field(..., alias="422")
    field_200: Field200260 = Field(..., alias="200")


class Post65(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter279]
    responses: Responses299
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSamlTestConfigs(BaseModel):
    post: Post65


class Schema1360(Items):
    pass


class Parameter280(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1360 = Field(..., alias="schema")


class Schema1361(Schema):
    pass


class Field400293(BaseModel):
    description: str
    schema_: Schema1361 = Field(..., alias="schema")


class Field40339(Field400293):
    pass


class Field404300(Field400293):
    pass


class Field200261(Field400293):
    pass


class Responses300(BaseModel):
    field_400: Field400293 = Field(..., alias="400")
    field_403: Field40339 = Field(..., alias="403")
    field_404: Field404300 = Field(..., alias="404")
    field_200: Field200261 = Field(..., alias="200")


class Post66(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    consumes: List[str]
    parameters: List[Parameter280]
    responses: Responses300
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldParseSamlIdpMetadata(BaseModel):
    post: Post66


class Schema1365(Items):
    pass


class Parameter281(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1365 = Field(..., alias="schema")


class Schema1366(Schema):
    pass


class Field400294(BaseModel):
    description: str
    schema_: Schema1366 = Field(..., alias="schema")


class Field40340(Field400294):
    pass


class Field404301(Field400294):
    pass


class Field200262(Field400294):
    pass


class Responses301(BaseModel):
    field_400: Field400294 = Field(..., alias="400")
    field_403: Field40340 = Field(..., alias="403")
    field_404: Field404301 = Field(..., alias="404")
    field_200: Field200262 = Field(..., alias="200")


class Post67(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    consumes: List[str]
    parameters: List[Parameter281]
    responses: Responses301
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldFetchAndParseSamlIdpMetadata(BaseModel):
    post: Post67


class Parameter282(Parameter2):
    pass


class Field400295(Field400294):
    pass


class Field404302(Field400294):
    pass


class Schema1372(Schema1331):
    pass


class Field200263(BaseModel):
    description: str
    schema_: Schema1372 = Field(..., alias="schema")


class Responses302(BaseModel):
    field_400: Field400295 = Field(..., alias="400")
    field_404: Field404302 = Field(..., alias="404")
    field_200: Field200263 = Field(..., alias="200")


class Get140(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter282]
    responses: Responses302
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldScheduledPlansSpaceSpaceId(BaseModel):
    get: Get140


class Schema1373(Schema):
    pass


class Field400296(BaseModel):
    description: str
    schema_: Schema1373 = Field(..., alias="schema")


class Field404303(Field400296):
    pass


class Field429146(Field400296):
    pass


class Schema1376(Items):
    pass


class Field20449(BaseModel):
    description: str
    schema_: Schema1376 = Field(..., alias="schema")


class Responses303(BaseModel):
    field_400: Field400296 = Field(..., alias="400")
    field_404: Field404303 = Field(..., alias="404")
    field_429: Field429146 = Field(..., alias="429")
    field_204: Field20449 = Field(..., alias="204")


class Delete38(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter282]
    responses: Responses303
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1377(Schema):
    pass


class Parameter284(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1377] = Field(None, alias="schema")


class Field400297(BaseModel):
    description: str
    schema_: Schema1377 = Field(..., alias="schema")


class Field404304(Field400297):
    pass


class Field422134(Field400297):
    pass


class Field429147(Field400297):
    pass


class Field200264(Field400297):
    pass


class Responses304(BaseModel):
    field_400: Field400297 = Field(..., alias="400")
    field_404: Field404304 = Field(..., alias="404")
    field_422: Field422134 = Field(..., alias="422")
    field_429: Field429147 = Field(..., alias="429")
    field_200: Field200264 = Field(..., alias="200")


class Patch39(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter284]
    responses: Responses304
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter285(Parameter2):
    pass


class Field400298(Field400297):
    pass


class Field404305(Field400297):
    pass


class Field200265(Field400297):
    pass


class Responses305(BaseModel):
    field_400: Field400298 = Field(..., alias="400")
    field_404: Field404305 = Field(..., alias="404")
    field_200: Field200265 = Field(..., alias="200")


class Get141(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter285]
    responses: Responses305
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldScheduledPlansScheduledPlanId(BaseModel):
    delete: Delete38
    patch: Patch39
    get: Get141


class Parameter286(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1377 = Field(..., alias="schema")


class Field400299(Field400297):
    pass


class Field404306(Field400297):
    pass


class Field40946(Field400297):
    pass


class Field422135(Field400297):
    pass


class Field429148(Field400297):
    pass


class Field200266(Field400297):
    pass


class Responses306(BaseModel):
    field_400: Field400299 = Field(..., alias="400")
    field_404: Field404306 = Field(..., alias="404")
    field_409: Field40946 = Field(..., alias="409")
    field_422: Field422135 = Field(..., alias="422")
    field_429: Field429148 = Field(..., alias="429")
    field_200: Field200266 = Field(..., alias="200")


class Post68(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter286]
    responses: Responses306
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter287(Parameter2):
    pass


class Field400300(Field400297):
    pass


class Field404307(Field400297):
    pass


class Field422136(Field400297):
    pass


class Schema1396(Schema1331):
    pass


class Field200267(BaseModel):
    description: str
    schema_: Schema1396 = Field(..., alias="schema")


class Responses307(BaseModel):
    field_400: Field400300 = Field(..., alias="400")
    field_404: Field404307 = Field(..., alias="404")
    field_422: Field422136 = Field(..., alias="422")
    field_200: Field200267 = Field(..., alias="200")


class Get142(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter287]
    responses: Responses307
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldScheduledPlans(BaseModel):
    post: Post68
    get: Get142


class Schema1397(Schema):
    pass


class Parameter288(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1397 = Field(..., alias="schema")


class Field400301(BaseModel):
    description: str
    schema_: Schema1397 = Field(..., alias="schema")


class Field404308(Field400301):
    pass


class Field40947(Field400301):
    pass


class Field422137(Field400301):
    pass


class Field429149(Field400301):
    pass


class Field200268(Field400301):
    pass


class Responses308(BaseModel):
    field_400: Field400301 = Field(..., alias="400")
    field_404: Field404308 = Field(..., alias="404")
    field_409: Field40947 = Field(..., alias="409")
    field_422: Field422137 = Field(..., alias="422")
    field_429: Field429149 = Field(..., alias="429")
    field_200: Field200268 = Field(..., alias="200")


class Post69(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter288]
    responses: Responses308
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldScheduledPlansRunOnce(BaseModel):
    post: Post69


class Parameter289(Parameter2):
    pass


class Field400302(Field400301):
    pass


class Field404309(Field400301):
    pass


class Schema1406(Schema1331):
    pass


class Field200269(BaseModel):
    description: str
    schema_: Schema1406 = Field(..., alias="schema")


class Responses309(BaseModel):
    field_400: Field400302 = Field(..., alias="400")
    field_404: Field404309 = Field(..., alias="404")
    field_200: Field200269 = Field(..., alias="200")


class Get143(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter289]
    responses: Responses309
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldScheduledPlansLookLookId(BaseModel):
    get: Get143


class Schema1407(Schema):
    pass


class Field400303(BaseModel):
    description: str
    schema_: Schema1407 = Field(..., alias="schema")


class Field404310(Field400303):
    pass


class Schema1409(Schema1331):
    pass


class Field200270(BaseModel):
    description: str
    schema_: Schema1409 = Field(..., alias="schema")


class Responses310(BaseModel):
    field_400: Field400303 = Field(..., alias="400")
    field_404: Field404310 = Field(..., alias="404")
    field_200: Field200270 = Field(..., alias="200")


class Get144(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter289]
    responses: Responses310
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldScheduledPlansDashboardDashboardId(BaseModel):
    get: Get144


class Schema1410(Schema):
    pass


class Field400304(BaseModel):
    description: str
    schema_: Schema1410 = Field(..., alias="schema")


class Field404311(Field400304):
    pass


class Schema1412(Schema1331):
    pass


class Field200271(BaseModel):
    description: str
    schema_: Schema1412 = Field(..., alias="schema")


class Responses311(BaseModel):
    field_400: Field400304 = Field(..., alias="400")
    field_404: Field404311 = Field(..., alias="404")
    field_200: Field200271 = Field(..., alias="200")


class Get145(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter289]
    responses: Responses311
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldScheduledPlansLookmlDashboardLookmlDashboardId(BaseModel):
    get: Get145


class Schema1413(Schema):
    pass


class Parameter292(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1413] = Field(None, alias="schema")


class Field400305(BaseModel):
    description: str
    schema_: Schema1413 = Field(..., alias="schema")


class Field404312(Field400305):
    pass


class Field40948(Field400305):
    pass


class Field422138(Field400305):
    pass


class Field429150(Field400305):
    pass


class Field200272(Field400305):
    pass


class Responses312(BaseModel):
    field_400: Field400305 = Field(..., alias="400")
    field_404: Field404312 = Field(..., alias="404")
    field_409: Field40948 = Field(..., alias="409")
    field_422: Field422138 = Field(..., alias="422")
    field_429: Field429150 = Field(..., alias="429")
    field_200: Field200272 = Field(..., alias="200")


class Post70(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter292]
    responses: Responses312
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldScheduledPlansScheduledPlanIdRunOnce(BaseModel):
    post: Post70


class Field400306(Field400305):
    pass


class Field404313(Field400305):
    pass


class Field200273(Field400305):
    pass


class Responses313(BaseModel):
    field_400: Field400306 = Field(..., alias="400")
    field_404: Field404313 = Field(..., alias="404")
    field_200: Field200273 = Field(..., alias="200")


class Get146(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses313
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter293(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1413 = Field(..., alias="schema")


class Field400307(Field400305):
    pass


class Field404314(Field400305):
    pass


class Field422139(Field400305):
    pass


class Field429151(Field400305):
    pass


class Field200274(Field400305):
    pass


class Responses314(BaseModel):
    field_400: Field400307 = Field(..., alias="400")
    field_404: Field404314 = Field(..., alias="404")
    field_422: Field422139 = Field(..., alias="422")
    field_429: Field429151 = Field(..., alias="429")
    field_200: Field200274 = Field(..., alias="200")


class Patch40(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter293]
    responses: Responses314
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSessionConfig(BaseModel):
    get: Get146
    patch: Patch40


class Field400308(Field400305):
    pass


class Field404315(Field400305):
    pass


class Field200275(Field400305):
    pass


class Responses315(BaseModel):
    field_400: Field400308 = Field(..., alias="400")
    field_404: Field404315 = Field(..., alias="404")
    field_200: Field200275 = Field(..., alias="200")


class Get147(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses315
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter294(Parameter293):
    pass


class Field400309(Field400305):
    pass


class Field404316(Field400305):
    pass


class Field422140(Field400305):
    pass


class Field429152(Field400305):
    pass


class Field200276(Field400305):
    pass


class Responses316(BaseModel):
    field_400: Field400309 = Field(..., alias="400")
    field_404: Field404316 = Field(..., alias="404")
    field_422: Field422140 = Field(..., alias="422")
    field_429: Field429152 = Field(..., alias="429")
    field_200: Field200276 = Field(..., alias="200")


class Patch41(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter294]
    responses: Responses316
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSession(BaseModel):
    get: Get147
    patch: Patch41


class Parameter295(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema1413] = Field(None, alias="schema")
    type: Optional[str] = None


class Field400310(Field400305):
    pass


class Field404317(Field400305):
    pass


class Field422141(Field400305):
    pass


class Field429153(Field400305):
    pass


class Field200277(Field400305):
    pass


class Field40341(Field400305):
    pass


class Field40949(Field400305):
    pass


class Responses317(BaseModel):
    field_400: Field400310 = Field(..., alias="400")
    field_404: Field404317 = Field(..., alias="404")
    field_422: Field422141 = Field(..., alias="422")
    field_429: Field429153 = Field(..., alias="429")
    field_200: Field200277 = Field(..., alias="200")
    field_403: Field40341 = Field(..., alias="403")
    field_409: Field40949 = Field(..., alias="409")


class Patch42(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter295]
    responses: Responses317
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter296(Parameter2):
    pass


class Field400311(Field400305):
    pass


class Field404318(Field400305):
    pass


class Field200278(Field400305):
    pass


class Field422142(Field400305):
    pass


class Responses318(BaseModel):
    field_400: Field400311 = Field(..., alias="400")
    field_404: Field404318 = Field(..., alias="404")
    field_200: Field200278 = Field(..., alias="200")
    field_422: Field422142 = Field(..., alias="422")


class Get148(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter296]
    responses: Responses318
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSetting(BaseModel):
    patch: Patch42
    get: Get148


class Parameter297(Parameter293):
    pass


class Field400312(Field400305):
    pass


class Field404319(Field400305):
    pass


class Field40950(Field400305):
    pass


class Field422143(Field400305):
    pass


class Field429154(Field400305):
    pass


class Field40342(Field400305):
    pass


class Field20450(Field2042):
    pass


class Responses319(BaseModel):
    field_400: Field400312 = Field(..., alias="400")
    field_404: Field404319 = Field(..., alias="404")
    field_409: Field40950 = Field(..., alias="409")
    field_422: Field422143 = Field(..., alias="422")
    field_429: Field429154 = Field(..., alias="429")
    field_403: Field40342 = Field(..., alias="403")
    field_204: Field20450 = Field(..., alias="204")


class Post71(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter297]
    responses: Responses319
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")
    x_looker_rate_limited: bool = Field(..., alias="x-looker-rate-limited")


class FieldSmtpSettings(BaseModel):
    post: Post71


class Parameter298(Parameter2):
    pass


class Field400313(Field400305):
    pass


class Field404320(Field400305):
    pass


class Field200279(Field400305):
    pass


class Responses320(BaseModel):
    field_400: Field400313 = Field(..., alias="400")
    field_404: Field404320 = Field(..., alias="404")
    field_200: Field200279 = Field(..., alias="200")


class Get149(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter298]
    responses: Responses320
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSmtpStatus(BaseModel):
    get: Get149


class Parameter299(Parameter60):
    pass


class Field400314(Field400305):
    pass


class Field404321(Field400305):
    pass


class Schema1462(Schema1331):
    pass


class Field200280(BaseModel):
    description: str
    schema_: Schema1462 = Field(..., alias="schema")


class Responses321(BaseModel):
    field_400: Field400314 = Field(..., alias="400")
    field_404: Field404321 = Field(..., alias="404")
    field_200: Field200280 = Field(..., alias="200")


class Get150(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter299]
    responses: Responses321
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldFoldersSearch(BaseModel):
    get: Get150


class Parameter300(Parameter2):
    pass


class Schema1463(Schema):
    pass


class Field400315(BaseModel):
    description: str
    schema_: Schema1463 = Field(..., alias="schema")


class Field404322(Field400315):
    pass


class Field200281(Field400315):
    pass


class Responses322(BaseModel):
    field_400: Field400315 = Field(..., alias="400")
    field_404: Field404322 = Field(..., alias="404")
    field_200: Field200281 = Field(..., alias="200")


class Get151(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter300]
    responses: Responses322
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400316(Field400315):
    pass


class Field404323(Field400315):
    pass


class Field429155(Field400315):
    pass


class Schema1469(Items):
    pass


class Field20451(BaseModel):
    description: str
    schema_: Schema1469 = Field(..., alias="schema")


class Responses323(BaseModel):
    field_400: Field400316 = Field(..., alias="400")
    field_404: Field404323 = Field(..., alias="404")
    field_429: Field429155 = Field(..., alias="429")
    field_204: Field20451 = Field(..., alias="204")


class Delete39(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter300]
    responses: Responses323
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1470(Schema):
    pass


class Parameter302(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1470] = Field(None, alias="schema")


class Field400317(BaseModel):
    description: str
    schema_: Schema1470 = Field(..., alias="schema")


class Field404324(Field400317):
    pass


class Field422144(Field400317):
    pass


class Field429156(Field400317):
    pass


class Field200282(Field400317):
    pass


class Responses324(BaseModel):
    field_400: Field400317 = Field(..., alias="400")
    field_404: Field404324 = Field(..., alias="404")
    field_422: Field422144 = Field(..., alias="422")
    field_429: Field429156 = Field(..., alias="429")
    field_200: Field200282 = Field(..., alias="200")


class Patch43(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter302]
    responses: Responses324
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldFoldersFolderId(BaseModel):
    get: Get151
    delete: Delete39
    patch: Patch43


class Parameter303(Parameter2):
    pass


class Field400318(Field400317):
    pass


class Field404325(Field400317):
    pass


class Schema1478(Schema1331):
    pass


class Field200283(BaseModel):
    description: str
    schema_: Schema1478 = Field(..., alias="schema")


class Responses325(BaseModel):
    field_400: Field400318 = Field(..., alias="400")
    field_404: Field404325 = Field(..., alias="404")
    field_200: Field200283 = Field(..., alias="200")


class Get152(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter303]
    responses: Responses325
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1479(Schema):
    pass


class Parameter304(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1479 = Field(..., alias="schema")


class Field400319(BaseModel):
    description: str
    schema_: Schema1479 = Field(..., alias="schema")


class Field404326(Field400319):
    pass


class Field40951(Field400319):
    pass


class Field422145(Field400319):
    pass


class Field429157(Field400319):
    pass


class Field200284(Field400319):
    pass


class Responses326(BaseModel):
    field_400: Field400319 = Field(..., alias="400")
    field_404: Field404326 = Field(..., alias="404")
    field_409: Field40951 = Field(..., alias="409")
    field_422: Field422145 = Field(..., alias="422")
    field_429: Field429157 = Field(..., alias="429")
    field_200: Field200284 = Field(..., alias="200")


class Post72(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter304]
    responses: Responses326
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldFolders(BaseModel):
    get: Get152
    post: Post72


class Parameter305(Parameter60):
    pass


class Field400320(Field400319):
    pass


class Field404327(Field400319):
    pass


class Schema1488(Schema1331):
    pass


class Field200285(BaseModel):
    description: str
    schema_: Schema1488 = Field(..., alias="schema")


class Responses327(BaseModel):
    field_400: Field400320 = Field(..., alias="400")
    field_404: Field404327 = Field(..., alias="404")
    field_200: Field200285 = Field(..., alias="200")


class Get153(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter305]
    responses: Responses327
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldFoldersFolderIdChildren(BaseModel):
    get: Get153


class Parameter306(Parameter2):
    pass


class Schema1489(Schema):
    pass


class Field400321(BaseModel):
    description: str
    schema_: Schema1489 = Field(..., alias="schema")


class Field404328(Field400321):
    pass


class Schema1491(Schema1331):
    pass


class Field200286(BaseModel):
    description: str
    schema_: Schema1491 = Field(..., alias="schema")


class Responses328(BaseModel):
    field_400: Field400321 = Field(..., alias="400")
    field_404: Field404328 = Field(..., alias="404")
    field_200: Field200286 = Field(..., alias="200")


class Get154(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter306]
    responses: Responses328
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldFoldersFolderIdChildrenSearch(BaseModel):
    get: Get154


class Schema1492(Schema):
    pass


class Field400322(BaseModel):
    description: str
    schema_: Schema1492 = Field(..., alias="schema")


class Field404329(Field400322):
    pass


class Field200287(Field400322):
    pass


class Responses329(BaseModel):
    field_400: Field400322 = Field(..., alias="400")
    field_404: Field404329 = Field(..., alias="404")
    field_200: Field200287 = Field(..., alias="200")


class Get155(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter306]
    responses: Responses329
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldFoldersFolderIdParent(BaseModel):
    get: Get155


class Field400323(Field400322):
    pass


class Field404330(Field400322):
    pass


class Schema1497(Schema1331):
    pass


class Field200288(BaseModel):
    description: str
    schema_: Schema1497 = Field(..., alias="schema")


class Responses330(BaseModel):
    field_400: Field400323 = Field(..., alias="400")
    field_404: Field404330 = Field(..., alias="404")
    field_200: Field200288 = Field(..., alias="200")


class Get156(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter306]
    responses: Responses330
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldFoldersFolderIdAncestors(BaseModel):
    get: Get156


class Schema1498(Schema):
    pass


class Field400324(BaseModel):
    description: str
    schema_: Schema1498 = Field(..., alias="schema")


class Field404331(Field400324):
    pass


class Schema1500(Schema1331):
    pass


class Field200289(BaseModel):
    description: str
    schema_: Schema1500 = Field(..., alias="schema")


class Responses331(BaseModel):
    field_400: Field400324 = Field(..., alias="400")
    field_404: Field404331 = Field(..., alias="404")
    field_200: Field200289 = Field(..., alias="200")


class Get157(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter306]
    responses: Responses331
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldFoldersFolderIdLooks(BaseModel):
    get: Get157


class Schema1501(Schema):
    pass


class Field400325(BaseModel):
    description: str
    schema_: Schema1501 = Field(..., alias="schema")


class Field404332(Field400325):
    pass


class Schema1503(Schema1331):
    pass


class Field200290(BaseModel):
    description: str
    schema_: Schema1503 = Field(..., alias="schema")


class Responses332(BaseModel):
    field_400: Field400325 = Field(..., alias="400")
    field_404: Field404332 = Field(..., alias="404")
    field_200: Field200290 = Field(..., alias="200")


class Get158(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter306]
    responses: Responses332
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldFoldersFolderIdDashboards(BaseModel):
    get: Get158


class Schema1504(Schema):
    pass


class Field400326(BaseModel):
    description: str
    schema_: Schema1504 = Field(..., alias="schema")


class Field404333(Field400326):
    pass


class Field200291(Field400326):
    pass


class Responses333(BaseModel):
    field_400: Field400326 = Field(..., alias="400")
    field_404: Field404333 = Field(..., alias="404")
    field_200: Field200291 = Field(..., alias="200")


class Get159(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter306]
    responses: Responses333
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSqlQueriesSlug(BaseModel):
    get: Get159


class Parameter312(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1504 = Field(..., alias="schema")


class Field400327(Field400326):
    pass


class Field404334(Field400326):
    pass


class Field40952(Field400326):
    pass


class Field422146(Field400326):
    pass


class Field429158(Field400326):
    pass


class Field200292(Field400326):
    pass


class Responses334(BaseModel):
    field_400: Field400327 = Field(..., alias="400")
    field_404: Field404334 = Field(..., alias="404")
    field_409: Field40952 = Field(..., alias="409")
    field_422: Field422146 = Field(..., alias="422")
    field_429: Field429158 = Field(..., alias="429")
    field_200: Field200292 = Field(..., alias="200")


class Post73(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter312]
    responses: Responses334
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSqlQueries(BaseModel):
    post: Post73


class Parameter313(Parameter2):
    pass


class Field400328(Field400326):
    pass


class Field404335(Field400326):
    pass


class Field422147(Field400326):
    pass


class Field429159(Field400326):
    pass


class Schema1518(Items):
    pass


class Field200293(BaseModel):
    description: str
    schema_: Schema1518 = Field(..., alias="schema")


class Responses335(BaseModel):
    field_400: Field400328 = Field(..., alias="400")
    field_404: Field404335 = Field(..., alias="404")
    field_422: Field422147 = Field(..., alias="422")
    field_429: Field429159 = Field(..., alias="429")
    field_200: Field200293 = Field(..., alias="200")


class Post74(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter313]
    responses: Responses335
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSqlQueriesSlugRunResultFormat(BaseModel):
    post: Post74


class Schema1519(Schema):
    pass


class Field400329(BaseModel):
    description: str
    schema_: Schema1519 = Field(..., alias="schema")


class Field40343(Field400329):
    pass


class Field404336(Field400329):
    pass


class Schema1522(Schema1331):
    pass


class Field200294(BaseModel):
    description: str
    schema_: Schema1522 = Field(..., alias="schema")


class Responses336(BaseModel):
    field_400: Field400329 = Field(..., alias="400")
    field_403: Field40343 = Field(..., alias="403")
    field_404: Field404336 = Field(..., alias="404")
    field_200: Field200294 = Field(..., alias="200")


class Get160(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter313]
    responses: Responses336
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1523(Schema):
    pass


class Parameter315(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1523 = Field(..., alias="schema")


class Field400330(BaseModel):
    description: str
    schema_: Schema1523 = Field(..., alias="schema")


class Field40344(Field400330):
    pass


class Field404337(Field400330):
    pass


class Field422148(Field400330):
    pass


class Field429160(Field400330):
    pass


class Schema1529(Schema1331):
    pass


class Field200295(BaseModel):
    description: str
    schema_: Schema1529 = Field(..., alias="schema")


class Responses337(BaseModel):
    field_400: Field400330 = Field(..., alias="400")
    field_403: Field40344 = Field(..., alias="403")
    field_404: Field404337 = Field(..., alias="404")
    field_422: Field422148 = Field(..., alias="422")
    field_429: Field429160 = Field(..., alias="429")
    field_200: Field200295 = Field(..., alias="200")


class Post75(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter315]
    responses: Responses337
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSupportAccessAllowlist(BaseModel):
    get: Get160
    post: Post75


class Parameter316(Parameter2):
    pass


class Schema1530(Schema):
    pass


class Field400331(BaseModel):
    description: str
    schema_: Schema1530 = Field(..., alias="schema")


class Field40345(Field400331):
    pass


class Field404338(Field400331):
    pass


class Field429161(Field400331):
    pass


class Schema1534(Items):
    pass


class Field20452(BaseModel):
    description: str
    schema_: Schema1534 = Field(..., alias="schema")


class Responses338(BaseModel):
    field_400: Field400331 = Field(..., alias="400")
    field_403: Field40345 = Field(..., alias="403")
    field_404: Field404338 = Field(..., alias="404")
    field_429: Field429161 = Field(..., alias="429")
    field_204: Field20452 = Field(..., alias="204")


class Delete40(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter316]
    responses: Responses338
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSupportAccessAllowlistEntryId(BaseModel):
    delete: Delete40


class Schema1535(Schema):
    pass


class Parameter317(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1535 = Field(..., alias="schema")


class Field400332(BaseModel):
    description: str
    schema_: Schema1535 = Field(..., alias="schema")


class Field40346(Field400332):
    pass


class Field404339(Field400332):
    pass


class Field40953(Field400332):
    pass


class Field422149(Field400332):
    pass


class Field429162(Field400332):
    pass


class Field200296(Field400332):
    pass


class Responses339(BaseModel):
    field_400: Field400332 = Field(..., alias="400")
    field_403: Field40346 = Field(..., alias="403")
    field_404: Field404339 = Field(..., alias="404")
    field_409: Field40953 = Field(..., alias="409")
    field_422: Field422149 = Field(..., alias="422")
    field_429: Field429162 = Field(..., alias="429")
    field_200: Field200296 = Field(..., alias="200")


class Put17(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter317]
    responses: Responses339
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSupportAccessEnable(BaseModel):
    put: Put17


class Field400333(Field400332):
    pass


class Field40347(Field400332):
    pass


class Field404340(Field400332):
    pass


class Field429163(Field400332):
    pass


class Field200297(Field400332):
    pass


class Responses340(BaseModel):
    field_400: Field400333 = Field(..., alias="400")
    field_403: Field40347 = Field(..., alias="403")
    field_404: Field404340 = Field(..., alias="404")
    field_429: Field429163 = Field(..., alias="429")
    field_200: Field200297 = Field(..., alias="200")


class Put18(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses340
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSupportAccessDisable(BaseModel):
    put: Put18


class Field400334(Field400332):
    pass


class Field40348(Field400332):
    pass


class Field404341(Field400332):
    pass


class Field200298(Field400332):
    pass


class Responses341(BaseModel):
    field_400: Field400334 = Field(..., alias="400")
    field_403: Field40348 = Field(..., alias="403")
    field_404: Field404341 = Field(..., alias="404")
    field_200: Field200298 = Field(..., alias="200")


class Get161(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses341
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSupportAccessStatus(BaseModel):
    get: Get161


class Parameter318(Parameter2):
    pass


class Field400335(Field400332):
    pass


class Field404342(Field400332):
    pass


class Schema1554(Schema1331):
    pass


class Field200299(BaseModel):
    description: str
    schema_: Schema1554 = Field(..., alias="schema")


class Responses342(BaseModel):
    field_400: Field400335 = Field(..., alias="400")
    field_404: Field404342 = Field(..., alias="404")
    field_200: Field200299 = Field(..., alias="200")


class Get162(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter318]
    responses: Responses342
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1555(Schema):
    pass


class Parameter319(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1555 = Field(..., alias="schema")


class Field400336(BaseModel):
    description: str
    schema_: Schema1555 = Field(..., alias="schema")


class Field404343(Field400336):
    pass


class Field40954(Field400336):
    pass


class Field422150(Field400336):
    pass


class Field429164(Field400336):
    pass


class Field200300(Field400336):
    pass


class Responses343(BaseModel):
    field_400: Field400336 = Field(..., alias="400")
    field_404: Field404343 = Field(..., alias="404")
    field_409: Field40954 = Field(..., alias="409")
    field_422: Field422150 = Field(..., alias="422")
    field_429: Field429164 = Field(..., alias="429")
    field_200: Field200300 = Field(..., alias="200")


class Post76(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter319]
    responses: Responses343
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldThemes(BaseModel):
    get: Get162
    post: Post76


class Parameter320(Parameter7):
    pass


class Field400337(Field400336):
    pass


class Field404344(Field400336):
    pass


class Schema1564(Schema1331):
    pass


class Field200301(BaseModel):
    description: str
    schema_: Schema1564 = Field(..., alias="schema")


class Responses344(BaseModel):
    field_400: Field400337 = Field(..., alias="400")
    field_404: Field404344 = Field(..., alias="404")
    field_200: Field200301 = Field(..., alias="200")


class Get163(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter320]
    responses: Responses344
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldThemesSearch(BaseModel):
    get: Get163


class Parameter321(Parameter21):
    pass


class Schema1565(Schema):
    pass


class Field400338(BaseModel):
    description: str
    schema_: Schema1565 = Field(..., alias="schema")


class Field404345(Field400338):
    pass


class Field200302(Field400338):
    pass


class Responses345(BaseModel):
    field_400: Field400338 = Field(..., alias="400")
    field_404: Field404345 = Field(..., alias="404")
    field_200: Field200302 = Field(..., alias="200")


class Get164(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter321]
    responses: Responses345
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter322(Parameter2):
    pass


class Field400339(Field400338):
    pass


class Field404346(Field400338):
    pass


class Field422151(Field400338):
    pass


class Field429165(Field400338):
    pass


class Field200303(Field400338):
    pass


class Responses346(BaseModel):
    field_400: Field400339 = Field(..., alias="400")
    field_404: Field404346 = Field(..., alias="404")
    field_422: Field422151 = Field(..., alias="422")
    field_429: Field429165 = Field(..., alias="429")
    field_200: Field200303 = Field(..., alias="200")


class Put19(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter322]
    responses: Responses346
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldThemesDefault(BaseModel):
    get: Get164
    put: Put19


class Parameter323(Parameter7):
    pass


class Field400340(Field400338):
    pass


class Field404347(Field400338):
    pass


class Schema1575(Schema1331):
    pass


class Field200304(BaseModel):
    description: str
    schema_: Schema1575 = Field(..., alias="schema")


class Responses347(BaseModel):
    field_400: Field400340 = Field(..., alias="400")
    field_404: Field404347 = Field(..., alias="404")
    field_200: Field200304 = Field(..., alias="200")


class Get165(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter323]
    responses: Responses347
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldThemesActive(BaseModel):
    get: Get165


class Schema1576(Schema):
    pass


class Field400341(BaseModel):
    description: str
    schema_: Schema1576 = Field(..., alias="schema")


class Field404348(Field400341):
    pass


class Field200305(Field400341):
    pass


class Responses348(BaseModel):
    field_400: Field400341 = Field(..., alias="400")
    field_404: Field404348 = Field(..., alias="404")
    field_200: Field200305 = Field(..., alias="200")


class Get166(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter323]
    responses: Responses348
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldThemesThemeOrDefault(BaseModel):
    get: Get166


class Parameter325(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1576 = Field(..., alias="schema")


class Field400342(Field400341):
    pass


class Field404349(Field400341):
    pass


class Field40955(Field400341):
    pass


class Field422152(Field400341):
    pass


class Field429166(Field400341):
    pass


class Field200306(Field400341):
    pass


class Schema1586(Items):
    pass


class Field20453(BaseModel):
    description: str
    schema_: Schema1586 = Field(..., alias="schema")


class Responses349(BaseModel):
    field_400: Field400342 = Field(..., alias="400")
    field_404: Field404349 = Field(..., alias="404")
    field_409: Field40955 = Field(..., alias="409")
    field_422: Field422152 = Field(..., alias="422")
    field_429: Field429166 = Field(..., alias="429")
    field_200: Field200306 = Field(..., alias="200")
    field_204: Field20453 = Field(..., alias="204")


class Post77(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter325]
    responses: Responses349
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldThemesValidate(BaseModel):
    post: Post77


class Parameter326(Parameter2):
    pass


class Schema1587(Schema):
    pass


class Field400343(BaseModel):
    description: str
    schema_: Schema1587 = Field(..., alias="schema")


class Field404350(Field400343):
    pass


class Field200307(Field400343):
    pass


class Responses350(BaseModel):
    field_400: Field400343 = Field(..., alias="400")
    field_404: Field404350 = Field(..., alias="404")
    field_200: Field200307 = Field(..., alias="200")


class Get167(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter326]
    responses: Responses350
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter327(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1587] = Field(None, alias="schema")


class Field400344(Field400343):
    pass


class Field404351(Field400343):
    pass


class Field422153(Field400343):
    pass


class Field429167(Field400343):
    pass


class Field200308(Field400343):
    pass


class Responses351(BaseModel):
    field_400: Field400344 = Field(..., alias="400")
    field_404: Field404351 = Field(..., alias="404")
    field_422: Field422153 = Field(..., alias="422")
    field_429: Field429167 = Field(..., alias="429")
    field_200: Field200308 = Field(..., alias="200")


class Patch44(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter327]
    responses: Responses351
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter328(Parameter2):
    pass


class Field400345(Field400343):
    pass


class Field404352(Field400343):
    pass


class Field429168(Field400343):
    pass


class Schema1599(Items):
    pass


class Field20454(BaseModel):
    description: str
    schema_: Schema1599 = Field(..., alias="schema")


class Responses352(BaseModel):
    field_400: Field400345 = Field(..., alias="400")
    field_404: Field404352 = Field(..., alias="404")
    field_429: Field429168 = Field(..., alias="429")
    field_204: Field20454 = Field(..., alias="204")


class Delete41(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter328]
    responses: Responses352
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldThemesThemeId(BaseModel):
    get: Get167
    patch: Patch44
    delete: Delete41


class Schema1600(Schema):
    pass


class Field400346(BaseModel):
    description: str
    schema_: Schema1600 = Field(..., alias="schema")


class Field404353(Field400346):
    pass


class Schema1602(Schema1331):
    pass


class Field200309(BaseModel):
    description: str
    schema_: Schema1602 = Field(..., alias="schema")


class Responses353(BaseModel):
    field_400: Field400346 = Field(..., alias="400")
    field_404: Field404353 = Field(..., alias="404")
    field_200: Field200309 = Field(..., alias="200")


class Get168(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses353
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldTimezones(BaseModel):
    get: Get168


class Schema1603(Schema):
    pass


class Field400347(BaseModel):
    description: str
    schema_: Schema1603 = Field(..., alias="schema")


class Field404354(Field400347):
    pass


class Schema1605(Schema1331):
    pass


class Field200310(BaseModel):
    description: str
    schema_: Schema1605 = Field(..., alias="schema")


class Responses354(BaseModel):
    field_400: Field400347 = Field(..., alias="400")
    field_404: Field404354 = Field(..., alias="404")
    field_200: Field200310 = Field(..., alias="200")


class Get169(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter328]
    responses: Responses354
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1606(Schema):
    pass


class Parameter330(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1606 = Field(..., alias="schema")


class Field400348(BaseModel):
    description: str
    schema_: Schema1606 = Field(..., alias="schema")


class Field404355(Field400348):
    pass


class Field40956(Field400348):
    pass


class Field422154(Field400348):
    pass


class Field429169(Field400348):
    pass


class Field200311(Field400348):
    pass


class Responses355(BaseModel):
    field_400: Field400348 = Field(..., alias="400")
    field_404: Field404355 = Field(..., alias="404")
    field_409: Field40956 = Field(..., alias="409")
    field_422: Field422154 = Field(..., alias="422")
    field_429: Field429169 = Field(..., alias="429")
    field_200: Field200311 = Field(..., alias="200")


class Post78(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter330]
    responses: Responses355
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSshServers(BaseModel):
    get: Get169
    post: Post78


class Parameter331(Parameter2):
    pass


class Field400349(Field400348):
    pass


class Field404356(Field400348):
    pass


class Field200312(Field400348):
    pass


class Responses356(BaseModel):
    field_400: Field400349 = Field(..., alias="400")
    field_404: Field404356 = Field(..., alias="404")
    field_200: Field200312 = Field(..., alias="200")


class Get170(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter331]
    responses: Responses356
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter332(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1606] = Field(None, alias="schema")


class Field400350(Field400348):
    pass


class Field404357(Field400348):
    pass


class Field422155(Field400348):
    pass


class Field429170(Field400348):
    pass


class Field200313(Field400348):
    pass


class Responses357(BaseModel):
    field_400: Field400350 = Field(..., alias="400")
    field_404: Field404357 = Field(..., alias="404")
    field_422: Field422155 = Field(..., alias="422")
    field_429: Field429170 = Field(..., alias="429")
    field_200: Field200313 = Field(..., alias="200")


class Patch45(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter332]
    responses: Responses357
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter333(Parameter2):
    pass


class Field400351(Field400348):
    pass


class Field404358(Field400348):
    pass


class Field429171(Field400348):
    pass


class Schema1625(Items):
    pass


class Field20455(BaseModel):
    description: str
    schema_: Schema1625 = Field(..., alias="schema")


class Responses358(BaseModel):
    field_400: Field400351 = Field(..., alias="400")
    field_404: Field404358 = Field(..., alias="404")
    field_429: Field429171 = Field(..., alias="429")
    field_204: Field20455 = Field(..., alias="204")


class Delete42(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter333]
    responses: Responses358
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSshServerSshServerId(BaseModel):
    get: Get170
    patch: Patch45
    delete: Delete42


class Schema1626(Schema):
    pass


class Field400352(BaseModel):
    description: str
    schema_: Schema1626 = Field(..., alias="schema")


class Field404359(Field400352):
    pass


class Field200314(Field400352):
    pass


class Responses359(BaseModel):
    field_400: Field400352 = Field(..., alias="400")
    field_404: Field404359 = Field(..., alias="404")
    field_200: Field200314 = Field(..., alias="200")


class Get171(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter333]
    responses: Responses359
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSshServerSshServerIdTest(BaseModel):
    get: Get171


class Field400353(Field400352):
    pass


class Field404360(Field400352):
    pass


class Schema1631(Schema1331):
    pass


class Field200315(BaseModel):
    description: str
    schema_: Schema1631 = Field(..., alias="schema")


class Responses360(BaseModel):
    field_400: Field400353 = Field(..., alias="400")
    field_404: Field404360 = Field(..., alias="404")
    field_200: Field200315 = Field(..., alias="200")


class Get172(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter333]
    responses: Responses360
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1632(Schema):
    pass


class Parameter336(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1632 = Field(..., alias="schema")


class Field400354(BaseModel):
    description: str
    schema_: Schema1632 = Field(..., alias="schema")


class Field404361(Field400354):
    pass


class Field40957(Field400354):
    pass


class Field422156(Field400354):
    pass


class Field429172(Field400354):
    pass


class Field200316(Field400354):
    pass


class Responses361(BaseModel):
    field_400: Field400354 = Field(..., alias="400")
    field_404: Field404361 = Field(..., alias="404")
    field_409: Field40957 = Field(..., alias="409")
    field_422: Field422156 = Field(..., alias="422")
    field_429: Field429172 = Field(..., alias="429")
    field_200: Field200316 = Field(..., alias="200")


class Post79(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter336]
    responses: Responses361
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSshTunnels(BaseModel):
    get: Get172
    post: Post79


class Parameter337(Parameter2):
    pass


class Field400355(Field400354):
    pass


class Field404362(Field400354):
    pass


class Field200317(Field400354):
    pass


class Responses362(BaseModel):
    field_400: Field400355 = Field(..., alias="400")
    field_404: Field404362 = Field(..., alias="404")
    field_200: Field200317 = Field(..., alias="200")


class Get173(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter337]
    responses: Responses362
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter338(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1632] = Field(None, alias="schema")


class Field400356(Field400354):
    pass


class Field404363(Field400354):
    pass


class Field422157(Field400354):
    pass


class Field429173(Field400354):
    pass


class Field200318(Field400354):
    pass


class Responses363(BaseModel):
    field_400: Field400356 = Field(..., alias="400")
    field_404: Field404363 = Field(..., alias="404")
    field_422: Field422157 = Field(..., alias="422")
    field_429: Field429173 = Field(..., alias="429")
    field_200: Field200318 = Field(..., alias="200")


class Patch46(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter338]
    responses: Responses363
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter339(Parameter2):
    pass


class Field400357(Field400354):
    pass


class Field404364(Field400354):
    pass


class Field429174(Field400354):
    pass


class Schema1651(Items):
    pass


class Field20456(BaseModel):
    description: str
    schema_: Schema1651 = Field(..., alias="schema")


class Responses364(BaseModel):
    field_400: Field400357 = Field(..., alias="400")
    field_404: Field404364 = Field(..., alias="404")
    field_429: Field429174 = Field(..., alias="429")
    field_204: Field20456 = Field(..., alias="204")


class Delete43(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter339]
    responses: Responses364
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSshTunnelSshTunnelId(BaseModel):
    get: Get173
    patch: Patch46
    delete: Delete43


class Schema1652(Schema):
    pass


class Field400358(BaseModel):
    description: str
    schema_: Schema1652 = Field(..., alias="schema")


class Field404365(Field400358):
    pass


class Field200319(Field400358):
    pass


class Responses365(BaseModel):
    field_400: Field400358 = Field(..., alias="400")
    field_404: Field404365 = Field(..., alias="404")
    field_200: Field200319 = Field(..., alias="200")


class Get174(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter339]
    responses: Responses365
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSshTunnelSshTunnelIdTest(BaseModel):
    get: Get174


class Field400359(Field400358):
    pass


class Field404366(Field400358):
    pass


class Field200320(Field400358):
    pass


class Responses366(BaseModel):
    field_400: Field400359 = Field(..., alias="400")
    field_404: Field404366 = Field(..., alias="404")
    field_200: Field200320 = Field(..., alias="200")


class Get175(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses366
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldSshPublicKey(BaseModel):
    get: Get175


class Field400360(Field400358):
    pass


class Field404367(Field400358):
    pass


class Schema1660(Schema1331):
    pass


class Field200321(BaseModel):
    description: str
    schema_: Schema1660 = Field(..., alias="schema")


class Responses367(BaseModel):
    field_400: Field400360 = Field(..., alias="400")
    field_404: Field404367 = Field(..., alias="404")
    field_200: Field200321 = Field(..., alias="200")


class Get176(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter339]
    responses: Responses367
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1661(Schema):
    pass


class Parameter342(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema1661] = Field(None, alias="schema")
    type: Optional[str] = None


class Field400361(BaseModel):
    description: str
    schema_: Schema1661 = Field(..., alias="schema")


class Field404368(Field400361):
    pass


class Field40958(Field400361):
    pass


class Field422158(Field400361):
    pass


class Field429175(Field400361):
    pass


class Field200322(Field400361):
    pass


class Responses368(BaseModel):
    field_400: Field400361 = Field(..., alias="400")
    field_404: Field404368 = Field(..., alias="404")
    field_409: Field40958 = Field(..., alias="409")
    field_422: Field422158 = Field(..., alias="422")
    field_429: Field429175 = Field(..., alias="429")
    field_200: Field200322 = Field(..., alias="200")


class Post80(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter342]
    responses: Responses368
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUserAttributes(BaseModel):
    get: Get176
    post: Post80


class Parameter343(Parameter2):
    pass


class Field400362(Field400361):
    pass


class Field404369(Field400361):
    pass


class Field200323(Field400361):
    pass


class Responses369(BaseModel):
    field_400: Field400362 = Field(..., alias="400")
    field_404: Field404369 = Field(..., alias="404")
    field_200: Field200323 = Field(..., alias="200")


class Get177(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter343]
    responses: Responses369
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter344(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1661] = Field(None, alias="schema")


class Field400363(Field400361):
    pass


class Field404370(Field400361):
    pass


class Field422159(Field400361):
    pass


class Field429176(Field400361):
    pass


class Field200324(Field400361):
    pass


class Responses370(BaseModel):
    field_400: Field400363 = Field(..., alias="400")
    field_404: Field404370 = Field(..., alias="404")
    field_422: Field422159 = Field(..., alias="422")
    field_429: Field429176 = Field(..., alias="429")
    field_200: Field200324 = Field(..., alias="200")


class Patch47(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter344]
    responses: Responses370
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter345(Parameter2):
    pass


class Field400364(Field400361):
    pass


class Field404371(Field400361):
    pass


class Field429177(Field400361):
    pass


class Schema1680(Items):
    pass


class Field20457(BaseModel):
    description: str
    schema_: Schema1680 = Field(..., alias="schema")


class Responses371(BaseModel):
    field_400: Field400364 = Field(..., alias="400")
    field_404: Field404371 = Field(..., alias="404")
    field_429: Field429177 = Field(..., alias="429")
    field_204: Field20457 = Field(..., alias="204")


class Delete44(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter345]
    responses: Responses371
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUserAttributesUserAttributeId(BaseModel):
    get: Get177
    patch: Patch47
    delete: Delete44


class Schema1681(Schema):
    pass


class Field400365(BaseModel):
    description: str
    schema_: Schema1681 = Field(..., alias="schema")


class Field404372(Field400365):
    pass


class Schema1683(Schema1331):
    pass


class Field200325(BaseModel):
    description: str
    schema_: Schema1683 = Field(..., alias="schema")


class Responses372(BaseModel):
    field_400: Field400365 = Field(..., alias="400")
    field_404: Field404372 = Field(..., alias="404")
    field_200: Field200325 = Field(..., alias="200")


class Get178(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter345]
    responses: Responses372
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1684(Schema1331):
    pass


class Parameter347(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1684] = Field(None, alias="schema")


class Schema1685(Schema):
    pass


class Field400366(BaseModel):
    description: str
    schema_: Schema1685 = Field(..., alias="schema")


class Field404373(Field400366):
    pass


class Field40959(Field400366):
    pass


class Field422160(Field400366):
    pass


class Field429178(Field400366):
    pass


class Schema1690(Schema1331):
    pass


class Field200326(BaseModel):
    description: str
    schema_: Schema1690 = Field(..., alias="schema")


class Responses373(BaseModel):
    field_400: Field400366 = Field(..., alias="400")
    field_404: Field404373 = Field(..., alias="404")
    field_409: Field40959 = Field(..., alias="409")
    field_422: Field422160 = Field(..., alias="422")
    field_429: Field429178 = Field(..., alias="429")
    field_200: Field200326 = Field(..., alias="200")


class Post81(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter347]
    responses: Responses373
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUserAttributesUserAttributeIdGroupValues(BaseModel):
    get: Get178
    post: Post81


class Parameter348(Parameter2):
    pass


class Schema1691(Schema):
    pass


class Field400367(BaseModel):
    description: str
    schema_: Schema1691 = Field(..., alias="schema")


class Field404374(Field400367):
    pass


class Schema1693(Schema1331):
    pass


class Field200327(BaseModel):
    description: str
    schema_: Schema1693 = Field(..., alias="schema")


class Responses374(BaseModel):
    field_400: Field400367 = Field(..., alias="400")
    field_404: Field404374 = Field(..., alias="404")
    field_200: Field200327 = Field(..., alias="200")


class Get179(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter348]
    responses: Responses374
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUserLoginLockouts(BaseModel):
    get: Get179


class Parameter349(Parameter60):
    pass


class Schema1694(Schema):
    pass


class Field400368(BaseModel):
    description: str
    schema_: Schema1694 = Field(..., alias="schema")


class Field404375(Field400368):
    pass


class Schema1696(Schema1331):
    pass


class Field200328(BaseModel):
    description: str
    schema_: Schema1696 = Field(..., alias="schema")


class Responses375(BaseModel):
    field_400: Field400368 = Field(..., alias="400")
    field_404: Field404375 = Field(..., alias="404")
    field_200: Field200328 = Field(..., alias="200")


class Get180(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter349]
    responses: Responses375
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUserLoginLockoutsSearch(BaseModel):
    get: Get180


class Parameter350(Parameter2):
    pass


class Schema1697(Schema):
    pass


class Field400369(BaseModel):
    description: str
    schema_: Schema1697 = Field(..., alias="schema")


class Field404376(Field400369):
    pass


class Field429179(Field400369):
    pass


class Schema1700(Items):
    pass


class Field20458(BaseModel):
    description: str
    schema_: Schema1700 = Field(..., alias="schema")


class Responses376(BaseModel):
    field_400: Field400369 = Field(..., alias="400")
    field_404: Field404376 = Field(..., alias="404")
    field_429: Field429179 = Field(..., alias="429")
    field_204: Field20458 = Field(..., alias="204")


class Delete45(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter350]
    responses: Responses376
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUserLoginLockoutKey(BaseModel):
    delete: Delete45


class Schema1701(Schema):
    pass


class Field404377(BaseModel):
    description: str
    schema_: Schema1701 = Field(..., alias="schema")


class Field200329(Field404377):
    pass


class Responses377(BaseModel):
    field_404: Field404377 = Field(..., alias="404")
    field_200: Field200329 = Field(..., alias="200")


class Get181(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter350]
    responses: Responses377
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUser(BaseModel):
    get: Get181


class Items107(Items):
    pass


class Parameter352(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str
    format: Optional[str] = None
    x_looker_deprecated: Optional[bool] = Field(None, alias="x-looker-deprecated")
    items: Optional[Items107] = None
    collectionFormat: Optional[str] = None


class Field400370(Field404377):
    pass


class Field404378(Field404377):
    pass


class Items108(Schema):
    pass


class Schema1705(BaseModel):
    type: str
    items: Items108


class Field200330(BaseModel):
    description: str
    schema_: Schema1705 = Field(..., alias="schema")


class Responses378(BaseModel):
    field_400: Field400370 = Field(..., alias="400")
    field_404: Field404378 = Field(..., alias="404")
    field_200: Field200330 = Field(..., alias="200")


class Get182(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter352]
    responses: Responses378
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1706(Schema):
    pass


class Parameter353(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Optional[Schema1706] = Field(None, alias="schema")
    type: Optional[str] = None


class Field400371(BaseModel):
    description: str
    schema_: Schema1706 = Field(..., alias="schema")


class Field404379(Field400371):
    pass


class Field40960(Field400371):
    pass


class Field422161(Field400371):
    pass


class Field200331(Field400371):
    pass


class Responses379(BaseModel):
    field_400: Field400371 = Field(..., alias="400")
    field_404: Field404379 = Field(..., alias="404")
    field_409: Field40960 = Field(..., alias="409")
    field_422: Field422161 = Field(..., alias="422")
    field_200: Field200331 = Field(..., alias="200")


class Post82(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter353]
    responses: Responses379
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsers(BaseModel):
    get: Get182
    post: Post82


class Parameter354(Parameter60):
    pass


class Field400372(Field400371):
    pass


class Field404380(Field400371):
    pass


class Schema1714(Schema1705):
    pass


class Field200332(BaseModel):
    description: str
    schema_: Schema1714 = Field(..., alias="schema")


class Responses380(BaseModel):
    field_400: Field400372 = Field(..., alias="400")
    field_404: Field404380 = Field(..., alias="404")
    field_200: Field200332 = Field(..., alias="200")


class Get183(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter354]
    responses: Responses380
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersSearch(BaseModel):
    get: Get183


class Schema1715(Schema):
    pass


class Field400373(BaseModel):
    description: str
    schema_: Schema1715 = Field(..., alias="schema")


class Field404381(Field400373):
    pass


class Schema1717(Schema1705):
    pass


class Field200333(BaseModel):
    description: str
    schema_: Schema1717 = Field(..., alias="schema")


class Responses381(BaseModel):
    field_400: Field400373 = Field(..., alias="400")
    field_404: Field404381 = Field(..., alias="404")
    field_200: Field200333 = Field(..., alias="200")


class Get184(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter354]
    responses: Responses381
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersSearchNamesPattern(BaseModel):
    get: Get184


class Parameter356(Parameter2):
    pass


class Schema1718(Schema):
    pass


class Field400374(BaseModel):
    description: str
    schema_: Schema1718 = Field(..., alias="schema")


class Field404382(Field400374):
    pass


class Field200334(Field400374):
    pass


class Responses382(BaseModel):
    field_400: Field400374 = Field(..., alias="400")
    field_404: Field404382 = Field(..., alias="404")
    field_200: Field200334 = Field(..., alias="200")


class Get185(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter356]
    responses: Responses382
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter357(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1718] = Field(None, alias="schema")


class Field400375(Field400374):
    pass


class Field40349(Field400374):
    pass


class Field404383(Field400374):
    pass


class Field422162(Field400374):
    pass


class Field200335(Field400374):
    pass


class Responses383(BaseModel):
    field_400: Field400375 = Field(..., alias="400")
    field_403: Field40349 = Field(..., alias="403")
    field_404: Field404383 = Field(..., alias="404")
    field_422: Field422162 = Field(..., alias="422")
    field_200: Field200335 = Field(..., alias="200")


class Patch48(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter357]
    responses: Responses383
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter358(Parameter2):
    pass


class Field400376(Field400374):
    pass


class Field40350(Field400374):
    pass


class Field404384(Field400374):
    pass


class Schema1730(Items):
    pass


class Field20459(BaseModel):
    description: str
    schema_: Schema1730 = Field(..., alias="schema")


class Responses384(BaseModel):
    field_400: Field400376 = Field(..., alias="400")
    field_403: Field40350 = Field(..., alias="403")
    field_404: Field404384 = Field(..., alias="404")
    field_204: Field20459 = Field(..., alias="204")


class Delete46(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter358]
    responses: Responses384
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserId(BaseModel):
    get: Get185
    patch: Patch48
    delete: Delete46


class Schema1731(Schema):
    pass


class Field400377(BaseModel):
    description: str
    schema_: Schema1731 = Field(..., alias="schema")


class Field40351(Field400377):
    pass


class Field404385(Field400377):
    pass


class Field200336(Field400377):
    pass


class Responses385(BaseModel):
    field_400: Field400377 = Field(..., alias="400")
    field_403: Field40351 = Field(..., alias="403")
    field_404: Field404385 = Field(..., alias="404")
    field_200: Field200336 = Field(..., alias="200")


class Get186(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter358]
    responses: Responses385
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersCredentialCredentialTypeCredentialId(BaseModel):
    get: Get186


class Field400378(Field400377):
    pass


class Field404386(Field400377):
    pass


class Field40352(Field400377):
    pass


class Field200337(Field400377):
    pass


class Responses386(BaseModel):
    field_400: Field400378 = Field(..., alias="400")
    field_404: Field404386 = Field(..., alias="404")
    field_403: Field40352 = Field(..., alias="403")
    field_200: Field200337 = Field(..., alias="200")


class Get187(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter358]
    responses: Responses386
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter361(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1731] = Field(None, alias="schema")


class Field400379(Field400377):
    pass


class Field404387(Field400377):
    pass


class Field40961(Field400377):
    pass


class Field422163(Field400377):
    pass


class Field429180(Field400377):
    pass


class Field40353(Field400377):
    pass


class Field200338(Field400377):
    pass


class Responses387(BaseModel):
    field_400: Field400379 = Field(..., alias="400")
    field_404: Field404387 = Field(..., alias="404")
    field_409: Field40961 = Field(..., alias="409")
    field_422: Field422163 = Field(..., alias="422")
    field_429: Field429180 = Field(..., alias="429")
    field_403: Field40353 = Field(..., alias="403")
    field_200: Field200338 = Field(..., alias="200")


class Post83(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter361]
    responses: Responses387
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter362(Parameter361):
    pass


class Field400380(Field400377):
    pass


class Field404388(Field400377):
    pass


class Field422164(Field400377):
    pass


class Field429181(Field400377):
    pass


class Field40354(Field400377):
    pass


class Field200339(Field400377):
    pass


class Responses388(BaseModel):
    field_400: Field400380 = Field(..., alias="400")
    field_404: Field404388 = Field(..., alias="404")
    field_422: Field422164 = Field(..., alias="422")
    field_429: Field429181 = Field(..., alias="429")
    field_403: Field40354 = Field(..., alias="403")
    field_200: Field200339 = Field(..., alias="200")


class Patch49(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter362]
    responses: Responses388
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter363(Parameter2):
    pass


class Field400381(Field400377):
    pass


class Field404389(Field400377):
    pass


class Field429182(Field400377):
    pass


class Field40355(Field400377):
    pass


class Schema1758(Items):
    pass


class Field20460(BaseModel):
    description: str
    schema_: Schema1758 = Field(..., alias="schema")


class Responses389(BaseModel):
    field_400: Field400381 = Field(..., alias="400")
    field_404: Field404389 = Field(..., alias="404")
    field_429: Field429182 = Field(..., alias="429")
    field_403: Field40355 = Field(..., alias="403")
    field_204: Field20460 = Field(..., alias="204")


class Delete47(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter363]
    responses: Responses389
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsEmail(BaseModel):
    get: Get187
    post: Post83
    patch: Patch49
    delete: Delete47


class Schema1759(Schema):
    pass


class Field400382(BaseModel):
    description: str
    schema_: Schema1759 = Field(..., alias="schema")


class Field404390(Field400382):
    pass


class Field40356(Field400382):
    pass


class Field200340(Field400382):
    pass


class Responses390(BaseModel):
    field_400: Field400382 = Field(..., alias="400")
    field_404: Field404390 = Field(..., alias="404")
    field_403: Field40356 = Field(..., alias="403")
    field_200: Field200340 = Field(..., alias="200")


class Get188(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter363]
    responses: Responses390
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter365(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1759] = Field(None, alias="schema")


class Field400383(Field400382):
    pass


class Field404391(Field400382):
    pass


class Field40962(Field400382):
    pass


class Field422165(Field400382):
    pass


class Field429183(Field400382):
    pass


class Field40357(Field400382):
    pass


class Field200341(Field400382):
    pass


class Responses391(BaseModel):
    field_400: Field400383 = Field(..., alias="400")
    field_404: Field404391 = Field(..., alias="404")
    field_409: Field40962 = Field(..., alias="409")
    field_422: Field422165 = Field(..., alias="422")
    field_429: Field429183 = Field(..., alias="429")
    field_403: Field40357 = Field(..., alias="403")
    field_200: Field200341 = Field(..., alias="200")


class Post84(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter365]
    responses: Responses391
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter366(Parameter2):
    pass


class Field400384(Field400382):
    pass


class Field404392(Field400382):
    pass


class Field429184(Field400382):
    pass


class Field40358(Field400382):
    pass


class Schema1775(Items):
    pass


class Field20461(BaseModel):
    description: str
    schema_: Schema1775 = Field(..., alias="schema")


class Responses392(BaseModel):
    field_400: Field400384 = Field(..., alias="400")
    field_404: Field404392 = Field(..., alias="404")
    field_429: Field429184 = Field(..., alias="429")
    field_403: Field40358 = Field(..., alias="403")
    field_204: Field20461 = Field(..., alias="204")


class Delete48(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses392
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsTotp(BaseModel):
    get: Get188
    post: Post84
    delete: Delete48


class Schema1776(Schema):
    pass


class Field400385(BaseModel):
    description: str
    schema_: Schema1776 = Field(..., alias="schema")


class Field404393(Field400385):
    pass


class Field40359(Field400385):
    pass


class Field200342(Field400385):
    pass


class Responses393(BaseModel):
    field_400: Field400385 = Field(..., alias="400")
    field_404: Field404393 = Field(..., alias="404")
    field_403: Field40359 = Field(..., alias="403")
    field_200: Field200342 = Field(..., alias="200")


class Get189(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses393
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400386(Field400385):
    pass


class Field404394(Field400385):
    pass


class Field429185(Field400385):
    pass


class Field40360(Field400385):
    pass


class Schema1784(Items):
    pass


class Field20462(BaseModel):
    description: str
    schema_: Schema1784 = Field(..., alias="schema")


class Responses394(BaseModel):
    field_400: Field400386 = Field(..., alias="400")
    field_404: Field404394 = Field(..., alias="404")
    field_429: Field429185 = Field(..., alias="429")
    field_403: Field40360 = Field(..., alias="403")
    field_204: Field20462 = Field(..., alias="204")


class Delete49(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses394
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsLdap(BaseModel):
    get: Get189
    delete: Delete49


class Schema1785(Schema):
    pass


class Field400387(BaseModel):
    description: str
    schema_: Schema1785 = Field(..., alias="schema")


class Field404395(Field400387):
    pass


class Field40361(Field400387):
    pass


class Field200343(Field400387):
    pass


class Responses395(BaseModel):
    field_400: Field400387 = Field(..., alias="400")
    field_404: Field404395 = Field(..., alias="404")
    field_403: Field40361 = Field(..., alias="403")
    field_200: Field200343 = Field(..., alias="200")


class Get190(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses395
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400388(Field400387):
    pass


class Field404396(Field400387):
    pass


class Field429186(Field400387):
    pass


class Field40362(Field400387):
    pass


class Schema1793(Items):
    pass


class Field20463(BaseModel):
    description: str
    schema_: Schema1793 = Field(..., alias="schema")


class Responses396(BaseModel):
    field_400: Field400388 = Field(..., alias="400")
    field_404: Field404396 = Field(..., alias="404")
    field_429: Field429186 = Field(..., alias="429")
    field_403: Field40362 = Field(..., alias="403")
    field_204: Field20463 = Field(..., alias="204")


class Delete50(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses396
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsGoogle(BaseModel):
    get: Get190
    delete: Delete50


class Schema1794(Schema):
    pass


class Field400389(BaseModel):
    description: str
    schema_: Schema1794 = Field(..., alias="schema")


class Field404397(Field400389):
    pass


class Field40363(Field400389):
    pass


class Field200344(Field400389):
    pass


class Responses397(BaseModel):
    field_400: Field400389 = Field(..., alias="400")
    field_404: Field404397 = Field(..., alias="404")
    field_403: Field40363 = Field(..., alias="403")
    field_200: Field200344 = Field(..., alias="200")


class Get191(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses397
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400390(Field400389):
    pass


class Field404398(Field400389):
    pass


class Field429187(Field400389):
    pass


class Field40364(Field400389):
    pass


class Schema1802(Items):
    pass


class Field20464(BaseModel):
    description: str
    schema_: Schema1802 = Field(..., alias="schema")


class Responses398(BaseModel):
    field_400: Field400390 = Field(..., alias="400")
    field_404: Field404398 = Field(..., alias="404")
    field_429: Field429187 = Field(..., alias="429")
    field_403: Field40364 = Field(..., alias="403")
    field_204: Field20464 = Field(..., alias="204")


class Delete51(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses398
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsSaml(BaseModel):
    get: Get191
    delete: Delete51


class Schema1803(Schema):
    pass


class Field400391(BaseModel):
    description: str
    schema_: Schema1803 = Field(..., alias="schema")


class Field404399(Field400391):
    pass


class Field40365(Field400391):
    pass


class Field200345(Field400391):
    pass


class Responses399(BaseModel):
    field_400: Field400391 = Field(..., alias="400")
    field_404: Field404399 = Field(..., alias="404")
    field_403: Field40365 = Field(..., alias="403")
    field_200: Field200345 = Field(..., alias="200")


class Get192(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses399
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400392(Field400391):
    pass


class Field404400(Field400391):
    pass


class Field429188(Field400391):
    pass


class Field40366(Field400391):
    pass


class Schema1811(Items):
    pass


class Field20465(BaseModel):
    description: str
    schema_: Schema1811 = Field(..., alias="schema")


class Responses400(BaseModel):
    field_400: Field400392 = Field(..., alias="400")
    field_404: Field404400 = Field(..., alias="404")
    field_429: Field429188 = Field(..., alias="429")
    field_403: Field40366 = Field(..., alias="403")
    field_204: Field20465 = Field(..., alias="204")


class Delete52(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses400
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsOidc(BaseModel):
    get: Get192
    delete: Delete52


class Schema1812(Schema):
    pass


class Field400393(BaseModel):
    description: str
    schema_: Schema1812 = Field(..., alias="schema")


class Field404401(Field400393):
    pass


class Field40367(Field400393):
    pass


class Field200346(Field400393):
    pass


class Responses401(BaseModel):
    field_400: Field400393 = Field(..., alias="400")
    field_404: Field404401 = Field(..., alias="404")
    field_403: Field40367 = Field(..., alias="403")
    field_200: Field200346 = Field(..., alias="200")


class Get193(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses401
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400394(Field400393):
    pass


class Field404402(Field400393):
    pass


class Field429189(Field400393):
    pass


class Field40368(Field400393):
    pass


class Schema1820(Items):
    pass


class Field20466(BaseModel):
    description: str
    schema_: Schema1820 = Field(..., alias="schema")


class Responses402(BaseModel):
    field_400: Field400394 = Field(..., alias="400")
    field_404: Field404402 = Field(..., alias="404")
    field_429: Field429189 = Field(..., alias="429")
    field_403: Field40368 = Field(..., alias="403")
    field_204: Field20466 = Field(..., alias="204")


class Delete53(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses402
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsApi3CredentialsApi3Id(BaseModel):
    get: Get193
    delete: Delete53


class Schema1821(Schema):
    pass


class Field400395(BaseModel):
    description: str
    schema_: Schema1821 = Field(..., alias="schema")


class Field404403(Field400395):
    pass


class Field40369(Field400395):
    pass


class Schema1824(Schema1705):
    pass


class Field200347(BaseModel):
    description: str
    schema_: Schema1824 = Field(..., alias="schema")


class Responses403(BaseModel):
    field_400: Field400395 = Field(..., alias="400")
    field_404: Field404403 = Field(..., alias="404")
    field_403: Field40369 = Field(..., alias="403")
    field_200: Field200347 = Field(..., alias="200")


class Get194(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses403
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Schema1825(Schema):
    pass


class Field400396(BaseModel):
    description: str
    schema_: Schema1825 = Field(..., alias="schema")


class Field404404(Field400396):
    pass


class Field40963(Field400396):
    pass


class Field422166(Field400396):
    pass


class Field429190(Field400396):
    pass


class Field40370(Field400396):
    pass


class Field200348(Field400396):
    pass


class Responses404(BaseModel):
    field_400: Field400396 = Field(..., alias="400")
    field_404: Field404404 = Field(..., alias="404")
    field_409: Field40963 = Field(..., alias="409")
    field_422: Field422166 = Field(..., alias="422")
    field_429: Field429190 = Field(..., alias="429")
    field_403: Field40370 = Field(..., alias="403")
    field_200: Field200348 = Field(..., alias="200")


class Post85(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses404
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsApi3(BaseModel):
    get: Get194
    post: Post85


class Field400397(Field400396):
    pass


class Field404405(Field400396):
    pass


class Field40371(Field400396):
    pass


class Field200349(Field400396):
    pass


class Responses405(BaseModel):
    field_400: Field400397 = Field(..., alias="400")
    field_404: Field404405 = Field(..., alias="404")
    field_403: Field40371 = Field(..., alias="403")
    field_200: Field200349 = Field(..., alias="200")


class Get195(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses405
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400398(Field400396):
    pass


class Field404406(Field400396):
    pass


class Field429191(Field400396):
    pass


class Field40372(Field400396):
    pass


class Schema1840(Items):
    pass


class Field20467(BaseModel):
    description: str
    schema_: Schema1840 = Field(..., alias="schema")


class Responses406(BaseModel):
    field_400: Field400398 = Field(..., alias="400")
    field_404: Field404406 = Field(..., alias="404")
    field_429: Field429191 = Field(..., alias="429")
    field_403: Field40372 = Field(..., alias="403")
    field_204: Field20467 = Field(..., alias="204")


class Delete54(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses406
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsEmbedCredentialsEmbedId(BaseModel):
    get: Get195
    delete: Delete54


class Schema1841(Schema):
    pass


class Field400399(BaseModel):
    description: str
    schema_: Schema1841 = Field(..., alias="schema")


class Field404407(Field400399):
    pass


class Field40373(Field400399):
    pass


class Schema1844(Schema1705):
    pass


class Field200350(BaseModel):
    description: str
    schema_: Schema1844 = Field(..., alias="schema")


class Responses407(BaseModel):
    field_400: Field400399 = Field(..., alias="400")
    field_404: Field404407 = Field(..., alias="404")
    field_403: Field40373 = Field(..., alias="403")
    field_200: Field200350 = Field(..., alias="200")


class Get196(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses407
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsEmbed(BaseModel):
    get: Get196


class Schema1845(Schema):
    pass


class Field400400(BaseModel):
    description: str
    schema_: Schema1845 = Field(..., alias="schema")


class Field404408(Field400400):
    pass


class Field40374(Field400400):
    pass


class Field200351(Field400400):
    pass


class Responses408(BaseModel):
    field_400: Field400400 = Field(..., alias="400")
    field_404: Field404408 = Field(..., alias="404")
    field_403: Field40374 = Field(..., alias="403")
    field_200: Field200351 = Field(..., alias="200")


class Get197(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses408
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400401(Field400400):
    pass


class Field404409(Field400400):
    pass


class Field429192(Field400400):
    pass


class Field40375(Field400400):
    pass


class Schema1853(Items):
    pass


class Field20468(BaseModel):
    description: str
    schema_: Schema1853 = Field(..., alias="schema")


class Responses409(BaseModel):
    field_400: Field400401 = Field(..., alias="400")
    field_404: Field404409 = Field(..., alias="404")
    field_429: Field429192 = Field(..., alias="429")
    field_403: Field40375 = Field(..., alias="403")
    field_204: Field20468 = Field(..., alias="204")


class Delete55(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses409
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsLookerOpenid(BaseModel):
    get: Get197
    delete: Delete55


class Schema1854(Schema):
    pass


class Field400402(BaseModel):
    description: str
    schema_: Schema1854 = Field(..., alias="schema")


class Field404410(Field400402):
    pass


class Field40376(Field400402):
    pass


class Field200352(Field400402):
    pass


class Responses410(BaseModel):
    field_400: Field400402 = Field(..., alias="400")
    field_404: Field404410 = Field(..., alias="404")
    field_403: Field40376 = Field(..., alias="403")
    field_200: Field200352 = Field(..., alias="200")


class Get198(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses410
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Field400403(Field400402):
    pass


class Field404411(Field400402):
    pass


class Field429193(Field400402):
    pass


class Field40377(Field400402):
    pass


class Schema1862(Items):
    pass


class Field20469(BaseModel):
    description: str
    schema_: Schema1862 = Field(..., alias="schema")


class Responses411(BaseModel):
    field_400: Field400403 = Field(..., alias="400")
    field_404: Field404411 = Field(..., alias="404")
    field_429: Field429193 = Field(..., alias="429")
    field_403: Field40377 = Field(..., alias="403")
    field_204: Field20469 = Field(..., alias="204")


class Delete56(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses411
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdSessionsSessionId(BaseModel):
    get: Get198
    delete: Delete56


class Schema1863(Schema):
    pass


class Field400404(BaseModel):
    description: str
    schema_: Schema1863 = Field(..., alias="schema")


class Field404412(Field400404):
    pass


class Field40378(Field400404):
    pass


class Schema1866(Schema1705):
    pass


class Field200353(BaseModel):
    description: str
    schema_: Schema1866 = Field(..., alias="schema")


class Responses412(BaseModel):
    field_400: Field400404 = Field(..., alias="400")
    field_404: Field404412 = Field(..., alias="404")
    field_403: Field40378 = Field(..., alias="403")
    field_200: Field200353 = Field(..., alias="200")


class Get199(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses412
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdSessions(BaseModel):
    get: Get199


class Schema1867(Schema):
    pass


class Field400405(BaseModel):
    description: str
    schema_: Schema1867 = Field(..., alias="schema")


class Field40379(Field400405):
    pass


class Field404413(Field400405):
    pass


class Field200354(Field400405):
    pass


class Responses413(BaseModel):
    field_400: Field400405 = Field(..., alias="400")
    field_403: Field40379 = Field(..., alias="403")
    field_404: Field404413 = Field(..., alias="404")
    field_200: Field200354 = Field(..., alias="200")


class Post86(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses413
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsEmailPasswordReset(BaseModel):
    post: Post86


class Field400406(Field400405):
    pass


class Field404414(Field400405):
    pass


class Schema1873(Schema1705):
    pass


class Field200355(BaseModel):
    description: str
    schema_: Schema1873 = Field(..., alias="schema")


class Responses414(BaseModel):
    field_400: Field400406 = Field(..., alias="400")
    field_404: Field404414 = Field(..., alias="404")
    field_200: Field200355 = Field(..., alias="200")


class Get200(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter366]
    responses: Responses414
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Items115(Items):
    pass


class Schema1874(BaseModel):
    type: str
    items: Items115


class Parameter389(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1874] = Field(None, alias="schema")


class Schema1875(Schema):
    pass


class Field400407(BaseModel):
    description: str
    schema_: Schema1875 = Field(..., alias="schema")


class Field40380(Field400407):
    pass


class Field404415(Field400407):
    pass


class Items116(Schema):
    pass


class Schema1878(BaseModel):
    type: str
    items: Items116


class Field200356(BaseModel):
    description: str
    schema_: Schema1878 = Field(..., alias="schema")


class Responses415(BaseModel):
    field_400: Field400407 = Field(..., alias="400")
    field_403: Field40380 = Field(..., alias="403")
    field_404: Field404415 = Field(..., alias="404")
    field_200: Field200356 = Field(..., alias="200")


class Put20(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter389]
    responses: Responses415
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdRoles(BaseModel):
    get: Get200
    put: Put20


class Items117(Items):
    pass


class Parameter390(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: str
    items: Optional[Items117] = None
    collectionFormat: Optional[str] = None


class Schema1879(Schema):
    pass


class Field404416(BaseModel):
    description: str
    schema_: Schema1879 = Field(..., alias="schema")


class Items118(Schema):
    pass


class Schema1880(BaseModel):
    type: str
    items: Items118


class Field200357(BaseModel):
    description: str
    schema_: Schema1880 = Field(..., alias="schema")


class Responses416(BaseModel):
    field_404: Field404416 = Field(..., alias="404")
    field_200: Field200357 = Field(..., alias="200")


class Get201(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter390]
    responses: Responses416
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdAttributeValues(BaseModel):
    get: Get201


class Schema1881(Schema):
    pass


class Parameter391(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1881] = Field(None, alias="schema")


class Field400408(BaseModel):
    description: str
    schema_: Schema1881 = Field(..., alias="schema")


class Field40381(Field400408):
    pass


class Field404417(Field400408):
    pass


class Field422167(Field400408):
    pass


class Field200358(Field400408):
    pass


class Responses417(BaseModel):
    field_400: Field400408 = Field(..., alias="400")
    field_403: Field40381 = Field(..., alias="403")
    field_404: Field404417 = Field(..., alias="404")
    field_422: Field422167 = Field(..., alias="422")
    field_200: Field200358 = Field(..., alias="200")


class Patch50(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter391]
    responses: Responses417
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter392(Parameter2):
    pass


class Field20470(Field2042):
    pass


class Field400409(Field400408):
    pass


class Field40382(Field400408):
    pass


class Field404418(Field400408):
    pass


class Responses418(BaseModel):
    field_204: Field20470 = Field(..., alias="204")
    field_400: Field400409 = Field(..., alias="400")
    field_403: Field40382 = Field(..., alias="403")
    field_404: Field404418 = Field(..., alias="404")


class Delete57(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter392]
    responses: Responses418
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdAttributeValuesUserAttributeId(BaseModel):
    patch: Patch50
    delete: Delete57


class Field400410(Field400408):
    pass


class Field40383(Field400408):
    pass


class Field404419(Field400408):
    pass


class Field200359(Field400408):
    pass


class Responses419(BaseModel):
    field_400: Field400410 = Field(..., alias="400")
    field_403: Field40383 = Field(..., alias="403")
    field_404: Field404419 = Field(..., alias="404")
    field_200: Field200359 = Field(..., alias="200")


class Post87(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter392]
    responses: Responses419
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdCredentialsEmailSendPasswordReset(BaseModel):
    post: Post87


class Parameter394(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: Optional[str]
    required: bool
    type: Optional[str] = None
    schema_: Optional[Schema1881] = Field(None, alias="schema")


class Field400411(Field400408):
    pass


class Field40384(Field400408):
    pass


class Field404420(Field400408):
    pass


class Field40964(Field400408):
    pass


class Field422168(Field400408):
    pass


class Field200360(Field400408):
    pass


class Responses420(BaseModel):
    field_400: Field400411 = Field(..., alias="400")
    field_403: Field40384 = Field(..., alias="403")
    field_404: Field404420 = Field(..., alias="404")
    field_409: Field40964 = Field(..., alias="409")
    field_422: Field422168 = Field(..., alias="422")
    field_200: Field200360 = Field(..., alias="200")


class Post88(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter394]
    responses: Responses420
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersUserIdUpdateEmails(BaseModel):
    post: Post88


class Parameter395(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: None
    required: bool
    schema_: Schema1881 = Field(..., alias="schema")


class Field400412(Field400408):
    pass


class Field404421(Field400408):
    pass


class Field200361(Field400408):
    pass


class Responses421(BaseModel):
    field_400: Field400412 = Field(..., alias="400")
    field_404: Field404421 = Field(..., alias="404")
    field_200: Field200361 = Field(..., alias="200")


class Post89(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter395]
    responses: Responses421
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldUsersEmbedUser(BaseModel):
    post: Post89


class Parameter396(Parameter2):
    pass


class Field400413(Field400408):
    pass


class Field404422(Field400408):
    pass


class Schema1907(Items):
    pass


class Field200362(BaseModel):
    description: str
    schema_: Schema1907 = Field(..., alias="schema")


class Responses422(BaseModel):
    field_400: Field400413 = Field(..., alias="400")
    field_404: Field404422 = Field(..., alias="404")
    field_200: Field200362 = Field(..., alias="200")


class Get202(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    produces: List[str]
    parameters: List[Parameter396]
    responses: Responses422
    deprecated: bool
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldVectorThumbnailTypeResourceId(BaseModel):
    get: Get202


class Schema1908(Schema):
    pass


class Field400414(BaseModel):
    description: str
    schema_: Schema1908 = Field(..., alias="schema")


class Field404423(Field400414):
    pass


class Field200363(Field400414):
    pass


class Responses423(BaseModel):
    field_400: Field400414 = Field(..., alias="400")
    field_404: Field404423 = Field(..., alias="404")
    field_200: Field200363 = Field(..., alias="200")


class Get203(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter396]
    responses: Responses423
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldVersions(BaseModel):
    get: Get203


class Schema1911(AdditionalProperties):
    pass


class Field200364(BaseModel):
    description: str
    schema_: Schema1911 = Field(..., alias="schema")


class Schema1912(Schema):
    pass


class Field400415(BaseModel):
    description: str
    schema_: Schema1912 = Field(..., alias="schema")


class Field404424(Field400415):
    pass


class Responses424(BaseModel):
    field_200: Field200364 = Field(..., alias="200")
    field_400: Field400415 = Field(..., alias="400")
    field_404: Field404424 = Field(..., alias="404")


class Get204(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter396]
    responses: Responses424
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldApiSpecApiVersionSpecification(BaseModel):
    get: Get204


class Field400416(Field400415):
    pass


class Field404425(Field400415):
    pass


class Field200365(Field400415):
    pass


class Responses425(BaseModel):
    field_400: Field400416 = Field(..., alias="400")
    field_404: Field404425 = Field(..., alias="404")
    field_200: Field200365 = Field(..., alias="200")


class Get205(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter396]
    responses: Responses425
    deprecated: bool
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class Parameter400(BaseModel):
    name: str
    in_: str = Field(..., alias="in")
    description: str
    required: bool
    schema_: Schema1912 = Field(..., alias="schema")


class Field400417(Field400415):
    pass


class Field404426(Field400415):
    pass


class Field422169(Field400415):
    pass


class Field429194(Field400415):
    pass


class Field200366(Field400415):
    pass


class Responses426(BaseModel):
    field_400: Field400417 = Field(..., alias="400")
    field_404: Field404426 = Field(..., alias="404")
    field_422: Field422169 = Field(..., alias="422")
    field_429: Field429194 = Field(..., alias="429")
    field_200: Field200366 = Field(..., alias="200")


class Put21(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter400]
    responses: Responses426
    deprecated: bool
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldWhitelabelConfiguration(BaseModel):
    get: Get205
    put: Put21


class Field400418(Field400415):
    pass


class Field404427(Field400415):
    pass


class Schema1925(Schema1880):
    pass


class Field200367(BaseModel):
    description: str
    schema_: Schema1925 = Field(..., alias="schema")


class Responses427(BaseModel):
    field_400: Field400418 = Field(..., alias="400")
    field_404: Field404427 = Field(..., alias="404")
    field_200: Field200367 = Field(..., alias="200")


class Get206(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    responses: Responses427
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldWorkspaces(BaseModel):
    get: Get206


class Parameter401(Parameter2):
    pass


class Schema1926(Schema):
    pass


class Field400419(BaseModel):
    description: str
    schema_: Schema1926 = Field(..., alias="schema")


class Field404428(Field400419):
    pass


class Field200368(Field400419):
    pass


class Responses428(BaseModel):
    field_400: Field400419 = Field(..., alias="400")
    field_404: Field404428 = Field(..., alias="404")
    field_200: Field200368 = Field(..., alias="200")


class Get207(BaseModel):
    tags: List[str]
    operationId: str
    summary: str
    description: str
    parameters: List[Parameter401]
    responses: Responses428
    x_looker_status: str = Field(..., alias="x-looker-status")
    x_looker_activity_type: str = Field(..., alias="x-looker-activity-type")


class FieldWorkspacesWorkspaceId(BaseModel):
    get: Get207


class Paths(BaseModel):
    field_query_tasks: FieldQueryTasks = Field(..., alias="/query_tasks")
    field_query_tasks_multi_results: FieldQueryTasksMultiResults = Field(..., alias="/query_tasks/multi_results")
    field_query_tasks__query_task_id_: FieldQueryTasksQueryTaskId = Field(..., alias="/query_tasks/{query_task_id}")
    field_query_tasks__query_task_id__results: FieldQueryTasksQueryTaskIdResults = Field(
        ..., alias="/query_tasks/{query_task_id}/results"
    )
    field_queries__query_id_: FieldQueriesQueryId = Field(..., alias="/queries/{query_id}")
    field_queries_slug__slug_: FieldQueriesSlugSlug = Field(..., alias="/queries/slug/{slug}")
    field_queries: FieldQueries = Field(..., alias="/queries")
    field_queries__query_id__run__result_format_: FieldQueriesQueryIdRunResultFormat = Field(
        ..., alias="/queries/{query_id}/run/{result_format}"
    )
    field_queries_run__result_format_: FieldQueriesRunResultFormat = Field(..., alias="/queries/run/{result_format}")
    field_queries_models__model_name__views__view_name__run__result_format_: FieldQueriesModelsModelNameViewsViewNameRunResultFormat = Field(
        ..., alias="/queries/models/{model_name}/views/{view_name}/run/{result_format}"
    )
    field_login: FieldLogin = Field(..., alias="/login")
    field_login__user_id_: FieldLoginUserId = Field(..., alias="/login/{user_id}")
    field_logout: FieldLogout = Field(..., alias="/logout")
    field_alerts__alert_id__follow: FieldAlertsAlertIdFollow = Field(..., alias="/alerts/{alert_id}/follow")
    field_alerts_search: FieldAlertsSearch = Field(..., alias="/alerts/search")
    field_alerts__alert_id_: FieldAlertsAlertId = Field(..., alias="/alerts/{alert_id}")
    field_alerts: FieldAlerts = Field(..., alias="/alerts")
    field_alerts__alert_id__enqueue: FieldAlertsAlertIdEnqueue = Field(..., alias="/alerts/{alert_id}/enqueue")
    field_alert_notifications: FieldAlertNotifications = Field(..., alias="/alert_notifications")
    field_alert_notifications__alert_notification_id_: FieldAlertNotificationsAlertNotificationId = Field(
        ..., alias="/alert_notifications/{alert_notification_id}"
    )
    field_artifact_usage: FieldArtifactUsage = Field(..., alias="/artifact/usage")
    field_artifact_namespaces: FieldArtifactNamespaces = Field(..., alias="/artifact/namespaces")
    field_artifact__namespace__value: FieldArtifactNamespaceValue = Field(..., alias="/artifact/{namespace}/value")
    field_artifact__namespace__purge: FieldArtifactNamespacePurge = Field(..., alias="/artifact/{namespace}/purge")
    field_artifact__namespace__search: FieldArtifactNamespaceSearch = Field(..., alias="/artifact/{namespace}/search")
    field_artifact__namespace_: FieldArtifactNamespace = Field(..., alias="/artifact/{namespace}")
    field_artifacts__namespace_: FieldArtifactsNamespace = Field(..., alias="/artifacts/{namespace}")
    field_cloud_storage: FieldCloudStorage = Field(..., alias="/cloud_storage")
    field_color_collections: FieldColorCollections = Field(..., alias="/color_collections")
    field_color_collections_custom: FieldColorCollectionsCustom = Field(..., alias="/color_collections/custom")
    field_color_collections_standard: FieldColorCollectionsStandard = Field(..., alias="/color_collections/standard")
    field_color_collections_default: FieldColorCollectionsDefault = Field(..., alias="/color_collections/default")
    field_color_collections__collection_id_: FieldColorCollectionsCollectionId = Field(
        ..., alias="/color_collections/{collection_id}"
    )
    field_content_favorite_search: FieldContentFavoriteSearch = Field(..., alias="/content_favorite/search")
    field_content_favorite__content_favorite_id_: FieldContentFavoriteContentFavoriteId = Field(
        ..., alias="/content_favorite/{content_favorite_id}"
    )
    field_content_favorite: FieldContentFavorite = Field(..., alias="/content_favorite")
    field_content_metadata: FieldContentMetadata = Field(..., alias="/content_metadata")
    field_content_metadata__content_metadata_id_: FieldContentMetadataContentMetadataId = Field(
        ..., alias="/content_metadata/{content_metadata_id}"
    )
    field_content_metadata_access: FieldContentMetadataAccess = Field(..., alias="/content_metadata_access")
    field_content_metadata_access__content_metadata_access_id_: FieldContentMetadataAccessContentMetadataAccessId = (
        Field(..., alias="/content_metadata_access/{content_metadata_access_id}")
    )
    field_content__terms_: FieldContentTerms = Field(..., alias="/content/{terms}")
    field_content_thumbnail__type___resource_id_: FieldContentThumbnailTypeResourceId = Field(
        ..., alias="/content_thumbnail/{type}/{resource_id}"
    )
    field_content_validation: FieldContentValidation = Field(..., alias="/content_validation")
    field_content_view_search: FieldContentViewSearch = Field(..., alias="/content_view/search")
    field_credentials_email_search: FieldCredentialsEmailSearch = Field(..., alias="/credentials_email/search")
    field_custom_welcome_email: FieldCustomWelcomeEmail = Field(..., alias="/custom_welcome_email")
    field_custom_welcome_email_test: FieldCustomWelcomeEmailTest = Field(..., alias="/custom_welcome_email_test")
    field_dashboards: FieldDashboards = Field(..., alias="/dashboards")
    field_dashboards_search: FieldDashboardsSearch = Field(..., alias="/dashboards/search")
    field_dashboards__lookml_dashboard_id__import__space_id_: FieldDashboardsLookmlDashboardIdImportSpaceId = Field(
        ..., alias="/dashboards/{lookml_dashboard_id}/import/{space_id}"
    )
    field_dashboards__lookml_dashboard_id__sync: FieldDashboardsLookmlDashboardIdSync = Field(
        ..., alias="/dashboards/{lookml_dashboard_id}/sync"
    )
    field_dashboards__dashboard_id_: FieldDashboardsDashboardId = Field(..., alias="/dashboards/{dashboard_id}")
    field_dashboards_aggregate_table_lookml__dashboard_id_: FieldDashboardsAggregateTableLookmlDashboardId = Field(
        ..., alias="/dashboards/aggregate_table_lookml/{dashboard_id}"
    )
    field_dashboards_lookml__dashboard_id_: FieldDashboardsLookmlDashboardId = Field(
        ..., alias="/dashboards/lookml/{dashboard_id}"
    )
    field_dashboards__dashboard_id__move: FieldDashboardsDashboardIdMove = Field(
        ..., alias="/dashboards/{dashboard_id}/move"
    )
    field_dashboards_lookml: FieldDashboardsLookml = Field(..., alias="/dashboards/lookml")
    field_dashboards_from_lookml: FieldDashboardsFromLookml = Field(..., alias="/dashboards/from_lookml")
    field_dashboards__dashboard_id__copy: FieldDashboardsDashboardIdCopy = Field(
        ..., alias="/dashboards/{dashboard_id}/copy"
    )
    field_dashboard_elements_search: FieldDashboardElementsSearch = Field(..., alias="/dashboard_elements/search")
    field_dashboard_elements__dashboard_element_id_: FieldDashboardElementsDashboardElementId = Field(
        ..., alias="/dashboard_elements/{dashboard_element_id}"
    )
    field_dashboards__dashboard_id__dashboard_elements: FieldDashboardsDashboardIdDashboardElements = Field(
        ..., alias="/dashboards/{dashboard_id}/dashboard_elements"
    )
    field_dashboard_elements: FieldDashboardElements = Field(..., alias="/dashboard_elements")
    field_dashboard_filters__dashboard_filter_id_: FieldDashboardFiltersDashboardFilterId = Field(
        ..., alias="/dashboard_filters/{dashboard_filter_id}"
    )
    field_dashboards__dashboard_id__dashboard_filters: FieldDashboardsDashboardIdDashboardFilters = Field(
        ..., alias="/dashboards/{dashboard_id}/dashboard_filters"
    )
    field_dashboard_filters: FieldDashboardFilters = Field(..., alias="/dashboard_filters")
    field_dashboard_layout_components__dashboard_layout_component_id_: FieldDashboardLayoutComponentsDashboardLayoutComponentId = Field(
        ..., alias="/dashboard_layout_components/{dashboard_layout_component_id}"
    )
    field_dashboard_layouts__dashboard_layout_id__dashboard_layout_components: FieldDashboardLayoutsDashboardLayoutIdDashboardLayoutComponents = Field(
        ..., alias="/dashboard_layouts/{dashboard_layout_id}/dashboard_layout_components"
    )
    field_dashboard_layouts__dashboard_layout_id_: FieldDashboardLayoutsDashboardLayoutId = Field(
        ..., alias="/dashboard_layouts/{dashboard_layout_id}"
    )
    field_dashboards__dashboard_id__dashboard_layouts: FieldDashboardsDashboardIdDashboardLayouts = Field(
        ..., alias="/dashboards/{dashboard_id}/dashboard_layouts"
    )
    field_dashboard_layouts: FieldDashboardLayouts = Field(..., alias="/dashboard_layouts")
    field_data_actions: FieldDataActions = Field(..., alias="/data_actions")
    field_data_actions_form: FieldDataActionsForm = Field(..., alias="/data_actions/form")
    field_datagroups: FieldDatagroups = Field(..., alias="/datagroups")
    field_datagroups__datagroup_id_: FieldDatagroupsDatagroupId = Field(..., alias="/datagroups/{datagroup_id}")
    field_connections: FieldConnections = Field(..., alias="/connections")
    field_connections__connection_name_: FieldConnectionsConnectionName = Field(
        ..., alias="/connections/{connection_name}"
    )
    field_connections__connection_name__connection_override__override_context_: FieldConnectionsConnectionNameConnectionOverrideOverrideContext = Field(
        ..., alias="/connections/{connection_name}/connection_override/{override_context}"
    )
    field_connections__connection_name__test: FieldConnectionsConnectionNameTest = Field(
        ..., alias="/connections/{connection_name}/test"
    )
    field_connections_test: FieldConnectionsTest = Field(..., alias="/connections/test")
    field_projects__project_id__manifest_lock_all: FieldProjectsProjectIdManifestLockAll = Field(
        ..., alias="/projects/{project_id}/manifest/lock_all"
    )
    field_derived_table_graph_model__model_: FieldDerivedTableGraphModelModel = Field(
        ..., alias="/derived_table/graph/model/{model}"
    )
    field_derived_table_graph_view__view_: FieldDerivedTableGraphViewView = Field(
        ..., alias="/derived_table/graph/view/{view}"
    )
    field_derived_table__model_name___view_name__start: FieldDerivedTableModelNameViewNameStart = Field(
        ..., alias="/derived_table/{model_name}/{view_name}/start"
    )
    field_derived_table__materialization_id__status: FieldDerivedTableMaterializationIdStatus = Field(
        ..., alias="/derived_table/{materialization_id}/status"
    )
    field_derived_table__materialization_id__stop: FieldDerivedTableMaterializationIdStop = Field(
        ..., alias="/derived_table/{materialization_id}/stop"
    )
    field_dialect_info: FieldDialectInfo = Field(..., alias="/dialect_info")
    field_digest_emails_enabled: FieldDigestEmailsEnabled = Field(..., alias="/digest_emails_enabled")
    field_digest_email_send: FieldDigestEmailSend = Field(..., alias="/digest_email_send")
    field_public_egress_ip_addresses: FieldPublicEgressIpAddresses = Field(..., alias="/public_egress_ip_addresses")
    field_embed_config_secrets: FieldEmbedConfigSecrets = Field(..., alias="/embed_config/secrets")
    field_embed_config_secrets__embed_secret_id_: FieldEmbedConfigSecretsEmbedSecretId = Field(
        ..., alias="/embed_config/secrets/{embed_secret_id}"
    )
    field_embed_sso_url: FieldEmbedSsoUrl = Field(..., alias="/embed/sso_url")
    field_embed_token_url_me: FieldEmbedTokenUrlMe = Field(..., alias="/embed/token_url/me")
    field_embed_cookieless_session_acquire: FieldEmbedCookielessSessionAcquire = Field(
        ..., alias="/embed/cookieless_session/acquire"
    )
    field_embed_cookieless_session__session_reference_token_: FieldEmbedCookielessSessionSessionReferenceToken = Field(
        ..., alias="/embed/cookieless_session/{session_reference_token}"
    )
    field_embed_cookieless_session_generate_tokens: FieldEmbedCookielessSessionGenerateTokens = Field(
        ..., alias="/embed/cookieless_session/generate_tokens"
    )
    field_external_oauth_applications: FieldExternalOauthApplications = Field(..., alias="/external_oauth_applications")
    field_external_oauth_applications_user_state: FieldExternalOauthApplicationsUserState = Field(
        ..., alias="/external_oauth_applications/user_state"
    )
    field_projects__project_id__git_branches: FieldProjectsProjectIdGitBranches = Field(
        ..., alias="/projects/{project_id}/git_branches"
    )
    field_projects__project_id__git_branch: FieldProjectsProjectIdGitBranch = Field(
        ..., alias="/projects/{project_id}/git_branch"
    )
    field_projects__project_id__git_branch__branch_name_: FieldProjectsProjectIdGitBranchBranchName = Field(
        ..., alias="/projects/{project_id}/git_branch/{branch_name}"
    )
    field_groups: FieldGroups = Field(..., alias="/groups")
    field_groups_search: FieldGroupsSearch = Field(..., alias="/groups/search")
    field_groups_search_with_roles: FieldGroupsSearchWithRoles = Field(..., alias="/groups/search/with_roles")
    field_groups_search_with_hierarchy: FieldGroupsSearchWithHierarchy = Field(
        ..., alias="/groups/search/with_hierarchy"
    )
    field_groups__group_id_: FieldGroupsGroupId = Field(..., alias="/groups/{group_id}")
    field_groups__group_id__groups: FieldGroupsGroupIdGroups = Field(..., alias="/groups/{group_id}/groups")
    field_groups__group_id__users: FieldGroupsGroupIdUsers = Field(..., alias="/groups/{group_id}/users")
    field_groups__group_id__users__user_id_: FieldGroupsGroupIdUsersUserId = Field(
        ..., alias="/groups/{group_id}/users/{user_id}"
    )
    field_groups__group_id__groups__deleting_group_id_: FieldGroupsGroupIdGroupsDeletingGroupId = Field(
        ..., alias="/groups/{group_id}/groups/{deleting_group_id}"
    )
    field_groups__group_id__attribute_values__user_attribute_id_: FieldGroupsGroupIdAttributeValuesUserAttributeId = (
        Field(..., alias="/groups/{group_id}/attribute_values/{user_attribute_id}")
    )
    field_boards: FieldBoards = Field(..., alias="/boards")
    field_boards_search: FieldBoardsSearch = Field(..., alias="/boards/search")
    field_boards__board_id_: FieldBoardsBoardId = Field(..., alias="/boards/{board_id}")
    field_board_items: FieldBoardItems = Field(..., alias="/board_items")
    field_board_items__board_item_id_: FieldBoardItemsBoardItemId = Field(..., alias="/board_items/{board_item_id}")
    field_primary_homepage_sections: FieldPrimaryHomepageSections = Field(..., alias="/primary_homepage_sections")
    field_board_sections: FieldBoardSections = Field(..., alias="/board_sections")
    field_board_sections__board_section_id_: FieldBoardSectionsBoardSectionId = Field(
        ..., alias="/board_sections/{board_section_id}"
    )
    field_integration_hubs: FieldIntegrationHubs = Field(..., alias="/integration_hubs")
    field_integration_hubs__integration_hub_id_: FieldIntegrationHubsIntegrationHubId = Field(
        ..., alias="/integration_hubs/{integration_hub_id}"
    )
    field_integration_hubs__integration_hub_id__accept_legal_agreement: FieldIntegrationHubsIntegrationHubIdAcceptLegalAgreement = Field(
        ..., alias="/integration_hubs/{integration_hub_id}/accept_legal_agreement"
    )
    field_integrations: FieldIntegrations = Field(..., alias="/integrations")
    field_integrations__integration_id_: FieldIntegrationsIntegrationId = Field(
        ..., alias="/integrations/{integration_id}"
    )
    field_integrations__integration_id__form: FieldIntegrationsIntegrationIdForm = Field(
        ..., alias="/integrations/{integration_id}/form"
    )
    field_integrations__integration_id__test: FieldIntegrationsIntegrationIdTest = Field(
        ..., alias="/integrations/{integration_id}/test"
    )
    field_internal_help_resources_content: FieldInternalHelpResourcesContent = Field(
        ..., alias="/internal_help_resources_content"
    )
    field_internal_help_resources_enabled: FieldInternalHelpResourcesEnabled = Field(
        ..., alias="/internal_help_resources_enabled"
    )
    field_internal_help_resources: FieldInternalHelpResources = Field(..., alias="/internal_help_resources")
    field_ldap_config: FieldLdapConfig = Field(..., alias="/ldap_config")
    field_ldap_config_test_connection: FieldLdapConfigTestConnection = Field(..., alias="/ldap_config/test_connection")
    field_ldap_config_test_auth: FieldLdapConfigTestAuth = Field(..., alias="/ldap_config/test_auth")
    field_ldap_config_test_user_info: FieldLdapConfigTestUserInfo = Field(..., alias="/ldap_config/test_user_info")
    field_ldap_config_test_user_auth: FieldLdapConfigTestUserAuth = Field(..., alias="/ldap_config/test_user_auth")
    field_legacy_features: FieldLegacyFeatures = Field(..., alias="/legacy_features")
    field_legacy_features__legacy_feature_id_: FieldLegacyFeaturesLegacyFeatureId = Field(
        ..., alias="/legacy_features/{legacy_feature_id}"
    )
    field_locales: FieldLocales = Field(..., alias="/locales")
    field_looks: FieldLooks = Field(..., alias="/looks")
    field_looks_search: FieldLooksSearch = Field(..., alias="/looks/search")
    field_looks__look_id_: FieldLooksLookId = Field(..., alias="/looks/{look_id}")
    field_looks__look_id__run__result_format_: FieldLooksLookIdRunResultFormat = Field(
        ..., alias="/looks/{look_id}/run/{result_format}"
    )
    field_looks__look_id__copy: FieldLooksLookIdCopy = Field(..., alias="/looks/{look_id}/copy")
    field_looks__look_id__move: FieldLooksLookIdMove = Field(..., alias="/looks/{look_id}/move")
    field_lookml_models: FieldLookmlModels = Field(..., alias="/lookml_models")
    field_lookml_models__lookml_model_name_: FieldLookmlModelsLookmlModelName = Field(
        ..., alias="/lookml_models/{lookml_model_name}"
    )
    field_lookml_models__lookml_model_name__explores__explore_name_: FieldLookmlModelsLookmlModelNameExploresExploreName = Field(
        ..., alias="/lookml_models/{lookml_model_name}/explores/{explore_name}"
    )
    field_merge_queries__merge_query_id_: FieldMergeQueriesMergeQueryId = Field(
        ..., alias="/merge_queries/{merge_query_id}"
    )
    field_merge_queries: FieldMergeQueries = Field(..., alias="/merge_queries")
    field_models__model_name__views__view_name__fields__field_name__suggestions: FieldModelsModelNameViewsViewNameFieldsFieldNameSuggestions = Field(
        ..., alias="/models/{model_name}/views/{view_name}/fields/{field_name}/suggestions"
    )
    field_models__model_name_: FieldModelsModelName = Field(..., alias="/models/{model_name}")
    field_connections__connection_name__databases: FieldConnectionsConnectionNameDatabases = Field(
        ..., alias="/connections/{connection_name}/databases"
    )
    field_connections__connection_name__features: FieldConnectionsConnectionNameFeatures = Field(
        ..., alias="/connections/{connection_name}/features"
    )
    field_connections__connection_name__schemas: FieldConnectionsConnectionNameSchemas = Field(
        ..., alias="/connections/{connection_name}/schemas"
    )
    field_connections__connection_name__tables: FieldConnectionsConnectionNameTables = Field(
        ..., alias="/connections/{connection_name}/tables"
    )
    field_connections__connection_name__columns: FieldConnectionsConnectionNameColumns = Field(
        ..., alias="/connections/{connection_name}/columns"
    )
    field_connections__connection_name__search_columns: FieldConnectionsConnectionNameSearchColumns = Field(
        ..., alias="/connections/{connection_name}/search_columns"
    )
    field_connections__connection_name__cost_estimate: FieldConnectionsConnectionNameCostEstimate = Field(
        ..., alias="/connections/{connection_name}/cost_estimate"
    )
    field_mobile_settings: FieldMobileSettings = Field(..., alias="/mobile/settings")
    field_mobile_device: FieldMobileDevice = Field(..., alias="/mobile/device")
    field_mobile_device__device_id_: FieldMobileDeviceDeviceId = Field(..., alias="/mobile/device/{device_id}")
    field_model_sets_search: FieldModelSetsSearch = Field(..., alias="/model_sets/search")
    field_model_sets__model_set_id_: FieldModelSetsModelSetId = Field(..., alias="/model_sets/{model_set_id}")
    field_model_sets: FieldModelSets = Field(..., alias="/model_sets")
    field_oauth_client_apps: FieldOauthClientApps = Field(..., alias="/oauth_client_apps")
    field_oauth_client_apps__client_guid_: FieldOauthClientAppsClientGuid = Field(
        ..., alias="/oauth_client_apps/{client_guid}"
    )
    field_oauth_client_apps__client_guid__tokens: FieldOauthClientAppsClientGuidTokens = Field(
        ..., alias="/oauth_client_apps/{client_guid}/tokens"
    )
    field_oauth_client_apps__client_guid__users__user_id_: FieldOauthClientAppsClientGuidUsersUserId = Field(
        ..., alias="/oauth_client_apps/{client_guid}/users/{user_id}"
    )
    field_oidc_config: FieldOidcConfig = Field(..., alias="/oidc_config")
    field_oidc_test_configs__test_slug_: FieldOidcTestConfigsTestSlug = Field(
        ..., alias="/oidc_test_configs/{test_slug}"
    )
    field_oidc_test_configs: FieldOidcTestConfigs = Field(..., alias="/oidc_test_configs")
    field_password_config: FieldPasswordConfig = Field(..., alias="/password_config")
    field_password_config_force_password_reset_at_next_login_for_all_users: FieldPasswordConfigForcePasswordResetAtNextLoginForAllUsers = Field(
        ..., alias="/password_config/force_password_reset_at_next_login_for_all_users"
    )
    field_permissions: FieldPermissions = Field(..., alias="/permissions")
    field_permission_sets_search: FieldPermissionSetsSearch = Field(..., alias="/permission_sets/search")
    field_permission_sets__permission_set_id_: FieldPermissionSetsPermissionSetId = Field(
        ..., alias="/permission_sets/{permission_set_id}"
    )
    field_permission_sets: FieldPermissionSets = Field(..., alias="/permission_sets")
    field_projects__project_id__deploy_ref_to_production: FieldProjectsProjectIdDeployRefToProduction = Field(
        ..., alias="/projects/{project_id}/deploy_ref_to_production"
    )
    field_projects__project_id__deploy_to_production: FieldProjectsProjectIdDeployToProduction = Field(
        ..., alias="/projects/{project_id}/deploy_to_production"
    )
    field_projects__project_id__reset_to_production: FieldProjectsProjectIdResetToProduction = Field(
        ..., alias="/projects/{project_id}/reset_to_production"
    )
    field_projects__project_id__reset_to_remote: FieldProjectsProjectIdResetToRemote = Field(
        ..., alias="/projects/{project_id}/reset_to_remote"
    )
    field_projects: FieldProjects = Field(..., alias="/projects")
    field_projects__project_id_: FieldProjectsProjectId = Field(..., alias="/projects/{project_id}")
    field_projects__project_id__manifest: FieldProjectsProjectIdManifest = Field(
        ..., alias="/projects/{project_id}/manifest"
    )
    field_projects__project_id__git_deploy_key: FieldProjectsProjectIdGitDeployKey = Field(
        ..., alias="/projects/{project_id}/git/deploy_key"
    )
    field_projects__project_id__validate: FieldProjectsProjectIdValidate = Field(
        ..., alias="/projects/{project_id}/validate"
    )
    field_projects__project_id__current_workspace: FieldProjectsProjectIdCurrentWorkspace = Field(
        ..., alias="/projects/{project_id}/current_workspace"
    )
    field_projects__project_id__files: FieldProjectsProjectIdFiles = Field(..., alias="/projects/{project_id}/files")
    field_projects__project_id__files_file: FieldProjectsProjectIdFilesFile = Field(
        ..., alias="/projects/{project_id}/files/file"
    )
    field_projects__project_id__git_connection_tests: FieldProjectsProjectIdGitConnectionTests = Field(
        ..., alias="/projects/{project_id}/git_connection_tests"
    )
    field_projects__project_id__git_connection_tests__test_id_: FieldProjectsProjectIdGitConnectionTestsTestId = Field(
        ..., alias="/projects/{project_id}/git_connection_tests/{test_id}"
    )
    field_projects__project_id__lookml_tests: FieldProjectsProjectIdLookmlTests = Field(
        ..., alias="/projects/{project_id}/lookml_tests"
    )
    field_projects__project_id__lookml_tests_run: FieldProjectsProjectIdLookmlTestsRun = Field(
        ..., alias="/projects/{project_id}/lookml_tests/run"
    )
    field_projects__project_id__tag: FieldProjectsProjectIdTag = Field(..., alias="/projects/{project_id}/tag")
    field_render_tasks_looks__look_id___result_format_: FieldRenderTasksLooksLookIdResultFormat = Field(
        ..., alias="/render_tasks/looks/{look_id}/{result_format}"
    )
    field_render_tasks_queries__query_id___result_format_: FieldRenderTasksQueriesQueryIdResultFormat = Field(
        ..., alias="/render_tasks/queries/{query_id}/{result_format}"
    )
    field_render_tasks_dashboards__dashboard_id___result_format_: FieldRenderTasksDashboardsDashboardIdResultFormat = (
        Field(..., alias="/render_tasks/dashboards/{dashboard_id}/{result_format}")
    )
    field_render_tasks__render_task_id_: FieldRenderTasksRenderTaskId = Field(
        ..., alias="/render_tasks/{render_task_id}"
    )
    field_render_tasks__render_task_id__results: FieldRenderTasksRenderTaskIdResults = Field(
        ..., alias="/render_tasks/{render_task_id}/results"
    )
    field_render_tasks_dashboard_elements__dashboard_element_id___result_format_: FieldRenderTasksDashboardElementsDashboardElementIdResultFormat = Field(
        ..., alias="/render_tasks/dashboard_elements/{dashboard_element_id}/{result_format}"
    )
    field_projects__root_project_id__credential__credential_id_: FieldProjectsRootProjectIdCredentialCredentialId = (
        Field(..., alias="/projects/{root_project_id}/credential/{credential_id}")
    )
    field_projects__root_project_id__credentials: FieldProjectsRootProjectIdCredentials = Field(
        ..., alias="/projects/{root_project_id}/credentials"
    )
    field_roles: FieldRoles = Field(..., alias="/roles")
    field_roles_search: FieldRolesSearch = Field(..., alias="/roles/search")
    field_roles_search_with_user_count: FieldRolesSearchWithUserCount = Field(
        ..., alias="/roles/search/with_user_count"
    )
    field_roles__role_id_: FieldRolesRoleId = Field(..., alias="/roles/{role_id}")
    field_roles__role_id__groups: FieldRolesRoleIdGroups = Field(..., alias="/roles/{role_id}/groups")
    field_roles__role_id__users: FieldRolesRoleIdUsers = Field(..., alias="/roles/{role_id}/users")
    field_running_queries: FieldRunningQueries = Field(..., alias="/running_queries")
    field_running_queries__query_task_id_: FieldRunningQueriesQueryTaskId = Field(
        ..., alias="/running_queries/{query_task_id}"
    )
    field_saml_config: FieldSamlConfig = Field(..., alias="/saml_config")
    field_saml_test_configs__test_slug_: FieldSamlTestConfigsTestSlug = Field(
        ..., alias="/saml_test_configs/{test_slug}"
    )
    field_saml_test_configs: FieldSamlTestConfigs = Field(..., alias="/saml_test_configs")
    field_parse_saml_idp_metadata: FieldParseSamlIdpMetadata = Field(..., alias="/parse_saml_idp_metadata")
    field_fetch_and_parse_saml_idp_metadata: FieldFetchAndParseSamlIdpMetadata = Field(
        ..., alias="/fetch_and_parse_saml_idp_metadata"
    )
    field_scheduled_plans_space__space_id_: FieldScheduledPlansSpaceSpaceId = Field(
        ..., alias="/scheduled_plans/space/{space_id}"
    )
    field_scheduled_plans__scheduled_plan_id_: FieldScheduledPlansScheduledPlanId = Field(
        ..., alias="/scheduled_plans/{scheduled_plan_id}"
    )
    field_scheduled_plans: FieldScheduledPlans = Field(..., alias="/scheduled_plans")
    field_scheduled_plans_run_once: FieldScheduledPlansRunOnce = Field(..., alias="/scheduled_plans/run_once")
    field_scheduled_plans_look__look_id_: FieldScheduledPlansLookLookId = Field(
        ..., alias="/scheduled_plans/look/{look_id}"
    )
    field_scheduled_plans_dashboard__dashboard_id_: FieldScheduledPlansDashboardDashboardId = Field(
        ..., alias="/scheduled_plans/dashboard/{dashboard_id}"
    )
    field_scheduled_plans_lookml_dashboard__lookml_dashboard_id_: FieldScheduledPlansLookmlDashboardLookmlDashboardId = Field(
        ..., alias="/scheduled_plans/lookml_dashboard/{lookml_dashboard_id}"
    )
    field_scheduled_plans__scheduled_plan_id__run_once: FieldScheduledPlansScheduledPlanIdRunOnce = Field(
        ..., alias="/scheduled_plans/{scheduled_plan_id}/run_once"
    )
    field_session_config: FieldSessionConfig = Field(..., alias="/session_config")
    field_session: FieldSession = Field(..., alias="/session")
    field_setting: FieldSetting = Field(..., alias="/setting")
    field_smtp_settings: FieldSmtpSettings = Field(..., alias="/smtp_settings")
    field_smtp_status: FieldSmtpStatus = Field(..., alias="/smtp_status")
    field_folders_search: FieldFoldersSearch = Field(..., alias="/folders/search")
    field_folders__folder_id_: FieldFoldersFolderId = Field(..., alias="/folders/{folder_id}")
    field_folders: FieldFolders = Field(..., alias="/folders")
    field_folders__folder_id__children: FieldFoldersFolderIdChildren = Field(..., alias="/folders/{folder_id}/children")
    field_folders__folder_id__children_search: FieldFoldersFolderIdChildrenSearch = Field(
        ..., alias="/folders/{folder_id}/children/search"
    )
    field_folders__folder_id__parent: FieldFoldersFolderIdParent = Field(..., alias="/folders/{folder_id}/parent")
    field_folders__folder_id__ancestors: FieldFoldersFolderIdAncestors = Field(
        ..., alias="/folders/{folder_id}/ancestors"
    )
    field_folders__folder_id__looks: FieldFoldersFolderIdLooks = Field(..., alias="/folders/{folder_id}/looks")
    field_folders__folder_id__dashboards: FieldFoldersFolderIdDashboards = Field(
        ..., alias="/folders/{folder_id}/dashboards"
    )
    field_sql_queries__slug_: FieldSqlQueriesSlug = Field(..., alias="/sql_queries/{slug}")
    field_sql_queries: FieldSqlQueries = Field(..., alias="/sql_queries")
    field_sql_queries__slug__run__result_format_: FieldSqlQueriesSlugRunResultFormat = Field(
        ..., alias="/sql_queries/{slug}/run/{result_format}"
    )
    field_support_access_allowlist: FieldSupportAccessAllowlist = Field(..., alias="/support_access/allowlist")
    field_support_access_allowlist__entry_id_: FieldSupportAccessAllowlistEntryId = Field(
        ..., alias="/support_access/allowlist/{entry_id}"
    )
    field_support_access_enable: FieldSupportAccessEnable = Field(..., alias="/support_access/enable")
    field_support_access_disable: FieldSupportAccessDisable = Field(..., alias="/support_access/disable")
    field_support_access_status: FieldSupportAccessStatus = Field(..., alias="/support_access/status")
    field_themes: FieldThemes = Field(..., alias="/themes")
    field_themes_search: FieldThemesSearch = Field(..., alias="/themes/search")
    field_themes_default: FieldThemesDefault = Field(..., alias="/themes/default")
    field_themes_active: FieldThemesActive = Field(..., alias="/themes/active")
    field_themes_theme_or_default: FieldThemesThemeOrDefault = Field(..., alias="/themes/theme_or_default")
    field_themes_validate: FieldThemesValidate = Field(..., alias="/themes/validate")
    field_themes__theme_id_: FieldThemesThemeId = Field(..., alias="/themes/{theme_id}")
    field_timezones: FieldTimezones = Field(..., alias="/timezones")
    field_ssh_servers: FieldSshServers = Field(..., alias="/ssh_servers")
    field_ssh_server__ssh_server_id_: FieldSshServerSshServerId = Field(..., alias="/ssh_server/{ssh_server_id}")
    field_ssh_server__ssh_server_id__test: FieldSshServerSshServerIdTest = Field(
        ..., alias="/ssh_server/{ssh_server_id}/test"
    )
    field_ssh_tunnels: FieldSshTunnels = Field(..., alias="/ssh_tunnels")
    field_ssh_tunnel__ssh_tunnel_id_: FieldSshTunnelSshTunnelId = Field(..., alias="/ssh_tunnel/{ssh_tunnel_id}")
    field_ssh_tunnel__ssh_tunnel_id__test: FieldSshTunnelSshTunnelIdTest = Field(
        ..., alias="/ssh_tunnel/{ssh_tunnel_id}/test"
    )
    field_ssh_public_key: FieldSshPublicKey = Field(..., alias="/ssh_public_key")
    field_user_attributes: FieldUserAttributes = Field(..., alias="/user_attributes")
    field_user_attributes__user_attribute_id_: FieldUserAttributesUserAttributeId = Field(
        ..., alias="/user_attributes/{user_attribute_id}"
    )
    field_user_attributes__user_attribute_id__group_values: FieldUserAttributesUserAttributeIdGroupValues = Field(
        ..., alias="/user_attributes/{user_attribute_id}/group_values"
    )
    field_user_login_lockouts: FieldUserLoginLockouts = Field(..., alias="/user_login_lockouts")
    field_user_login_lockouts_search: FieldUserLoginLockoutsSearch = Field(..., alias="/user_login_lockouts/search")
    field_user_login_lockout__key_: FieldUserLoginLockoutKey = Field(..., alias="/user_login_lockout/{key}")
    field_user: FieldUser = Field(..., alias="/user")
    field_users: FieldUsers = Field(..., alias="/users")
    field_users_search: FieldUsersSearch = Field(..., alias="/users/search")
    field_users_search_names__pattern_: FieldUsersSearchNamesPattern = Field(..., alias="/users/search/names/{pattern}")
    field_users__user_id_: FieldUsersUserId = Field(..., alias="/users/{user_id}")
    field_users_credential__credential_type___credential_id_: FieldUsersCredentialCredentialTypeCredentialId = Field(
        ..., alias="/users/credential/{credential_type}/{credential_id}"
    )
    field_users__user_id__credentials_email: FieldUsersUserIdCredentialsEmail = Field(
        ..., alias="/users/{user_id}/credentials_email"
    )
    field_users__user_id__credentials_totp: FieldUsersUserIdCredentialsTotp = Field(
        ..., alias="/users/{user_id}/credentials_totp"
    )
    field_users__user_id__credentials_ldap: FieldUsersUserIdCredentialsLdap = Field(
        ..., alias="/users/{user_id}/credentials_ldap"
    )
    field_users__user_id__credentials_google: FieldUsersUserIdCredentialsGoogle = Field(
        ..., alias="/users/{user_id}/credentials_google"
    )
    field_users__user_id__credentials_saml: FieldUsersUserIdCredentialsSaml = Field(
        ..., alias="/users/{user_id}/credentials_saml"
    )
    field_users__user_id__credentials_oidc: FieldUsersUserIdCredentialsOidc = Field(
        ..., alias="/users/{user_id}/credentials_oidc"
    )
    field_users__user_id__credentials_api3__credentials_api3_id_: FieldUsersUserIdCredentialsApi3CredentialsApi3Id = (
        Field(..., alias="/users/{user_id}/credentials_api3/{credentials_api3_id}")
    )
    field_users__user_id__credentials_api3: FieldUsersUserIdCredentialsApi3 = Field(
        ..., alias="/users/{user_id}/credentials_api3"
    )
    field_users__user_id__credentials_embed__credentials_embed_id_: FieldUsersUserIdCredentialsEmbedCredentialsEmbedId = Field(
        ..., alias="/users/{user_id}/credentials_embed/{credentials_embed_id}"
    )
    field_users__user_id__credentials_embed: FieldUsersUserIdCredentialsEmbed = Field(
        ..., alias="/users/{user_id}/credentials_embed"
    )
    field_users__user_id__credentials_looker_openid: FieldUsersUserIdCredentialsLookerOpenid = Field(
        ..., alias="/users/{user_id}/credentials_looker_openid"
    )
    field_users__user_id__sessions__session_id_: FieldUsersUserIdSessionsSessionId = Field(
        ..., alias="/users/{user_id}/sessions/{session_id}"
    )
    field_users__user_id__sessions: FieldUsersUserIdSessions = Field(..., alias="/users/{user_id}/sessions")
    field_users__user_id__credentials_email_password_reset: FieldUsersUserIdCredentialsEmailPasswordReset = Field(
        ..., alias="/users/{user_id}/credentials_email/password_reset"
    )
    field_users__user_id__roles: FieldUsersUserIdRoles = Field(..., alias="/users/{user_id}/roles")
    field_users__user_id__attribute_values: FieldUsersUserIdAttributeValues = Field(
        ..., alias="/users/{user_id}/attribute_values"
    )
    field_users__user_id__attribute_values__user_attribute_id_: FieldUsersUserIdAttributeValuesUserAttributeId = Field(
        ..., alias="/users/{user_id}/attribute_values/{user_attribute_id}"
    )
    field_users__user_id__credentials_email_send_password_reset: FieldUsersUserIdCredentialsEmailSendPasswordReset = (
        Field(..., alias="/users/{user_id}/credentials_email/send_password_reset")
    )
    field_users__user_id__update_emails: FieldUsersUserIdUpdateEmails = Field(
        ..., alias="/users/{user_id}/update_emails"
    )
    field_users_embed_user: FieldUsersEmbedUser = Field(..., alias="/users/embed_user")
    field_vector_thumbnail__type___resource_id_: FieldVectorThumbnailTypeResourceId = Field(
        ..., alias="/vector_thumbnail/{type}/{resource_id}"
    )
    field_versions: FieldVersions = Field(..., alias="/versions")
    field_api_spec__api_version___specification_: FieldApiSpecApiVersionSpecification = Field(
        ..., alias="/api_spec/{api_version}/{specification}"
    )
    field_whitelabel_configuration: FieldWhitelabelConfiguration = Field(..., alias="/whitelabel_configuration")
    field_workspaces: FieldWorkspaces = Field(..., alias="/workspaces")
    field_workspaces__workspace_id_: FieldWorkspacesWorkspaceId = Field(..., alias="/workspaces/{workspace_id}")


class Message(BaseModel):
    type: str
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class DocumentationUrl(BaseModel):
    type: str
    format: str
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties(BaseModel):
    message: Message
    documentation_url: DocumentationUrl


class Error(BaseModel):
    properties: Properties
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Can(BaseModel):
    type: str
    additionalProperties: AdditionalProperties1
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class ContentFavoriteId(Message):
    pass


class ContentMetadataId(Message):
    pass


class Description(Message):
    pass


class Hidden(Message):
    pass


class Id(Message):
    pass


class Model1(BaseModel):
    field_ref: str = Field(..., alias="$ref")
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class QueryTimezone(Message):
    pass


class Readonly(Message):
    pass


class RefreshInterval(Message):
    pass


class RefreshIntervalToI(DocumentationUrl):
    pass


class Folder(Model1):
    pass


class Title(Message):
    pass


class UserId(Message):
    pass


class Slug(Message):
    pass


class PreferredViewer(Message):
    pass


class Properties1(BaseModel):
    can: Can
    content_favorite_id: ContentFavoriteId
    content_metadata_id: ContentMetadataId
    description: Description
    hidden: Hidden
    id: Id
    model: Model1
    query_timezone: QueryTimezone
    readonly: Readonly
    refresh_interval: RefreshInterval
    refresh_interval_to_i: RefreshIntervalToI
    folder: Folder
    title: Title
    user_id: UserId
    slug: Slug
    preferred_viewer: PreferredViewer


class DashboardBase(BaseModel):
    properties: Properties1
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can1(Can):
    pass


class CreatedAt(DocumentationUrl):
    pass


class DeletedAt(BaseModel):
    type: str
    format: str
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class DetailUrl(Message):
    pass


class HomepageId(BaseModel):
    type: str
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class HomepageItems(BaseModel):
    type: str
    items: Items118
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class IsHeader(Message):
    pass


class Items121(Items):
    pass


class ItemOrder(BaseModel):
    type: str
    items: Items121
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Title1(HomepageId):
    pass


class UpdatedAt(DocumentationUrl):
    pass


class Description1(HomepageId):
    pass


class VisibleItemOrder(BaseModel):
    type: str
    items: Items121
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties2(BaseModel):
    can: Can1
    created_at: CreatedAt
    deleted_at: DeletedAt
    detail_url: DetailUrl
    homepage_id: HomepageId
    homepage_items: HomepageItems
    id: Id
    is_header: IsHeader
    item_order: ItemOrder
    title: Title1
    updated_at: UpdatedAt
    description: Description1
    visible_item_order: VisibleItemOrder


class HomepageSection(BaseModel):
    properties: Properties2
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can2(Can):
    pass


class ContentCreatedBy(Message):
    pass


class ContentUpdatedAt(Message):
    pass


class CustomDescription(HomepageId):
    pass


class CustomImageDataBase64(BaseModel):
    type: str
    x_looker_write_only: bool = Field(..., alias="x-looker-write-only")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class CustomImageUrl(Message):
    pass


class CustomTitle(HomepageId):
    pass


class CustomUrl(HomepageId):
    pass


class DashboardId(HomepageId):
    pass


class Description2(Message):
    pass


class FavoriteCount(DocumentationUrl):
    pass


class HomepageSectionId(HomepageId):
    pass


class ImageUrl(Message):
    pass


class Location(Message):
    pass


class LookId(HomepageId):
    pass


class LookmlDashboardId(HomepageId):
    pass


class Order(DeletedAt):
    pass


class SectionFetchTime(DocumentationUrl):
    pass


class Title2(Message):
    pass


class Url(Message):
    pass


class UseCustomDescription(HomepageId):
    pass


class UseCustomImage(HomepageId):
    pass


class UseCustomTitle(HomepageId):
    pass


class UseCustomUrl(HomepageId):
    pass


class ViewCount(DocumentationUrl):
    pass


class Properties3(BaseModel):
    can: Can2
    content_created_by: ContentCreatedBy
    content_favorite_id: ContentFavoriteId
    content_metadata_id: ContentMetadataId
    content_updated_at: ContentUpdatedAt
    custom_description: CustomDescription
    custom_image_data_base64: CustomImageDataBase64
    custom_image_url: CustomImageUrl
    custom_title: CustomTitle
    custom_url: CustomUrl
    dashboard_id: DashboardId
    description: Description2
    favorite_count: FavoriteCount
    homepage_section_id: HomepageSectionId
    id: Id
    image_url: ImageUrl
    location: Location
    look_id: LookId
    lookml_dashboard_id: LookmlDashboardId
    order: Order
    section_fetch_time: SectionFetchTime
    title: Title2
    url: Url
    use_custom_description: UseCustomDescription
    use_custom_image: UseCustomImage
    use_custom_title: UseCustomTitle
    use_custom_url: UseCustomUrl
    view_count: ViewCount


class HomepageItem(BaseModel):
    properties: Properties3
    x_looker_status: str = Field(..., alias="x-looker-status")


class Items123(Schema):
    pass


class Errors(BaseModel):
    type: str
    items: Items123
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties4(BaseModel):
    message: Message
    errors: Errors
    documentation_url: DocumentationUrl


class ValidationError(BaseModel):
    properties: Properties4
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class FieldModel(Message):
    pass


class Code(Message):
    pass


class Properties5(BaseModel):
    field: FieldModel
    code: Code
    message: Message
    documentation_url: DocumentationUrl


class ValidationErrorDetail(BaseModel):
    properties: Properties5
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class AccessToken1(Message):
    pass


class TokenType(Message):
    pass


class ExpiresIn(DocumentationUrl):
    pass


class RefreshToken(Message):
    pass


class Properties6(BaseModel):
    access_token: AccessToken1
    token_type: TokenType
    expires_in: ExpiresIn
    refresh_token: RefreshToken


class AccessToken(BaseModel):
    properties: Properties6
    x_looker_status: str = Field(..., alias="x-looker-status")


class FieldName(HomepageId):
    pass


class FieldValue(DeletedAt):
    pass


class FilterValue(HomepageId):
    pass


class Properties7(BaseModel):
    field_name: FieldName
    field_value: FieldValue
    filter_value: FilterValue


class AlertFieldFilter(BaseModel):
    properties: Properties7
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class FilterTitle(HomepageId):
    pass


class FilterDescription(Message):
    pass


class Properties8(BaseModel):
    filter_title: FilterTitle
    field_name: FieldName
    filter_value: FilterValue
    filter_description: FilterDescription


class AlertAppliedDashboardFilter(BaseModel):
    properties: Properties8
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Title3(HomepageId):
    pass


class Name(HomepageId):
    pass


class Filter(BaseModel):
    type: str
    items: Items123
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties9(BaseModel):
    title: Title3
    name: Name
    filter: Filter


class AlertField(BaseModel):
    properties: Properties9
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class PreviousTimeSeriesId(CustomImageDataBase64):
    pass


class LatestTimeSeriesId(CustomImageDataBase64):
    pass


class Properties10(BaseModel):
    previous_time_series_id: PreviousTimeSeriesId
    latest_time_series_id: LatestTimeSeriesId


class AlertConditionState(BaseModel):
    properties: Properties10
    x_looker_status: str = Field(..., alias="x-looker-status")


class AppliedDashboardFilters(Filter):
    pass


class ComparisonType(BaseModel):
    type: str
    x_looker_values: List[str] = Field(..., alias="x-looker-values")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Cron(HomepageId):
    pass


class CustomUrlBase(HomepageId):
    pass


class CustomUrlParams(HomepageId):
    pass


class CustomUrlLabel(HomepageId):
    pass


class ShowCustomUrl(HomepageId):
    pass


class DashboardElementId(HomepageId):
    pass


class Description3(HomepageId):
    pass


class Destinations(Filter):
    pass


class Field(BaseModel):
    field_ref: str = Field(..., alias="$ref")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Followed(Message):
    pass


class Followable(Message):
    pass


class IsDisabled(HomepageId):
    pass


class DisabledReason(HomepageId):
    pass


class IsPublic(HomepageId):
    pass


class InvestigativeContentType(ComparisonType):
    pass


class InvestigativeContentId(HomepageId):
    pass


class InvestigativeContentTitle(Message):
    pass


class LookmlLinkId(HomepageId):
    pass


class OwnerId(HomepageId):
    pass


class OwnerDisplayName(Message):
    pass


class Threshold(DeletedAt):
    pass


class TimeSeriesConditionState(BaseModel):
    field_ref: str = Field(..., alias="$ref")
    x_looker_write_only: bool = Field(..., alias="x-looker-write-only")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties11(BaseModel):
    applied_dashboard_filters: AppliedDashboardFilters
    comparison_type: ComparisonType
    cron: Cron
    custom_url_base: CustomUrlBase
    custom_url_params: CustomUrlParams
    custom_url_label: CustomUrlLabel
    show_custom_url: ShowCustomUrl
    custom_title: CustomTitle
    dashboard_element_id: DashboardElementId
    description: Description3
    destinations: Destinations
    field: Field
    followed: Followed
    followable: Followable
    id: Id
    is_disabled: IsDisabled
    disabled_reason: DisabledReason
    is_public: IsPublic
    investigative_content_type: InvestigativeContentType
    investigative_content_id: InvestigativeContentId
    investigative_content_title: InvestigativeContentTitle
    lookml_dashboard_id: LookmlDashboardId
    lookml_link_id: LookmlLinkId
    owner_id: OwnerId
    owner_display_name: OwnerDisplayName
    threshold: Threshold
    time_series_condition_state: TimeSeriesConditionState


class Alert(BaseModel):
    properties: Properties11
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Title4(Message):
    pass


class AlertId(Message):
    pass


class InvestigativeContentId1(Message):
    pass


class DashboardName(Message):
    pass


class DashboardId1(Message):
    pass


class QuerySlug(Message):
    pass


class Properties12(BaseModel):
    title: Title4
    alert_id: AlertId
    investigative_content_id: InvestigativeContentId1
    dashboard_name: DashboardName
    dashboard_id: DashboardId1
    query_slug: QuerySlug


class MobilePayload(BaseModel):
    properties: Properties12
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class NotificationId(Message):
    pass


class AlertConditionId(Message):
    pass


class IsRead(HomepageId):
    pass


class FieldValue1(DocumentationUrl):
    pass


class ThresholdValue(DocumentationUrl):
    pass


class RanAt(Message):
    pass


class Alert1(Model1):
    pass


class Properties13(BaseModel):
    notification_id: NotificationId
    alert_condition_id: AlertConditionId
    user_id: UserId
    is_read: IsRead
    field_value: FieldValue1
    threshold_value: ThresholdValue
    ran_at: RanAt
    alert: Alert1


class AlertNotifications(BaseModel):
    properties: Properties13
    x_looker_status: str = Field(..., alias="x-looker-status")


class DestinationType(ComparisonType):
    pass


class EmailAddress(HomepageId):
    pass


class ActionHubIntegrationId(HomepageId):
    pass


class ActionHubFormParamsJson(HomepageId):
    pass


class Properties14(BaseModel):
    destination_type: DestinationType
    email_address: EmailAddress
    action_hub_integration_id: ActionHubIntegrationId
    action_hub_form_params_json: ActionHubFormParamsJson


class AlertDestination(BaseModel):
    properties: Properties14
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Properties15(BaseModel):
    owner_id: OwnerId
    is_disabled: IsDisabled
    disabled_reason: DisabledReason
    is_public: IsPublic
    threshold: Threshold


class AlertPatch(BaseModel):
    properties: Properties15
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can3(Can):
    pass


class WorkspaceId(HomepageId):
    pass


class SudoUserId(Message):
    pass


class Properties16(BaseModel):
    can: Can3
    workspace_id: WorkspaceId
    sudo_user_id: SudoUserId


class ApiSession(BaseModel):
    properties: Properties16
    x_looker_status: str = Field(..., alias="x-looker-status")


class MaxSize(DocumentationUrl):
    pass


class Usage(DocumentationUrl):
    pass


class Properties17(BaseModel):
    max_size: MaxSize
    usage: Usage


class ArtifactUsage(BaseModel):
    properties: Properties17
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Namespace(Message):
    pass


class Count(DocumentationUrl):
    pass


class Properties18(BaseModel):
    namespace: Namespace
    count: Count


class ArtifactNamespace(BaseModel):
    properties: Properties18
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Key(HomepageId):
    pass


class Value(HomepageId):
    pass


class ContentType(HomepageId):
    pass


class Version(DocumentationUrl):
    pass


class Properties19(BaseModel):
    key: Key
    value: Value
    content_type: ContentType
    version: Version


class UpdateArtifact(BaseModel):
    properties: Properties19
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class ValueSize(DocumentationUrl):
    pass


class CreatedByUserid(Message):
    pass


class UpdatedByUserid(Message):
    pass


class Properties20(BaseModel):
    key: Key
    value: Value
    content_type: ContentType
    version: Version
    namespace: Namespace
    created_at: CreatedAt
    updated_at: UpdatedAt
    value_size: ValueSize
    created_by_userid: CreatedByUserid
    updated_by_userid: UpdatedByUserid


class Artifact(BaseModel):
    properties: Properties20
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Can4(Can):
    pass


class Type(HomepageId):
    pass


class CustomS3Bucket(HomepageId):
    pass


class CustomS3BucketRegion(HomepageId):
    pass


class CustomS3Key(CustomImageDataBase64):
    pass


class CustomS3Secret(CustomImageDataBase64):
    pass


class Url1(DocumentationUrl):
    pass


class Properties21(BaseModel):
    can: Can4
    type: Type
    custom_s3_bucket: CustomS3Bucket
    custom_s3_bucket_region: CustomS3BucketRegion
    custom_s3_key: CustomS3Key
    custom_s3_secret: CustomS3Secret
    url: Url1


class BackupConfiguration(BaseModel):
    properties: Properties21
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can5(Can):
    pass


class DashboardId2(HomepageId):
    pass


class Description4(Message):
    pass


class BoardSectionId(HomepageId):
    pass


class Url2(Message):
    pass


class Properties22(BaseModel):
    can: Can5
    content_created_by: ContentCreatedBy
    content_favorite_id: ContentFavoriteId
    content_metadata_id: ContentMetadataId
    content_updated_at: ContentUpdatedAt
    custom_description: CustomDescription
    custom_title: CustomTitle
    custom_url: CustomUrl
    dashboard_id: DashboardId2
    description: Description4
    favorite_count: FavoriteCount
    board_section_id: BoardSectionId
    id: Id
    image_url: ImageUrl
    location: Location
    look_id: LookId
    lookml_dashboard_id: LookmlDashboardId
    order: Order
    title: Title4
    url: Url2
    use_custom_description: UseCustomDescription
    use_custom_title: UseCustomTitle
    use_custom_url: UseCustomUrl
    view_count: ViewCount
    custom_image_data_base64: CustomImageDataBase64
    custom_image_url: CustomImageUrl
    use_custom_image: UseCustomImage


class BoardItem(BaseModel):
    properties: Properties22
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can6(Can):
    pass


class Description5(HomepageId):
    pass


class BoardSections(Errors):
    pass


class Items128(Items):
    pass


class SectionOrder(BaseModel):
    type: str
    items: Items128
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Title6(HomepageId):
    pass


class PrimaryHomepage(Message):
    pass


class Properties23(BaseModel):
    can: Can6
    content_metadata_id: ContentMetadataId
    created_at: CreatedAt
    deleted_at: DeletedAt
    description: Description5
    board_sections: BoardSections
    id: Id
    section_order: SectionOrder
    title: Title6
    updated_at: UpdatedAt
    user_id: UserId
    primary_homepage: PrimaryHomepage


class Board(BaseModel):
    properties: Properties23
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can7(Can):
    pass


class BoardId(HomepageId):
    pass


class Items129(Schema):
    pass


class BoardItems(BaseModel):
    type: str
    items: Items129
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items130(Items):
    pass


class ItemOrder1(BaseModel):
    type: str
    items: Items130
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class VisibleItemOrder1(BaseModel):
    type: str
    items: Items130
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties24(BaseModel):
    can: Can7
    created_at: CreatedAt
    deleted_at: DeletedAt
    description: Description5
    board_id: BoardId
    board_items: BoardItems
    id: Id
    item_order: ItemOrder1
    visible_item_order: VisibleItemOrder1
    title: Title6
    updated_at: UpdatedAt


class BoardSection(BaseModel):
    properties: Properties24
    x_looker_status: str = Field(..., alias="x-looker-status")


class Color(HomepageId):
    pass


class Offset(DeletedAt):
    pass


class Properties25(BaseModel):
    color: Color
    offset: Offset


class ColorStop(BaseModel):
    properties: Properties25
    x_looker_status: str = Field(..., alias="x-looker-status")


class Label(HomepageId):
    pass


class Items132(Schema):
    pass


class Stops(BaseModel):
    type: str
    items: Items132
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties26(BaseModel):
    id: Id
    label: Label
    type: Type
    stops: Stops


class ContinuousPalette(BaseModel):
    properties: Properties26
    x_looker_status: str = Field(..., alias="x-looker-status")


class Items133(Items):
    pass


class Colors(BaseModel):
    type: str
    items: Items133
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties27(BaseModel):
    id: Id
    label: Label
    type: Type
    colors: Colors


class DiscretePalette(BaseModel):
    properties: Properties27
    x_looker_status: str = Field(..., alias="x-looker-status")


class Items134(Schema):
    pass


class CategoricalPalettes(BaseModel):
    type: str
    items: Items134
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class SequentialPalettes(CategoricalPalettes):
    pass


class DivergingPalettes(CategoricalPalettes):
    pass


class Properties28(BaseModel):
    id: Id
    label: Label
    categoricalPalettes: CategoricalPalettes
    sequentialPalettes: SequentialPalettes
    divergingPalettes: DivergingPalettes


class ColorCollection(BaseModel):
    properties: Properties28
    x_looker_status: str = Field(..., alias="x-looker-status")


class UserId3(HomepageId):
    pass


class ContentMetadataId4(HomepageId):
    pass


class LookId2(Message):
    pass


class DashboardId3(Message):
    pass


class Look(Model1):
    pass


class Dashboard(Model1):
    pass


class BoardId1(Message):
    pass


class Properties29(BaseModel):
    id: Id
    user_id: UserId3
    content_metadata_id: ContentMetadataId4
    look_id: LookId2
    dashboard_id: DashboardId3
    look: Look
    dashboard: Dashboard
    board_id: BoardId1


class ContentFavorite(BaseModel):
    properties: Properties29
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can8(Can):
    pass


class ContentMetadataId5(Message):
    pass


class PermissionType(BaseModel):
    type: str
    readOnly: bool
    x_looker_values: List[str] = Field(..., alias="x-looker-values")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class GroupId(Message):
    pass


class UserId4(Message):
    pass


class Properties30(BaseModel):
    can: Can8
    id: Id
    content_metadata_id: ContentMetadataId5
    permission_type: PermissionType
    group_id: GroupId
    user_id: UserId4


class ContentMetaGroupUser(BaseModel):
    properties: Properties30
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can9(Can):
    pass


class Name1(Message):
    pass


class ParentId(Message):
    pass


class FolderId(Message):
    pass


class ContentType2(Message):
    pass


class Inherits(HomepageId):
    pass


class InheritingId(Message):
    pass


class Properties31(BaseModel):
    can: Can9
    id: Id
    name: Name1
    parent_id: ParentId
    dashboard_id: DashboardId3
    look_id: LookId2
    folder_id: FolderId
    content_type: ContentType2
    inherits: Inherits
    inheriting_id: InheritingId
    slug: Slug


class ContentMeta(BaseModel):
    properties: Properties31
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can10(Can):
    pass


class ContentId(Message):
    pass


class Type3(Message):
    pass


class Title8(Message):
    pass


class Description7(Message):
    pass


class FolderName(Message):
    pass


class Properties32(BaseModel):
    can: Can10
    content_id: ContentId
    type: Type3
    title: Title8
    description: Description7
    folder_id: FolderId
    folder_name: FolderName
    view_count: ViewCount
    preferred_viewer: PreferredViewer


class ContentSearch(BaseModel):
    properties: Properties32
    x_looker_status: str = Field(..., alias="x-looker-status")


class ContentWithErrors(BaseModel):
    type: str
    items: Items134
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class ComputationTime(DocumentationUrl):
    pass


class TotalLooksValidated(DocumentationUrl):
    pass


class TotalDashboardElementsValidated(DocumentationUrl):
    pass


class TotalDashboardFiltersValidated(DocumentationUrl):
    pass


class TotalScheduledPlansValidated(DocumentationUrl):
    pass


class TotalAlertsValidated(DocumentationUrl):
    pass


class TotalExploresValidated(DocumentationUrl):
    pass


class Properties33(BaseModel):
    content_with_errors: ContentWithErrors
    computation_time: ComputationTime
    total_looks_validated: TotalLooksValidated
    total_dashboard_elements_validated: TotalDashboardElementsValidated
    total_dashboard_filters_validated: TotalDashboardFiltersValidated
    total_scheduled_plans_validated: TotalScheduledPlansValidated
    total_alerts_validated: TotalAlertsValidated
    total_explores_validated: TotalExploresValidated


class ContentValidation(BaseModel):
    properties: Properties33
    x_looker_status: str = Field(..., alias="x-looker-status")


class Look1(BaseModel):
    field_ref: str = Field(..., alias="$ref")
    readOnly: bool
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Dashboard1(Look1):
    pass


class DashboardElement(Look1):
    pass


class DashboardFilter(Look1):
    pass


class ScheduledPlan(Look1):
    pass


class Alert2(Look1):
    pass


class LookmlDashboard(Look1):
    pass


class LookmlDashboardElement(Look1):
    pass


class Errors1(ContentWithErrors):
    pass


class Properties34(BaseModel):
    look: Look1
    dashboard: Dashboard1
    dashboard_element: DashboardElement
    dashboard_filter: DashboardFilter
    scheduled_plan: ScheduledPlan
    alert: Alert2
    lookml_dashboard: LookmlDashboard
    lookml_dashboard_element: LookmlDashboardElement
    errors: Errors1
    id: Id


class ContentValidatorError(BaseModel):
    properties: Properties34
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name2(HomepageId):
    pass


class Properties35(BaseModel):
    name: Name2
    id: Id


class ContentValidationFolder(BaseModel):
    properties: Properties35
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Title9(HomepageId):
    pass


class ShortUrl(Message):
    pass


class Properties36(BaseModel):
    id: Id
    title: Title9
    short_url: ShortUrl
    folder: Folder


class ContentValidationLook(BaseModel):
    properties: Properties36
    x_looker_status: str = Field(..., alias="x-looker-status")


class Description8(HomepageId):
    pass


class Properties37(BaseModel):
    description: Description8
    id: Id
    folder: Folder
    title: Title9
    url: Url2


class ContentValidationDashboard(BaseModel):
    properties: Properties37
    x_looker_status: str = Field(..., alias="x-looker-status")


class BodyText(HomepageId):
    pass


class DashboardId5(HomepageId):
    pass


class LookId4(HomepageId):
    pass


class NoteDisplay(HomepageId):
    pass


class NoteState(HomepageId):
    pass


class NoteText(HomepageId):
    pass


class NoteTextAsHtml(Message):
    pass


class QueryId(HomepageId):
    pass


class SubtitleText(HomepageId):
    pass


class TitleHidden(HomepageId):
    pass


class TitleText(HomepageId):
    pass


class Type4(HomepageId):
    pass


class RichContentJson(HomepageId):
    pass


class ExtensionId(HomepageId):
    pass


class Properties38(BaseModel):
    body_text: BodyText
    dashboard_id: DashboardId5
    id: Id
    look_id: LookId4
    note_display: NoteDisplay
    note_state: NoteState
    note_text: NoteText
    note_text_as_html: NoteTextAsHtml
    query_id: QueryId
    subtitle_text: SubtitleText
    title: Title9
    title_hidden: TitleHidden
    title_text: TitleText
    type: Type4
    rich_content_json: RichContentJson
    extension_id: ExtensionId


class ContentValidationDashboardElement(BaseModel):
    properties: Properties38
    x_looker_status: str = Field(..., alias="x-looker-status")


class FieldName2(Message):
    pass


class ModelName(Message):
    pass


class ExploreName(Message):
    pass


class Removable(Message):
    pass


class Properties39(BaseModel):
    message: Message
    field_name: FieldName2
    model_name: ModelName
    explore_name: ExploreName
    removable: Removable


class ContentValidationError(BaseModel):
    properties: Properties39
    x_looker_status: str = Field(..., alias="x-looker-status")


class DashboardId6(Message):
    pass


class DefaultValue(HomepageId):
    pass


class Model2(HomepageId):
    pass


class Explore(HomepageId):
    pass


class Dimension(HomepageId):
    pass


class Properties40(BaseModel):
    id: Id
    dashboard_id: DashboardId6
    name: Name2
    title: Title9
    type: Type4
    default_value: DefaultValue
    model: Model2
    explore: Explore
    dimension: Dimension


class ContentValidationDashboardFilter(BaseModel):
    properties: Properties40
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties41(BaseModel):
    name: Name2
    look_id: LookId4
    id: Id


class ContentValidationScheduledPlan(BaseModel):
    properties: Properties41
    x_looker_status: str = Field(..., alias="x-looker-status")


class Id20(HomepageId):
    pass


class Properties42(BaseModel):
    id: Id20
    lookml_dashboard_id: LookmlDashboardId
    lookml_link_id: LookmlLinkId
    custom_url_base: CustomUrlBase
    custom_url_params: CustomUrlParams
    custom_url_label: CustomUrlLabel
    show_custom_url: ShowCustomUrl
    custom_title: CustomTitle


class ContentValidationAlert(BaseModel):
    properties: Properties42
    x_looker_status: str = Field(..., alias="x-looker-status")


class Id21(Message):
    pass


class Title13(Message):
    pass


class SpaceId(Message):
    pass


class Properties43(BaseModel):
    id: Id21
    title: Title13
    space_id: SpaceId


class ContentValidationLookMLDashboard(BaseModel):
    properties: Properties43
    x_looker_status: str = Field(..., alias="x-looker-status")


class LookmlLinkId2(Message):
    pass


class Properties44(BaseModel):
    lookml_link_id: LookmlLinkId2
    title: Title13


class ContentValidationLookMLDashboardElement(BaseModel):
    properties: Properties44
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can11(Can):
    pass


class LookId6(Message):
    pass


class LastViewedAt(Message):
    pass


class StartOfWeekDate(Message):
    pass


class Properties45(BaseModel):
    can: Can11
    id: Id21
    look_id: LookId6
    dashboard_id: DashboardId6
    title: Title13
    content_metadata_id: ContentMetadataId5
    user_id: UserId4
    group_id: GroupId
    view_count: ViewCount
    favorite_count: FavoriteCount
    last_viewed_at: LastViewedAt
    start_of_week_date: StartOfWeekDate


class ContentView(BaseModel):
    properties: Properties45
    x_looker_status: str = Field(..., alias="x-looker-status")


class ExternalUserId(BaseModel):
    type: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties46(BaseModel):
    external_user_id: ExternalUserId


class CreateEmbedUserRequest(BaseModel):
    properties: Properties46
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class UserId6(ExternalUserId):
    pass


class OauthApplicationId(ExternalUserId):
    pass


class AccessToken2(ExternalUserId):
    pass


class AccessTokenExpiresAt(BaseModel):
    type: str
    format: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class RefreshToken1(ExternalUserId):
    pass


class RefreshTokenExpiresAt(AccessTokenExpiresAt):
    pass


class Properties47(BaseModel):
    user_id: UserId6
    oauth_application_id: OauthApplicationId
    access_token: AccessToken2
    access_token_expires_at: AccessTokenExpiresAt
    refresh_token: RefreshToken1
    refresh_token_expires_at: RefreshTokenExpiresAt


class CreateOAuthApplicationUserStateRequest(BaseModel):
    properties: Properties47
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class UserId7(Message):
    pass


class OauthApplicationId1(Message):
    pass


class Properties48(BaseModel):
    user_id: UserId7
    oauth_application_id: OauthApplicationId1


class CreateOAuthApplicationUserStateResponse(BaseModel):
    properties: Properties48
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Can12(Can):
    pass


class ClientId(Message):
    pass


class CreatedAt4(Message):
    pass


class IsDisabled2(Message):
    pass


class Type6(Message):
    pass


class Url4(DocumentationUrl):
    pass


class Properties49(BaseModel):
    can: Can12
    id: Id21
    client_id: ClientId
    created_at: CreatedAt4
    is_disabled: IsDisabled2
    type: Type6
    url: Url4


class CredentialsApi3(BaseModel):
    properties: Properties49
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can13(Can):
    pass


class ClientSecret(Message):
    pass


class Properties50(BaseModel):
    can: Can13
    id: Id21
    client_id: ClientId
    created_at: CreatedAt4
    is_disabled: IsDisabled2
    type: Type6
    client_secret: ClientSecret
    url: Url4


class CreateCredentialsApi3(BaseModel):
    properties: Properties50
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can14(Can):
    pass


class Email(HomepageId):
    pass


class ForcedPasswordResetAtNextLogin(HomepageId):
    pass


class LoggedInAt(Message):
    pass


class PasswordResetUrl(Message):
    pass


class AccountSetupUrl(Message):
    pass


class UserUrl(DocumentationUrl):
    pass


class Properties51(BaseModel):
    can: Can14
    created_at: CreatedAt4
    email: Email
    forced_password_reset_at_next_login: ForcedPasswordResetAtNextLogin
    user_id: UserId7
    is_disabled: IsDisabled2
    logged_in_at: LoggedInAt
    password_reset_url: PasswordResetUrl
    account_setup_url: AccountSetupUrl
    type: Type6
    url: Url4
    user_url: UserUrl


class CredentialsEmail(BaseModel):
    properties: Properties51
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can15(Can):
    pass


class Properties52(BaseModel):
    can: Can15
    created_at: CreatedAt4
    email: Email
    forced_password_reset_at_next_login: ForcedPasswordResetAtNextLogin
    user_id: UserId7
    is_disabled: IsDisabled2
    logged_in_at: LoggedInAt
    password_reset_url: PasswordResetUrl
    account_setup_url: AccountSetupUrl
    type: Type6
    url: Url4
    user_url: UserUrl


class CredentialsEmailSearch(BaseModel):
    properties: Properties52
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can16(Can):
    pass


class ExternalGroupId(Message):
    pass


class ExternalUserId1(Message):
    pass


class Properties53(BaseModel):
    can: Can16
    created_at: CreatedAt4
    external_group_id: ExternalGroupId
    external_user_id: ExternalUserId1
    id: Id21
    is_disabled: IsDisabled2
    logged_in_at: LoggedInAt
    type: Type6
    url: Url4


class CredentialsEmbed(BaseModel):
    properties: Properties53
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can17(Can):
    pass


class Domain(Message):
    pass


class Email2(Message):
    pass


class GoogleUserId(Message):
    pass


class Properties54(BaseModel):
    can: Can17
    created_at: CreatedAt4
    domain: Domain
    email: Email2
    google_user_id: GoogleUserId
    is_disabled: IsDisabled2
    logged_in_at: LoggedInAt
    type: Type6
    url: Url4


class CredentialsGoogle(BaseModel):
    properties: Properties54
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can18(Can):
    pass


class LdapDn(Message):
    pass


class LdapId(Message):
    pass


class Properties55(BaseModel):
    can: Can18
    created_at: CreatedAt4
    email: Email2
    is_disabled: IsDisabled2
    ldap_dn: LdapDn
    ldap_id: LdapId
    logged_in_at: LoggedInAt
    type: Type6
    url: Url4


class CredentialsLDAP(BaseModel):
    properties: Properties55
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can19(Can):
    pass


class LoggedInIp(Message):
    pass


class Properties56(BaseModel):
    can: Can19
    created_at: CreatedAt4
    email: Email2
    is_disabled: IsDisabled2
    logged_in_at: LoggedInAt
    logged_in_ip: LoggedInIp
    type: Type6
    url: Url4
    user_url: UserUrl


class CredentialsLookerOpenid(BaseModel):
    properties: Properties56
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can20(Can):
    pass


class OidcUserId(Message):
    pass


class Properties57(BaseModel):
    can: Can20
    created_at: CreatedAt4
    email: Email2
    is_disabled: IsDisabled2
    logged_in_at: LoggedInAt
    oidc_user_id: OidcUserId
    type: Type6
    url: Url4


class CredentialsOIDC(BaseModel):
    properties: Properties57
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can21(Can):
    pass


class SamlUserId(Message):
    pass


class Properties58(BaseModel):
    can: Can21
    created_at: CreatedAt4
    email: Email2
    is_disabled: IsDisabled2
    logged_in_at: LoggedInAt
    saml_user_id: SamlUserId
    type: Type6
    url: Url4


class CredentialsSaml(BaseModel):
    properties: Properties58
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can22(Can):
    pass


class Verified(Message):
    pass


class Properties59(BaseModel):
    can: Can22
    created_at: CreatedAt4
    is_disabled: IsDisabled2
    type: Type6
    verified: Verified
    url: Url4


class CredentialsTotp(BaseModel):
    properties: Properties59
    x_looker_status: str = Field(..., alias="x-looker-status")


class Enabled(HomepageId):
    pass


class Content(HomepageId):
    pass


class Subject(HomepageId):
    pass


class Header(HomepageId):
    pass


class Properties60(BaseModel):
    enabled: Enabled
    content: Content
    subject: Subject
    header: Header


class CustomWelcomeEmail(BaseModel):
    properties: Properties60
    x_looker_status: str = Field(..., alias="x-looker-status")


class AggregateTableLookml(Message):
    pass


class Properties61(BaseModel):
    dashboard_id: DashboardId6
    aggregate_table_lookml: AggregateTableLookml


class DashboardAggregateTableLookml(BaseModel):
    properties: Properties61
    x_looker_status: str = Field(..., alias="x-looker-status")


class PageSideMargins(DeletedAt):
    pass


class PageBackgroundColor(HomepageId):
    pass


class TileTitleAlignment(HomepageId):
    pass


class TileSpaceBetween(DeletedAt):
    pass


class TileBackgroundColor(HomepageId):
    pass


class TileShadow(HomepageId):
    pass


class KeyColor(HomepageId):
    pass


class Properties62(BaseModel):
    page_side_margins: PageSideMargins
    page_background_color: PageBackgroundColor
    tile_title_alignment: TileTitleAlignment
    tile_space_between: TileSpaceBetween
    tile_background_color: TileBackgroundColor
    tile_shadow: TileShadow
    key_color: KeyColor


class DashboardAppearance(BaseModel):
    properties: Properties62
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can23(Can):
    pass


class BodyTextAsHtml(Message):
    pass


class DashboardId9(HomepageId):
    pass


class EditUri(DocumentationUrl):
    pass


class Look2(Model1):
    pass


class LookId7(HomepageId):
    pass


class MergeResultId(HomepageId):
    pass


class Query(Model1):
    pass


class RefreshInterval1(HomepageId):
    pass


class ResultMaker(Field):
    pass


class ResultMakerId(HomepageId):
    pass


class Title16(HomepageId):
    pass


class Type17(HomepageId):
    pass


class AlertCount(DocumentationUrl):
    pass


class TitleTextAsHtml(Message):
    pass


class SubtitleTextAsHtml(Message):
    pass


class Properties63(BaseModel):
    can: Can23
    body_text: BodyText
    body_text_as_html: BodyTextAsHtml
    dashboard_id: DashboardId9
    edit_uri: EditUri
    id: Id21
    look: Look2
    look_id: LookId7
    lookml_link_id: LookmlLinkId2
    merge_result_id: MergeResultId
    note_display: NoteDisplay
    note_state: NoteState
    note_text: NoteText
    note_text_as_html: NoteTextAsHtml
    query: Query
    query_id: QueryId
    refresh_interval: RefreshInterval1
    refresh_interval_to_i: RefreshIntervalToI
    result_maker: ResultMaker
    result_maker_id: ResultMakerId
    subtitle_text: SubtitleText
    title: Title16
    title_hidden: TitleHidden
    title_text: TitleText
    type: Type17
    alert_count: AlertCount
    rich_content_json: RichContentJson
    title_text_as_html: TitleTextAsHtml
    subtitle_text_as_html: SubtitleTextAsHtml
    extension_id: ExtensionId


class DashboardElement1(BaseModel):
    properties: Properties63
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can24(Can):
    pass


class DashboardId10(Message):
    pass


class AdditionalProperties29(AdditionalProperties):
    pass


class Field2(BaseModel):
    type: str
    additionalProperties: AdditionalProperties29
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Row(DeletedAt):
    pass


class Items139(Items):
    pass


class ListensToFilters(BaseModel):
    type: str
    items: Items139
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class AllowMultipleValues(HomepageId):
    pass


class Required(HomepageId):
    pass


class UiConfig(BaseModel):
    type: str
    additionalProperties: AdditionalProperties29
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties64(BaseModel):
    can: Can24
    id: Id21
    dashboard_id: DashboardId10
    name: Name2
    title: Title16
    type: Type17
    default_value: DefaultValue
    model: Model2
    explore: Explore
    dimension: Dimension
    field: Field2
    row: Row
    listens_to_filters: ListensToFilters
    allow_multiple_values: AllowMultipleValues
    required: Required
    ui_config: UiConfig


class DashboardFilter1(BaseModel):
    properties: Properties64
    x_looker_status: str = Field(..., alias="x-looker-status")


class DashboardId11(HomepageId):
    pass


class Field3(Field2):
    pass


class ListensToFilters1(ListensToFilters):
    pass


class UiConfig1(UiConfig):
    pass


class Properties65(BaseModel):
    id: Id21
    dashboard_id: DashboardId11
    name: Name2
    title: Title16
    type: Type17
    default_value: DefaultValue
    model: Model2
    explore: Explore
    dimension: Dimension
    field: Field3
    row: Row
    listens_to_filters: ListensToFilters1
    allow_multiple_values: AllowMultipleValues
    required: Required
    ui_config: UiConfig1


class CreateDashboardFilter(BaseModel):
    properties: Properties65
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class AdditionalProperties33(Items):
    pass


class Can25(BaseModel):
    type: str
    additionalProperties: AdditionalProperties33
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class DashboardLayoutId(HomepageId):
    pass


class Column(DeletedAt):
    pass


class Width(DeletedAt):
    pass


class Height(DeletedAt):
    pass


class Deleted(Message):
    pass


class ElementTitle(Message):
    pass


class ElementTitleHidden(Message):
    pass


class VisType(Message):
    pass


class Properties66(BaseModel):
    can: Can25
    id: Id21
    dashboard_layout_id: DashboardLayoutId
    dashboard_element_id: DashboardElementId
    row: Row
    column: Column
    width: Width
    height: Height
    deleted: Deleted
    element_title: ElementTitle
    element_title_hidden: ElementTitleHidden
    vis_type: VisType


class DashboardLayoutComponent(BaseModel):
    properties: Properties66
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can26(Can25):
    pass


class Active(HomepageId):
    pass


class ColumnWidth(DeletedAt):
    pass


class DashboardTitle(Message):
    pass


class Items141(Schema):
    pass


class DashboardLayoutComponents(BaseModel):
    type: str
    items: Items141
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties67(BaseModel):
    can: Can26
    id: Id21
    dashboard_id: DashboardId11
    type: Type17
    active: Active
    column_width: ColumnWidth
    width: Width
    deleted: Deleted
    dashboard_title: DashboardTitle
    dashboard_layout_components: DashboardLayoutComponents


class DashboardLayout(BaseModel):
    properties: Properties67
    x_looker_status: str = Field(..., alias="x-looker-status")


class DashboardId13(Message):
    pass


class FolderId2(CustomImageDataBase64):
    pass


class Lookml(HomepageId):
    pass


class Properties68(BaseModel):
    dashboard_id: DashboardId13
    folder_id: FolderId2
    lookml: Lookml


class DashboardLookml(BaseModel):
    properties: Properties68
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can27(Can25):
    pass


class Hidden1(HomepageId):
    pass


class Model5(Model1):
    pass


class QueryTimezone1(HomepageId):
    pass


class Slug2(HomepageId):
    pass


class PreferredViewer2(HomepageId):
    pass


class AlertSyncWithDashboardFilterEnabled(HomepageId):
    pass


class BackgroundColor(HomepageId):
    pass


class CreatedAt15(DocumentationUrl):
    pass


class CrossfilterEnabled(HomepageId):
    pass


class DashboardElements(DashboardLayoutComponents):
    pass


class DashboardFilters(DashboardLayoutComponents):
    pass


class DashboardLayouts(DashboardLayoutComponents):
    pass


class Deleted2(HomepageId):
    pass


class DeletedAt3(DocumentationUrl):
    pass


class DeleterId(Message):
    pass


class EnableVizFullScreen(HomepageId):
    pass


class FiltersBarCollapsed(HomepageId):
    pass


class FiltersLocationTop(HomepageId):
    pass


class LastAccessedAt(DocumentationUrl):
    pass


class LastViewedAt1(DocumentationUrl):
    pass


class LastUpdaterId(Message):
    pass


class LastUpdaterName(Message):
    pass


class UserName(Message):
    pass


class LoadConfiguration(HomepageId):
    pass


class LookmlLinkId4(HomepageId):
    pass


class ShowFiltersBar(HomepageId):
    pass


class ShowTitle(HomepageId):
    pass


class FolderId3(HomepageId):
    pass


class TextTileTextColor(HomepageId):
    pass


class TileTextColor(HomepageId):
    pass


class TitleColor(HomepageId):
    pass


class Appearance(Field):
    pass


class Url15(Message):
    pass


class Properties69(BaseModel):
    can: Can27
    content_favorite_id: ContentFavoriteId
    content_metadata_id: ContentMetadataId5
    description: Description8
    hidden: Hidden1
    id: Id21
    model: Model5
    query_timezone: QueryTimezone1
    readonly: Readonly
    refresh_interval: RefreshInterval1
    refresh_interval_to_i: RefreshIntervalToI
    folder: Folder
    title: Title16
    user_id: UserId7
    slug: Slug2
    preferred_viewer: PreferredViewer2
    alert_sync_with_dashboard_filter_enabled: AlertSyncWithDashboardFilterEnabled
    background_color: BackgroundColor
    created_at: CreatedAt15
    crossfilter_enabled: CrossfilterEnabled
    dashboard_elements: DashboardElements
    dashboard_filters: DashboardFilters
    dashboard_layouts: DashboardLayouts
    deleted: Deleted2
    deleted_at: DeletedAt3
    deleter_id: DeleterId
    edit_uri: EditUri
    enable_viz_full_screen: EnableVizFullScreen
    favorite_count: FavoriteCount
    filters_bar_collapsed: FiltersBarCollapsed
    filters_location_top: FiltersLocationTop
    last_accessed_at: LastAccessedAt
    last_viewed_at: LastViewedAt1
    updated_at: UpdatedAt
    last_updater_id: LastUpdaterId
    last_updater_name: LastUpdaterName
    user_name: UserName
    load_configuration: LoadConfiguration
    lookml_link_id: LookmlLinkId4
    show_filters_bar: ShowFiltersBar
    show_title: ShowTitle
    folder_id: FolderId3
    text_tile_text_color: TextTileTextColor
    tile_background_color: TileBackgroundColor
    tile_text_color: TileTextColor
    title_color: TitleColor
    view_count: ViewCount
    appearance: Appearance
    url: Url15


class Dashboard2(BaseModel):
    properties: Properties69
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name7(Message):
    pass


class Label3(Message):
    pass


class Description10(Message):
    pass


class Type21(Message):
    pass


class Default(Message):
    pass


class OauthUrl(Message):
    pass


class Interactive(Message):
    pass


class Required2(Message):
    pass


class Options(DashboardLayoutComponents):
    pass


class Properties70(BaseModel):
    name: Name7
    label: Label3
    description: Description10
    type: Type21
    default: Default
    oauth_url: OauthUrl
    interactive: Interactive
    required: Required2
    options: Options


class DataActionFormField(BaseModel):
    properties: Properties70
    x_looker_status: str = Field(..., alias="x-looker-status")


class State(Model1):
    pass


class Fields(DashboardLayoutComponents):
    pass


class Properties71(BaseModel):
    state: State
    fields: Fields


class DataActionForm(BaseModel):
    properties: Properties71
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties72(BaseModel):
    name: Name7
    label: Label3


class DataActionFormSelectOption(BaseModel):
    properties: Properties72
    x_looker_status: str = Field(..., alias="x-looker-status")


class AdditionalProperties36(AdditionalProperties):
    pass


class Action(BaseModel):
    type: str
    additionalProperties: AdditionalProperties36
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class AdditionalProperties37(Items):
    pass


class FormValues(BaseModel):
    type: str
    additionalProperties: AdditionalProperties37
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties73(BaseModel):
    action: Action
    form_values: FormValues


class DataActionRequest(BaseModel):
    properties: Properties73
    x_looker_status: str = Field(..., alias="x-looker-status")


class WebhookId(Message):
    pass


class Success(Message):
    pass


class RefreshQuery(Message):
    pass


class ValidationErrors(Model1):
    pass


class Properties74(BaseModel):
    webhook_id: WebhookId
    success: Success
    refresh_query: RefreshQuery
    validation_errors: ValidationErrors
    message: Message


class DataActionResponse(BaseModel):
    properties: Properties74
    x_looker_status: str = Field(..., alias="x-looker-status")


class Data(Message):
    pass


class RefreshTime(DocumentationUrl):
    pass


class Properties75(BaseModel):
    data: Data
    refresh_time: RefreshTime


class DataActionUserState(BaseModel):
    properties: Properties75
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can28(BaseModel):
    type: str
    additionalProperties: AdditionalProperties37
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class StaleBefore(DeletedAt):
    pass


class TriggerCheckAt(DocumentationUrl):
    pass


class TriggerError(Message):
    pass


class TriggerValue(Message):
    pass


class TriggeredAt(DeletedAt):
    pass


class Properties76(BaseModel):
    can: Can28
    created_at: CreatedAt15
    id: Id21
    model_name: ModelName
    name: Name7
    stale_before: StaleBefore
    trigger_check_at: TriggerCheckAt
    trigger_error: TriggerError
    trigger_value: TriggerValue
    triggered_at: TriggeredAt


class Datagroup(BaseModel):
    properties: Properties76
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can29(Can28):
    pass


class Dialect(Model1):
    pass


class Snippets(DashboardLayoutComponents):
    pass


class PdtsEnabled(Message):
    pass


class Properties77(BaseModel):
    can: Can29
    name: Name7
    dialect: Dialect
    snippets: Snippets
    pdts_enabled: PdtsEnabled


class DBConnectionBase(BaseModel):
    properties: Properties77
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can30(Can28):
    pass


class Name11(HomepageId):
    pass


class Snippets1(DashboardLayoutComponents):
    pass


class Host(HomepageId):
    pass


class Port(HomepageId):
    pass


class Username(HomepageId):
    pass


class Password(CustomImageDataBase64):
    pass


class UsesOauth(Message):
    pass


class Certificate(CustomImageDataBase64):
    pass


class FileType(CustomImageDataBase64):
    pass


class Database(HomepageId):
    pass


class DbTimezone(HomepageId):
    pass


class Schema1929(HomepageId):
    pass


class MaxConnections(DeletedAt):
    pass


class MaxBillingGigabytes(HomepageId):
    pass


class Ssl(HomepageId):
    pass


class VerifySsl(HomepageId):
    pass


class TmpDbName(HomepageId):
    pass


class JdbcAdditionalParams(HomepageId):
    pass


class PoolTimeout(DeletedAt):
    pass


class DialectName(HomepageId):
    pass


class SupportsDataStudioLink(Message):
    pass


class CreatedAt17(Message):
    pass


class Example(Message):
    pass


class UserDbCredentials(HomepageId):
    pass


class Items149(Items):
    pass


class UserAttributeFields(BaseModel):
    type: str
    items: Items149
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class MaintenanceCron(HomepageId):
    pass


class LastRegenAt(Message):
    pass


class LastReapAt(Message):
    pass


class SqlRunnerPrecacheTables(HomepageId):
    pass


class SqlWritingWithInfoSchema(HomepageId):
    pass


class AfterConnectStatements(HomepageId):
    pass


class PdtContextOverride(Field):
    pass


class Managed(Message):
    pass


class CustomLocalPort(BaseModel):
    type: str
    format: str
    description: str
    x_looker_undocumented: bool = Field(..., alias="x-looker-undocumented")
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class TunnelId(BaseModel):
    type: str
    description: str
    x_looker_undocumented: bool = Field(..., alias="x-looker-undocumented")
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class UsesTns(HomepageId):
    pass


class PdtConcurrency(DeletedAt):
    pass


class DisableContextComment(HomepageId):
    pass


class OauthApplicationId2(HomepageId):
    pass


class AlwaysRetryFailedBuilds(HomepageId):
    pass


class CostEstimateEnabled(HomepageId):
    pass


class PdtApiControlEnabled(HomepageId):
    pass


class Properties78(BaseModel):
    can: Can30
    name: Name11
    dialect: Dialect
    snippets: Snippets1
    pdts_enabled: PdtsEnabled
    host: Host
    port: Port
    username: Username
    password: Password
    uses_oauth: UsesOauth
    certificate: Certificate
    file_type: FileType
    database: Database
    db_timezone: DbTimezone
    query_timezone: QueryTimezone1
    schema_: Schema1929 = Field(..., alias="schema")
    max_connections: MaxConnections
    max_billing_gigabytes: MaxBillingGigabytes
    ssl: Ssl
    verify_ssl: VerifySsl
    tmp_db_name: TmpDbName
    jdbc_additional_params: JdbcAdditionalParams
    pool_timeout: PoolTimeout
    dialect_name: DialectName
    supports_data_studio_link: SupportsDataStudioLink
    created_at: CreatedAt17
    user_id: UserId7
    example: Example
    user_db_credentials: UserDbCredentials
    user_attribute_fields: UserAttributeFields
    maintenance_cron: MaintenanceCron
    last_regen_at: LastRegenAt
    last_reap_at: LastReapAt
    sql_runner_precache_tables: SqlRunnerPrecacheTables
    sql_writing_with_info_schema: SqlWritingWithInfoSchema
    after_connect_statements: AfterConnectStatements
    pdt_context_override: PdtContextOverride
    managed: Managed
    custom_local_port: CustomLocalPort
    tunnel_id: TunnelId
    uses_tns: UsesTns
    pdt_concurrency: PdtConcurrency
    disable_context_comment: DisableContextComment
    oauth_application_id: OauthApplicationId2
    always_retry_failed_builds: AlwaysRetryFailedBuilds
    cost_estimate_enabled: CostEstimateEnabled
    pdt_api_control_enabled: PdtApiControlEnabled


class DBConnection(BaseModel):
    properties: Properties78
    x_looker_status: str = Field(..., alias="x-looker-status")


class Context(HomepageId):
    pass


class HasPassword(Message):
    pass


class Properties79(BaseModel):
    context: Context
    host: Host
    port: Port
    username: Username
    password: Password
    has_password: HasPassword
    certificate: Certificate
    file_type: FileType
    database: Database
    schema_: Schema1929 = Field(..., alias="schema")
    jdbc_additional_params: JdbcAdditionalParams
    after_connect_statements: AfterConnectStatements


class DBConnectionOverride(BaseModel):
    properties: Properties79
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can31(Can28):
    pass


class ConnectionString(Message):
    pass


class Name12(Message):
    pass


class Status(Message):
    pass


class Properties80(BaseModel):
    can: Can31
    connection_string: ConnectionString
    message: Message
    name: Name12
    status: Status


class DBConnectionTestResult(BaseModel):
    properties: Properties80
    x_looker_status: str = Field(..., alias="x-looker-status")


class InstallationTargetId(Message):
    pass


class InstallationId(Message):
    pass


class Properties81(BaseModel):
    name: Name12
    installation_target_id: InstallationTargetId
    installation_id: InstallationId
    success: Success


class DelegateOauthTest(BaseModel):
    properties: Properties81
    x_looker_status: str = Field(..., alias="x-looker-status")


class GraphText(Message):
    pass


class Properties82(BaseModel):
    graph_text: GraphText


class DependencyGraph(BaseModel):
    properties: Properties82
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can32(Can28):
    pass


class DefaultMaxConnections(Message):
    pass


class DefaultPort(Message):
    pass


class Installed(Message):
    pass


class LabelForDatabaseEquivalent(Message):
    pass


class SupportedOptions(Model1):
    pass


class Properties83(BaseModel):
    can: Can32
    default_max_connections: DefaultMaxConnections
    default_port: DefaultPort
    installed: Installed
    label: Label3
    label_for_database_equivalent: LabelForDatabaseEquivalent
    name: Name12
    supported_options: SupportedOptions


class DialectInfo(BaseModel):
    properties: Properties83
    x_looker_status: str = Field(..., alias="x-looker-status")


class AdditionalParams(Message):
    pass


class AfterConnectStatements2(Message):
    pass


class AnalyticalViewDataset(Message):
    pass


class Auth(Message):
    pass


class CostEstimate(Message):
    pass


class DisableContextComment1(Message):
    pass


class Host2(Message):
    pass


class InstanceName(Message):
    pass


class MaxBillingGigabytes1(Message):
    pass


class OauthCredentials(Message):
    pass


class PdtsForOauth(Message):
    pass


class Port2(Message):
    pass


class ProjectName(Message):
    pass


class Schema1931(Message):
    pass


class ServiceAccountCredentials(Message):
    pass


class Ssl1(Message):
    pass


class Timezone(Message):
    pass


class TmpTable(Message):
    pass


class Tns(Message):
    pass


class Username2(Message):
    pass


class UsernameRequired(Message):
    pass


class Properties84(BaseModel):
    additional_params: AdditionalParams
    after_connect_statements: AfterConnectStatements2
    analytical_view_dataset: AnalyticalViewDataset
    auth: Auth
    cost_estimate: CostEstimate
    disable_context_comment: DisableContextComment1
    host: Host2
    instance_name: InstanceName
    max_billing_gigabytes: MaxBillingGigabytes1
    oauth_credentials: OauthCredentials
    pdts_for_oauth: PdtsForOauth
    port: Port2
    project_name: ProjectName
    schema_: Schema1931 = Field(..., alias="schema")
    service_account_credentials: ServiceAccountCredentials
    ssl: Ssl1
    timezone: Timezone
    tmp_table: TmpTable
    tns: Tns
    username: Username2
    username_required: UsernameRequired


class DialectInfoOptions(BaseModel):
    properties: Properties84
    x_looker_status: str = Field(..., alias="x-looker-status")


class SupportsCostEstimate(Message):
    pass


class CostEstimateStyle(Message):
    pass


class PersistentTableIndexes(Message):
    pass


class PersistentTableSortkeys(Message):
    pass


class PersistentTableDistkey(Message):
    pass


class SupportsStreaming(Message):
    pass


class AutomaticallyRunSqlRunnerSnippets(Message):
    pass


class ConnectionTests(BaseModel):
    type: str
    items: Items149
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class SupportsInducer(Message):
    pass


class SupportsMultipleDatabases(Message):
    pass


class SupportsPersistentDerivedTables(Message):
    pass


class HasSslSupport(Message):
    pass


class Properties85(BaseModel):
    name: Name12
    label: Label3
    supports_cost_estimate: SupportsCostEstimate
    cost_estimate_style: CostEstimateStyle
    persistent_table_indexes: PersistentTableIndexes
    persistent_table_sortkeys: PersistentTableSortkeys
    persistent_table_distkey: PersistentTableDistkey
    supports_streaming: SupportsStreaming
    automatically_run_sql_runner_snippets: AutomaticallyRunSqlRunnerSnippets
    connection_tests: ConnectionTests
    supports_inducer: SupportsInducer
    supports_multiple_databases: SupportsMultipleDatabases
    supports_persistent_derived_tables: SupportsPersistentDerivedTables
    has_ssl_support: HasSslSupport


class Dialect2(BaseModel):
    properties: Properties85
    x_looker_status: str = Field(..., alias="x-looker-status")


class ConfigurationDelivered(HomepageId):
    pass


class Properties86(BaseModel):
    configuration_delivered: ConfigurationDelivered


class DigestEmailSend(BaseModel):
    properties: Properties86
    x_looker_status: str = Field(..., alias="x-looker-status")


class IsEnabled(HomepageId):
    pass


class Properties87(BaseModel):
    is_enabled: IsEnabled


class DigestEmails(BaseModel):
    properties: Properties87
    x_looker_status: str = Field(..., alias="x-looker-status")


class EgressIpAddresses1(ConnectionTests):
    pass


class Properties88(BaseModel):
    egress_ip_addresses: EgressIpAddresses1


class EgressIpAddresses(BaseModel):
    properties: Properties88
    x_looker_status: str = Field(..., alias="x-looker-status")


class TargetUrl(DeletedAt):
    pass


class SessionLength(DeletedAt):
    pass


class ForceLogoutLogin(HomepageId):
    pass


class Properties89(BaseModel):
    target_url: TargetUrl
    session_length: SessionLength
    force_logout_login: ForceLogoutLogin


class EmbedParams(BaseModel):
    properties: Properties89
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class ExternalUserId2(HomepageId):
    pass


class FirstName(HomepageId):
    pass


class LastName(HomepageId):
    pass


class UserTimezone(HomepageId):
    pass


class Permissions(UserAttributeFields):
    pass


class Models(UserAttributeFields):
    pass


class GroupIds(UserAttributeFields):
    pass


class ExternalGroupId1(HomepageId):
    pass


class AdditionalProperties43(AdditionalProperties):
    pass


class UserAttributes(BaseModel):
    type: str
    additionalProperties: AdditionalProperties43
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class SecretId(HomepageId):
    pass


class Properties90(BaseModel):
    target_url: TargetUrl
    session_length: SessionLength
    force_logout_login: ForceLogoutLogin
    external_user_id: ExternalUserId2
    first_name: FirstName
    last_name: LastName
    user_timezone: UserTimezone
    permissions: Permissions
    models: Models
    group_ids: GroupIds
    external_group_id: ExternalGroupId1
    user_attributes: UserAttributes
    secret_id: SecretId


class EmbedSsoParams(BaseModel):
    properties: Properties90
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Permissions1(UserAttributeFields):
    pass


class Models1(UserAttributeFields):
    pass


class GroupIds1(UserAttributeFields):
    pass


class UserAttributes1(UserAttributes):
    pass


class SessionReferenceToken(HomepageId):
    pass


class EmbedDomain(HomepageId):
    pass


class Properties91(BaseModel):
    session_length: SessionLength
    force_logout_login: ForceLogoutLogin
    external_user_id: ExternalUserId2
    first_name: FirstName
    last_name: LastName
    user_timezone: UserTimezone
    permissions: Permissions1
    models: Models1
    group_ids: GroupIds1
    external_group_id: ExternalGroupId1
    user_attributes: UserAttributes1
    session_reference_token: SessionReferenceToken
    embed_domain: EmbedDomain


class EmbedCookielessSessionAcquire(BaseModel):
    properties: Properties91
    x_looker_status: str = Field(..., alias="x-looker-status")


class AuthenticationToken(HomepageId):
    pass


class AuthenticationTokenTtl(DeletedAt):
    pass


class NavigationToken(HomepageId):
    pass


class NavigationTokenTtl(DeletedAt):
    pass


class ApiToken(HomepageId):
    pass


class ApiTokenTtl(DeletedAt):
    pass


class SessionReferenceTokenTtl(DeletedAt):
    pass


class Properties92(BaseModel):
    authentication_token: AuthenticationToken
    authentication_token_ttl: AuthenticationTokenTtl
    navigation_token: NavigationToken
    navigation_token_ttl: NavigationTokenTtl
    api_token: ApiToken
    api_token_ttl: ApiTokenTtl
    session_reference_token: SessionReferenceToken
    session_reference_token_ttl: SessionReferenceTokenTtl


class EmbedCookielessSessionAcquireResponse(BaseModel):
    properties: Properties92
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties93(BaseModel):
    session_reference_token: SessionReferenceToken
    navigation_token: NavigationToken
    api_token: ApiToken


class EmbedCookielessSessionGenerateTokens(BaseModel):
    properties: Properties93
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Properties94(BaseModel):
    navigation_token: NavigationToken
    navigation_token_ttl: NavigationTokenTtl
    api_token: ApiToken
    api_token_ttl: ApiTokenTtl
    session_reference_token: SessionReferenceToken
    session_reference_token_ttl: SessionReferenceTokenTtl


class EmbedCookielessSessionGenerateTokensResponse(BaseModel):
    properties: Properties94
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Algorithm(HomepageId):
    pass


class Secret(Message):
    pass


class SecretType(ComparisonType):
    pass


class Properties95(BaseModel):
    algorithm: Algorithm
    created_at: CreatedAt17
    enabled: Enabled
    id: Id21
    secret: Secret
    user_id: UserId7
    secret_type: SecretType


class EmbedSecret(BaseModel):
    properties: Properties95
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties96(BaseModel):
    url: Url15


class EmbedUrlResponse(BaseModel):
    properties: Properties96
    x_looker_status: str = Field(..., alias="x-looker-status")


class Address(HomepageId):
    pass


class From(HomepageId):
    pass


class UserName1(HomepageId):
    pass


class Password2(HomepageId):
    pass


class Port3(DeletedAt):
    pass


class EnableStarttlsAuto(HomepageId):
    pass


class SslVersion(ComparisonType):
    pass


class DefaultSmtp(HomepageId):
    pass


class Properties97(BaseModel):
    address: Address
    from_: From = Field(..., alias="from")
    user_name: UserName1
    password: Password2
    port: Port3
    enable_starttls_auto: EnableStarttlsAuto
    ssl_version: SslVersion
    default_smtp: DefaultSmtp


class SmtpSettings(BaseModel):
    properties: Properties97
    x_looker_status: str = Field(..., alias="x-looker-status")


class AdditionalProperties45(Items):
    pass


class Can33(BaseModel):
    type: str
    additionalProperties: AdditionalProperties45
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Name16(HomepageId):
    pass


class ClientId2(HomepageId):
    pass


class ClientSecret1(CustomImageDataBase64):
    pass


class CreatedAt19(DocumentationUrl):
    pass


class Properties98(BaseModel):
    can: Can33
    id: Id21
    name: Name16
    client_id: ClientId2
    client_secret: ClientSecret1
    dialect_name: DialectName
    created_at: CreatedAt19


class ExternalOauthApplication(BaseModel):
    properties: Properties98
    x_looker_status: str = Field(..., alias="x-looker-status")


class ParentId1(HomepageId):
    pass


class CreatorId(Message):
    pass


class ChildCount(DocumentationUrl):
    pass


class ExternalId(Message):
    pass


class IsEmbed(Message):
    pass


class IsEmbedSharedRoot(Message):
    pass


class IsEmbedUsersRoot(Message):
    pass


class IsPersonal(Message):
    pass


class IsPersonalDescendant(Message):
    pass


class IsSharedRoot(Message):
    pass


class IsUsersRoot(Message):
    pass


class Can34(Can33):
    pass


class Properties99(BaseModel):
    name: Name16
    parent_id: ParentId1
    id: Id21
    content_metadata_id: ContentMetadataId5
    created_at: CreatedAt19
    creator_id: CreatorId
    child_count: ChildCount
    external_id: ExternalId
    is_embed: IsEmbed
    is_embed_shared_root: IsEmbedSharedRoot
    is_embed_users_root: IsEmbedUsersRoot
    is_personal: IsPersonal
    is_personal_descendant: IsPersonalDescendant
    is_shared_root: IsSharedRoot
    is_users_root: IsUsersRoot
    can: Can34


class FolderBase(BaseModel):
    properties: Properties99
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Properties100(BaseModel):
    name: Name16
    parent_id: ParentId1


class CreateFolder(BaseModel):
    properties: Properties100
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Properties101(Properties100):
    pass


class UpdateFolder(BaseModel):
    properties: Properties101
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can35(Can33):
    pass


class Items158(Schema):
    pass


class Dashboards(BaseModel):
    type: str
    items: Items158
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Looks(Dashboards):
    pass


class Properties102(BaseModel):
    name: Name16
    parent_id: ParentId1
    id: Id21
    content_metadata_id: ContentMetadataId5
    created_at: CreatedAt19
    creator_id: CreatorId
    child_count: ChildCount
    external_id: ExternalId
    is_embed: IsEmbed
    is_embed_shared_root: IsEmbedSharedRoot
    is_embed_users_root: IsEmbedUsersRoot
    is_personal: IsPersonal
    is_personal_descendant: IsPersonalDescendant
    is_shared_root: IsSharedRoot
    is_users_root: IsUsersRoot
    can: Can35
    dashboards: Dashboards
    looks: Looks


class Folder4(BaseModel):
    properties: Properties102
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Can36(Can33):
    pass


class Remote(Message):
    pass


class RemoteName(Message):
    pass


class Error1(Message):
    pass


class OwnerName(Message):
    pass


class Personal(Message):
    pass


class IsLocal(Message):
    pass


class IsRemote(Message):
    pass


class IsProduction(Message):
    pass


class AheadCount(DocumentationUrl):
    pass


class BehindCount(DocumentationUrl):
    pass


class CommitAt(DocumentationUrl):
    pass


class Ref(HomepageId):
    pass


class RemoteRef(Message):
    pass


class Properties103(BaseModel):
    can: Can36
    name: Name16
    remote: Remote
    remote_name: RemoteName
    error: Error1
    message: Message
    owner_name: OwnerName
    readonly: Readonly
    personal: Personal
    is_local: IsLocal
    is_remote: IsRemote
    is_production: IsProduction
    ahead_count: AheadCount
    behind_count: BehindCount
    commit_at: CommitAt
    ref: Ref
    remote_ref: RemoteRef


class GitBranch(BaseModel):
    properties: Properties103
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can37(Can33):
    pass


class Properties104(BaseModel):
    can: Can37
    description: Description10
    id: Id21


class GitConnectionTest(BaseModel):
    properties: Properties104
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can38(Can33):
    pass


class Properties105(BaseModel):
    can: Can38
    id: Id21
    message: Message
    status: Status


class GitConnectionTestResult(BaseModel):
    properties: Properties105
    x_looker_status: str = Field(..., alias="x-looker-status")


class Action1(Message):
    pass


class Conflict(Message):
    pass


class Revertable(Message):
    pass


class Text(Message):
    pass


class Properties106(BaseModel):
    action: Action1
    conflict: Conflict
    revertable: Revertable
    text: Text


class GitStatus(BaseModel):
    properties: Properties106
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties107(BaseModel):
    group_id: GroupId


class GroupIdForGroupInclusion(BaseModel):
    properties: Properties107
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can39(Can33):
    pass


class CanAddToContentMetadata(HomepageId):
    pass


class ContainsCurrentUser(Message):
    pass


class ExternalGroupId3(Message):
    pass


class ExternallyManaged(Message):
    pass


class IncludeByDefault(Message):
    pass


class UserCount(DocumentationUrl):
    pass


class Properties108(BaseModel):
    can: Can39
    can_add_to_content_metadata: CanAddToContentMetadata
    contains_current_user: ContainsCurrentUser
    external_group_id: ExternalGroupId3
    externally_managed: ExternallyManaged
    id: Id21
    include_by_default: IncludeByDefault
    name: Name16
    user_count: UserCount


class Group(BaseModel):
    properties: Properties108
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can40(Can33):
    pass


class Roles(Dashboards):
    pass


class Properties109(BaseModel):
    can: Can40
    can_add_to_content_metadata: CanAddToContentMetadata
    contains_current_user: ContainsCurrentUser
    external_group_id: ExternalGroupId3
    externally_managed: ExternallyManaged
    id: Id21
    include_by_default: IncludeByDefault
    name: Name16
    user_count: UserCount
    roles: Roles


class GroupSearch(BaseModel):
    properties: Properties109
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can41(Can33):
    pass


class Items161(Items):
    pass


class ParentGroupIds(BaseModel):
    type: str
    items: Items161
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class RoleIds(ParentGroupIds):
    pass


class Properties110(BaseModel):
    can: Can41
    can_add_to_content_metadata: CanAddToContentMetadata
    contains_current_user: ContainsCurrentUser
    external_group_id: ExternalGroupId3
    externally_managed: ExternallyManaged
    id: Id21
    include_by_default: IncludeByDefault
    name: Name16
    user_count: UserCount
    parent_group_ids: ParentGroupIds
    role_ids: RoleIds


class GroupHierarchy(BaseModel):
    properties: Properties110
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties111(BaseModel):
    user_id: UserId7


class GroupIdForGroupUserInclusion(BaseModel):
    properties: Properties111
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name25(Message):
    pass


class Ref1(Message):
    pass


class Properties112(BaseModel):
    name: Name25
    url: Url15
    ref: Ref1
    is_remote: IsRemote


class ImportedProject(BaseModel):
    properties: Properties112
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can42(Can33):
    pass


class Url18(HomepageId):
    pass


class Official(Message):
    pass


class FetchErrorMessage(Message):
    pass


class AuthorizationToken(CustomImageDataBase64):
    pass


class HasAuthorizationToken(Message):
    pass


class LegalAgreementSigned(Message):
    pass


class LegalAgreementRequired(Message):
    pass


class LegalAgreementText(Message):
    pass


class Properties113(BaseModel):
    can: Can42
    id: Id21
    url: Url18
    label: Label3
    official: Official
    fetch_error_message: FetchErrorMessage
    authorization_token: AuthorizationToken
    has_authorization_token: HasAuthorizationToken
    legal_agreement_signed: LegalAgreementSigned
    legal_agreement_required: LegalAgreementRequired
    legal_agreement_text: LegalAgreementText


class IntegrationHub(BaseModel):
    properties: Properties113
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can43(Can33):
    pass


class IntegrationHubId(Message):
    pass


class Items163(Schema):
    pass


class Params(BaseModel):
    type: str
    items: Items163
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items164(Items):
    pass


class SupportedFormats(BaseModel):
    type: str
    items: Items164
    readOnly: bool
    x_looker_values: List[str] = Field(..., alias="x-looker-values")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class SupportedActionTypes(SupportedFormats):
    pass


class SupportedFormattings(SupportedFormats):
    pass


class SupportedVisualizationFormattings(SupportedFormats):
    pass


class SupportedDownloadSettings(SupportedFormats):
    pass


class IconUrl(Message):
    pass


class Items169(Schema):
    pass


class RequiredFields(BaseModel):
    type: str
    items: Items169
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class PrivacyLink(Message):
    pass


class DelegateOauth(Message):
    pass


class Items170(Items):
    pass


class InstalledDelegateOauthTargets(BaseModel):
    type: str
    items: Items170
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties114(BaseModel):
    can: Can43
    id: Id21
    integration_hub_id: IntegrationHubId
    label: Label3
    description: Description10
    enabled: Enabled
    params: Params
    supported_formats: SupportedFormats
    supported_action_types: SupportedActionTypes
    supported_formattings: SupportedFormattings
    supported_visualization_formattings: SupportedVisualizationFormattings
    supported_download_settings: SupportedDownloadSettings
    icon_url: IconUrl
    uses_oauth: UsesOauth
    required_fields: RequiredFields
    privacy_link: PrivacyLink
    delegate_oauth: DelegateOauth
    installed_delegate_oauth_targets: InstalledDelegateOauthTargets


class Integration(BaseModel):
    properties: Properties114
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name26(HomepageId):
    pass


class HasValue(Message):
    pass


class UserAttributeName(HomepageId):
    pass


class Sensitive(Message):
    pass


class PerUser(Message):
    pass


class DelegateOauthUrl(Message):
    pass


class Properties115(BaseModel):
    name: Name26
    label: Label3
    description: Description10
    required: Required2
    has_value: HasValue
    value: Value
    user_attribute_name: UserAttributeName
    sensitive: Sensitive
    per_user: PerUser
    delegate_oauth_url: DelegateOauthUrl


class IntegrationParam(BaseModel):
    properties: Properties115
    x_looker_status: str = Field(..., alias="x-looker-status")


class Tag1(Message):
    pass


class AnyTag(BaseModel):
    type: str
    items: Items170
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class AllTags(AnyTag):
    pass


class Properties116(BaseModel):
    tag: Tag1
    any_tag: AnyTag
    all_tags: AllTags


class IntegrationRequiredField(BaseModel):
    properties: Properties116
    x_looker_status: str = Field(..., alias="x-looker-status")


class Items173(Schema):
    pass


class DelegateOauthResult(BaseModel):
    type: str
    items: Items173
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties117(BaseModel):
    success: Success
    message: Message
    delegate_oauth_result: DelegateOauthResult


class IntegrationTestResult(BaseModel):
    properties: Properties117
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can44(Can33):
    pass


class OrganizationName(HomepageId):
    pass


class MarkdownContent(HomepageId):
    pass


class Properties118(BaseModel):
    can: Can44
    organization_name: OrganizationName
    markdown_content: MarkdownContent


class InternalHelpResourcesContent(BaseModel):
    properties: Properties118
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can45(Can33):
    pass


class Properties119(BaseModel):
    can: Can45
    enabled: Enabled


class InternalHelpResources(BaseModel):
    properties: Properties119
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can46(Can33):
    pass


class AlternateEmailLoginAllowed(HomepageId):
    pass


class AuthPassword(CustomImageDataBase64):
    pass


class AuthRequiresRole(HomepageId):
    pass


class AuthUsername(HomepageId):
    pass


class ConnectionHost(HomepageId):
    pass


class ConnectionPort(HomepageId):
    pass


class ConnectionTls(HomepageId):
    pass


class ConnectionTlsNoVerify(HomepageId):
    pass


class Items174(Items):
    pass


class DefaultNewUserGroupIds(BaseModel):
    type: str
    items: Items174
    x_looker_write_only: bool = Field(..., alias="x-looker-write-only")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items175(Schema):
    pass


class DefaultNewUserGroups(BaseModel):
    type: str
    items: Items175
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items176(Items):
    pass


class DefaultNewUserRoleIds(BaseModel):
    type: str
    items: Items176
    x_looker_write_only: bool = Field(..., alias="x-looker-write-only")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items177(Schema):
    pass


class DefaultNewUserRoles(BaseModel):
    type: str
    items: Items177
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class ForceNoPage(HomepageId):
    pass


class Groups(DefaultNewUserRoles):
    pass


class GroupsBaseDn(HomepageId):
    pass


class GroupsFinderType(HomepageId):
    pass


class GroupsMemberAttribute(HomepageId):
    pass


class GroupsObjectclasses(HomepageId):
    pass


class GroupsUserAttribute(HomepageId):
    pass


class GroupsWithRoleIds(BaseModel):
    type: str
    items: Items177
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class HasAuthPassword(Message):
    pass


class MergeNewUsersByEmail(HomepageId):
    pass


class ModifiedAt(Message):
    pass


class ModifiedBy(Message):
    pass


class SetRolesFromGroups(HomepageId):
    pass


class TestLdapPassword(CustomImageDataBase64):
    pass


class TestLdapUser(CustomImageDataBase64):
    pass


class UserAttributeMapEmail(HomepageId):
    pass


class UserAttributeMapFirstName(HomepageId):
    pass


class UserAttributeMapLastName(HomepageId):
    pass


class UserAttributeMapLdapId(HomepageId):
    pass


class UserAttributes2(DefaultNewUserRoles):
    pass


class UserAttributesWithIds(GroupsWithRoleIds):
    pass


class UserBindBaseDn(HomepageId):
    pass


class UserCustomFilter(HomepageId):
    pass


class UserIdAttributeNames(HomepageId):
    pass


class UserObjectclass(HomepageId):
    pass


class AllowNormalGroupMembership(HomepageId):
    pass


class AllowRolesFromNormalGroups(HomepageId):
    pass


class AllowDirectRoles(HomepageId):
    pass


class Url19(DocumentationUrl):
    pass


class Properties120(BaseModel):
    can: Can46
    alternate_email_login_allowed: AlternateEmailLoginAllowed
    auth_password: AuthPassword
    auth_requires_role: AuthRequiresRole
    auth_username: AuthUsername
    connection_host: ConnectionHost
    connection_port: ConnectionPort
    connection_tls: ConnectionTls
    connection_tls_no_verify: ConnectionTlsNoVerify
    default_new_user_group_ids: DefaultNewUserGroupIds
    default_new_user_groups: DefaultNewUserGroups
    default_new_user_role_ids: DefaultNewUserRoleIds
    default_new_user_roles: DefaultNewUserRoles
    enabled: Enabled
    force_no_page: ForceNoPage
    groups: Groups
    groups_base_dn: GroupsBaseDn
    groups_finder_type: GroupsFinderType
    groups_member_attribute: GroupsMemberAttribute
    groups_objectclasses: GroupsObjectclasses
    groups_user_attribute: GroupsUserAttribute
    groups_with_role_ids: GroupsWithRoleIds
    has_auth_password: HasAuthPassword
    merge_new_users_by_email: MergeNewUsersByEmail
    modified_at: ModifiedAt
    modified_by: ModifiedBy
    set_roles_from_groups: SetRolesFromGroups
    test_ldap_password: TestLdapPassword
    test_ldap_user: TestLdapUser
    user_attribute_map_email: UserAttributeMapEmail
    user_attribute_map_first_name: UserAttributeMapFirstName
    user_attribute_map_last_name: UserAttributeMapLastName
    user_attribute_map_ldap_id: UserAttributeMapLdapId
    user_attributes: UserAttributes2
    user_attributes_with_ids: UserAttributesWithIds
    user_bind_base_dn: UserBindBaseDn
    user_custom_filter: UserCustomFilter
    user_id_attribute_names: UserIdAttributeNames
    user_objectclass: UserObjectclass
    allow_normal_group_membership: AllowNormalGroupMembership
    allow_roles_from_normal_groups: AllowRolesFromNormalGroups
    allow_direct_roles: AllowDirectRoles
    url: Url19


class LDAPConfig(BaseModel):
    properties: Properties120
    x_looker_status: str = Field(..., alias="x-looker-status")


class Details(Message):
    pass


class Issues(DefaultNewUserRoles):
    pass


class Trace(Message):
    pass


class User(Model1):
    pass


class Properties121(BaseModel):
    details: Details
    issues: Issues
    message: Message
    status: Status
    trace: Trace
    user: User
    url: Url19


class LDAPConfigTestResult(BaseModel):
    properties: Properties121
    x_looker_status: str = Field(..., alias="x-looker-status")


class Severity(Message):
    pass


class Properties122(BaseModel):
    severity: Severity
    message: Message


class LDAPConfigTestIssue(BaseModel):
    properties: Properties122
    x_looker_status: str = Field(..., alias="x-looker-status")


class LookerGroupId(Message):
    pass


class LookerGroupName(Message):
    pass


class Name27(Message):
    pass


class Roles1(DefaultNewUserRoles):
    pass


class Properties123(BaseModel):
    id: Id21
    looker_group_id: LookerGroupId
    looker_group_name: LookerGroupName
    name: Name27
    roles: Roles1
    url: Url19


class LDAPGroupRead(BaseModel):
    properties: Properties123
    x_looker_status: str = Field(..., alias="x-looker-status")


class Id45(HomepageId):
    pass


class LookerGroupName1(HomepageId):
    pass


class Name28(HomepageId):
    pass


class Items184(Items):
    pass


class RoleIds1(BaseModel):
    type: str
    items: Items184
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties124(BaseModel):
    id: Id45
    looker_group_id: LookerGroupId
    looker_group_name: LookerGroupName1
    name: Name28
    role_ids: RoleIds1
    url: Url19


class LDAPGroupWrite(BaseModel):
    properties: Properties124
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name29(Message):
    pass


class Items185(Schema):
    pass


class UserAttributes3(BaseModel):
    type: str
    items: Items185
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties125(BaseModel):
    name: Name29
    required: Required2
    user_attributes: UserAttributes3
    url: Url19


class LDAPUserAttributeRead(BaseModel):
    properties: Properties125
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name30(HomepageId):
    pass


class Required5(HomepageId):
    pass


class Items186(Items):
    pass


class UserAttributeIds(BaseModel):
    type: str
    items: Items186
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties126(BaseModel):
    name: Name30
    required: Required5
    user_attribute_ids: UserAttributeIds
    url: Url19


class LDAPUserAttributeWrite(BaseModel):
    properties: Properties126
    x_looker_status: str = Field(..., alias="x-looker-status")


class AllEmails(BaseModel):
    type: str
    items: Items186
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Attributes(Can33):
    pass


class FirstName2(Message):
    pass


class Groups1(AllEmails):
    pass


class LastName2(Message):
    pass


class Roles2(AllEmails):
    pass


class Properties127(BaseModel):
    all_emails: AllEmails
    attributes: Attributes
    email: Email2
    first_name: FirstName2
    groups: Groups1
    last_name: LastName2
    ldap_dn: LdapDn
    ldap_id: LdapId
    roles: Roles2
    url: Url19


class LDAPUser(BaseModel):
    properties: Properties127
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can47(Can33):
    pass


class Id46(Message):
    pass


class Name31(Message):
    pass


class EnabledLocally(HomepageId):
    pass


class Enabled5(Message):
    pass


class DisallowedAsOfVersion(Message):
    pass


class DisableOnUpgradeToVersion(Message):
    pass


class EndOfLifeVersion(Message):
    pass


class DocumentationUrl3(Message):
    pass


class ApproximateDisableDate(DocumentationUrl):
    pass


class ApproximateEndOfLifeDate(DocumentationUrl):
    pass


class HasDisabledOnUpgrade(Message):
    pass


class Properties128(BaseModel):
    can: Can47
    id: Id46
    name: Name31
    description: Description10
    enabled_locally: EnabledLocally
    enabled: Enabled5
    disallowed_as_of_version: DisallowedAsOfVersion
    disable_on_upgrade_to_version: DisableOnUpgradeToVersion
    end_of_life_version: EndOfLifeVersion
    documentation_url: DocumentationUrl3
    approximate_disable_date: ApproximateDisableDate
    approximate_end_of_life_date: ApproximateEndOfLifeDate
    has_disabled_on_upgrade: HasDisabledOnUpgrade


class LegacyFeature(BaseModel):
    properties: Properties128
    x_looker_status: str = Field(..., alias="x-looker-status")


class NativeName(Message):
    pass


class EnglishName(Message):
    pass


class Properties129(BaseModel):
    code: Code
    native_name: NativeName
    english_name: EnglishName


class Locale(BaseModel):
    properties: Properties129
    x_looker_status: str = Field(..., alias="x-looker-status")


class DefaultLocale(Message):
    pass


class LocalizationLevel(Message):
    pass


class Properties130(BaseModel):
    default_locale: DefaultLocale
    localization_level: LocalizationLevel


class LocalizationSettings(BaseModel):
    properties: Properties130
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can48(Can33):
    pass


class Title20(Message):
    pass


class UserId14(HomepageId):
    pass


class Properties131(BaseModel):
    can: Can48
    content_metadata_id: ContentMetadataId5
    id: Id46
    title: Title20
    user_id: UserId14


class LookBasic(BaseModel):
    properties: Properties131
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can49(Can33):
    pass


class Title21(HomepageId):
    pass


class Description15(HomepageId):
    pass


class EmbedUrl(Message):
    pass


class ExcelFileUrl(Message):
    pass


class GoogleSpreadsheetFormula(Message):
    pass


class ImageEmbedUrl(Message):
    pass


class IsRunOnLoad(HomepageId):
    pass


class Public(HomepageId):
    pass


class PublicSlug(Message):
    pass


class PublicUrl(Message):
    pass


class Folder5(Model1):
    pass


class Properties132(BaseModel):
    can: Can49
    content_metadata_id: ContentMetadataId5
    id: Id46
    title: Title21
    user_id: UserId14
    content_favorite_id: ContentFavoriteId
    created_at: CreatedAt19
    deleted: Deleted2
    deleted_at: DeletedAt3
    deleter_id: DeleterId
    description: Description15
    embed_url: EmbedUrl
    excel_file_url: ExcelFileUrl
    favorite_count: FavoriteCount
    google_spreadsheet_formula: GoogleSpreadsheetFormula
    image_embed_url: ImageEmbedUrl
    is_run_on_load: IsRunOnLoad
    last_accessed_at: LastAccessedAt
    last_updater_id: LastUpdaterId
    last_viewed_at: LastViewedAt1
    model: Model5
    public: Public
    public_slug: PublicSlug
    public_url: PublicUrl
    query_id: QueryId
    short_url: ShortUrl
    folder: Folder5
    folder_id: FolderId3
    updated_at: UpdatedAt
    view_count: ViewCount


class Look3(BaseModel):
    properties: Properties132
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can50(Can33):
    pass


class Url26(Message):
    pass


class Properties133(BaseModel):
    can: Can50
    content_metadata_id: ContentMetadataId5
    id: Id46
    title: Title21
    user_id: UserId14
    content_favorite_id: ContentFavoriteId
    created_at: CreatedAt19
    deleted: Deleted2
    deleted_at: DeletedAt3
    deleter_id: DeleterId
    description: Description15
    embed_url: EmbedUrl
    excel_file_url: ExcelFileUrl
    favorite_count: FavoriteCount
    google_spreadsheet_formula: GoogleSpreadsheetFormula
    image_embed_url: ImageEmbedUrl
    is_run_on_load: IsRunOnLoad
    last_accessed_at: LastAccessedAt
    last_updater_id: LastUpdaterId
    last_viewed_at: LastViewedAt1
    model: Model5
    public: Public
    public_slug: PublicSlug
    public_url: PublicUrl
    query_id: QueryId
    short_url: ShortUrl
    folder: Folder5
    folder_id: FolderId3
    updated_at: UpdatedAt
    view_count: ViewCount
    query: Query
    url: Url26


class LookWithQuery(BaseModel):
    properties: Properties133
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can51(Can33):
    pass


class Items190(Schema):
    pass


class Dashboards1(BaseModel):
    type: str
    items: Items190
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties134(BaseModel):
    can: Can51
    content_metadata_id: ContentMetadataId5
    id: Id46
    title: Title21
    user_id: UserId14
    content_favorite_id: ContentFavoriteId
    created_at: CreatedAt19
    deleted: Deleted2
    deleted_at: DeletedAt3
    deleter_id: DeleterId
    description: Description15
    embed_url: EmbedUrl
    excel_file_url: ExcelFileUrl
    favorite_count: FavoriteCount
    google_spreadsheet_formula: GoogleSpreadsheetFormula
    image_embed_url: ImageEmbedUrl
    is_run_on_load: IsRunOnLoad
    last_accessed_at: LastAccessedAt
    last_updater_id: LastUpdaterId
    last_viewed_at: LastViewedAt1
    model: Model5
    public: Public
    public_slug: PublicSlug
    public_url: PublicUrl
    query_id: QueryId
    short_url: ShortUrl
    folder: Folder5
    folder_id: FolderId3
    updated_at: UpdatedAt
    view_count: ViewCount
    dashboards: Dashboards1


class LookWithDashboards(BaseModel):
    properties: Properties134
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties135(BaseModel):
    id: Id46
    label: Label3


class LookModel(BaseModel):
    properties: Properties135
    x_looker_status: str = Field(..., alias="x-looker-status")


class Description18(Message):
    pass


class Hidden2(Message):
    pass


class GroupLabel(Message):
    pass


class Properties136(BaseModel):
    name: Name31
    description: Description18
    label: Label3
    hidden: Hidden2
    group_label: GroupLabel


class LookmlModelNavExplore(BaseModel):
    properties: Properties136
    x_looker_status: str = Field(..., alias="x-looker-status")


class Title24(BaseModel):
    type: str
    readOnly: bool
    description: str
    x_looker_undocumented: bool = Field(..., alias="x-looker-undocumented")
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items191(Items):
    pass


class Scopes(BaseModel):
    type: str
    items: Items191
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class CanTotal(Message):
    pass


class CanDevelop(Title24):
    pass


class CanSeeLookml(Title24):
    pass


class LookmlLink(Title24):
    pass


class CanSave(Message):
    pass


class CanExplain(Message):
    pass


class CanPivotInDb(Message):
    pass


class CanSubtotal(Message):
    pass


class HasTimezoneSupport(Message):
    pass


class ConnectionName(Message):
    pass


class NullSortTreatment(Message):
    pass


class Files(Scopes):
    pass


class SourceFile(Message):
    pass


class ViewName(Message):
    pass


class SqlTableName(Message):
    pass


class AccessFilterFields(BaseModel):
    type: str
    items: Items191
    readOnly: bool
    x_looker_deprecated: bool = Field(..., alias="x-looker-deprecated")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items194(Schema):
    pass


class AccessFilters(BaseModel):
    type: str
    items: Items194
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Aliases(AccessFilters):
    pass


class AlwaysFilter(AccessFilters):
    pass


class ConditionallyFilter(AccessFilters):
    pass


class Items198(Items):
    pass


class IndexFields(BaseModel):
    type: str
    items: Items198
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items199(Schema):
    pass


class Sets(BaseModel):
    type: str
    items: Items199
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items200(Items):
    pass


class Tags(BaseModel):
    type: str
    items: Items200
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items201(Schema):
    pass


class Errors2(BaseModel):
    type: str
    items: Items201
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Fields1(Model1):
    pass


class Joins(Errors2):
    pass


class SupportedMeasureTypes(Errors2):
    pass


class Items204(Items):
    pass


class AlwaysJoin(BaseModel):
    type: str
    items: Items204
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties137(BaseModel):
    id: Id46
    name: Name31
    description: Description18
    label: Label3
    title: Title24
    scopes: Scopes
    can_total: CanTotal
    can_develop: CanDevelop
    can_see_lookml: CanSeeLookml
    lookml_link: LookmlLink
    can_save: CanSave
    can_explain: CanExplain
    can_pivot_in_db: CanPivotInDb
    can_subtotal: CanSubtotal
    has_timezone_support: HasTimezoneSupport
    supports_cost_estimate: SupportsCostEstimate
    connection_name: ConnectionName
    null_sort_treatment: NullSortTreatment
    files: Files
    source_file: SourceFile
    project_name: ProjectName
    model_name: ModelName
    view_name: ViewName
    hidden: Hidden2
    sql_table_name: SqlTableName
    access_filter_fields: AccessFilterFields
    access_filters: AccessFilters
    aliases: Aliases
    always_filter: AlwaysFilter
    conditionally_filter: ConditionallyFilter
    index_fields: IndexFields
    sets: Sets
    tags: Tags
    errors: Errors2
    fields: Fields1
    joins: Joins
    group_label: GroupLabel
    supported_measure_types: SupportedMeasureTypes
    always_join: AlwaysJoin


class LookmlModelExplore(BaseModel):
    properties: Properties137
    x_looker_status: str = Field(..., alias="x-looker-status")


class DimensionType(BaseModel):
    type: str
    readOnly: bool
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class MeasureTypes(BaseModel):
    type: str
    items: Items204
    readOnly: bool
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties138(BaseModel):
    dimension_type: DimensionType
    measure_types: MeasureTypes


class LookmlModelExploreSupportedMeasureType(BaseModel):
    properties: Properties138
    x_looker_status: str = Field(..., alias="x-looker-status")


class Field4(Message):
    pass


class UserAttribute(Message):
    pass


class Properties139(BaseModel):
    field: Field4
    user_attribute: UserAttribute


class LookmlModelExploreAccessFilter(BaseModel):
    properties: Properties139
    x_looker_status: str = Field(..., alias="x-looker-status")


class Value3(Message):
    pass


class Properties140(BaseModel):
    name: Name31
    value: Value3


class LookmlModelExploreConditionallyFilter(BaseModel):
    properties: Properties140
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties141(Properties140):
    pass


class LookmlModelExploreAlwaysFilter(BaseModel):
    properties: Properties141
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties142(Properties140):
    pass


class LookmlModelExploreAlias(BaseModel):
    properties: Properties142
    x_looker_status: str = Field(..., alias="x-looker-status")


class Value6(AlwaysJoin):
    pass


class Properties143(BaseModel):
    name: Name31
    value: Value6


class LookmlModelExploreSet(BaseModel):
    properties: Properties143
    x_looker_status: str = Field(..., alias="x-looker-status")


class Details1(DocumentationUrl):
    pass


class ErrorPos(Message):
    pass


class FieldError(Message):
    pass


class Properties144(BaseModel):
    message: Message
    details: Details1
    error_pos: ErrorPos
    field_error: FieldError


class LookmlModelExploreError(BaseModel):
    properties: Properties144
    x_looker_status: str = Field(..., alias="x-looker-status")


class Items207(Schema):
    pass


class Dimensions(BaseModel):
    type: str
    items: Items207
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Measures(Dimensions):
    pass


class Filters(Dimensions):
    pass


class Parameters(Dimensions):
    pass


class Properties145(BaseModel):
    dimensions: Dimensions
    measures: Measures
    filters: Filters
    parameters: Parameters


class LookmlModelExploreFieldset(BaseModel):
    properties: Properties145
    x_looker_status: str = Field(..., alias="x-looker-status")


class Align(PermissionType):
    pass


class CanFilter(Message):
    pass


class Category(PermissionType):
    pass


class DefaultFilterValue(Message):
    pass


class DimensionGroup(Message):
    pass


class Enumerations(Dimensions):
    pass


class FieldGroupLabel(Message):
    pass


class FieldGroupVariant(Message):
    pass


class FillStyle(PermissionType):
    pass


class FiscalMonthOffset(DocumentationUrl):
    pass


class HasAllowedValues(Message):
    pass


class IsFilter(Message):
    pass


class IsFiscal(Message):
    pass


class IsNumeric(Message):
    pass


class IsTimeframe(Message):
    pass


class CanTimeFilter(Message):
    pass


class TimeInterval(Model1):
    pass


class LabelFromParameter(Message):
    pass


class LabelShort(Message):
    pass


class LookmlLink1(Message):
    pass


class MapLayer(Model1):
    pass


class Measure(Message):
    pass


class StrictValueFormat(Message):
    pass


class Parameter402(Message):
    pass


class Permanent(Message):
    pass


class PrimaryKey(Message):
    pass


class RequiresRefreshOnSort(Message):
    pass


class Scope(Message):
    pass


class Sortable(Message):
    pass


class SourceFilePath(Message):
    pass


class Sql(Message):
    pass


class SqlCase(Dimensions):
    pass


class Filters1(Dimensions):
    pass


class SuggestDimension(Message):
    pass


class SuggestExplore(Message):
    pass


class Suggestable(Message):
    pass


class Items214(Items):
    pass


class Suggestions(BaseModel):
    type: str
    items: Items214
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Tags1(Suggestions):
    pass


class UserAttributeFilterTypes(BaseModel):
    type: str
    items: Items214
    readOnly: bool
    x_looker_values: List[str] = Field(..., alias="x-looker-values")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class ValueFormat(Message):
    pass


class View(Message):
    pass


class ViewLabel(Message):
    pass


class Dynamic(Message):
    pass


class WeekStartDay(PermissionType):
    pass


class TimesUsed(DocumentationUrl):
    pass


class OriginalView(Message):
    pass


class Properties146(BaseModel):
    align: Align
    can_filter: CanFilter
    category: Category
    default_filter_value: DefaultFilterValue
    description: Description18
    dimension_group: DimensionGroup
    enumerations: Enumerations
    error: Error1
    field_group_label: FieldGroupLabel
    field_group_variant: FieldGroupVariant
    fill_style: FillStyle
    fiscal_month_offset: FiscalMonthOffset
    has_allowed_values: HasAllowedValues
    hidden: Hidden2
    is_filter: IsFilter
    is_fiscal: IsFiscal
    is_numeric: IsNumeric
    is_timeframe: IsTimeframe
    can_time_filter: CanTimeFilter
    time_interval: TimeInterval
    label: Label3
    label_from_parameter: LabelFromParameter
    label_short: LabelShort
    lookml_link: LookmlLink1
    map_layer: MapLayer
    measure: Measure
    name: Name31
    strict_value_format: StrictValueFormat
    parameter: Parameter402
    permanent: Permanent
    primary_key: PrimaryKey
    project_name: ProjectName
    requires_refresh_on_sort: RequiresRefreshOnSort
    scope: Scope
    sortable: Sortable
    source_file: SourceFile
    source_file_path: SourceFilePath
    sql: Sql
    sql_case: SqlCase
    filters: Filters1
    suggest_dimension: SuggestDimension
    suggest_explore: SuggestExplore
    suggestable: Suggestable
    suggestions: Suggestions
    tags: Tags1
    type: Type21
    user_attribute_filter_types: UserAttributeFilterTypes
    value_format: ValueFormat
    view: View
    view_label: ViewLabel
    dynamic: Dynamic
    week_start_day: WeekStartDay
    times_used: TimesUsed
    original_view: OriginalView


class LookmlModelExploreField(BaseModel):
    properties: Properties146
    x_looker_status: str = Field(..., alias="x-looker-status")


class XLookerPolymorphicType(BaseModel):
    type: str
    format: Optional[str] = None


class Value7(BaseModel):
    type: str
    format: str
    x_looker_polymorphic_types: List[XLookerPolymorphicType] = Field(..., alias="x-looker-polymorphic-types")
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties147(BaseModel):
    label: Label3
    value: Value7


class LookmlModelExploreFieldEnumeration(BaseModel):
    properties: Properties147
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name39(PermissionType):
    pass


class Properties148(BaseModel):
    name: Name39
    count: Count


class LookmlModelExploreFieldTimeInterval(BaseModel):
    properties: Properties148
    x_looker_status: str = Field(..., alias="x-looker-status")


class Value8(Message):
    pass


class Condition(Message):
    pass


class Properties149(BaseModel):
    value: Value8
    condition: Condition


class LookmlModelExploreFieldSqlCase(BaseModel):
    properties: Properties149
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties150(BaseModel):
    field: Field4
    condition: Condition


class LookmlModelExploreFieldMeasureFilters(BaseModel):
    properties: Properties150
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name40(Message):
    pass


class FeatureKey(Message):
    pass


class PropertyKey(Message):
    pass


class PropertyLabelKey(Message):
    pass


class Projection(Message):
    pass


class Format(PermissionType):
    pass


class ExtentsJsonUrl(Message):
    pass


class MaxZoomLevel(DocumentationUrl):
    pass


class MinZoomLevel(DocumentationUrl):
    pass


class Properties151(BaseModel):
    url: Url26
    name: Name40
    feature_key: FeatureKey
    property_key: PropertyKey
    property_label_key: PropertyLabelKey
    projection: Projection
    format: Format
    extents_json_url: ExtentsJsonUrl
    max_zoom_level: MaxZoomLevel
    min_zoom_level: MinZoomLevel


class LookmlModelExploreFieldMapLayer(BaseModel):
    properties: Properties151
    x_looker_status: str = Field(..., alias="x-looker-status")


class DependentFields(Suggestions):
    pass


class Fields2(Suggestions):
    pass


class ForeignKey(Message):
    pass


class From1(Message):
    pass


class OuterOnly(Message):
    pass


class Relationship(Message):
    pass


class RequiredJoins(Suggestions):
    pass


class SqlForeignKey(Message):
    pass


class SqlOn(Message):
    pass


class Properties152(BaseModel):
    name: Name40
    dependent_fields: DependentFields
    fields: Fields2
    foreign_key: ForeignKey
    from_: From1 = Field(..., alias="from")
    outer_only: OuterOnly
    relationship: Relationship
    required_joins: RequiredJoins
    sql_foreign_key: SqlForeignKey
    sql_on: SqlOn
    sql_table_name: SqlTableName
    type: Type21
    view_label: ViewLabel


class LookmlModelExploreJoins(BaseModel):
    properties: Properties152
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can52(Can33):
    pass


class AllowedDbConnectionNames(BaseModel):
    type: str
    items: Items214
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items221(Schema):
    pass


class Explores(BaseModel):
    type: str
    items: Items221
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class HasContent(Message):
    pass


class Name42(HomepageId):
    pass


class ProjectName3(HomepageId):
    pass


class UnlimitedDbConnections(HomepageId):
    pass


class Properties153(BaseModel):
    can: Can52
    allowed_db_connection_names: AllowedDbConnectionNames
    explores: Explores
    has_content: HasContent
    label: Label3
    name: Name42
    project_name: ProjectName3
    unlimited_db_connections: UnlimitedDbConnections


class LookmlModel(BaseModel):
    properties: Properties153
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can53(Can33):
    pass


class Name43(Message):
    pass


class QueryUrlParams(Message):
    pass


class File(Message):
    pass


class Line(DocumentationUrl):
    pass


class Properties154(BaseModel):
    can: Can53
    model_name: ModelName
    name: Name43
    explore_name: ExploreName
    query_url_params: QueryUrlParams
    file: File
    line: Line


class LookmlTest(BaseModel):
    properties: Properties154
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can54(Can33):
    pass


class TestName(Message):
    pass


class AssertionsCount(DocumentationUrl):
    pass


class AssertionsFailed(DocumentationUrl):
    pass


class Errors3(Explores):
    pass


class Warnings(Explores):
    pass


class Properties155(BaseModel):
    can: Can54
    model_name: ModelName
    test_name: TestName
    assertions_count: AssertionsCount
    assertions_failed: AssertionsFailed
    errors: Errors3
    warnings: Warnings
    success: Success


class LookmlTestResult(BaseModel):
    properties: Properties155
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can55(Can33):
    pass


class Imports(Explores):
    pass


class LocalizationSettings1(Model1):
    pass


class Properties156(BaseModel):
    can: Can55
    name: Name43
    imports: Imports
    localization_settings: LocalizationSettings1


class Manifest(BaseModel):
    properties: Properties156
    x_looker_status: str = Field(..., alias="x-looker-status")


class MaterializationId(Message):
    pass


class RespText(Message):
    pass


class Properties157(BaseModel):
    materialization_id: MaterializationId
    resp_text: RespText


class MaterializePDT(BaseModel):
    properties: Properties157
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can56(Can33):
    pass


class ColumnLimit(HomepageId):
    pass


class DynamicFields(HomepageId):
    pass


class Items225(Items):
    pass


class Pivots(BaseModel):
    type: str
    items: Items225
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class ResultMakerId1(Message):
    pass


class Sorts(Pivots):
    pass


class Items227(Schema):
    pass


class SourceQueries(BaseModel):
    type: str
    items: Items227
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Total(HomepageId):
    pass


class VisConfig(BaseModel):
    type: str
    additionalProperties: AdditionalProperties45
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties158(BaseModel):
    can: Can56
    column_limit: ColumnLimit
    dynamic_fields: DynamicFields
    id: Id46
    pivots: Pivots
    result_maker_id: ResultMakerId1
    sorts: Sorts
    source_queries: SourceQueries
    total: Total
    vis_config: VisConfig


class MergeQuery(BaseModel):
    properties: Properties158
    x_looker_status: str = Field(..., alias="x-looker-status")


class MergeFields(SourceQueries):
    pass


class Name45(HomepageId):
    pass


class Properties159(BaseModel):
    merge_fields: MergeFields
    name: Name45
    query_id: QueryId


class MergeQuerySourceQuery(BaseModel):
    properties: Properties159
    x_looker_status: str = Field(..., alias="x-looker-status")


class FieldName3(HomepageId):
    pass


class SourceFieldName(HomepageId):
    pass


class Properties160(BaseModel):
    field_name: FieldName3
    source_field_name: SourceFieldName


class MergeFields1(BaseModel):
    properties: Properties160
    x_looker_status: str = Field(..., alias="x-looker-status")


class MobileForceAuthentication(Message):
    pass


class MobileAppIntegration(Message):
    pass


class MobileFeatureFlags(BaseModel):
    type: str
    items: Items227
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties161(BaseModel):
    mobile_force_authentication: MobileForceAuthentication
    mobile_app_integration: MobileAppIntegration
    mobile_feature_flags: MobileFeatureFlags


class MobileSettings(BaseModel):
    properties: Properties161
    x_looker_status: str = Field(..., alias="x-looker-status")


class FeatureFlagName(Message):
    pass


class FeatureFlagState(Message):
    pass


class Properties162(BaseModel):
    feature_flag_name: FeatureFlagName
    feature_flag_state: FeatureFlagState


class MobileFeatureFlags1(BaseModel):
    properties: Properties162
    x_looker_status: str = Field(..., alias="x-looker-status")


class DeviceToken(HomepageId):
    pass


class DeviceType(ComparisonType):
    pass


class Properties163(BaseModel):
    id: Id46
    device_token: DeviceToken
    device_type: DeviceType


class MobileToken(BaseModel):
    properties: Properties163
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Can57(Can33):
    pass


class AllAccess(DimensionType):
    pass


class BuiltIn(DimensionType):
    pass


class Items230(Items):
    pass


class Models2(BaseModel):
    type: str
    items: Items230
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Url28(DocumentationUrl):
    pass


class Properties164(BaseModel):
    can: Can57
    all_access: AllAccess
    built_in: BuiltIn
    id: Id46
    models: Models2
    name: Name45
    url: Url28


class ModelSet(BaseModel):
    properties: Properties164
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can58(Can33):
    pass


class ClientGuid(Message):
    pass


class RedirectUri(HomepageId):
    pass


class DisplayName(HomepageId):
    pass


class Description21(HomepageId):
    pass


class Enabled6(HomepageId):
    pass


class GroupId3(HomepageId):
    pass


class TokensInvalidBefore(DocumentationUrl):
    pass


class Items231(Schema):
    pass


class ActivatedUsers(BaseModel):
    type: str
    items: Items231
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties165(BaseModel):
    can: Can58
    client_guid: ClientGuid
    redirect_uri: RedirectUri
    display_name: DisplayName
    description: Description21
    enabled: Enabled6
    group_id: GroupId3
    tokens_invalid_before: TokensInvalidBefore
    activated_users: ActivatedUsers


class OauthClientApp(BaseModel):
    properties: Properties165
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can59(Can33):
    pass


class Audience(HomepageId):
    pass


class AuthorizationEndpoint(DeletedAt):
    pass


class Items232(Items):
    pass


class DefaultNewUserGroupIds1(BaseModel):
    type: str
    items: Items232
    x_looker_write_only: bool = Field(..., alias="x-looker-write-only")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items233(Schema):
    pass


class DefaultNewUserGroups1(BaseModel):
    type: str
    items: Items233
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items234(Items):
    pass


class DefaultNewUserRoleIds1(BaseModel):
    type: str
    items: Items234
    x_looker_write_only: bool = Field(..., alias="x-looker-write-only")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Items235(Schema):
    pass


class DefaultNewUserRoles1(BaseModel):
    type: str
    items: Items235
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Groups2(DefaultNewUserRoles1):
    pass


class GroupsAttribute(HomepageId):
    pass


class GroupsWithRoleIds1(BaseModel):
    type: str
    items: Items235
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Identifier(HomepageId):
    pass


class Issuer(HomepageId):
    pass


class ModifiedAt1(DocumentationUrl):
    pass


class NewUserMigrationTypes(HomepageId):
    pass


class Items238(Items):
    pass


class Scopes1(BaseModel):
    type: str
    items: Items238
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Secret1(CustomImageDataBase64):
    pass


class TestSlug(Message):
    pass


class TokenEndpoint(HomepageId):
    pass


class Items239(Schema):
    pass


class UserAttributes4(BaseModel):
    type: str
    items: Items239
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class UserAttributesWithIds1(BaseModel):
    type: str
    items: Items239
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class UserinfoEndpoint(DeletedAt):
    pass


class Properties166(BaseModel):
    can: Can59
    alternate_email_login_allowed: AlternateEmailLoginAllowed
    audience: Audience
    auth_requires_role: AuthRequiresRole
    authorization_endpoint: AuthorizationEndpoint
    default_new_user_group_ids: DefaultNewUserGroupIds1
    default_new_user_groups: DefaultNewUserGroups1
    default_new_user_role_ids: DefaultNewUserRoleIds1
    default_new_user_roles: DefaultNewUserRoles1
    enabled: Enabled6
    groups: Groups2
    groups_attribute: GroupsAttribute
    groups_with_role_ids: GroupsWithRoleIds1
    identifier: Identifier
    issuer: Issuer
    modified_at: ModifiedAt1
    modified_by: ModifiedBy
    new_user_migration_types: NewUserMigrationTypes
    scopes: Scopes1
    secret: Secret1
    set_roles_from_groups: SetRolesFromGroups
    test_slug: TestSlug
    token_endpoint: TokenEndpoint
    user_attribute_map_email: UserAttributeMapEmail
    user_attribute_map_first_name: UserAttributeMapFirstName
    user_attribute_map_last_name: UserAttributeMapLastName
    user_attributes: UserAttributes4
    user_attributes_with_ids: UserAttributesWithIds1
    userinfo_endpoint: UserinfoEndpoint
    allow_normal_group_membership: AllowNormalGroupMembership
    allow_roles_from_normal_groups: AllowRolesFromNormalGroups
    allow_direct_roles: AllowDirectRoles
    url: Url28


class OIDCConfig(BaseModel):
    properties: Properties166
    x_looker_status: str = Field(..., alias="x-looker-status")


class LookerGroupName2(Message):
    pass


class Name47(Message):
    pass


class Roles3(UserAttributes4):
    pass


class Properties167(BaseModel):
    id: Id46
    looker_group_id: LookerGroupId
    looker_group_name: LookerGroupName2
    name: Name47
    roles: Roles3


class OIDCGroupRead(BaseModel):
    properties: Properties167
    x_looker_status: str = Field(..., alias="x-looker-status")


class Id57(HomepageId):
    pass


class LookerGroupName3(HomepageId):
    pass


class Name48(HomepageId):
    pass


class Items242(Items):
    pass


class RoleIds2(BaseModel):
    type: str
    items: Items242
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties168(BaseModel):
    id: Id57
    looker_group_id: LookerGroupId
    looker_group_name: LookerGroupName3
    name: Name48
    role_ids: RoleIds2


class OIDCGroupWrite(BaseModel):
    properties: Properties168
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name49(Message):
    pass


class Required6(Message):
    pass


class Items243(Schema):
    pass


class UserAttributes5(BaseModel):
    type: str
    items: Items243
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties169(BaseModel):
    name: Name49
    required: Required6
    user_attributes: UserAttributes5


class OIDCUserAttributeRead(BaseModel):
    properties: Properties169
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name50(HomepageId):
    pass


class Required7(HomepageId):
    pass


class Items244(Items):
    pass


class UserAttributeIds1(BaseModel):
    type: str
    items: Items244
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties170(BaseModel):
    name: Name50
    required: Required7
    user_attribute_ids: UserAttributeIds1


class OIDCUserAttributeWrite(BaseModel):
    properties: Properties170
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can60(Can33):
    pass


class MinLength(DeletedAt):
    pass


class RequireNumeric(HomepageId):
    pass


class RequireUpperlower(HomepageId):
    pass


class RequireSpecial(HomepageId):
    pass


class Properties171(BaseModel):
    can: Can60
    min_length: MinLength
    require_numeric: RequireNumeric
    require_upperlower: RequireUpperlower
    require_special: RequireSpecial


class PasswordConfig(BaseModel):
    properties: Properties171
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can61(Can33):
    pass


class Permission1(Message):
    pass


class Parent(Message):
    pass


class Description22(Message):
    pass


class Properties172(BaseModel):
    can: Can61
    permission: Permission1
    parent: Parent
    description: Description22


class Permission(BaseModel):
    properties: Properties172
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can62(Can33):
    pass


class Id58(Message):
    pass


class Permissions2(BaseModel):
    type: str
    items: Items244
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties173(BaseModel):
    can: Can62
    all_access: AllAccess
    built_in: BuiltIn
    id: Id58
    name: Name50
    permissions: Permissions2
    url: Url28


class PermissionSet(BaseModel):
    properties: Properties173
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can63(Can33):
    pass


class Path(Message):
    pass


class Title25(Message):
    pass


class Extension(Message):
    pass


class MimeType(Message):
    pass


class Editable(Message):
    pass


class GitStatus1(Model1):
    pass


class Properties174(BaseModel):
    can: Can63
    id: Id58
    path: Path
    title: Title25
    type: Type21
    extension: Extension
    mime_type: MimeType
    editable: Editable
    git_status: GitStatus1


class ProjectFile(BaseModel):
    properties: Properties174
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can64(Can33):
    pass


class UsesGit(Message):
    pass


class GitRemoteUrl(HomepageId):
    pass


class GitUsername(HomepageId):
    pass


class GitPassword(CustomImageDataBase64):
    pass


class GitProductionBranchName(HomepageId):
    pass


class UseGitCookieAuth(HomepageId):
    pass


class GitUsernameUserAttribute(HomepageId):
    pass


class GitPasswordUserAttribute(HomepageId):
    pass


class GitServiceName(HomepageId):
    pass


class GitApplicationServerHttpPort(DeletedAt):
    pass


class GitApplicationServerHttpScheme(HomepageId):
    pass


class DeploySecret(CustomImageDataBase64):
    pass


class UnsetDeploySecret(CustomImageDataBase64):
    pass


class PullRequestMode(ComparisonType):
    pass


class ValidationRequired(HomepageId):
    pass


class GitReleaseMgmtEnabled(HomepageId):
    pass


class AllowWarnings(HomepageId):
    pass


class IsExample(Message):
    pass


class DependencyStatus(HomepageId):
    pass


class Properties175(BaseModel):
    can: Can64
    id: Id58
    name: Name50
    uses_git: UsesGit
    git_remote_url: GitRemoteUrl
    git_username: GitUsername
    git_password: GitPassword
    git_production_branch_name: GitProductionBranchName
    use_git_cookie_auth: UseGitCookieAuth
    git_username_user_attribute: GitUsernameUserAttribute
    git_password_user_attribute: GitPasswordUserAttribute
    git_service_name: GitServiceName
    git_application_server_http_port: GitApplicationServerHttpPort
    git_application_server_http_scheme: GitApplicationServerHttpScheme
    deploy_secret: DeploySecret
    unset_deploy_secret: UnsetDeploySecret
    pull_request_mode: PullRequestMode
    validation_required: ValidationRequired
    git_release_mgmt_enabled: GitReleaseMgmtEnabled
    allow_warnings: AllowWarnings
    is_example: IsExample
    dependency_status: DependencyStatus


class Project(BaseModel):
    properties: Properties175
    x_looker_status: str = Field(..., alias="x-looker-status")


class Kind(Message):
    pass


class FieldName4(Message):
    pass


class FilePath(Message):
    pass


class LineNumber(DocumentationUrl):
    pass


class ModelId(Message):
    pass


class Explore3(Message):
    pass


class HelpUrl(Message):
    pass


class AdditionalProperties79(AdditionalProperties):
    pass


class Params1(BaseModel):
    type: str
    additionalProperties: AdditionalProperties79
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class SanitizedMessage(Message):
    pass


class Properties176(BaseModel):
    code: Code
    severity: Severity
    kind: Kind
    message: Message
    field_name: FieldName4
    file_path: FilePath
    line_number: LineNumber
    model_id: ModelId
    explore: Explore3
    help_url: HelpUrl
    params: Params1
    sanitized_message: SanitizedMessage


class ProjectError(BaseModel):
    properties: Properties176
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name53(Message):
    pass


class ProjectFileId(Message):
    pass


class Properties177(BaseModel):
    name: Name53
    project_file_id: ProjectFileId


class ModelsNotValidated(BaseModel):
    properties: Properties177
    x_looker_status: str = Field(..., alias="x-looker-status")


class Items246(Schema):
    pass


class Errors4(BaseModel):
    type: str
    items: Items246
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class ProjectDigest(Message):
    pass


class ModelsNotValidated1(Errors4):
    pass


class Properties178(BaseModel):
    errors: Errors4
    project_digest: ProjectDigest
    models_not_validated: ModelsNotValidated1
    computation_time: ComputationTime


class ProjectValidation(BaseModel):
    properties: Properties178
    x_looker_status: str = Field(..., alias="x-looker-status")


class Errors5(Errors4):
    pass


class ModelsNotValidated2(Errors4):
    pass


class Stale(Message):
    pass


class Properties179(BaseModel):
    errors: Errors5
    project_digest: ProjectDigest
    models_not_validated: ModelsNotValidated2
    computation_time: ComputationTime
    stale: Stale


class ProjectValidationCache(BaseModel):
    properties: Properties179
    x_looker_status: str = Field(..., alias="x-looker-status")


class AdditionalProperties80(Items):
    pass


class Can65(BaseModel):
    type: str
    additionalProperties: AdditionalProperties80
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class ProjectId(Message):
    pass


class WorkspaceId1(Message):
    pass


class GitStatus2(Message):
    pass


class GitHead(Message):
    pass


class DependencyStatus1(PermissionType):
    pass


class GitBranch1(Model1):
    pass


class LookmlType(Message):
    pass


class Properties180(BaseModel):
    can: Can65
    project_id: ProjectId
    workspace_id: WorkspaceId1
    git_status: GitStatus2
    git_head: GitHead
    dependency_status: DependencyStatus1
    git_branch: GitBranch1
    lookml_type: LookmlType


class ProjectWorkspace(BaseModel):
    properties: Properties180
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can66(Can65):
    pass


class Model9(HomepageId):
    pass


class View1(HomepageId):
    pass


class Items250(Items):
    pass


class Fields3(BaseModel):
    type: str
    items: Items250
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Pivots1(Fields3):
    pass


class FillFields(Fields3):
    pass


class Filters2(BaseModel):
    type: str
    additionalProperties: AdditionalProperties80
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class FilterExpression(HomepageId):
    pass


class Sorts1(Fields3):
    pass


class Limit(HomepageId):
    pass


class RowTotal(HomepageId):
    pass


class Subtotals(Fields3):
    pass


class AdditionalProperties83(AdditionalProperties):
    pass


class VisConfig1(BaseModel):
    type: str
    additionalProperties: AdditionalProperties83
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class FilterConfig(VisConfig1):
    pass


class VisibleUiSections(HomepageId):
    pass


class Slug3(Message):
    pass


class ShareUrl(Message):
    pass


class ExpandedShareUrl(Message):
    pass


class Url31(Message):
    pass


class HasTableCalculations(Message):
    pass


class Properties181(BaseModel):
    can: Can66
    id: Id58
    model: Model9
    view: View1
    fields: Fields3
    pivots: Pivots1
    fill_fields: FillFields
    filters: Filters2
    filter_expression: FilterExpression
    sorts: Sorts1
    limit: Limit
    column_limit: ColumnLimit
    total: Total
    row_total: RowTotal
    subtotals: Subtotals
    vis_config: VisConfig1
    filter_config: FilterConfig
    visible_ui_sections: VisibleUiSections
    slug: Slug3
    dynamic_fields: DynamicFields
    client_id: ClientId2
    share_url: ShareUrl
    expanded_share_url: ExpandedShareUrl
    url: Url31
    query_timezone: QueryTimezone1
    has_table_calculations: HasTableCalculations


class Query2(BaseModel):
    properties: Properties181
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class AdditionalProperties85(Items):
    pass


class Can67(BaseModel):
    type: str
    additionalProperties: AdditionalProperties85
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class ResultFormat(ComparisonType):
    pass


class Source(HomepageId):
    pass


class Deferred(HomepageId):
    pass


class DashboardId14(HomepageId):
    pass


class Properties182(BaseModel):
    can: Can67
    query_id: QueryId
    result_format: ResultFormat
    source: Source
    deferred: Deferred
    look_id: LookId7
    dashboard_id: DashboardId14


class CreateQueryTask(BaseModel):
    properties: Properties182
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Can68(Can67):
    pass


class Query3(Model1):
    pass


class GenerateLinks(HomepageId):
    pass


class ForceProduction(HomepageId):
    pass


class PathPrefix(HomepageId):
    pass


class Cache(HomepageId):
    pass


class ServerTableCalcs(HomepageId):
    pass


class CacheOnly(HomepageId):
    pass


class CacheKey(Message):
    pass


class Status3(HomepageId):
    pass


class Runtime(DocumentationUrl):
    pass


class RebuildPdts(HomepageId):
    pass


class ResultSource(Message):
    pass


class ResultFormat1(Message):
    pass


class Properties183(BaseModel):
    can: Can68
    id: Id58
    query_id: QueryId
    query: Query3
    generate_links: GenerateLinks
    force_production: ForceProduction
    path_prefix: PathPrefix
    cache: Cache
    server_table_calcs: ServerTableCalcs
    cache_only: CacheOnly
    cache_key: CacheKey
    status: Status3
    source: Source
    runtime: Runtime
    rebuild_pdts: RebuildPdts
    result_source: ResultSource
    look_id: LookId7
    dashboard_id: DashboardId14
    result_format: ResultFormat1


class QueryTask(BaseModel):
    properties: Properties183
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can69(Can67):
    pass


class CreatedAt25(Message):
    pass


class DashboardFilters1(Message):
    pass


class DashboardId16(Message):
    pass


class DashboardStyle(Message):
    pass


class FinalizedAt(Message):
    pass


class Height1(DocumentationUrl):
    pass


class LookId10(Message):
    pass


class LookmlDashboardId4(Message):
    pass


class QueryId8(Message):
    pass


class DashboardElementId2(Message):
    pass


class QueryRuntime(DocumentationUrl):
    pass


class RenderRuntime(DocumentationUrl):
    pass


class Status4(Message):
    pass


class StatusDetail(Message):
    pass


class UserId18(Message):
    pass


class Width2(DocumentationUrl):
    pass


class Properties184(BaseModel):
    can: Can69
    created_at: CreatedAt25
    dashboard_filters: DashboardFilters1
    dashboard_id: DashboardId16
    dashboard_style: DashboardStyle
    finalized_at: FinalizedAt
    height: Height1
    id: Id58
    look_id: LookId10
    lookml_dashboard_id: LookmlDashboardId4
    query_id: QueryId8
    dashboard_element_id: DashboardElementId2
    query_runtime: QueryRuntime
    render_runtime: RenderRuntime
    result_format: ResultFormat1
    runtime: Runtime
    status: Status4
    status_detail: StatusDetail
    user_id: UserId18
    width: Width2


class RenderTask(BaseModel):
    properties: Properties184
    x_looker_status: str = Field(..., alias="x-looker-status")


class DashboardFilters2(HomepageId):
    pass


class DashboardStyle1(HomepageId):
    pass


class Properties185(BaseModel):
    dashboard_filters: DashboardFilters2
    dashboard_style: DashboardStyle1


class CreateDashboardRenderTask(BaseModel):
    properties: Properties185
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can70(Can67):
    pass


class RootProjectId(Message):
    pass


class RemoteUrl(Message):
    pass


class SshPublicKey(HomepageId):
    pass


class IsConfigured(Message):
    pass


class Properties186(BaseModel):
    can: Can70
    id: Id58
    root_project_id: RootProjectId
    remote_url: RemoteUrl
    git_username: GitUsername
    git_password: GitPassword
    ssh_public_key: SshPublicKey
    is_configured: IsConfigured


class RepositoryCredential(BaseModel):
    properties: Properties186
    x_looker_status: str = Field(..., alias="x-looker-status")


class DashboardFilterName(HomepageId):
    pass


class Field6(HomepageId):
    pass


class Properties187(BaseModel):
    dashboard_filter_name: DashboardFilterName
    field: Field6


class ResultMakerFilterablesListen(BaseModel):
    properties: Properties187
    x_looker_status: str = Field(..., alias="x-looker-status")


class Model10(Message):
    pass


class View2(Message):
    pass


class Items255(Schema):
    pass


class Listen(BaseModel):
    type: str
    items: Items255
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties188(BaseModel):
    model: Model10
    view: View2
    name: Name53
    listen: Listen


class ResultMakerFilterables(BaseModel):
    properties: Properties188
    x_looker_status: str = Field(..., alias="x-looker-status")


class DynamicFields2(Message):
    pass


class Filterables(Listen):
    pass


class Items257(Items):
    pass


class Sorts2(BaseModel):
    type: str
    items: Items257
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class MergeResultId1(Message):
    pass


class Total2(Message):
    pass


class SqlQueryId(Message):
    pass


class AdditionalProperties89(AdditionalProperties):
    pass


class VisConfig2(BaseModel):
    type: str
    additionalProperties: AdditionalProperties89
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties189(BaseModel):
    id: Id58
    dynamic_fields: DynamicFields2
    filterables: Filterables
    sorts: Sorts2
    merge_result_id: MergeResultId1
    total: Total2
    query_id: QueryId8
    sql_query_id: SqlQueryId
    query: Query3
    vis_config: VisConfig2


class ResultMakerWithIdVisConfigAndDynamicFields(BaseModel):
    properties: Properties189
    x_looker_status: str = Field(..., alias="x-looker-status")


class AdditionalProperties90(Items):
    pass


class Can71(BaseModel):
    type: str
    additionalProperties: AdditionalProperties90
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Name55(HomepageId):
    pass


class PermissionSet1(Model1):
    pass


class PermissionSetId(CustomImageDataBase64):
    pass


class ModelSet1(Model1):
    pass


class ModelSetId(CustomImageDataBase64):
    pass


class Url32(DocumentationUrl):
    pass


class UsersUrl(DocumentationUrl):
    pass


class Properties190(BaseModel):
    can: Can71
    id: Id58
    name: Name55
    permission_set: PermissionSet1
    permission_set_id: PermissionSetId
    model_set: ModelSet1
    model_set_id: ModelSetId
    url: Url32
    users_url: UsersUrl


class Role(BaseModel):
    properties: Properties190
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can72(Can71):
    pass


class Properties191(BaseModel):
    can: Can72
    id: Id58
    name: Name55
    permission_set: PermissionSet1
    permission_set_id: PermissionSetId
    model_set: ModelSet1
    model_set_id: ModelSetId
    user_count: UserCount
    url: Url32
    users_url: UsersUrl


class RoleSearch(BaseModel):
    properties: Properties191
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can73(Can71):
    pass


class SqlQuery(Model1):
    pass


class Look4(Model1):
    pass


class CompletedAt(Message):
    pass


class Source2(Message):
    pass


class NodeId(Message):
    pass


class QueryTaskId(Message):
    pass


class Dialect3(Message):
    pass


class ConnectionId(Message):
    pass


class Properties192(BaseModel):
    can: Can73
    id: Id58
    user: User
    query: Query3
    sql_query: SqlQuery
    look: Look4
    created_at: CreatedAt25
    completed_at: CompletedAt
    query_id: QueryId8
    source: Source2
    node_id: NodeId
    slug: Slug3
    query_task_id: QueryTaskId
    cache_key: CacheKey
    connection_name: ConnectionName
    dialect: Dialect3
    connection_id: ConnectionId
    message: Message
    status: Status4
    runtime: Runtime
    sql: Sql


class RunningQueries(BaseModel):
    properties: Properties192
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can74(Can71):
    pass


class IdpCert(HomepageId):
    pass


class IdpUrl(HomepageId):
    pass


class IdpIssuer(HomepageId):
    pass


class IdpAudience(HomepageId):
    pass


class AllowedClockDrift(DeletedAt):
    pass


class ModifiedAt2(Message):
    pass


class Items258(Schema):
    pass


class DefaultNewUserRoles2(BaseModel):
    type: str
    items: Items258
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class DefaultNewUserGroups2(DefaultNewUserRoles2):
    pass


class Items260(Items):
    pass


class DefaultNewUserRoleIds2(BaseModel):
    type: str
    items: Items260
    x_looker_write_only: bool = Field(..., alias="x-looker-write-only")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class DefaultNewUserGroupIds2(DefaultNewUserRoleIds2):
    pass


class Items262(Schema):
    pass


class Groups3(BaseModel):
    type: str
    items: Items262
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class GroupsWithRoleIds2(BaseModel):
    type: str
    items: Items262
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class UserAttributes6(Groups3):
    pass


class UserAttributesWithIds2(GroupsWithRoleIds2):
    pass


class GroupsMemberValue(HomepageId):
    pass


class BypassLoginPage(HomepageId):
    pass


class Properties193(BaseModel):
    can: Can74
    enabled: Enabled6
    idp_cert: IdpCert
    idp_url: IdpUrl
    idp_issuer: IdpIssuer
    idp_audience: IdpAudience
    allowed_clock_drift: AllowedClockDrift
    user_attribute_map_email: UserAttributeMapEmail
    user_attribute_map_first_name: UserAttributeMapFirstName
    user_attribute_map_last_name: UserAttributeMapLastName
    new_user_migration_types: NewUserMigrationTypes
    alternate_email_login_allowed: AlternateEmailLoginAllowed
    test_slug: TestSlug
    modified_at: ModifiedAt2
    modified_by: ModifiedBy
    default_new_user_roles: DefaultNewUserRoles2
    default_new_user_groups: DefaultNewUserGroups2
    default_new_user_role_ids: DefaultNewUserRoleIds2
    default_new_user_group_ids: DefaultNewUserGroupIds2
    set_roles_from_groups: SetRolesFromGroups
    groups_attribute: GroupsAttribute
    groups: Groups3
    groups_with_role_ids: GroupsWithRoleIds2
    auth_requires_role: AuthRequiresRole
    user_attributes: UserAttributes6
    user_attributes_with_ids: UserAttributesWithIds2
    groups_finder_type: GroupsFinderType
    groups_member_value: GroupsMemberValue
    bypass_login_page: BypassLoginPage
    allow_normal_group_membership: AllowNormalGroupMembership
    allow_roles_from_normal_groups: AllowRolesFromNormalGroups
    allow_direct_roles: AllowDirectRoles
    url: Url32


class SamlConfig(BaseModel):
    properties: Properties193
    x_looker_status: str = Field(..., alias="x-looker-status")


class LookerGroupName4(Message):
    pass


class Name57(Message):
    pass


class Roles4(Groups3):
    pass


class Properties194(BaseModel):
    id: Id58
    looker_group_id: LookerGroupId
    looker_group_name: LookerGroupName4
    name: Name57
    roles: Roles4
    url: Url32


class SamlGroupRead(BaseModel):
    properties: Properties194
    x_looker_status: str = Field(..., alias="x-looker-status")


class Id70(HomepageId):
    pass


class LookerGroupName5(HomepageId):
    pass


class Name58(HomepageId):
    pass


class Items267(Items):
    pass


class RoleIds3(BaseModel):
    type: str
    items: Items267
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties195(BaseModel):
    id: Id70
    looker_group_id: LookerGroupId
    looker_group_name: LookerGroupName5
    name: Name58
    role_ids: RoleIds3
    url: Url32


class SamlGroupWrite(BaseModel):
    properties: Properties195
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can75(Can71):
    pass


class IdpIssuer1(Message):
    pass


class IdpUrl1(Message):
    pass


class IdpCert1(Message):
    pass


class Properties196(BaseModel):
    can: Can75
    idp_issuer: IdpIssuer1
    idp_url: IdpUrl1
    idp_cert: IdpCert1


class SamlMetadataParseResult(BaseModel):
    properties: Properties196
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name59(Message):
    pass


class Required8(Message):
    pass


class Items268(Schema):
    pass


class UserAttributes7(BaseModel):
    type: str
    items: Items268
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties197(BaseModel):
    name: Name59
    required: Required8
    user_attributes: UserAttributes7
    url: Url32


class SamlUserAttributeRead(BaseModel):
    properties: Properties197
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name60(HomepageId):
    pass


class Required9(HomepageId):
    pass


class Items269(Items):
    pass


class UserAttributeIds2(BaseModel):
    type: str
    items: Items269
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties198(BaseModel):
    name: Name60
    required: Required9
    user_attribute_ids: UserAttributeIds2
    url: Url32


class SamlUserAttributeWrite(BaseModel):
    properties: Properties198
    x_looker_status: str = Field(..., alias="x-looker-status")


class Id71(Message):
    pass


class ScheduledPlanId(HomepageId):
    pass


class Format1(HomepageId):
    pass


class ApplyFormatting(HomepageId):
    pass


class ApplyVis(HomepageId):
    pass


class LookerRecipient(Message):
    pass


class Type25(HomepageId):
    pass


class Parameters1(HomepageId):
    pass


class SecretParameters(CustomImageDataBase64):
    pass


class Message14(HomepageId):
    pass


class Properties199(BaseModel):
    id: Id71
    scheduled_plan_id: ScheduledPlanId
    format: Format1
    apply_formatting: ApplyFormatting
    apply_vis: ApplyVis
    address: Address
    looker_recipient: LookerRecipient
    type: Type25
    parameters: Parameters1
    secret_parameters: SecretParameters
    message: Message14


class ScheduledPlanDestination(BaseModel):
    properties: Properties199
    x_looker_status: str = Field(..., alias="x-looker-status")


class UserId19(HomepageId):
    pass


class RunAsRecipient(HomepageId):
    pass


class LookId11(HomepageId):
    pass


class DashboardId17(HomepageId):
    pass


class LookmlDashboardId5(HomepageId):
    pass


class FiltersString(HomepageId):
    pass


class DashboardFilters3(BaseModel):
    type: str
    x_looker_deprecated: bool = Field(..., alias="x-looker-deprecated")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class RequireResults(HomepageId):
    pass


class RequireNoResults(HomepageId):
    pass


class RequireChange(HomepageId):
    pass


class SendAllResults(HomepageId):
    pass


class Crontab(HomepageId):
    pass


class Datagroup1(HomepageId):
    pass


class Timezone1(HomepageId):
    pass


class QueryId11(HomepageId):
    pass


class Items270(Schema):
    pass


class ScheduledPlanDestination1(BaseModel):
    type: str
    items: Items270
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class RunOnce(HomepageId):
    pass


class IncludeLinks(HomepageId):
    pass


class PdfPaperSize(HomepageId):
    pass


class PdfLandscape(HomepageId):
    pass


class Embed(HomepageId):
    pass


class ColorTheme(HomepageId):
    pass


class LongTables(HomepageId):
    pass


class InlineTableWidth(DeletedAt):
    pass


class Properties200(BaseModel):
    name: Name60
    user_id: UserId19
    run_as_recipient: RunAsRecipient
    enabled: Enabled6
    look_id: LookId11
    dashboard_id: DashboardId17
    lookml_dashboard_id: LookmlDashboardId5
    filters_string: FiltersString
    dashboard_filters: DashboardFilters3
    require_results: RequireResults
    require_no_results: RequireNoResults
    require_change: RequireChange
    send_all_results: SendAllResults
    crontab: Crontab
    datagroup: Datagroup1
    timezone: Timezone1
    query_id: QueryId11
    scheduled_plan_destination: ScheduledPlanDestination1
    run_once: RunOnce
    include_links: IncludeLinks
    custom_url_base: CustomUrlBase
    custom_url_params: CustomUrlParams
    custom_url_label: CustomUrlLabel
    show_custom_url: ShowCustomUrl
    pdf_paper_size: PdfPaperSize
    pdf_landscape: PdfLandscape
    embed: Embed
    color_theme: ColorTheme
    long_tables: LongTables
    inline_table_width: InlineTableWidth


class WriteScheduledPlan(BaseModel):
    properties: Properties200
    x_looker_status: str = Field(..., alias="x-looker-status")


class ScheduledPlanDestination2(ScheduledPlanDestination1):
    pass


class CreatedAt27(DocumentationUrl):
    pass


class NextRunAt(DocumentationUrl):
    pass


class LastRunAt(DocumentationUrl):
    pass


class Can76(Can71):
    pass


class Properties201(BaseModel):
    name: Name60
    user_id: UserId19
    run_as_recipient: RunAsRecipient
    enabled: Enabled6
    look_id: LookId11
    dashboard_id: DashboardId17
    lookml_dashboard_id: LookmlDashboardId5
    filters_string: FiltersString
    dashboard_filters: DashboardFilters3
    require_results: RequireResults
    require_no_results: RequireNoResults
    require_change: RequireChange
    send_all_results: SendAllResults
    crontab: Crontab
    datagroup: Datagroup1
    timezone: Timezone1
    query_id: QueryId11
    scheduled_plan_destination: ScheduledPlanDestination2
    run_once: RunOnce
    include_links: IncludeLinks
    custom_url_base: CustomUrlBase
    custom_url_params: CustomUrlParams
    custom_url_label: CustomUrlLabel
    show_custom_url: ShowCustomUrl
    pdf_paper_size: PdfPaperSize
    pdf_landscape: PdfLandscape
    embed: Embed
    color_theme: ColorTheme
    long_tables: LongTables
    inline_table_width: InlineTableWidth
    id: Id71
    created_at: CreatedAt27
    updated_at: UpdatedAt
    title: Title25
    user: User
    next_run_at: NextRunAt
    last_run_at: LastRunAt
    can: Can76


class ScheduledPlan1(BaseModel):
    properties: Properties201
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name63(Message):
    pass


class SqlEscapedName(Message):
    pass


class SchemaName(Message):
    pass


class DataTypeDatabase(Message):
    pass


class DataType(Message):
    pass


class DataTypeLooker(Message):
    pass


class ColumnSize(DocumentationUrl):
    pass


class Snippets2(BaseModel):
    type: str
    items: Items270
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties202(BaseModel):
    name: Name63
    sql_escaped_name: SqlEscapedName
    schema_name: SchemaName
    data_type_database: DataTypeDatabase
    data_type: DataType
    data_type_looker: DataTypeLooker
    description: Description22
    column_size: ColumnSize
    snippets: Snippets2


class SchemaColumn(BaseModel):
    properties: Properties202
    x_looker_status: str = Field(..., alias="x-looker-status")


class Rows(DocumentationUrl):
    pass


class External(Message):
    pass


class Snippets3(Snippets2):
    pass


class Properties203(BaseModel):
    name: Name63
    sql_escaped_name: SqlEscapedName
    schema_name: SchemaName
    rows: Rows
    external: External
    snippets: Snippets3


class SchemaTable(BaseModel):
    properties: Properties203
    x_looker_status: str = Field(..., alias="x-looker-status")


class DialectName2(Message):
    pass


class MultipleDatabases(Message):
    pass


class ColumnSearch(Message):
    pass


class PersistentDerivedTables(Message):
    pass


class Turtles(Message):
    pass


class Percentile(Message):
    pass


class DistinctPercentile(Message):
    pass


class StableViews(Message):
    pass


class Milliseconds(Message):
    pass


class Microseconds(Message):
    pass


class Subtotals1(Message):
    pass


class Timezone3(Message):
    pass


class ConnectionPooling(Message):
    pass


class Properties204(BaseModel):
    dialect_name: DialectName2
    cost_estimate: CostEstimate
    multiple_databases: MultipleDatabases
    column_search: ColumnSearch
    persistent_table_indexes: PersistentTableIndexes
    persistent_derived_tables: PersistentDerivedTables
    turtles: Turtles
    percentile: Percentile
    distinct_percentile: DistinctPercentile
    stable_views: StableViews
    milliseconds: Milliseconds
    microseconds: Microseconds
    subtotals: Subtotals1
    location: Location
    timezone: Timezone3
    connection_pooling: ConnectionPooling


class ConnectionFeatures(BaseModel):
    properties: Properties204
    x_looker_status: str = Field(..., alias="x-looker-status")


class IsDefault(Message):
    pass


class Properties205(BaseModel):
    name: Name63
    is_default: IsDefault


class Schema1932(BaseModel):
    properties: Properties205
    x_looker_status: str = Field(..., alias="x-looker-status")


class Tables(Snippets2):
    pass


class TableLimitHit(Message):
    pass


class Properties206(BaseModel):
    name: Name63
    is_default: IsDefault
    tables: Tables
    table_limit_hit: TableLimitHit


class SchemaTables(BaseModel):
    properties: Properties206
    x_looker_status: str = Field(..., alias="x-looker-status")


class Columns(Snippets2):
    pass


class Properties207(BaseModel):
    name: Name63
    sql_escaped_name: SqlEscapedName
    schema_name: SchemaName
    columns: Columns


class SchemaColumns(BaseModel):
    properties: Properties207
    x_looker_status: str = Field(..., alias="x-looker-status")


class TableName(Message):
    pass


class ColumnName(Message):
    pass


class Properties208(BaseModel):
    schema_name: SchemaName
    table_name: TableName
    column_name: ColumnName
    data_type: DataType


class ColumnSearch1(BaseModel):
    properties: Properties208
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties209(BaseModel):
    sql: Sql


class CreateCostEstimate(BaseModel):
    properties: Properties209
    x_looker_status: str = Field(..., alias="x-looker-status")


class Cost(DocumentationUrl):
    pass


class CacheHit(Message):
    pass


class CostUnit(Message):
    pass


class Message15(Message):
    pass


class Properties210(BaseModel):
    cost: Cost
    cache_hit: CacheHit
    cost_unit: CostUnit
    message: Message15


class CostEstimate2(BaseModel):
    properties: Properties210
    x_looker_status: str = Field(..., alias="x-looker-status")


class Items276(Items):
    pass


class Suggestions1(BaseModel):
    type: str
    items: Items276
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class FromCache(Message):
    pass


class HitLimit(Message):
    pass


class UsedCalciteMaterialization(Message):
    pass


class Properties211(BaseModel):
    suggestions: Suggestions1
    error: Error1
    from_cache: FromCache
    hit_limit: HitLimit
    used_calcite_materialization: UsedCalciteMaterialization


class ModelFieldSuggestions(BaseModel):
    properties: Properties211
    x_looker_status: str = Field(..., alias="x-looker-status")


class FormatString(DimensionType):
    pass


class Label16(DimensionType):
    pass


class Name68(DimensionType):
    pass


class StrictValueFormat1(DimensionType):
    pass


class Properties212(BaseModel):
    format_string: FormatString
    label: Label16
    name: Name68
    strict_value_format: StrictValueFormat1


class ModelNamedValueFormats(BaseModel):
    properties: Properties212
    x_looker_status: str = Field(..., alias="x-looker-status")


class Connection(DimensionType):
    pass


class Items277(Schema):
    pass


class ValueFormats(BaseModel):
    type: str
    items: Items277
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties213(BaseModel):
    connection: Connection
    name: Name68
    value_formats: ValueFormats


class Model11(BaseModel):
    properties: Properties213
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can77(Can71):
    pass


class AllowPersistentSessions(HomepageId):
    pass


class SessionMinutes(DeletedAt):
    pass


class UnlimitedSessionsPerUser(HomepageId):
    pass


class UseInactivityBasedLogout(HomepageId):
    pass


class TrackSessionLocation(HomepageId):
    pass


class Properties214(BaseModel):
    can: Can77
    allow_persistent_sessions: AllowPersistentSessions
    session_minutes: SessionMinutes
    unlimited_sessions_per_user: UnlimitedSessionsPerUser
    use_inactivity_based_logout: UseInactivityBasedLogout
    track_session_location: TrackSessionLocation


class SessionConfig(BaseModel):
    properties: Properties214
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can78(Can71):
    pass


class IpAddress(Message):
    pass


class Browser(Message):
    pass


class OperatingSystem(Message):
    pass


class City(Message):
    pass


class State1(Message):
    pass


class Country(Message):
    pass


class CredentialsType(Message):
    pass


class ExtendedAt(Message):
    pass


class ExtendedCount(DocumentationUrl):
    pass


class CreatedAt28(Message):
    pass


class ExpiresAt(Message):
    pass


class Properties215(BaseModel):
    can: Can78
    id: Id71
    ip_address: IpAddress
    browser: Browser
    operating_system: OperatingSystem
    city: City
    state: State1
    country: Country
    credentials_type: CredentialsType
    extended_at: ExtendedAt
    extended_count: ExtendedCount
    sudo_user_id: SudoUserId
    created_at: CreatedAt28
    expires_at: ExpiresAt
    url: Url32


class Session(BaseModel):
    properties: Properties215
    x_looker_status: str = Field(..., alias="x-looker-status")


class ExtensionFrameworkEnabled(HomepageId):
    pass


class ExtensionLoadUrlEnabled(DashboardFilters3):
    pass


class MarketplaceAutoInstallEnabled(HomepageId):
    pass


class MarketplaceEnabled(HomepageId):
    pass


class PrivatelabelConfiguration(Model1):
    pass


class CustomWelcomeEmail1(Model1):
    pass


class OnboardingEnabled(HomepageId):
    pass


class Timezone4(HomepageId):
    pass


class AllowUserTimezones(HomepageId):
    pass


class DataConnectorDefaultEnabled(HomepageId):
    pass


class HostUrl(HomepageId):
    pass


class OverrideWarnings(CustomImageDataBase64):
    pass


class Items278(Items):
    pass


class EmailDomainAllowlist(BaseModel):
    type: str
    items: Items278
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class EmbedCookielessV2(HomepageId):
    pass


class EmbedEnabled(Message):
    pass


class Properties216(BaseModel):
    extension_framework_enabled: ExtensionFrameworkEnabled
    extension_load_url_enabled: ExtensionLoadUrlEnabled
    marketplace_auto_install_enabled: MarketplaceAutoInstallEnabled
    marketplace_enabled: MarketplaceEnabled
    privatelabel_configuration: PrivatelabelConfiguration
    custom_welcome_email: CustomWelcomeEmail1
    onboarding_enabled: OnboardingEnabled
    timezone: Timezone4
    allow_user_timezones: AllowUserTimezones
    data_connector_default_enabled: DataConnectorDefaultEnabled
    host_url: HostUrl
    override_warnings: OverrideWarnings
    email_domain_allowlist: EmailDomainAllowlist
    embed_cookieless_v2: EmbedCookielessV2
    embed_enabled: EmbedEnabled


class Setting(BaseModel):
    properties: Properties216
    x_looker_status: str = Field(..., alias="x-looker-status")


class IsValid(Message):
    pass


class NodeCount(DocumentationUrl):
    pass


class Items279(Schema):
    pass


class NodeStatus(BaseModel):
    type: str
    items: Items279
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties217(BaseModel):
    is_valid: IsValid
    node_count: NodeCount
    node_status: NodeStatus


class SmtpStatus(BaseModel):
    properties: Properties217
    x_looker_status: str = Field(..., alias="x-looker-status")


class Hostname(Message):
    pass


class Properties218(BaseModel):
    is_valid: IsValid
    message: Message15
    hostname: Hostname


class SmtpNodeStatus(BaseModel):
    properties: Properties218
    x_looker_status: str = Field(..., alias="x-looker-status")


class Name70(Message):
    pass


class Label17(Message):
    pass


class Properties219(BaseModel):
    name: Name70
    label: Label17
    sql: Sql


class Snippet(BaseModel):
    properties: Properties219
    x_looker_status: str = Field(..., alias="x-looker-status")


class ConnectionName2(HomepageId):
    pass


class ConnectionId1(DashboardFilters3):
    pass


class ModelName5(HomepageId):
    pass


class Sql4(HomepageId):
    pass


class AdditionalProperties98(AdditionalProperties):
    pass


class VisConfig3(BaseModel):
    type: str
    additionalProperties: AdditionalProperties98
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties220(BaseModel):
    connection_name: ConnectionName2
    connection_id: ConnectionId1
    model_name: ModelName5
    sql: Sql4
    vis_config: VisConfig3


class SqlQueryCreate(BaseModel):
    properties: Properties220
    x_looker_status: str = Field(..., alias="x-looker-status")


class AdditionalProperties99(Items):
    pass


class Can79(BaseModel):
    type: str
    additionalProperties: AdditionalProperties99
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class LastRuntime(DocumentationUrl):
    pass


class RunCount(DocumentationUrl):
    pass


class BrowserLimit(DocumentationUrl):
    pass


class Sql5(Message):
    pass


class LastRunAt1(Message):
    pass


class Connection1(Model1):
    pass


class ModelName6(Message):
    pass


class Creator(Model1):
    pass


class ExploreUrl(Message):
    pass


class Plaintext(Message):
    pass


class AdditionalProperties100(AdditionalProperties):
    pass


class VisConfig4(BaseModel):
    type: str
    additionalProperties: AdditionalProperties100
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class ResultMakerId2(HomepageId):
    pass


class Properties221(BaseModel):
    can: Can79
    slug: Slug3
    last_runtime: LastRuntime
    run_count: RunCount
    browser_limit: BrowserLimit
    sql: Sql5
    last_run_at: LastRunAt1
    connection: Connection1
    model_name: ModelName6
    creator: Creator
    explore_url: ExploreUrl
    plaintext: Plaintext
    vis_config: VisConfig4
    result_maker_id: ResultMakerId2


class SqlQuery1(BaseModel):
    properties: Properties221
    x_looker_status: str = Field(..., alias="x-looker-status")


class PublicKey(Message):
    pass


class Properties222(BaseModel):
    public_key: PublicKey


class SshPublicKey1(BaseModel):
    properties: Properties222
    x_looker_status: str = Field(..., alias="x-looker-status")


class SshServerId(Message):
    pass


class SshServerName(HomepageId):
    pass


class SshServerHost(HomepageId):
    pass


class SshServerPort(DeletedAt):
    pass


class SshServerUser(HomepageId):
    pass


class FingerPrint(Message):
    pass


class ShaFingerPrint(Message):
    pass


class Properties223(BaseModel):
    ssh_server_id: SshServerId
    ssh_server_name: SshServerName
    ssh_server_host: SshServerHost
    ssh_server_port: SshServerPort
    ssh_server_user: SshServerUser
    finger_print: FingerPrint
    sha_finger_print: ShaFingerPrint
    public_key: PublicKey
    status: Status4


class SshServer(BaseModel):
    properties: Properties223
    x_looker_status: str = Field(..., alias="x-looker-status")


class TunnelId1(Message):
    pass


class SshServerId1(HomepageId):
    pass


class SshServerName1(Message):
    pass


class SshServerHost1(Message):
    pass


class SshServerPort1(DocumentationUrl):
    pass


class SshServerUser1(Message):
    pass


class LastAttempt(Message):
    pass


class LocalHostPort(DeletedAt):
    pass


class DatabaseHost(HomepageId):
    pass


class DatabasePort(DeletedAt):
    pass


class Properties224(BaseModel):
    tunnel_id: TunnelId1
    ssh_server_id: SshServerId1
    ssh_server_name: SshServerName1
    ssh_server_host: SshServerHost1
    ssh_server_port: SshServerPort1
    ssh_server_user: SshServerUser1
    last_attempt: LastAttempt
    local_host_port: LocalHostPort
    database_host: DatabaseHost
    database_port: DatabasePort
    status: Status4


class SshTunnel(BaseModel):
    properties: Properties224
    x_looker_status: str = Field(..., alias="x-looker-status")


class Email8(HomepageId):
    pass


class FullName(Message):
    pass


class Reason(HomepageId):
    pass


class CreatedDate(DocumentationUrl):
    pass


class Properties225(BaseModel):
    id: Id71
    email: Email8
    full_name: FullName
    reason: Reason
    created_date: CreatedDate


class SupportAccessAllowlistEntry(BaseModel):
    properties: Properties225
    x_looker_status: str = Field(..., alias="x-looker-status")


class Items280(Items):
    pass


class Emails(BaseModel):
    type: str
    items: Items280
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class Properties226(BaseModel):
    emails: Emails
    reason: Reason


class SupportAccessAddEntries(BaseModel):
    properties: Properties226
    x_looker_status: str = Field(..., alias="x-looker-status")


class DurationInSeconds(DeletedAt):
    pass


class Properties227(BaseModel):
    duration_in_seconds: DurationInSeconds


class SupportAccessEnable(BaseModel):
    properties: Properties227
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Open(Message):
    pass


class OpenUntil(DocumentationUrl):
    pass


class Properties228(BaseModel):
    open: Open
    open_until: OpenUntil


class SupportAccessStatus(BaseModel):
    properties: Properties228
    x_looker_status: str = Field(..., alias="x-looker-status")


class BaseFontSize(HomepageId):
    pass


class ColorCollectionId(HomepageId):
    pass


class FontColor(HomepageId):
    pass


class FontFamily(HomepageId):
    pass


class FontSource(HomepageId):
    pass


class InfoButtonColor(DashboardFilters3):
    pass


class PrimaryButtonColor(HomepageId):
    pass


class TextTileBackgroundColor(HomepageId):
    pass


class WarnButtonColor(DashboardFilters3):
    pass


class ShowLastUpdatedIndicator(HomepageId):
    pass


class ShowReloadDataIcon(HomepageId):
    pass


class ShowDashboardMenu(HomepageId):
    pass


class ShowFiltersToggle(HomepageId):
    pass


class ShowDashboardHeader(HomepageId):
    pass


class CenterDashboardTitle(HomepageId):
    pass


class DashboardTitleFontSize(HomepageId):
    pass


class BoxShadow(HomepageId):
    pass


class PageMarginTop(HomepageId):
    pass


class PageMarginBottom(HomepageId):
    pass


class PageMarginSides(HomepageId):
    pass


class ShowExploreHeader(HomepageId):
    pass


class ShowExploreTitle(HomepageId):
    pass


class ShowExploreLastRun(HomepageId):
    pass


class ShowExploreTimezone(HomepageId):
    pass


class ShowExploreRunStopButton(HomepageId):
    pass


class ShowExploreActionsButton(HomepageId):
    pass


class ShowLookHeader(HomepageId):
    pass


class ShowLookTitle(HomepageId):
    pass


class ShowLookLastRun(HomepageId):
    pass


class ShowLookTimezone(HomepageId):
    pass


class ShowLookRunStopButton(HomepageId):
    pass


class ShowLookActionsButton(HomepageId):
    pass


class TileTitleFontSize(HomepageId):
    pass


class ColumnGapSize(HomepageId):
    pass


class RowGapSize(HomepageId):
    pass


class Properties229(BaseModel):
    background_color: BackgroundColor
    base_font_size: BaseFontSize
    color_collection_id: ColorCollectionId
    font_color: FontColor
    font_family: FontFamily
    font_source: FontSource
    info_button_color: InfoButtonColor
    primary_button_color: PrimaryButtonColor
    show_filters_bar: ShowFiltersBar
    show_title: ShowTitle
    text_tile_text_color: TextTileTextColor
    tile_background_color: TileBackgroundColor
    text_tile_background_color: TextTileBackgroundColor
    tile_text_color: TileTextColor
    title_color: TitleColor
    warn_button_color: WarnButtonColor
    tile_title_alignment: TileTitleAlignment
    tile_shadow: TileShadow
    show_last_updated_indicator: ShowLastUpdatedIndicator
    show_reload_data_icon: ShowReloadDataIcon
    show_dashboard_menu: ShowDashboardMenu
    show_filters_toggle: ShowFiltersToggle
    show_dashboard_header: ShowDashboardHeader
    center_dashboard_title: CenterDashboardTitle
    dashboard_title_font_size: DashboardTitleFontSize
    box_shadow: BoxShadow
    page_margin_top: PageMarginTop
    page_margin_bottom: PageMarginBottom
    page_margin_sides: PageMarginSides
    show_explore_header: ShowExploreHeader
    show_explore_title: ShowExploreTitle
    show_explore_last_run: ShowExploreLastRun
    show_explore_timezone: ShowExploreTimezone
    show_explore_run_stop_button: ShowExploreRunStopButton
    show_explore_actions_button: ShowExploreActionsButton
    show_look_header: ShowLookHeader
    show_look_title: ShowLookTitle
    show_look_last_run: ShowLookLastRun
    show_look_timezone: ShowLookTimezone
    show_look_run_stop_button: ShowLookRunStopButton
    show_look_actions_button: ShowLookActionsButton
    tile_title_font_size: TileTitleFontSize
    column_gap_size: ColumnGapSize
    row_gap_size: RowGapSize


class ThemeSettings(BaseModel):
    properties: Properties229
    x_looker_status: str = Field(..., alias="x-looker-status")


class AdditionalProperties101(Items):
    pass


class Can80(BaseModel):
    type: str
    additionalProperties: AdditionalProperties101
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class BeginAt(DeletedAt):
    pass


class EndAt(DeletedAt):
    pass


class Name71(HomepageId):
    pass


class Settings(Field):
    pass


class Properties230(BaseModel):
    can: Can80
    begin_at: BeginAt
    end_at: EndAt
    id: Id71
    name: Name71
    settings: Settings


class Theme(BaseModel):
    properties: Properties230
    x_looker_status: str = Field(..., alias="x-looker-status")


class Group1(Message):
    pass


class Properties231(BaseModel):
    value: Value8
    label: Label17
    group: Group1


class Timezone5(BaseModel):
    properties: Properties231
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can81(Can80):
    pass


class Label19(HomepageId):
    pass


class IsSystem(Message):
    pass


class IsPermanent(Message):
    pass


class ValueIsHidden(HomepageId):
    pass


class UserCanView(HomepageId):
    pass


class UserCanEdit(HomepageId):
    pass


class HiddenValueDomainWhitelist(HomepageId):
    pass


class Properties232(BaseModel):
    can: Can81
    id: Id71
    name: Name71
    label: Label19
    type: Type25
    default_value: DefaultValue
    is_system: IsSystem
    is_permanent: IsPermanent
    value_is_hidden: ValueIsHidden
    user_can_view: UserCanView
    user_can_edit: UserCanEdit
    hidden_value_domain_whitelist: HiddenValueDomainWhitelist


class UserAttribute1(BaseModel):
    properties: Properties232
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Can82(Can80):
    pass


class GroupId4(Message):
    pass


class UserAttributeId(Message):
    pass


class ValueIsHidden1(Message):
    pass


class Rank(DocumentationUrl):
    pass


class Properties233(BaseModel):
    can: Can82
    id: Id71
    group_id: GroupId4
    user_attribute_id: UserAttributeId
    value_is_hidden: ValueIsHidden1
    rank: Rank
    value: Value8


class UserAttributeGroupValue(BaseModel):
    properties: Properties233
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can83(Can80):
    pass


class Name73(Message):
    pass


class Label20(Message):
    pass


class Value11(HomepageId):
    pass


class UserId21(Message):
    pass


class UserCanEdit1(Message):
    pass


class HiddenValueDomainWhitelist1(Message):
    pass


class Properties234(BaseModel):
    can: Can83
    name: Name73
    label: Label20
    rank: Rank
    value: Value11
    user_id: UserId21
    user_can_edit: UserCanEdit1
    value_is_hidden: ValueIsHidden1
    user_attribute_id: UserAttributeId
    source: Source2
    hidden_value_domain_whitelist: HiddenValueDomainWhitelist1


class UserAttributeWithValue(BaseModel):
    properties: Properties234
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties235(BaseModel):
    email: Email8


class UserEmailOnly(BaseModel):
    properties: Properties235
    x_looker_status: str = Field(..., alias="x-looker-status")
    required: List[str]


class Can84(Can80):
    pass


class Key2(Message):
    pass


class AuthType(Message):
    pass


class Ip(Message):
    pass


class RemoteId(Message):
    pass


class Email10(Message):
    pass


class FailCount(DocumentationUrl):
    pass


class LockoutAt(DocumentationUrl):
    pass


class Properties236(BaseModel):
    can: Can84
    key: Key2
    auth_type: AuthType
    ip: Ip
    user_id: UserId21
    remote_id: RemoteId
    full_name: FullName
    email: Email10
    fail_count: FailCount
    lockout_at: LockoutAt


class UserLoginLockout(BaseModel):
    properties: Properties236
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can85(Can80):
    pass


class AvatarUrl(DocumentationUrl):
    pass


class AvatarUrlWithoutSizing(DocumentationUrl):
    pass


class Items281(Schema):
    pass


class CredentialsApi31(BaseModel):
    type: str
    items: Items281
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class CredentialsEmail1(Model1):
    pass


class CredentialsEmbed1(CredentialsApi31):
    pass


class CredentialsGoogle1(Model1):
    pass


class CredentialsLdap(Model1):
    pass


class CredentialsLookerOpenid1(Model1):
    pass


class CredentialsOidc(Model1):
    pass


class CredentialsSaml1(Model1):
    pass


class CredentialsTotp1(Model1):
    pass


class DisplayName1(Message):
    pass


class EmbedGroupSpaceId(BaseModel):
    type: str
    readOnly: bool
    x_looker_deprecated: bool = Field(..., alias="x-looker-deprecated")
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class FirstName3(HomepageId):
    pass


class Items283(Items):
    pass


class GroupIds2(BaseModel):
    type: str
    items: Items283
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class HomeFolderId(HomepageId):
    pass


class IsDisabled13(HomepageId):
    pass


class LastName3(HomepageId):
    pass


class Locale1(HomepageId):
    pass


class LookerVersions(GroupIds2):
    pass


class ModelsDirValidated(HomepageId):
    pass


class PersonalFolderId(Message):
    pass


class PresumedLookerEmployee(Message):
    pass


class RoleIds4(GroupIds2):
    pass


class Items286(Schema):
    pass


class Sessions(BaseModel):
    type: str
    items: Items286
    readOnly: bool
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class UiState(BaseModel):
    type: str
    additionalProperties: AdditionalProperties101
    description: str
    x_looker_nullable: bool = Field(..., alias="x-looker-nullable")


class VerifiedLookerEmployee(Message):
    pass


class RolesExternallyManaged(Message):
    pass


class AllowDirectRoles3(Message):
    pass


class AllowNormalGroupMembership3(Message):
    pass


class AllowRolesFromNormalGroups3(Message):
    pass


class EmbedGroupFolderId(Message):
    pass


class IsIamAdmin(Message):
    pass


class Properties237(BaseModel):
    can: Can85
    avatar_url: AvatarUrl
    avatar_url_without_sizing: AvatarUrlWithoutSizing
    credentials_api3: CredentialsApi31
    credentials_email: CredentialsEmail1
    credentials_embed: CredentialsEmbed1
    credentials_google: CredentialsGoogle1
    credentials_ldap: CredentialsLdap
    credentials_looker_openid: CredentialsLookerOpenid1
    credentials_oidc: CredentialsOidc
    credentials_saml: CredentialsSaml1
    credentials_totp: CredentialsTotp1
    display_name: DisplayName1
    email: Email10
    embed_group_space_id: EmbedGroupSpaceId
    first_name: FirstName3
    group_ids: GroupIds2
    home_folder_id: HomeFolderId
    id: Id71
    is_disabled: IsDisabled13
    last_name: LastName3
    locale: Locale1
    looker_versions: LookerVersions
    models_dir_validated: ModelsDirValidated
    personal_folder_id: PersonalFolderId
    presumed_looker_employee: PresumedLookerEmployee
    role_ids: RoleIds4
    sessions: Sessions
    ui_state: UiState
    verified_looker_employee: VerifiedLookerEmployee
    roles_externally_managed: RolesExternallyManaged
    allow_direct_roles: AllowDirectRoles3
    allow_normal_group_membership: AllowNormalGroupMembership3
    allow_roles_from_normal_groups: AllowRolesFromNormalGroups3
    embed_group_folder_id: EmbedGroupFolderId
    is_iam_admin: IsIamAdmin
    url: Url32


class User3(BaseModel):
    properties: Properties237
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can86(Can80):
    pass


class FirstName4(Message):
    pass


class LastName4(Message):
    pass


class Properties238(BaseModel):
    can: Can86
    id: Id71
    first_name: FirstName4
    last_name: LastName4
    display_name: DisplayName1
    avatar_url: AvatarUrl
    url: Url32


class UserPublic(BaseModel):
    properties: Properties238
    x_looker_status: str = Field(..., alias="x-looker-status")


class Version2(Message):
    pass


class FullVersion(Message):
    pass


class SwaggerUrl(DocumentationUrl):
    pass


class Properties239(BaseModel):
    version: Version2
    full_version: FullVersion
    status: Status4
    swagger_url: SwaggerUrl


class ApiVersionElement(BaseModel):
    properties: Properties239
    x_looker_status: str = Field(..., alias="x-looker-status")


class LookerReleaseVersion(Message):
    pass


class CurrentVersion(Model1):
    pass


class SupportedVersions(Sessions):
    pass


class ApiServerUrl(Message):
    pass


class WebServerUrl(Message):
    pass


class Properties240(BaseModel):
    looker_release_version: LookerReleaseVersion
    current_version: CurrentVersion
    supported_versions: SupportedVersions
    api_server_url: ApiServerUrl
    web_server_url: WebServerUrl


class ApiVersion(BaseModel):
    properties: Properties240
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties241(BaseModel):
    content: Content
    subject: Subject
    header: Header


class WelcomeEmailTest(BaseModel):
    properties: Properties241
    x_looker_status: str = Field(..., alias="x-looker-status")


class LogoFile(HomepageId):
    pass


class LogoUrl(Message):
    pass


class FaviconFile(HomepageId):
    pass


class FaviconUrl(Message):
    pass


class DefaultTitle(HomepageId):
    pass


class ShowHelpMenu(HomepageId):
    pass


class ShowDocs(HomepageId):
    pass


class ShowEmailSubOptions(HomepageId):
    pass


class AllowLookerMentions(HomepageId):
    pass


class AllowLookerLinks(HomepageId):
    pass


class CustomWelcomeEmailAdvanced(HomepageId):
    pass


class SetupMentions(HomepageId):
    pass


class AlertsLogo(HomepageId):
    pass


class AlertsLinks(HomepageId):
    pass


class FoldersMentions(HomepageId):
    pass


class Properties242(BaseModel):
    id: Id71
    logo_file: LogoFile
    logo_url: LogoUrl
    favicon_file: FaviconFile
    favicon_url: FaviconUrl
    default_title: DefaultTitle
    show_help_menu: ShowHelpMenu
    show_docs: ShowDocs
    show_email_sub_options: ShowEmailSubOptions
    allow_looker_mentions: AllowLookerMentions
    allow_looker_links: AllowLookerLinks
    custom_welcome_email_advanced: CustomWelcomeEmailAdvanced
    setup_mentions: SetupMentions
    alerts_logo: AlertsLogo
    alerts_links: AlertsLinks
    folders_mentions: FoldersMentions


class WhitelabelConfiguration(BaseModel):
    properties: Properties242
    x_looker_status: str = Field(..., alias="x-looker-status")


class Properties243(BaseModel):
    logo_file: LogoFile
    logo_url: LogoUrl
    favicon_file: FaviconFile
    favicon_url: FaviconUrl
    default_title: DefaultTitle
    show_help_menu: ShowHelpMenu
    show_docs: ShowDocs
    show_email_sub_options: ShowEmailSubOptions
    allow_looker_mentions: AllowLookerMentions
    allow_looker_links: AllowLookerLinks
    custom_welcome_email_advanced: CustomWelcomeEmailAdvanced
    setup_mentions: SetupMentions
    alerts_logo: AlertsLogo
    alerts_links: AlertsLinks
    folders_mentions: FoldersMentions


class PrivatelabelConfiguration1(BaseModel):
    properties: Properties243
    x_looker_status: str = Field(..., alias="x-looker-status")


class Can87(Can80):
    pass


class Projects(Sessions):
    pass


class Properties244(BaseModel):
    can: Can87
    id: Id71
    projects: Projects


class Workspace(BaseModel):
    properties: Properties244
    x_looker_status: str = Field(..., alias="x-looker-status")


class Definitions(BaseModel):
    Error: Error
    DashboardBase: DashboardBase
    HomepageSection: HomepageSection
    HomepageItem: HomepageItem
    ValidationError: ValidationError
    ValidationErrorDetail: ValidationErrorDetail
    AccessToken: AccessToken
    AlertFieldFilter: AlertFieldFilter
    AlertAppliedDashboardFilter: AlertAppliedDashboardFilter
    AlertField: AlertField
    AlertConditionState: AlertConditionState
    Alert: Alert
    MobilePayload: MobilePayload
    AlertNotifications: AlertNotifications
    AlertDestination: AlertDestination
    AlertPatch: AlertPatch
    ApiSession: ApiSession
    ArtifactUsage: ArtifactUsage
    ArtifactNamespace: ArtifactNamespace
    UpdateArtifact: UpdateArtifact
    Artifact: Artifact
    BackupConfiguration: BackupConfiguration
    BoardItem: BoardItem
    Board: Board
    BoardSection: BoardSection
    ColorStop: ColorStop
    ContinuousPalette: ContinuousPalette
    DiscretePalette: DiscretePalette
    ColorCollection: ColorCollection
    ContentFavorite: ContentFavorite
    ContentMetaGroupUser: ContentMetaGroupUser
    ContentMeta: ContentMeta
    ContentSearch: ContentSearch
    ContentValidation: ContentValidation
    ContentValidatorError: ContentValidatorError
    ContentValidationFolder: ContentValidationFolder
    ContentValidationLook: ContentValidationLook
    ContentValidationDashboard: ContentValidationDashboard
    ContentValidationDashboardElement: ContentValidationDashboardElement
    ContentValidationError: ContentValidationError
    ContentValidationDashboardFilter: ContentValidationDashboardFilter
    ContentValidationScheduledPlan: ContentValidationScheduledPlan
    ContentValidationAlert: ContentValidationAlert
    ContentValidationLookMLDashboard: ContentValidationLookMLDashboard
    ContentValidationLookMLDashboardElement: ContentValidationLookMLDashboardElement
    ContentView: ContentView
    CreateEmbedUserRequest: CreateEmbedUserRequest
    CreateOAuthApplicationUserStateRequest: CreateOAuthApplicationUserStateRequest
    CreateOAuthApplicationUserStateResponse: CreateOAuthApplicationUserStateResponse
    CredentialsApi3: CredentialsApi3
    CreateCredentialsApi3: CreateCredentialsApi3
    CredentialsEmail: CredentialsEmail
    CredentialsEmailSearch: CredentialsEmailSearch
    CredentialsEmbed: CredentialsEmbed
    CredentialsGoogle: CredentialsGoogle
    CredentialsLDAP: CredentialsLDAP
    CredentialsLookerOpenid: CredentialsLookerOpenid
    CredentialsOIDC: CredentialsOIDC
    CredentialsSaml: CredentialsSaml
    CredentialsTotp: CredentialsTotp
    CustomWelcomeEmail: CustomWelcomeEmail
    DashboardAggregateTableLookml: DashboardAggregateTableLookml
    DashboardAppearance: DashboardAppearance
    DashboardElement: DashboardElement1
    DashboardFilter: DashboardFilter1
    CreateDashboardFilter: CreateDashboardFilter
    DashboardLayoutComponent: DashboardLayoutComponent
    DashboardLayout: DashboardLayout
    DashboardLookml: DashboardLookml
    Dashboard: Dashboard2
    DataActionFormField: DataActionFormField
    DataActionForm: DataActionForm
    DataActionFormSelectOption: DataActionFormSelectOption
    DataActionRequest: DataActionRequest
    DataActionResponse: DataActionResponse
    DataActionUserState: DataActionUserState
    Datagroup: Datagroup
    DBConnectionBase: DBConnectionBase
    DBConnection: DBConnection
    DBConnectionOverride: DBConnectionOverride
    DBConnectionTestResult: DBConnectionTestResult
    DelegateOauthTest: DelegateOauthTest
    DependencyGraph: DependencyGraph
    DialectInfo: DialectInfo
    DialectInfoOptions: DialectInfoOptions
    Dialect: Dialect2
    DigestEmailSend: DigestEmailSend
    DigestEmails: DigestEmails
    EgressIpAddresses: EgressIpAddresses
    EmbedParams: EmbedParams
    EmbedSsoParams: EmbedSsoParams
    EmbedCookielessSessionAcquire: EmbedCookielessSessionAcquire
    EmbedCookielessSessionAcquireResponse: EmbedCookielessSessionAcquireResponse
    EmbedCookielessSessionGenerateTokens: EmbedCookielessSessionGenerateTokens
    EmbedCookielessSessionGenerateTokensResponse: EmbedCookielessSessionGenerateTokensResponse
    EmbedSecret: EmbedSecret
    EmbedUrlResponse: EmbedUrlResponse
    SmtpSettings: SmtpSettings
    ExternalOauthApplication: ExternalOauthApplication
    FolderBase: FolderBase
    CreateFolder: CreateFolder
    UpdateFolder: UpdateFolder
    Folder: Folder4
    GitBranch: GitBranch
    GitConnectionTest: GitConnectionTest
    GitConnectionTestResult: GitConnectionTestResult
    GitStatus: GitStatus
    GroupIdForGroupInclusion: GroupIdForGroupInclusion
    Group: Group
    GroupSearch: GroupSearch
    GroupHierarchy: GroupHierarchy
    GroupIdForGroupUserInclusion: GroupIdForGroupUserInclusion
    ImportedProject: ImportedProject
    IntegrationHub: IntegrationHub
    Integration: Integration
    IntegrationParam: IntegrationParam
    IntegrationRequiredField: IntegrationRequiredField
    IntegrationTestResult: IntegrationTestResult
    InternalHelpResourcesContent: InternalHelpResourcesContent
    InternalHelpResources: InternalHelpResources
    LDAPConfig: LDAPConfig
    LDAPConfigTestResult: LDAPConfigTestResult
    LDAPConfigTestIssue: LDAPConfigTestIssue
    LDAPGroupRead: LDAPGroupRead
    LDAPGroupWrite: LDAPGroupWrite
    LDAPUserAttributeRead: LDAPUserAttributeRead
    LDAPUserAttributeWrite: LDAPUserAttributeWrite
    LDAPUser: LDAPUser
    LegacyFeature: LegacyFeature
    Locale: Locale
    LocalizationSettings: LocalizationSettings
    LookBasic: LookBasic
    Look: Look3
    LookWithQuery: LookWithQuery
    LookWithDashboards: LookWithDashboards
    LookModel: LookModel
    LookmlModelNavExplore: LookmlModelNavExplore
    LookmlModelExplore: LookmlModelExplore
    LookmlModelExploreSupportedMeasureType: LookmlModelExploreSupportedMeasureType
    LookmlModelExploreAccessFilter: LookmlModelExploreAccessFilter
    LookmlModelExploreConditionallyFilter: LookmlModelExploreConditionallyFilter
    LookmlModelExploreAlwaysFilter: LookmlModelExploreAlwaysFilter
    LookmlModelExploreAlias: LookmlModelExploreAlias
    LookmlModelExploreSet: LookmlModelExploreSet
    LookmlModelExploreError: LookmlModelExploreError
    LookmlModelExploreFieldset: LookmlModelExploreFieldset
    LookmlModelExploreField: LookmlModelExploreField
    LookmlModelExploreFieldEnumeration: LookmlModelExploreFieldEnumeration
    LookmlModelExploreFieldTimeInterval: LookmlModelExploreFieldTimeInterval
    LookmlModelExploreFieldSqlCase: LookmlModelExploreFieldSqlCase
    LookmlModelExploreFieldMeasureFilters: LookmlModelExploreFieldMeasureFilters
    LookmlModelExploreFieldMapLayer: LookmlModelExploreFieldMapLayer
    LookmlModelExploreJoins: LookmlModelExploreJoins
    LookmlModel: LookmlModel
    LookmlTest: LookmlTest
    LookmlTestResult: LookmlTestResult
    Manifest: Manifest
    MaterializePDT: MaterializePDT
    MergeQuery: MergeQuery
    MergeQuerySourceQuery: MergeQuerySourceQuery
    MergeFields: MergeFields1
    MobileSettings: MobileSettings
    MobileFeatureFlags: MobileFeatureFlags1
    MobileToken: MobileToken
    ModelSet: ModelSet
    OauthClientApp: OauthClientApp
    OIDCConfig: OIDCConfig
    OIDCGroupRead: OIDCGroupRead
    OIDCGroupWrite: OIDCGroupWrite
    OIDCUserAttributeRead: OIDCUserAttributeRead
    OIDCUserAttributeWrite: OIDCUserAttributeWrite
    PasswordConfig: PasswordConfig
    Permission: Permission
    PermissionSet: PermissionSet
    ProjectFile: ProjectFile
    Project: Project
    ProjectError: ProjectError
    ModelsNotValidated: ModelsNotValidated
    ProjectValidation: ProjectValidation
    ProjectValidationCache: ProjectValidationCache
    ProjectWorkspace: ProjectWorkspace
    Query: Query2
    CreateQueryTask: CreateQueryTask
    QueryTask: QueryTask
    RenderTask: RenderTask
    CreateDashboardRenderTask: CreateDashboardRenderTask
    RepositoryCredential: RepositoryCredential
    ResultMakerFilterablesListen: ResultMakerFilterablesListen
    ResultMakerFilterables: ResultMakerFilterables
    ResultMakerWithIdVisConfigAndDynamicFields: ResultMakerWithIdVisConfigAndDynamicFields
    Role: Role
    RoleSearch: RoleSearch
    RunningQueries: RunningQueries
    SamlConfig: SamlConfig
    SamlGroupRead: SamlGroupRead
    SamlGroupWrite: SamlGroupWrite
    SamlMetadataParseResult: SamlMetadataParseResult
    SamlUserAttributeRead: SamlUserAttributeRead
    SamlUserAttributeWrite: SamlUserAttributeWrite
    ScheduledPlanDestination: ScheduledPlanDestination
    WriteScheduledPlan: WriteScheduledPlan
    ScheduledPlan: ScheduledPlan1
    SchemaColumn: SchemaColumn
    SchemaTable: SchemaTable
    ConnectionFeatures: ConnectionFeatures
    Schema: Schema1932
    SchemaTables: SchemaTables
    SchemaColumns: SchemaColumns
    ColumnSearch: ColumnSearch1
    CreateCostEstimate: CreateCostEstimate
    CostEstimate: CostEstimate2
    ModelFieldSuggestions: ModelFieldSuggestions
    ModelNamedValueFormats: ModelNamedValueFormats
    Model: Model11
    SessionConfig: SessionConfig
    Session: Session
    Setting: Setting
    SmtpStatus: SmtpStatus
    SmtpNodeStatus: SmtpNodeStatus
    Snippet: Snippet
    SqlQueryCreate: SqlQueryCreate
    SqlQuery: SqlQuery1
    SshPublicKey: SshPublicKey1
    SshServer: SshServer
    SshTunnel: SshTunnel
    SupportAccessAllowlistEntry: SupportAccessAllowlistEntry
    SupportAccessAddEntries: SupportAccessAddEntries
    SupportAccessEnable: SupportAccessEnable
    SupportAccessStatus: SupportAccessStatus
    ThemeSettings: ThemeSettings
    Theme: Theme
    Timezone: Timezone5
    UserAttribute: UserAttribute1
    UserAttributeGroupValue: UserAttributeGroupValue
    UserAttributeWithValue: UserAttributeWithValue
    UserEmailOnly: UserEmailOnly
    UserLoginLockout: UserLoginLockout
    User: User3
    UserPublic: UserPublic
    ApiVersionElement: ApiVersionElement
    ApiVersion: ApiVersion
    WelcomeEmailTest: WelcomeEmailTest
    WhitelabelConfiguration: WhitelabelConfiguration
    PrivatelabelConfiguration: PrivatelabelConfiguration1
    Workspace: Workspace


class Model(BaseModel):
    swagger: str
    info: Info
    basePath: str
    consumes: List[str]
    produces: List[str]
    host: str
    schemes: List[str]
    tags: List[Tag]
    paths: Paths
    definitions: Definitions
